// Copyright (C) 2006 Ken Eguro, Scott Hauck, Akshay Sharma
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

/****************************************************************
Class definition of a generic net. A net has a single source 
terminal and one or more sink terminals.
****************************************************************/    

#ifndef NET_H
#define NET_H

#include "Terminal.h"
class PipeliningNet;
extern int routeBoundingBox;

//Note the difference between a Net and a PipeliningNet.  
//A regular net is one that is created at parse time and is persistent 
//until execution is done.  This is a net that has a source and sinks that
//might or might not have a pipelining latency.
//A pipelining net is one that is virtually created when a pipelined signal 
//is routed.  For example, a regular net with 2 one latency
//sinks might have two pipelining nets, if the sinks share the register, or
//three pipelining nets, if the sinks do not share the register.
//These will be created and destroyed each time the net is re-routed
class Net{
 public:
  //char name [MAX_STR_LEN];
  string name;
  Terminal* source;
  vector<Terminal*> sinkVector;  //vector of logic blocks in this netlist
  int numSinks;
  routeTree rTree;
  bool isCounted;
  int magicNum;

  //Timing-driven routing stuff
  vector <double> linkDelayVector; //vector of delay between this net's source and sink i
  vector <double> oldLinkDelayVector;
  vector <double> slackVector; //vector of longest path delay containing arc (source, sink i)

  //Pipeline routing stuff
  bool pipelined;

  //This is a vector that indicates the latency of each sink
  vector<int> latencyVector;
  int netMaxLatency;

  list<PipeliningNet *> pipelineNetList;
  list<Terminal *> pipelineRegisterList;

  //ID to keep consistent number of pipeline registers during placement
  long registerID;

  Net(){
    magicNum=17111;
    source=0;
    numSinks=0;
    isCounted=false;
    pipelined = false;
    netMaxLatency = 0;
    wireCost = 0.0;
    timingCost = 0.0;
    minRow = -1;
    maxRow = -1;
    minCol = -1;
    maxCol = -1;
    registerID = 0;
  }

  ~Net(){
    if(magicNum!=17111)
      ERROR("trying to double delete a node...");
    else
      magicNum=0;
  }

  void set_name(string n);  //give this net a name
  void set_source(Terminal* s);  //set this net's source terminal
  //void addto_sinkList(Terminal* s);  //add a terminal to this net's sink list
  void addto_sinkVector(Terminal* s);  //add a terminal to this net's sink list
  
  void init_rTreeRoot();  //make the source routing node the root of rTree
  void rip_rTree();  //rip up the existing routing tree, and decrement sharing

  //get a pointer to this net's route tree
  routeTree* get_rTree();

  //rip up the branch of the current route tree that contains sink_rnode
  void rip_rTreeBranch(routingNode* sink_rnode);

  void set_isCounted(bool c);  //whether this net's cost contributions have already
                                 //been considered during an annealing move

  string get_name();  //get name of this net
  Terminal* get_source();  //get this net's source terminal
  int get_numSinks();  //get num sinks of this net
  treeNode* find_treeNode(routingNode* rnode);  //return tree node that holds the routing
                                                //node rnode
  bool get_isCounted();  //get whether this net's cost contributions have already
                         //been considered during an annealing move
  void print_Info(bool pipeline_aware);  //print details about this net
  void print_rTree();  //print the current route tree of this net
  int get_rNodeCount();  //get the total number of nodes in the current route tree

  //count number of nodes in the current route tree
  void countNodes(treeNode* r,int* count_ptr);
  
  //calculate the sharing and wire cost of this net's current route
  void calcCosts(int* sh_cost,int* wr_cost); 
  
  //recursively calculate wire and sharing cost
  void wsCostCalc(treeNode* r,int* sh_cost,int* wr_cost);

  //reconnect the sub-tree rooted at join_node to this net's current route tree.
  //increment occupancies of the routing nodes that constitute the subtree.
  int reconnect(treeNode* join_node);

  //increment occupancies of the routing nodes that constitute sub-tree rooted at r
  void incrementOccupancies(treeNode* r,int how);

  //Reset the routing nodes in the tree to not counted in sharing calculations
  void reset_counted();

  //initial creation of vector of sinks
  void create_sinkVector();

  //sort vector of sinks in increasing order of separation from source...A*
  void sort_sinkVectorDistance();

  //sort vector of sinks in decreasing order of slack
  void sort_sinkVectorTiming();

  //We have placement-locked registers, so sort all of the pipelined nets
  void sortLockedRegs(bool sortByTiming);

  //Timing-driven stuff  
  //intial creation of slack vector
  void create_slackVector();
  //initial creation of linkDelayVector
  void create_linkDelayVector();
  //Calculates the new delay for this net
  //void updatePathDelay();

  void clearLinkDelayAndSlackVectors();

  //Pipeline routing stuff
  void addto_latencyVector(int latency);
  void ripUpPipelineInformation();
  void reCalculateOccupancy(bool pipeline_aware);

  bool followSinks(bool pipeline_aware);

  void calculatePipelinedSlacks();

  void printCriticalPath(bool pipeline_aware);

  // Pipelined placement stuff
  //void createZipped3PipelinedNets();
  void createZippedPipelinedNets();
  void cleanUpPipelinePlacementInformation();

  void updatePlacementTiming();

  double wireCost;
  double timingCost;
  double newWireCost;
  double newTimingCost;
  double newTimingCostPrime;
  void calculatePlacementCost(bool timing_driven, double criticalPathDelay);

  int minRow, maxRow, minCol, maxCol;

  int returnNumPipeliningRegisters(){
    return pipelineRegisterList.size();
  }

  int getNumPipelinedNets(){
    return (pipelineNetList.size());
  }
  
  void updateSinkDelayAndSlack(Terminal *sink);
  void updateSourceDelaysAndSlacks();
  void reCalculatePlacementCost(bool timing_driven, double criticalPathDelay);
  double getDeltaWireCost();
  void overWriteOldWireCost();  
  double getDeltaTimingCost();
  double getDeltaTimingCostPrime();
  void overWriteOldTimingCost();
  void clearTimingCostPrime();

  void zeroSlackVectors();

  void removePipelineRegFromPRList(Terminal *pipeReg);
  void removePipelineNetFromPNList(PipeliningNet *pNet);

  void subtractLatencyFromAllSinks(int numRemoved);
  void subtractLatencyFromThisSink(Terminal *sink, int numRemoved);

  PipeliningNet* makePipelined(Netlist *netlist);
  void increaseSinkLatency(Terminal *sink, int additionalLatency);
  void increaseAllSinkLatency(int additionalLatency);

  int returnLatencyOfSink(Terminal *sink);
  void setLatencyOfSink(Terminal *sink, int lat);
  void updateLatencyAllSinks();

  void checkNet();
  void printVPRPlacement(FILE *fp);
  
  bool checkNodeRoutingBoundingBox(routingNode *node){
    //Notice that node numbers are taken by the wires to the right and down of an array location
    if(node->row_high < minRow - routeBoundingBox - 1 || 
       node->col_high < minCol - routeBoundingBox - 1 || 
       node->row_low > maxRow + routeBoundingBox ||
       node->col_low > maxCol + routeBoundingBox){
      return false;
    }
    return true;
  }

  //See if this is one of the sinks of this net
  bool foundNewSink(routingNode *node);

  void addShiftRegistersAndPipelineRegisters(list <Terminal*> *shiftRegisterList, 
					     list <int> *shiftRegisterDepthList, 
					     list <Terminal*> *PRList);

  void printDistanceToAllSinks();
  int minDistanceToAllSinks();

  double returnSinkSlack(Terminal *sink);
  double returnMinSinkSlack();
  bool drivesSomethingIn(int row, int col);

  void printRTreeNodesFrom(Terminal *sink);

  void printAllSinkCriticalities(double criticalPathDelay, bool pipeline_aware);
};

class PipeliningNet{
 public:
  //char name [MAX_STR_LEN];
  string name;
  int magicNum;

  Net *parentNet;

  int latency;

  Terminal* source;
  vector<Terminal*> sinkVector;
  routeTree rTree;

  //Timing-driven routing stuff
  vector <double> linkDelayVector; //vector of delay between this net's source and sink i
  vector <double> oldLinkDelayVector;
  vector <double> slackVector; //vector of longest path delay containing arc (source, sink i)

  //Pipelined-Routing stuff
  routeTree tempRTree;

  PipeliningNet(string str, int lat, Net* parent){
    //strcpy(name, str);
    name = str;
    latency = lat;
    magicNum = 32983;
    wireCost = 0.0;
    timingCost = 0.0;
    minRow = -1;
    maxRow = -1;
    minCol = -1;
    maxCol = -1;
    parentNet = parent;
  }

  PipeliningNet(string str, Terminal* sourceP, int lat, Net* parent){
    //strcpy(name, str);
    name = str;
    source = sourceP;
    latency = lat;
    magicNum=32983;
    wireCost = 0.0;
    timingCost = 0.0;
    minRow = -1;
    maxRow = -1;
    minCol = -1;
    maxCol = -1;
    parentNet = parent;
  }

  ~PipeliningNet(){
    if(magicNum!=32983)
      ERROR("trying to double delete tree here...");
    else
      magicNum=0;
  }

  void init_rTreeRoot();  //make the source routing node the root of rTree
  treeNode* find_treeNode(routingNode* rnode);  //return tree node that holds the routing
                                                //node rnode  
  void print_rTree();  //print the current route tree of this net
  int get_rNodeCount();  //get the total number of nodes in the current route tree
  void countNodes(treeNode* r,int* count_ptr);    //count number of nodes in the current route tree
  void copyRTree();
  void print_Info();
  void ripUpRTree(bool countSource);
  void reCalculateOccupancy();

  void updatePlacementTiming();

  void printCriticalPath();

  double wireCost;
  double timingCost;

  double newWireCost;
  double newTimingCost;
  double newTimingCostPrime;

  void updateSinkDelayAndSlack(Terminal *sink);
  void updateSourceDelaysAndSlacks();
  void updateSourceDelaysAndSlacksForwards();

  void calculatePlacementCost(bool timing_driven, double criticalPathDelay);
  void reCalculatePlacementCost(bool timing_driven, double criticalPathDelay);
  void reCalculatePlacementCostForwards(bool timing_driven, double criticalPathDelay);

  int minRow, maxRow, minCol, maxCol;

  double getDeltaWireCost();
  void overWriteOldWireCost();
  double getDeltaTimingCost();
  double getDeltaTimingCostPrime();
  void overWriteOldTimingCost();
  void clearTimingCostPrime();

  void zeroSlackVectors();

  void sortLockedRegs(bool sortByTiming);
  //sort vector of sinks in increasing order of separation from source...A*
  void sort_sinkVectorDistance();

  //sort vector of sinks in decreasing order of slack
  void sort_sinkVectorTiming();

  void rip_rTree();  //rip up the existing routing tree, and decrement sharing
  routeTree* get_rTree();

  void pushSinkPlusData(Terminal *sink, double linkDelay, double slack);
  void removeSinkPlusData(Terminal *sink);
  void removeSinkPlusData(int select);

  Terminal* getSink(int i){return sinkVector.at(i);}
  double getLinkDelay(int i){return linkDelayVector.at(i);}
  double getSlack(int i){return slackVector.at(i);};

  void detachAndReattachHereForward(PipeliningNet *recvPipeliningNet, Netlist *netlist, int numMove);
  void reduceLatency(int amount);
  void updateAllLatencies(int lat);
  void insertRegistersExistingSinkBanks(int numRegs, Netlist *netlist, double criticalPathDelay);
  bool insertRegistersNewBanks(int numRegs, Netlist *netlist, double criticalPathDelay);

  void addBackNetAndRegistersForwards(Netlist *netlist, int numMove);
  void addBackNetAndRegistersBackwards(Terminal *sink, Netlist *netlist);

  void decrementNerfedSpanForwards();
  void decrementNerfedSpanBackwards(Terminal *sink);
  double returnNerfedWireCostForwards();
  double returnNerfedWireCostBackwards(bool timing_driven, double criticalPathDelay, Terminal *sink);
  double returnNerfedTimingCostForwards();
  double returnNerfedTimingCostBackwards(Terminal *sink);

  void deleteNerfedNetsAndRegistersForwards();
  void deleteNerfedNetsAndRegistersBackwards(Terminal *sink);

  void resetValues();

  void checkSourcesAndSinks();

  bool hasSink(Terminal *sink);

  string get_name(){
    return name;
  };

  bool checkNodeRoutingBoundingBox(routingNode *node);
  //See if this is one of the sinks of this net
  bool foundNewSink(routingNode *node);
  void printDistanceToAllSinks();
  int minDistanceToAllSinks();

  double returnSinkSlack(Terminal *sink);
  double returnMinSinkSlack();
  bool drivesSomethingIn(int row, int col);

  void calculateMaxSourceInputOutputCriticality(double criticalPathDelay);
  void calculateMaxSinkInputOutputCriticality(double criticalPathDelay);

  void printRTreeNodesFrom(Terminal *sink);

  void printAllSinkCriticalities(double criticalPathDelay, bool pipeline_aware);
};

#endif


