// Copyright (C) 2006 Ken Eguro, Scott Hauck, Akshay Sharma
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

/****************************************************************
Class definition for a generic netlist. A netlist consists of
a list of all logic blocks in the netlist, and a separate list
of the nets that connect the logic blocks. Further, we also keep
a record of the number of IO blocks, logic blocks and nets in 
the netlist.
****************************************************************/ 
#ifndef NETLIST_H
#define NETLIST_H

class logicBlock;
class Net;
class c_clb;
//class unZipWidget;

class Netlist{
 public:
  //char name [MAX_STR_LEN];  //name of netlist
  string name;  //name of netlist
  list<c_clb *> clbList;
  list<logicBlock*> lbList;  //list of logic blocks in netlist
  list<Net*> netList;  //list of nets in netlist
  int numlogicBlocks;  //num logic blocks in netlist
  int numInputs;  //num input-only IO blocks
  int numOutputs;  //num output-only IO blocks
  int numNets;  //num nets in netlist
  int numBlocks;
  int numBlocksAndBanks;

  double wireCostSurplus;
  double timingCostSurplus;

  vector<logicBlock*>* lbVector;  //pointer to vector of logic blocks in this netlist

  list<logicBlock*>::iterator* lb_itr;  //pointer to lbList iterator
  list<Net*>::iterator* net_itr;  //pointer to netList iterator

  int magicNum;

  //Pipeline Aware Stuff
  Netlist(){
    numlogicBlocks=0;
    numNets=0;
    numInputs=numOutputs=0;
    numBlocks = 0;
    numBlocksAndBanks = 0;
    lb_itr=0;net_itr=0;
    lbVector=0;
    magicNum=18111;
    wireCost = 0.0;
    timingCost = 0.0;
    maxCut = 0.0;
    aveCut = 0.0;
    wireCostSurplus = 0.0;
    timingCostSurplus = 0.0;
  }

  ~Netlist(){
    if(magicNum!=18111)
      ERROR("trying to double delete a node...");
    else
      magicNum=0;
    
    //Delete all of the clbs, lb and nets
    for(list<c_clb *>::iterator clbIter = clbList.begin(); clbIter != clbList.end(); clbIter++){
      delete (*clbIter);
    }
    for(list<logicBlock*>::iterator lbIter = lbList.begin(); lbIter != lbList.end(); lbIter++){
      delete (*lbIter);
    }
    for(list<Net*>::iterator netIter = netList.begin(); netIter!= netList.end(); netIter++){
      delete (*netIter);
    }
  }

  void create_lbVector();  //create vector of pointers to logic blocks that comprise 
                              //this netlist. note that this function can be called
                              //only AFTER lbList has been completely populated.
  logicBlock* get_logicBlock(int i); //return logic block at lbVector[i]
  void create_lbListItr();  //create a new iterator for lbList
  void create_netListItr();  //create a new iterator for netList
  void remove_lbListItr();  //remove lbList iterator
  void remove_netListItr();  //remove netList iterator
  logicBlock* getnext_lbList();  //get next element from lbList
  Net* getnext_netList();  //get next element from netList

  void set_name(string n);  //set the name of netlist
  void addto_lbList(logicBlock* p_block);  //add a logic block to lbList
  void addto_netList(Net* p_net);  //add a net to netList

  int get_numlogicBlocks();  //get num logic blocks in netlist
  int countLUT();
  int countMult();
  int countFixedReg();
  int countRAM();
  int get_numNets();  //get num nets in netlist
  string get_name();  //get the name of mcnc netlist
  int get_numInputs();  //get num input-only IO blocks
  int get_numOutputs();  //get num output-only IO blocks

  void print_lbList();  //print the entire list of logic blocks
  void print_netList(bool pipeline_aware);  //print the entire list of nets
  void print_name();  //print the name of netlist
  void print_logicBlock(logicBlock* p_block);  //print details of logic block
  void print_Net(Net* p_net, bool pipeline_aware);    //print details of net pointed to by p_net

  logicBlock* found_logicBlock(char* block_name);  //look for logic block in lbList by name.
  Net* found_Net(string net_name);  //look for net in netList by name.

  void calculateNumBlocks(){numBlocks = numlogicBlocks + numInputs+ numOutputs;}
  int get_numCLBs(){return clbList.size();};
  //int getNumLogicBlocks(){return numlogicBlocks;};
  int getNumBlocks(){return numBlocks;}
  int getNumPipeliningRegisters();

  // for each net in this netlist, create vector of sinks, and then populate
  //void create_netSinkVectors();

  // for each net in this netlist, sort the sinks in increasing order of separation from source  
  void sort_netSinksDistance();

  //Retiming stuff
  void calculateRegPushes();
  
  //Timing-driven stuff
  void create_slackVectors();
  void create_linkDelayVectors();
  void initTimingGraph(bool pipeline_aware);
  void calculateArrivalDepartureTimes(bool pipeline_aware);
  double returnCriticalPathDelay(bool pipeline_aware);
  void calculateRequiredTimes(double criticalPath, bool pipeline_aware);
  void calculateSlack(double criticalPath, bool pipeline_aware);
  void sort_netSinksTiming();

  void printCriticalPath(bool pipeline_aware);

  void clearLinkDelayAndSlackVectors();

  //Pipeline stuff
  double overallCost;
  double wireCost;
  double timingCost;
  double maxCut;
  double aveCut;

  double getWireCost(){return wireCost;};
  void setWireCost(double wc){wireCost = wc;};
  double getTimingCost(){return timingCost;};
  void setTimingCost(double tc){timingCost = tc;};

  //double getCost(bool timing_driven);

  int countDelayReg();
  void sort_netPipelineSinks(bool sortByTiming);
  void sort_netPipelineSinks_LockedRegs(bool sortByTiming);

  void reCalculateOccupancy(bool pipeline_aware);

  double calculateFakePipelinedDelay();

  void calculatePlacementCost(bool timing_driven, double criticalPathDelay);

  //double determineDeltaCostAllNets(bool timing_driven, double criticalPathDelay);
  
  //void fillUnZipVector(vector<unZipWidget> *unZipVector);

  void removePlacement();

  void updatePlacementTiming();

  int calculateNumBlocksAndPipeliningRegisters();
  int calculateNumNetsAndPNets();

  //This retrieves a random logic block.
  //It could be of any type (INPUTPAD, OUTPUTPAD, REGISTER_BANK, LUT, RAM, MULTIPLIER)
  logicBlock* getRandomLogicBlock(){
    return (lbVector->at(lrand48() % lbVector->size()));
  };
  Terminal* getRandomPipeliningRegister();

  vector <Terminal *> placementPipelineRegisterVector;
 
  void addPipelinedRegistersToPlacementList(Net *net);
  
  void zeroSlackVectors();

  void printPlacement();

  void removePipelineRegisterFromPlacementPRVector(Terminal *pipeReg);

  void checkNets();

  void printNewNetlist(int randomSeed, char* rrGraphName, char* arguments);

  void printNetWireCost(bool pipeline_aware);

  void calculateAvgMaxMinCrits(double criticalPathDelay);

  bool needClock();

  void checkLogicBlockMappings();

  void calculateMaxInputOutputCriticalities(double criticalPathDelay);

  void printAllSinkCriticalities(double criticalPathDelay, bool pipeline_aware);
};

#endif


