// Copyright (C) 2006 Ken Eguro, Scott Hauck, Akshay Sharma
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

/****************************************************************
Class definition of a generic logic block in a netlist. The 
Independence algo will operate on objects of type 'logicBlock'
'Net' and 'Terminal', rather than on class types that are SPECIFIC 
to the benchmark set. It is the user's responsibility to convert 
a netlist in a particular benchmark set to objects of the generic 
classes 'logicBlock', 'Net' and 'Terminal'.
****************************************************************/    

#ifndef LOGIC_BLOCK_H
#define LOGIC_BLOCK_H

class registerBank;
class Netlist;
class PipeliningNet;
class vprLUT;
#include "clb.h"

class logicBlock{
 public:
  //char name [MAX_STR_LEN];  //name of logic block
  string name;  //name of logic block

  enum e_genericBlockType blockType;
  list<Terminal*> inputList;  //list of input terminals
  list<Terminal*> outputList;  //list of output terminals

  list<Terminal*>::iterator* interm_itr;  //pointer to inputList iterator
  list<Terminal*>::iterator* outterm_itr;  //pointer to outputList iterator

  int numInputs;  //num input terminals
  int numOutputs;  //num output terminals
  int magicNum;

  // If this is a register bank, what architecture feature is it mapped to?
  registerBank *rb;

  //If this is a LUT or a register bank, what CLB does it belong to?
  c_clb *parentCLB;

  //Stuff we need for retiming
  int numRegCanPushForward;
  int numRegCanPushBackward;

  //Stuff we need to keep track of how critical this block is
  double maxInputCriticality;
  double maxOutputCriticality;

  //If this is a IOB, what row and column is it mapped to?
  int row,col;

  //If this is a IOB, LUT, or register bank, what specific subblock is it mapped to?
  int ptc_index;

  logicBlock(){
    blockType = GBT_NO_TYPE;
    //functionType=NO_TYPE;
    //isLogic=isInput=isOutput=isBidirectional=isRegisterBank=false;
    numInputs=numOutputs=0;
    interm_itr=0;outterm_itr=0;
    ptc_index=-1;
    magicNum=15111;
    row = -1;
    col = -1;
    rb = 0;

    numRegCanPushForward = 0;
    numRegCanPushBackward = 0;
    maxInputCriticality = -10000.0;
    maxOutputCriticality = -10000.0;
  }

  ~logicBlock();

  void create_inputListItr();  //create inputList iterator
  void create_outputListItr();  //create outputList iterator
  void remove_inputListItr();  //remove inputList iterator
  void remove_outputListItr();  //remove outputList iterator
  Terminal* getnext_inputList();  //get next element from inputList
  Terminal* getnext_outputList();  //get next element from outputList

  //void set_functionType(int f);  //set the funtionality type of this block
  void set_blockType(enum e_genericBlockType t);
  enum e_genericBlockType get_blockType();
  void set_name(string n);  //give this block a name
  void set_ptc_index(int n);

  void addto_inputList(Terminal* t);  //add a terminal to the input terminal list
  void addto_outputList(Terminal* t);  //add a terminal to the output terminal list

  string get_name();  //get name of this block

  int get_numInputs();  //get num input terminals
  int get_numOutputs();  //get num output terminals
  void print_Info();  //print details about this logic block

  //void initOutputDependency(); //Creates the correct size outputDependency vector
  //void initInputFanOut();  //Creates the correct size inputFandOut vector
  void setOutputDependency(); //Sets all outputs to depend upon all inputs
  void setInputFanOut(); //Sets all input to fan out to all outputs

  void set_registerBank(registerBank *regBank);
  void unset_registerBank(registerBank *regBank);

  //void setOutputDependency(int outputNum, int inputNum);  //Sets output i to depend upon input j
  //void setInputFanOut(int inputNum, int outputNum); //Sets input i to fan out to output j
  void updateSinkDelaysAndSlacks();
  void updateSinkDelaysAndSlacksForAllCLBLUTs();
  void updateSinkDelaysAndSlacksForEntireCLB();
  void updateSinkDelaysAndSlacksBackwards();
  void updateSinkDelaysAndSlacksForwards();
  void updateWireAndTimingCosts(bool timing_driven, double criticalPathDelay);
  void updateWireAndTimingCostsForEntireCLB(bool timing_driven, double criticalPathDelay);

  void updateCostsAndDeltas(bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming);
  void updateCostsAndDeltasCostDeltaDelay(bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming, double *deltaTimingPrime);

  void updateWireAndTimingCostsBackwards(bool timing_driven, double criticalPathDelay);
  void updateWireAndTimingCostsForwards(bool timing_driven, double criticalPathDelay);

  double determineDeltaWireCost();
  double determineDeltaTimingCost();
  double determineDeltaTimingCostPrime();
  double determineDeltaWireCostForEntireCLB();
  double determineDeltaTimingCostForEntireCLB();

  double determineDeltaWireCostBackwards();
  double determineDeltaWireCostForwards();
  double determineDeltaTimingCostBackwards();
  double determineDeltaTimingCostForwards();

  void calculateRegPushBackward();
  void calculateRegPushForward();
  void calculateMaxInputCriticality(double criticalPathDelay);
  void calculateMaxOutputCriticality(double criticalPathDelay);
  void calculateMaxInputOutputCriticality(double criticalPathDelay);
  void calculateMaxInputOutputCriticalitiesForwards(double criticalPathDelay);
  void calculateMaxInputOutputCriticalitiesBackwards(double criticalPathDelay);
  
  bool retimeRegistersBackwards(Netlist *netlist, int numMove, double criticalPathDelay, list <PipeliningNet *> *retimeUndoList);
  bool retimeRegistersForwards(Netlist *netlist, int numMove, double criticalPathDelay, list <PipeliningNet *> *retimeUndoList);
  void undoRetimeRegistersBackwards(Netlist *netlist, int numMove, double criticalPathDelay, list <PipeliningNet *> *retimeUndoList);
  void undoRetimeRegistersForwards(Netlist *netlist, int numMove, double criticalPathDelay, list <PipeliningNet *> *retimeUndoList);
  void abortRetimeRegistersBackwards(Netlist *netlist, int numMove, double criticalPathDelay, list <PipeliningNet *> *retimeUndoList);
  void abortRetimeRegistersForwards(Netlist *netlist, int numMove, double criticalPathDelay, list <PipeliningNet *> *retimeUndoList);
  
  void clearNerfedBlocksBackward();
  void clearNerfedBlocksForward();

  void checkPushForwardsAndBackwards();

  void updateCLBLocation(registerBank *block){
    parentCLB->updateLocation(block);
  }
  void updateCLBLocation(vprLUT *block){
    parentCLB->updateLocation(block);
  }

  void clearCLBLocation(registerBank *block){
    parentCLB->clearLocation(block);
  }

  void clearCLBLocation(vprLUT *block){
    parentCLB->clearLocation(block);
  }

  void placeEntireCLB(registerBank *block);
  void placeEntireCLB(vprLUT *block);

  void placeAllCLBLUTs(vprLUT *block);
  void unPlaceAllCLBLUTs(vprLUT *block);

  logicBlock* getRegisterSource();
  
  enum e_genericBlockType returnDriverType();
  void placeWithParent();
  bool parentPlaced();
  bool isSelfFeeding();
  void printDistanceToAllSinks();
  int minDistanceToAllSinks();
  bool drivesSomethingIn(int row, int col);
};

#endif
