// Copyright (C) 2006 Ken Eguro, Scott Hauck, Akshay Sharma
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

#include "include.h"
extern vprFPGA* FPGA;
extern bool deltaDelayCost;

extern bool lut_level_placement;
extern bool clb_level_placement;
extern double ff_level_placement_activation;
extern double ff_level_placement_crit;
extern double ff_level_placement_probability;
extern double ff_level_placement_homing;
extern bool lutStickiness;
extern bool tvpack_initial_placement;

extern bool placementRetiming;
extern double retime_placement_activation;
extern double retime_placement_crit;
extern double retime_placement_probability;

extern bool splitDelayBlocks;
extern double splitDelayBlocks_activation;
extern double splitDelayBlocks_crit;
extern double splitDelayBlocks_probability;

extern bool unzipNets;
extern double unzipNets_activation;
extern double unzipNets_crit;
extern double unzipNets_probability;

extern double tdPlacementLambda;
extern double tdPlacementUpdateInterval;
extern double tdPlacementCritExponentFinal;
extern double tdPlacementCritExponent;
extern bool noWindowing;

extern bool maxCutCost;
extern bool aveCutCost;
extern bool noWindowing;
extern double placeInnerLoop;
extern double averageManhattanDistance;

extern bool testInitialPlacementOnly;

extern bool noIOAnneal;

extern bool lotsTimingInfo;

#ifdef PARANOID_CHECK_PIPEPLACE
extern int numRegistersEstimatedFreed;
extern int numRegistersEstimatedNeeded;
#endif

int numSplitCalls = 0;
int numSplitSwapCalls = 0;
int numSplits = 0;
int numMergeCalls = 0;
int numMergeSwapCalls = 0;
int numMerges = 0;

int numAttemptedLUTMoves;
int numAttemptedRegMoves;
int numAttemptedIOMoves;
//int averageAttemptedLUTMoveManhattanDistance;
//int averageAttemptedRegMoveManhattanDistance;
//int averageAttemptedIOMoveManhattanDistance;

// Perform the pipeline-aware placement
// We assume that the netlist already has an initial placement for the
//  logic, I/O, and delay blocks
void pipeline_placement(Netlist *netlist, bool timing_driven, bool performInitialPlacement){
  double criticalPathDelay;
  double placementCost;
  //vector <zipWidget> zipVector;
  //vector <unZipWidget> unZipVector;
  double temperature;

  //Do the inital placement
  if(performInitialPlacement){
    init_pipeline_placement(netlist);
  }
  else{
    init_pipeline_placement_OnlyPipeliningRegisters(netlist);
  }
  
  if(testInitialPlacementOnly){
    return;
  }
  
  //printPlacement(netlist);

  //Now, calculate the critical path for the netlist given the initial placement
  netlist->updatePlacementTiming();
  
  criticalPathDelay = calculateCriticalPathDelay(netlist, true);
  
  //Now calculate the cost of the placement
  netlist->calculateMaxInputOutputCriticalities(criticalPathDelay);
  netlist->calculatePlacementCost(timing_driven, criticalPathDelay);

  printf("Initial placement Wire Cost = %f\n", netlist->wireCost);
  printf("Initial placement Timing Cost = %e\n", netlist->timingCost);
  printf("Starting critical path delay = %e\n", criticalPathDelay);

  if(maxCutCost){
    printf("Initial placement Max Cut Cost = %f\n", netlist->maxCut);
  }
  if(aveCutCost){
    printf("Initial placement Ave Cut Cost = %f\n", netlist->aveCut);
  }

  //printf("Pre-initial placement placement\n");
  //netlist->printPlacement();
  //netlist->print_netList(true);

  temperature = pipelineDetermineAnnealingTemp(netlist, timing_driven, &criticalPathDelay);
  printf("Initial Temp = %e\n", temperature);

  //Start the annealing
  pipelineAnneal(netlist, timing_driven, temperature, (float) max(FPGA->get_numRows(), FPGA->get_numCols()));

  //printf("Final placement Wire Cost = %f\n", netlist->wireCost);
  //printf("Final placement Timing Cost = %e\n", netlist->timingCost);
  //printf("Final critical path delay = %e\n", criticalPathDelay);  

  //netlist->printPlacement();
  //netlist->print_netList(true);

  if(maxCutCost){
    printf("Final placement Max Cut Cost = %f\n", netlist->maxCut);
  }
  if(aveCutCost){
    printf("Final placement Ave Cut Cost = %f\n", netlist->aveCut);
  }
}


//Perform a few iterations of low-temp annealing on the existing placement
void pipelineLowTempAnneal(Netlist *netlist, bool timing_driven){
 double criticalPathDelay;
  double placementCost;
  double temperature;

  if(ff_level_placement_activation >= 0.0 && clb_level_placement){
    cout << "FF-level placement activated!" << endl;
    clb_level_placement = false;
  }
  if(retime_placement_activation >= 0.0 && !placementRetiming){
    cout << "Retiming activated!" << endl;
    placementRetiming = true;
  }
  if(splitDelayBlocks_activation >= 0.0 && !splitDelayBlocks){
    cout << "Splitting of register banks activated!" << endl;
    splitDelayBlocks = true;
  }

  if(unzipNets_activation >= 0.0 && !unzipNets){
    cout << "Register zipping/unzipping activated!" << endl;
    splitDelayBlocks = true;
    unzipNets = true;
  }

  //Calculate the critical path and cost for the netlist's existing placement
  criticalPathDelay = calculateCriticalPathDelay(netlist, true);
  netlist->calculateMaxInputOutputCriticalities(criticalPathDelay);
  netlist->calculatePlacementCost(timing_driven, criticalPathDelay);

  temperature = 1.0E-5;
  printf("Low temp anneal starts at = %e\n", temperature);

  //Start the annealing
  pipelineAnneal(netlist, timing_driven, temperature, 1.0);
}

// Perform pipeline-aware initialization of the placement
// 1) Place all of the I/O and LUTs.
// 2) Place all register banks and delay registers based on the assumption that all nets share
//    as many pipelining registers as possible. (Fully zipped)
void init_pipeline_placement(Netlist *netlist){
  vprLUT* lut;
  vprIOB* io_block;
  registerBank *rb;
  Net* net;
  Terminal* outTerm;
  logicBlock* lb;
  Terminal* pr;
  list <logicBlock*> registerBankList;
  list <Terminal*> pipelineRegisterList;
  list <Terminal*> shiftRegisterList;
  list <int> shiftRegisterDepthList;
  int counter;
  bool done;
  int shiftRegisterLength;
  int numLUTs;

  // Go through netlist and sort all net sinks by non-decreasing latency
  netlist->sort_netPipelineSinks(true);

  //First, place all of the IOBs
  netlist->create_lbListItr();
  for(lb=netlist->getnext_lbList();lb!=0;lb=netlist->getnext_lbList()){
    if(lb->get_blockType() == INPUTPAD || lb->get_blockType() == OUTPUTPAD){
      //Check to see if it already has been placed
      if(lb->row != -1 || lb->col != -1){
	ERROR("IO block should not have a placement yet!");
      }
      done = false;
      while(!done){
	io_block = FPGA->iobList.at(lrand48() % FPGA->iobList.size());
	
	//See if this io block has any empty locations
	for(int i = 0; i < io_block->get_numIOs(); i++){
	  if(!io_block->getfrom_currBlockArray(i)){
	    // Place the io pin in the empty location
	    placeBlock(lb,(void *)(io_block), IOPIN, i);
	    done = true;
	    break;
	  }
	}
      }
    }
  }
  netlist->remove_lbListItr();  

  //Next, let's place all of the array CLBs.
  //Let's see if we should be respecting the original packing for the netlist
  if(tvpack_initial_placement){
    //If so, we don't have to worry about the LUT packing, we will just
    // get it from the netlist itself.
    netlist->create_lbListItr();
    for(lb=netlist->getnext_lbList();lb!=0;lb=netlist->getnext_lbList()){
      if(lb->get_blockType() == LUT || lb->get_blockType() == REGISTER_BANK){
	if(lb->row != -1 || lb->col != -1){
	  //We can move onto the next logic block because this one has been placed already
	  // because something else in the same CLB has been placed
	  continue;
	}

	done = false;
	while(!done){
	  if(lb->get_blockType() == REGISTER_BANK){
	    //We have to place all of the items in the same logical CLB into the same physical CLB.
	    //Thus, the entire CLB has to be empty.
	    rb = FPGA->getRandomFreeSingleRegBank();
	    if(rb->emptyCLB()){
	      lb->placeEntireCLB(rb);
	      done = true;
	    }
	  }
	  else{
	    if(lb->get_blockType() == LUT){
	      lut = FPGA->getRandomLUT();
	    }
	    else if(lb->get_blockType() == RAM_BLOCK){
	      lut = FPGA->getRandomRam();
	    }
	    else if(lb->get_blockType() == MULTIPLIER){
	      lut = FPGA->getRandomMult();
	    }
	    else{
	      ERROR("Unknown logic block type!");
	    }
      
	    //See if this CLB is already occupied
	    if(lut->emptyCLB()){
	      //If we are doing individual block level placement, just place all of the LUTs into the same physical CLB
	      lb->placeEntireCLB(lut);
	      done = true;
	    }
	  }
	}
      }
    }
    netlist->remove_lbListItr();  
  }
  else{
    //If we are not following the original packing for the netlist
    // we have to pack the luts by order of reverse number of luts in the clb
    // (all clbs that have 4 luts, 3 luts, etc.)
    //If so, we don't have to worry about the LUT packing, we will just
    // get it from the netlist itself.
    for(numLUTs = 4; numLUTs >= 0; numLUTs--){
      netlist->create_lbListItr();
      for(lb=netlist->getnext_lbList();lb!=0;lb=netlist->getnext_lbList()){
	//Only place the netlist CLBs that have the exact number of luts we are looking for
	if((lb->get_blockType() == LUT || lb->get_blockType() == REGISTER_BANK) &&
	   lb->parentCLB->getNumLUTs() == numLUTs){
	  if(lb->row != -1 || lb->col != -1){
	    //We can move onto the next logic block because this one has been placed already
	    // because something else in the same CLB has been placed
	    continue;
	  }
	  
	  //Make sure that there are enough CLBs available that can
	  // accomodate the required number of LUTs & registers
	  if(FPGA->getNumberCLBsEmptyLUTandRegs(lb->parentCLB->getNumLUTs(), lb->parentCLB->getNumRegisters()) == 0){
	    cout << "Ran out of LUTs or Registers, requesting " << lb->parentCLB->getNumLUTs();
	    cout << " luts and " << lb->parentCLB->getNumRegisters() << " registers" <<endl;
	    ERROR("Ran out of LUTs or Registers");
	  }

	  done = false;
	  while(!done){
	    if(lb->get_blockType() == REGISTER_BANK){
	      rb = FPGA->getRandomFreeSingleRegBank();
	      //Make sure that this CLB will fit in the given CLB
	      if(rb->getHostCLB()->getNumEmptyLUTs() >= lb->parentCLB->getNumLUTs() &&
		 rb->getHostCLB()->getNumEmptyRegBank(1) >= lb->parentCLB->getNumRegisters()){
		lb->placeEntireCLB(rb);
		done = true;
	      }
	    }
	    else{
	      if(lb->get_blockType() == LUT){
		lut = FPGA->getRandomLUT();
	      }
	      else if(lb->get_blockType() == RAM_BLOCK){
		lut = FPGA->getRandomRam();
	      }
	      else if(lb->get_blockType() == MULTIPLIER){
		lut = FPGA->getRandomMult();
	      }
	      else{
		ERROR("Unknown logic block type!");
	      }
	      
	      //See if this CLB is already occupied
	      //Make sure that this CLB will fit in the given CLB
	      if(lut->getHostCLB()->getNumEmptyLUTs() >= lb->parentCLB->getNumLUTs() &&
		 lut->getHostCLB()->getNumEmptyRegBank(1) >= lb->parentCLB->getNumRegisters()){
		lb->placeEntireCLB(lut);
		done = true;
	      }
	    }
	  }
	}
      }
      netlist->remove_lbListItr();  
    }
  }

  // Now we have to figure out what to do with any pipelining registers that might be on the output of input blocks or 
  //   on logic block outputs
  // We assume that all signals are fully "zipped" so we have maximum sharing of any pipelining registers.
  netlist->create_lbListItr();
  for(lb=netlist->getnext_lbList();lb!=0;lb=netlist->getnext_lbList()){
    // This test shouldn't be necessary since output pins shouldn't have any output terminals
    if(lb->get_blockType() != OUTPUTPAD){
      lb->create_outputListItr();
      for(outTerm = lb->getnext_outputList(); outTerm != 0; outTerm = lb->getnext_outputList()){
	// See if the net associated with this output terminal is pipelined
	net = outTerm->get_net();
	if(net->pipelined){
	  // For the moment, we'll assume that all nets are fully zipped.
	  // Create pipelined nets that are fully zipped, but don't worry about how many
	  // delays we can pick up at a given delay block.  We'll worry about packing them smartly
	  // when we actually place them.
	  net->createZippedPipelinedNets();
	  
	  net->addShiftRegistersAndPipelineRegisters(&shiftRegisterList, &shiftRegisterDepthList, &pipelineRegisterList);
	  //net->addPipelineRegistersToList(&pipelineRegisterList);
	  netlist->addPipelinedRegistersToPlacementList(net);
	}
      }
      lb->remove_outputListItr();
    }
  }
  netlist->remove_lbListItr();

  //Any pipelining registers in the system have not been placed yet
  //We'll do this in stages to maximize the likelihood that these are packed with their parent.
  //First, go through the respective lists once and place any shift registers, register banks or pipelining 
  // registers that are directly driven by a LUT.
  counter = registerBankList.size();
  //  cout << "Netlist needs at least " << counter << " explicit register banks" << endl;
  for(int i = 0; i < counter; i++){
    //Grab the first item in the list
    lb = registerBankList.front();
    registerBankList.pop_front();

    if(lb->returnDriverType() == LUT){
      //If it's parent is a LUT, place this logic block      
      lb->placeWithParent();
    }
    else{
      //If not, add it to the back of the list;
      registerBankList.push_back(lb);
    }
  }
  
  counter = pipelineRegisterList.size();
  cout << "Netlist needs at least " << counter << " single register banks" << endl;
  for(int i = 0; i < counter; i++){
    //Grab the first item in the list
    pr = pipelineRegisterList.front();
    pipelineRegisterList.pop_front();
    
    if(pr->returnDriverType() == LUT){
      //If it's parent is a LUT, place this pipelining register
      pr->placeWithParent();
    }
    else{
      //If not, add it to the back of the list;
	pipelineRegisterList.push_back(pr);
    }
  }

  counter = shiftRegisterList.size();
  cout << "Netlist needs at least " << counter << " shift register banks" << endl;
  for(int i = 0; i < counter; i++){
    //Grab the first item in the list
    pr = shiftRegisterList.front();
    shiftRegisterList.pop_front();
    shiftRegisterLength = shiftRegisterDepthList.front();
    shiftRegisterDepthList.pop_front();
    
    if(pr->returnDriverType() == LUT){
      //If it's parent is a LUT, place this shift register
      pr->placeWithParent(shiftRegisterLength);
    }
    else{
      //If not, add it to the back of the list;
      shiftRegisterList.push_back(pr);
      shiftRegisterDepthList.push_back(shiftRegisterLength);
    }
  }   

  //Next, place any register banks or pipelining registers that are attached to other
  // register banks or pipelining registers that have placements.
  //We will go through these lists until we no longer place any new items.
  done = false;
  while(!done){
    done = true;
    
    counter = registerBankList.size();
    for(int i = 0; i < counter; i++){
      //Grab the first item in the list
      lb = registerBankList.front();
      registerBankList.pop_front();
      
      if((lb->returnDriverType() == GBT_NO_TYPE || lb->returnDriverType() == REGISTER_BANK) && lb->parentPlaced()){
	//If it's parent is a register of some type and it's already been placed,
	// place this logic block.
	lb->placeWithParent();
      }
      else{
	//If not, add it to the back of the list;
	registerBankList.push_back(lb);
      }
    }
      
    counter = pipelineRegisterList.size();
    for(int i = 0; i < counter; i++){
      //Grab the first item in the list
      pr = pipelineRegisterList.front();
      pipelineRegisterList.pop_front();
	
      if((pr->returnDriverType() == GBT_NO_TYPE || pr->returnDriverType() == REGISTER_BANK) && pr->parentPlaced()){
	//If it's parent is a register of some type and it's already been placed,
	// place this pipelining register.
	pr->placeWithParent();
      }
      else{
	//If not, add it to the back of the list;
	pipelineRegisterList.push_back(pr);
      }
    }

    counter = shiftRegisterList.size();
    for(int i = 0; i < counter; i++){
      //Grab the first item in the list
      pr = shiftRegisterList.front();
      shiftRegisterList.pop_front();
      shiftRegisterLength = shiftRegisterDepthList.front();
      shiftRegisterDepthList.pop_front();
	
      if((pr->returnDriverType() == GBT_NO_TYPE || pr->returnDriverType() == REGISTER_BANK) && pr->parentPlaced()){
	//If it's parent is a register of some type and it's already been placed,
	// place this pipelining register.
	pr->placeWithParent(shiftRegisterLength);
      }
      else{
	//If not, add it to the back of the list;
	shiftRegisterList.push_back(pr);
	shiftRegisterDepthList.push_back(shiftRegisterLength);
      }
    }
  }
    
  //At this point all register banks and pipelining registers driven upstream at any point by any sort of LUT
  // should be placed.
  //We have NOT placed any registers/pipelining register that are attached to themselves (constant generators),
  // input pads, RAMs or multipliers.
  //Place constant generators.
  counter = registerBankList.size();
  for(int i = 0; i < counter; i++){
    //Grab the first item in the list
    lb = registerBankList.front();
    registerBankList.pop_front();
      
    if(lb->isSelfFeeding()){
      //If this is a constant generator, place this logic block someplace
      if(tvpack_initial_placement){
	//Notice, we are trying to replicate what T-VPack does, sort of.
	//Since this register bank is self feeding, we should put it in
	// the "closest" CLB that has an empty LUT.
	//~~~~~~~~~~~~~~~~~~~~~~~~~THIS IS A BIG ASSUMPTION! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	//However, since right now we are targeting unclustered architectures, "closeness" doesn't matter and
	// we can just put it into any completely free CLB.
	//On clustered architectures we could enter an infinite loop with this implementation...
	done = false;
	while(!done){
	  rb = FPGA->getRandomFreeSingleRegBank();
	  if(rb->emptyCLB()){
	    rb->placeLogicBlock(lb);
	    done = true;
	  }
	}
      }
      else{
	rb = FPGA->getRandomFreeSingleRegBank();
	rb->placeLogicBlock(lb);
      }
    }
    else{
      //If not, add it to the back of the list;
      registerBankList.push_back(lb);
    }
  }
  
    
  //Now, go through the lists and place any register bank or pipelining register whose parent has been placed
  while(pipelineRegisterList.size() != 0 || registerBankList.size() != 0 || shiftRegisterList.size() != 0){
    counter = registerBankList.size();
    for(int i = 0; i < counter; i++){
      //Grab the first item in the list
      lb = registerBankList.front();
      registerBankList.pop_front();
	
      if(lb->parentPlaced()){
	//If it's parent has been placed, place this logic block.
	lb->placeWithParent();
      }
      else{
	//If not, add it to the back of the list;
	registerBankList.push_back(lb);
      }
    }
      
    counter = pipelineRegisterList.size();
    for(int i = 0; i < counter; i++){
      //Grab the first item in the list
      pr = pipelineRegisterList.front();
      pipelineRegisterList.pop_front();
	
      if(pr->parentPlaced()){
	//If it's parent has been placed, place this pipelining register.
	pr->placeWithParent();
      }
      else{
	//If not, add it to the back of the list;
	pipelineRegisterList.push_back(pr);
      }
    }

    counter = shiftRegisterList.size();
    for(int i = 0; i < counter; i++){
      //Grab the first item in the list
      pr = shiftRegisterList.front();
      shiftRegisterList.pop_front();
      shiftRegisterLength = shiftRegisterDepthList.front();
      shiftRegisterDepthList.pop_front();
	
      if(pr->parentPlaced()){
	//If it's parent is a register of some type and it's already been placed,
	// place this pipelining register.
	pr->placeWithParent(shiftRegisterLength);
      }
      else{
	//If not, add it to the back of the list;
	shiftRegisterList.push_back(pr);
	shiftRegisterDepthList.push_back(shiftRegisterLength);
      }
    }
  }
}

// Perform half of the pipeline-aware initialization of the placement
// 1) All I/O blocks, LUTs and explicit registers should have been placed already.
// 2) Place all delay registers based on the assumption that all nets share
//    as many pipelining registers as possible. (Fully zipped)
void init_pipeline_placement_OnlyPipeliningRegisters(Netlist *netlist){
  vprLUT* lut;
  vprIOB* io_block;
  registerBank *rb;
  Net* net;
  Terminal* outTerm;
  logicBlock* lb;
  Terminal* pr;
  list <logicBlock*> registerBankList;
  list <Terminal*> pipelineRegisterList;
  list <Terminal*> shiftRegisterList;
  list <int> shiftRegisterDepthList;
  int counter;
  bool done;
  int shiftRegisterLength;
  int numLUTs;

  // Go through netlist and sort all net sinks by non-decreasing latency
  netlist->sort_netPipelineSinks(true);

  //First, place all of the IOBs
  netlist->create_lbListItr();
  for(lb=netlist->getnext_lbList();lb!=0;lb=netlist->getnext_lbList()){
    if(lb->get_blockType() == INPUTPAD || lb->get_blockType() == OUTPUTPAD){
      //Check to see if it already has been placed
      if(lb->row == -1 || lb->col == -1){
	ERROR("IO block should have a placement!");
      }
    }
    else if(lb->get_blockType() == LUT || lb->get_blockType() == REGISTER_BANK){
      //Check to see if it already has been placed
      if(lb->row == -1 || lb->col == -1){
	ERROR("LUTs and explict registers should have a placement!");
      }
    }
  }
  netlist->remove_lbListItr();  

  // Now we have to figure out what to do with any pipelining registers that might be on the output of input blocks or 
  //   on logic block outputs
  // We assume that all signals are fully "zipped" so we have maximum sharing of any pipelining registers.
  netlist->create_lbListItr();
  for(lb=netlist->getnext_lbList();lb!=0;lb=netlist->getnext_lbList()){
    // This test shouldn't be necessary since output pins shouldn't have any output terminals
    if(lb->get_blockType() != OUTPUTPAD){
      lb->create_outputListItr();
      for(outTerm = lb->getnext_outputList(); outTerm != 0; outTerm = lb->getnext_outputList()){
	// See if the net associated with this output terminal is pipelined
	net = outTerm->get_net();
	if(net->pipelined){
	  // For the moment, we'll assume that all nets are fully zipped.
	  // Create pipelined nets that are fully zipped, but don't worry about how many
	  // delays we can pick up at a given delay block.  We'll worry about packing them smartly
	  // when we actually place them.
	  net->createZippedPipelinedNets();
	  
	  net->addShiftRegistersAndPipelineRegisters(&shiftRegisterList, &shiftRegisterDepthList, &pipelineRegisterList);
	  //net->addPipelineRegistersToList(&pipelineRegisterList);
	  netlist->addPipelinedRegistersToPlacementList(net);
	}
      }
      lb->remove_outputListItr();
    }
  }
  netlist->remove_lbListItr();

  //Any pipelining registers in the system have not been placed yet
  //We'll do this in stages to maximize the likelihood that these are packed with their parent.
  //First, go through the respective lists once and place any shift registers, register banks or pipelining 
  // registers that are directly driven by a LUT.
  counter = registerBankList.size();
  //  cout << "Netlist needs at least " << counter << " explicit register banks" << endl;
  for(int i = 0; i < counter; i++){
    //Grab the first item in the list
    lb = registerBankList.front();
    registerBankList.pop_front();

    if(lb->returnDriverType() == LUT){
      //If it's parent is a LUT, place this logic block      
      lb->placeWithParent();
    }
    else{
      //If not, add it to the back of the list;
      registerBankList.push_back(lb);
    }
  }
  
  counter = pipelineRegisterList.size();
  cout << "Netlist needs at least " << counter << " single register banks" << endl;
  for(int i = 0; i < counter; i++){
    //Grab the first item in the list
    pr = pipelineRegisterList.front();
    pipelineRegisterList.pop_front();
    
    if(pr->returnDriverType() == LUT){
      //If it's parent is a LUT, place this pipelining register
      pr->placeWithParent();
    }
    else{
      //If not, add it to the back of the list;
	pipelineRegisterList.push_back(pr);
    }
  }

  counter = shiftRegisterList.size();
  cout << "Netlist needs at least " << counter << " shift register banks" << endl;
  for(int i = 0; i < counter; i++){
    //Grab the first item in the list
    pr = shiftRegisterList.front();
    shiftRegisterList.pop_front();
    shiftRegisterLength = shiftRegisterDepthList.front();
    shiftRegisterDepthList.pop_front();
    
    if(pr->returnDriverType() == LUT){
      //If it's parent is a LUT, place this shift register
      pr->placeWithParent(shiftRegisterLength);
    }
    else{
      //If not, add it to the back of the list;
      shiftRegisterList.push_back(pr);
      shiftRegisterDepthList.push_back(shiftRegisterLength);
    }
  }   

  //Next, place any register banks or pipelining registers that are attached to other
  // register banks or pipelining registers that have placements.
  //We will go through these lists until we no longer place any new items.
  done = false;
  while(!done){
    done = true;
    
    counter = registerBankList.size();
    for(int i = 0; i < counter; i++){
      //Grab the first item in the list
      lb = registerBankList.front();
      registerBankList.pop_front();
      
      if((lb->returnDriverType() == GBT_NO_TYPE || lb->returnDriverType() == REGISTER_BANK) && lb->parentPlaced()){
	//If it's parent is a register of some type and it's already been placed,
	// place this logic block.
	lb->placeWithParent();
      }
      else{
	//If not, add it to the back of the list;
	registerBankList.push_back(lb);
      }
    }
      
    counter = pipelineRegisterList.size();
    for(int i = 0; i < counter; i++){
      //Grab the first item in the list
      pr = pipelineRegisterList.front();
      pipelineRegisterList.pop_front();
	
      if((pr->returnDriverType() == GBT_NO_TYPE || pr->returnDriverType() == REGISTER_BANK) && pr->parentPlaced()){
	//If it's parent is a register of some type and it's already been placed,
	// place this pipelining register.
	pr->placeWithParent();
      }
      else{
	//If not, add it to the back of the list;
	pipelineRegisterList.push_back(pr);
      }
    }

    counter = shiftRegisterList.size();
    for(int i = 0; i < counter; i++){
      //Grab the first item in the list
      pr = shiftRegisterList.front();
      shiftRegisterList.pop_front();
      shiftRegisterLength = shiftRegisterDepthList.front();
      shiftRegisterDepthList.pop_front();
	
      if((pr->returnDriverType() == GBT_NO_TYPE || pr->returnDriverType() == REGISTER_BANK) && pr->parentPlaced()){
	//If it's parent is a register of some type and it's already been placed,
	// place this pipelining register.
	pr->placeWithParent(shiftRegisterLength);
      }
      else{
	//If not, add it to the back of the list;
	shiftRegisterList.push_back(pr);
	shiftRegisterDepthList.push_back(shiftRegisterLength);
      }
    }
  }
    
  //At this point all register banks and pipelining registers driven upstream at any point by any sort of LUT
  // should be placed.
  //We have NOT placed any registers/pipelining register that are attached to themselves (constant generators),
  // input pads, RAMs or multipliers.
  //Place constant generators.
  counter = registerBankList.size();
  for(int i = 0; i < counter; i++){
    //Grab the first item in the list
    lb = registerBankList.front();
    registerBankList.pop_front();
      
    if(lb->isSelfFeeding()){
      //If this is a constant generator, place this logic block someplace
      if(tvpack_initial_placement){
	//Notice, we are trying to replicate what T-VPack does, sort of.
	//Since this register bank is self feeding, we should put it in
	// the "closest" CLB that has an empty LUT.
	//~~~~~~~~~~~~~~~~~~~~~~~~~THIS IS A BIG ASSUMPTION! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	//However, since right now we are targeting unclustered architectures, "closeness" doesn't matter and
	// we can just put it into any completely free CLB.
	//On clustered architectures we could enter an infinite loop with this implementation...
	done = false;
	while(!done){
	  rb = FPGA->getRandomFreeSingleRegBank();
	  if(rb->emptyCLB()){
	    rb->placeLogicBlock(lb);
	    done = true;
	  }
	}
      }
      else{
	rb = FPGA->getRandomFreeSingleRegBank();
	rb->placeLogicBlock(lb);
      }
    }
    else{
      //If not, add it to the back of the list;
      registerBankList.push_back(lb);
    }
  }
  
    
  //Now, go through the lists and place any register bank or pipelining register whose parent has been placed
  while(pipelineRegisterList.size() != 0 || registerBankList.size() != 0 || shiftRegisterList.size() != 0){
    counter = registerBankList.size();
    for(int i = 0; i < counter; i++){
      //Grab the first item in the list
      lb = registerBankList.front();
      registerBankList.pop_front();
	
      if(lb->parentPlaced()){
	//If it's parent has been placed, place this logic block.
	lb->placeWithParent();
      }
      else{
	//If not, add it to the back of the list;
	registerBankList.push_back(lb);
      }
    }
      
    counter = pipelineRegisterList.size();
    for(int i = 0; i < counter; i++){
      //Grab the first item in the list
      pr = pipelineRegisterList.front();
      pipelineRegisterList.pop_front();
	
      if(pr->parentPlaced()){
	//If it's parent has been placed, place this pipelining register.
	pr->placeWithParent();
      }
      else{
	//If not, add it to the back of the list;
	pipelineRegisterList.push_back(pr);
      }
    }

    counter = shiftRegisterList.size();
    for(int i = 0; i < counter; i++){
      //Grab the first item in the list
      pr = shiftRegisterList.front();
      shiftRegisterList.pop_front();
      shiftRegisterLength = shiftRegisterDepthList.front();
      shiftRegisterDepthList.pop_front();
	
      if(pr->parentPlaced()){
	//If it's parent is a register of some type and it's already been placed,
	// place this pipelining register.
	pr->placeWithParent(shiftRegisterLength);
      }
      else{
	//If not, add it to the back of the list;
	shiftRegisterList.push_back(pr);
	shiftRegisterDepthList.push_back(shiftRegisterLength);
      }
    }
  }
}

// We have just created a single long string of pipelining registers for this net.
// To place them, we just need to find a random empty delay block bank, then
// put as many registers as we can into it.  We might be limited by a few things:
//   1) Of course, the capacity of the bank
//   2) Equally obvious, the number of registers needed by this net
//   3) We need have a sink that needs the signal at this latency
// We can just keep mapping successive terminals into the bank until we reach one of these
// limits.  If we hit condition 1 or 3, we should stop and find another random empty delay
// block bank.  This assumes that each bank of shift registers only has one output.
// That is, for example, a RaPiD bus connector cannot provide an output at 2 delays for one signal
// and an output at 3 delays for another.
// If we want to model that, we have to figure out something else entirely.
// void initialPlacePipelineRegs(Net *net){
//   registerBank *rb = NULL;
  
//   // This code makes the assumption that the list of pipelining registers contains
//   // only one continuous string of registers (no unzipped branches) and is in order of
//   // increasing latency.  Since we just created them in createZippedPipelinedNets,
//   // this is safe.
//   for(list<Terminal*>::iterator regTermIter = net->pipelineRegisterList.begin(); 
//       regTermIter !=  net->pipelineRegisterList.end(); regTermIter++){
//     if(!rb){
//       // If we don't have a register bank we are working on, grab an empty one
//       rb = FPGA->getRandomFreeRegBank();
//     }

//     //Put the current register in the bank
//     rb->placeTerminal(*regTermIter);

//     //We should now move onto another register bank if:
//     // 1) The bank is full
//     if(rb->bankFull()){
//       rb = NULL;
//     }
//     // Or 2) If the pipeline net sourced by this register has fanout
//     if((*regTermIter)->destPipeliningNet->sinkVector.size() > 1){
//       rb = NULL;
//     }
//   }
// }

void cleanUpPipelinePlacement(Netlist *netlist){
  // Go through all of the nets and remove the pipelining information 
  // we generated during placement before we do routing
  netlist->create_netListItr();
  for(Net* net=netlist->getnext_netList();net!=0;net=netlist->getnext_netList()){
    net->cleanUpPipelinePlacementInformation();
  }
  netlist->remove_netListItr();
}

// Try some random moves to see what affect this has on the cost
// We have to return not only the temperature we should use, but also
// update the final criticalPathDelay for the system we end up with.
double pipelineDetermineAnnealingTemp(Netlist *netlist, bool timing_driven, double *criticalPathDelay){
  void* block0;
  void* block1;
  int indexOrBank0, indexOrBank1;
  int moveType;
  enum e_timingMoveBlockType blockType;
  int numMove;
  double dCost;
  vector <double> costArray;
  double average = 0.0;
  double stdDev = 0.0;
  float rLim = (float) max(FPGA->get_numRows(), FPGA->get_numCols());
  list <PipeliningNet *> retimeUndoList;
  int moves = netlist->calculateNumBlocksAndPipeliningRegisters() * 100;

  averageManhattanDistance = 0.0;

#ifdef BREAKONANNEALNUMBER
  int numberMoves = 0;
#endif

#ifdef PARANOID_CHECK_PIPEPLACE
  double oldWireCost;
  double oldTimingCost;
  double oldMaxCut;
  double oldAveCut;
#endif

//   if(ff_level_placement_activation == 1.0){
//     cout << "FF-level placement activated!" << endl;
//     clb_level_placement = false;
//   }
//   if(retime_placement_activation == 1.0){
//     cout << "Retiming activated!" << endl;
//     placementRetiming = true;
//   }
//   if(unzipNets_activation == 1.0){
//     cout << "Register zipping/unzipping activated!" << endl;
//     splitDelayBlocks = true;
//     unzipNets = true;
//   }

  int numMovesBeforeUpdate = (int)(tdPlacementUpdateInterval * (double) moves);
  if(numMovesBeforeUpdate <= 0){
    numMovesBeforeUpdate = 1;
  }

#ifdef PARANOID_CHECK_PIPEPLACE
  numRegistersEstimatedFreed = 0;
  numRegistersEstimatedNeeded = 0;
#endif

  if(timing_driven){
    netlist->overallCost = 1.0;
  }
  else{
    netlist->overallCost = netlist->wireCost;
  }

  //Move each block an average of 100x
  for(int i = 0; i < moves; i++){
#if defined(PIPEPLACEDEBUGHEAVY)
    cout << "Wire Cost before move = " <<  netlist->wireCost <<endl;
    cout << "Timing Cost before move = " <<  netlist->timingCost <<endl;

    if(maxCutCost){
      cout << "Max Cut before move = " <<  netlist->maxCut <<endl;
    }
    if(aveCutCost){
      cout << "Ave Cut before move = " <<  netlist->aveCut <<endl;
    }
    //netlist->print_netList();
    //netlist->printPlacement();
#endif

#ifdef BREAKONANNEALNUMBER
    numberMoves ++;
    if(numberMoves == 68){
      //cout << "Before move!" << endl;
      //netlist->print_netList(true);
      //netlist->printPlacement();
    }
#endif
    
    //Generate a random move
    moveType = generatePipelineAwareMove(netlist, &block0, &block1, &indexOrBank0, &indexOrBank1, &blockType, &numMove, (int)(rLim + 0.5));

    //Make the move and update all the proper sink delays.
    //We are also making incremental changes to the slacks - they are inexact, but we'll see what
    // we can do with them
    //We don't care if the move failed, since we don't undo anything here
    makePipelineAwareMove(&moveType, netlist, &block0, &block1, indexOrBank0, indexOrBank1, 
			  blockType, numMove, *criticalPathDelay, &retimeUndoList);

#if defined (PARANOID_CHECK_PIPEPLACE)
    //Make sure that the mappings of the netlist logic block and the array block match
    netlist->checkLogicBlockMappings();

    //Check to make sure that the mappings of the array blocks match the logic blocks
    FPGA->checkArrayMappings();

#endif

    //Figure out the cost of the move
    if(deltaDelayCost){
      dCost = determineDeltaCostDeltaDelay(netlist, moveType, &block0, &block1, indexOrBank0, indexOrBank1, blockType, numMove, 
					   timing_driven, *criticalPathDelay, &retimeUndoList);
    }
    else{
      dCost = determineDeltaCost(netlist, moveType, &block0, &block1, indexOrBank0, indexOrBank1, blockType, numMove, 
				 timing_driven, *criticalPathDelay, &retimeUndoList);
    }
    netlist->overallCost += dCost;
    //netlist->printPlacement();

    //If we just zipped a net, we should go ahead and delete the registers in block1 now
    // we kept them around because of incremental cost calculation and just in case
    // we had to undo the move, but now we can get rid of them
    if(moveType == MOVEZIPPNET){
      ((registerBank*)block1)->freeBank();
    }

    else if(moveType == MOVEREGRETIMEBACK || moveType == MOVEREGRETIMEFORWARDABORT){
      clearRetimeUndoListForward(&retimeUndoList);
      clearNerfedBlocksBackward(block0, blockType);
    }

    else if(moveType == MOVEREGRETIMEFORWARD || moveType == MOVEREGRETIMEBACKABORT){
      clearRetimeUndoListBackward(&retimeUndoList, block0, blockType);
      clearNerfedBlocksForward(block0, blockType);
    }

#ifdef BREAKONANNEALNUMBER
    if(numberMoves == 68){
      cout << "After move!" << endl;
      netlist->print_netList(true);
      netlist->printPlacement();
    }
#endif

#if defined(PARANOID_CHECK_PIPEPLACE)
    oldWireCost = netlist->wireCost;
    oldTimingCost = netlist->timingCost;
    oldMaxCut = netlist->maxCut;
    oldAveCut = netlist->aveCut;
    
    netlist->calculatePlacementCost(timing_driven, *criticalPathDelay);
    if(fabs(oldWireCost - netlist->wireCost) / netlist->wireCost > 0.001){
      ERROR("Problem doing incremental cost update");
    }
    if(fabs(oldTimingCost - netlist->timingCost) / netlist->timingCost > 0.001 ){
      ERROR("Problem doing incremental cost update");
    }
    if(oldMaxCut != netlist->maxCut){
      ERROR("Problem doing incremental cost update");
    }
    if(oldAveCut != netlist->aveCut){
      ERROR("Problem doing incremental cost update");
    }

    //FPGA->checkFreeRegisters();
    netlist->checkNets();

    if(numRegistersEstimatedFreed != 0){
      //ERROR("Did not estimate # of freed registers correctly");
    }
    if(numRegistersEstimatedNeeded != 0){
      //ERROR("Did not estimate # of needed registers correctly");
    }
#endif

    //To be completely accurate, we should now recalculate the critical path delay and slacks
    // for the whole system.
    //However, this is too timing consuming to do after each move, so we don't do it all of the time
    // we only do it after every N moves
    if(i % numMovesBeforeUpdate == (numMovesBeforeUpdate - 1)){
      *criticalPathDelay =  calculateCriticalPathDelay(netlist, true);

      //Now, the costs for all of the nets might have changed, so calculate the placement cost from scratch      
      netlist->calculateMaxInputOutputCriticalities(*criticalPathDelay);
      netlist->calculatePlacementCost(timing_driven, *criticalPathDelay);
    }

    //Keep track of the cost
    //costArray.push_back(netlist->overallCost);
    average += netlist->overallCost;
    stdDev += (netlist->overallCost * netlist->overallCost);

    //netlist->printPlacement();
  }

  //Calculate the average delta cost
  average /= (double)moves;
  stdDev = (stdDev - ((double)moves * average * average)) / (double)(moves);
  stdDev = pow(stdDev, 0.5);
  
  //Initial temperature is 20 standard deviations of the expected initial delta cost
  return (stdDev * 20.0);
}


//Figure out what kind of move to make. First trying to select between 
// attempting to move a logic block, (should put an I/O pin here
// but RaPiD does not allow pins to move so it's absent for the momemt), 
// grouped delay block or free-moving register.  Once we make the decision
// what kind of block we want to move, we will figure out what kind of move
// to make.
int generatePipelineAwareMove(Netlist *netlist, void** block0, void** block1, int *index0, int *index1, 
			      enum e_timingMoveBlockType *blockType, int *numMove, int rLim){
  int moveType;

  //Should I move an explicitly defined logic block or an inplicitly defined pipelining register?
  moveType = selectPipelineAwareMove(netlist);

  if(moveType == MOVELOGICBLOCK){
    generatePipelineAwareMoveLogic(&moveType, netlist, block0, block1, index0, index1, blockType, numMove, rLim);
  }
  else if(moveType == MOVEPIPEREGSOMETHING){
    generatePipelineAwareMovePipeReg(&moveType, netlist, block0, block1, numMove, rLim);
  }
  else{
    ERROR("Unknown Move Type!");
  }

  return (moveType);
}

int selectPipelineAwareMove(Netlist *netlist){
  long choice;

  //Figure out if we should do something to a logic block, a delay block, or a pipeline register
  // We are choosing a random number between 0 and the number of (ALU, MULT, RAM, REG) + number of delay blocks 
  //      + number of pipelining registers used in pipelined nets
  // numBlocks includes logic blocks (luts, register banks, etc) and I/O pins.
  // numPipeliningRegisters only includes the pipelining registers created based on latency numbers.
  // We will only have pipelining registers when we are placing an un-register packed netlist.
  // numPipeliningRegisters is not calculated from scratch based on the current activity of the pipelined nets in the system.
  //  This means that we should keep the placementPipelineRegisterVector up-to-date when we zip or unzip registers.
  // *********************************************************
  if(noIOAnneal){
    choice = lrand48() % (netlist->getNumBlocks() - netlist->numInputs - netlist->numOutputs  + netlist->getNumPipeliningRegisters());

    if(choice < (netlist->getNumBlocks() - netlist->numInputs - netlist->numOutputs)){
      return (MOVELOGICBLOCK);
    }
    else{
      return (MOVEPIPEREGSOMETHING);
    }
  }
  else{
    choice = lrand48() % (netlist->getNumBlocks() + netlist->getNumPipeliningRegisters());
    
    if(choice < netlist->getNumBlocks()){
      return (MOVELOGICBLOCK);
    }
    else{
      return (MOVEPIPEREGSOMETHING);
    }
  }
}

void generatePipelineAwareMoveLogic(int *moveType, Netlist *netlist, void** block0, void** block1, int *index0, int *index1, 
				    enum e_timingMoveBlockType *blockType, int *numMove, int rLim){
  logicBlock* lb;
  int row0, col0, select, choices;
  
  if(noIOAnneal){
    while(1){
      lb = netlist->getRandomLogicBlock();
      if(lb->get_blockType() != INPUTPAD && lb->get_blockType() != OUTPUTPAD){
	break;
      }
    }
  }
  else{
    //Find a logic block in the netlist
    lb = netlist->getRandomLogicBlock();
  }
  
  if(lb->get_blockType() == INPUTPAD || lb->get_blockType() == OUTPUTPAD){
    //If we selected an input pad, we can only do a individual swap.
    //No retiming moves are possible
    *moveType = MOVELOGICBLOCKSELF;
    *blockType = IOPIN;

    row0 = lb->row;
    col0 = lb->col;
    *index0 = lb->ptc_index;
    *block0 = (void*) FPGA->getIOBAt(row0, col0);

#if defined(PIPEPLACEDEBUGHEAVY)
    printf("\n\nSelected block %s at %d, %d to start\n", lb->name.c_str(), ((vprIOB*)*block0)->get_col(), ((vprIOB*)*block0)->get_row());
    
#endif    
    //We are going to swap this iob with some other iob
    *block1 = *block0;
    *index1 = *index0;
    
    //Now find some other logic block in the array that matches the type of the one we picked
    while(((vprIOB*)*block0)->row == ((vprIOB*)*block1)->row &&
	  ((vprIOB*)*block0)->col == ((vprIOB*)*block1)->col){
      *block1 = (void *) FPGA->getOtherIOB(row0, col0, rLim);
      *index1 = ((vprIOB*)*block1)->getRandomIndex();
    }

#if defined(PIPEPLACEDEBUGHEAVY)
    if(((vprIOB*)(*block1))->getfrom_currBlockArray(*index1)){
      printf("Selected block %s at %d, %d index %d to swap\n", ((vprIOB*)(*block1))->getfrom_currBlockArray(*index1)->name.c_str(), 
	     ((vprIOB*)*block1)->get_col(), ((vprIOB*)*block1)->get_row(), *index1);
    }
    else{
      printf("Selected empty IOB at %d, %d index %d to swap\n", ((vprIOB*)*block1)->get_col(), 
	     ((vprIOB*)*block1)->get_row(), *index1);
    }
#endif
  }
  else if(lb->get_blockType() == REGISTER_BANK){
    //If we selected a register bank, we can do a swap or a retiming move.
    //A register bank represents a fixed register resource - we can't zip/unzip it.
    row0 = lb->row;
    col0 = lb->col;
    *block0 = (void*) lb->rb;
    *blockType = REGISTERBANK;
    
#if defined(PIPEPLACEDEBUGHEAVY)
    printf("\n\nSelected register bank %s at delay block %d, %d bank %d to start\n", lb->name.c_str(), ((registerBank*)*block0)->getHostCLB()->get_col(), ((registerBank*)*block0)->getHostCLB()->get_row(), ((registerBank*)*block0)->getHostBankNum());
    
    if(FPGA->getNumberofBlockType(REGISTER_BANK) == 1){
      // We might get into the situation in which we are unable to generate a move for this
      // netlist - there might be only one legal location for all of the blocks!
      // We should check for this before we get into this part of the code.
      // Can we also use this to help not make windowing make this run forever?      
      ERROR("Cannot handle move when there is only one register bank!");
    }
#endif

    //Swap this monolithic register bank with another one.
    if(clb_level_placement){
      *block1 = *block0;
      *moveType = MOVELOGICBLOCKCLB;

      //We are only able to do CLB-level placement, so keep going around till we find
      // a random register bank in some other CLB.
      //Find any register bank that is not in the same CLB.
      //NOTICE: If we have heterogeneous CLBs, we should check that the two are also compatible.
      while(((registerBank*)*block0)->getHostCLB() == ((registerBank*)*block1)->getHostCLB()){
	*block1 = (void*) FPGA->getOtherRegBank(row0, col0, rLim);
      }
    }
    else{
      //We are able to do FF-level placement, so we should do a homing test.
      //Figure out if this register bank is in the same CLB as it's parent - if not, figure out the "most eligable" register bank
      // in that CLB.
      //Notice that since this is a monolithic register bank, we don't have to consider
      // merging or zipping during the homing process.
      generateFFLevelHomingTest(lb, moveType, block0, block1, rLim);
    }
    
#if defined(PIPEPLACEDEBUGHEAVY)
    if(!((registerBank*)(*block1))->isBankAndLinkedLUTEmpty()){
      //See if this register bank is occupied by a logic block or a pipelinine register
      if(((registerBank*)*block1)->get_currBlock()){
	printf("Selected register bank with mapped logic block %s at delay block %d, %d bank to swap ", 
	       ((registerBank*)*block1)->get_currBlock()->name.c_str(), 
	       ((registerBank*)*block1)->getHostCLB()->get_col(), 
	       ((registerBank*)*block1)->getHostCLB()->get_row(),
	       ((registerBank*)*block1)->getHostBankNum());	
      }
      else{
	printf("Selected register bank with mapped pipelining registers at delay block %d, %d bank to swap ", 
	       ((registerBank*)*block1)->getHostCLB()->get_col(), 
	       ((registerBank*)*block1)->getHostCLB()->get_row(),
	       ((registerBank*)*block1)->getHostBankNum());	
      }
    }
    else{
      printf("Selected empty register bank at delay block %d, %d bank %d to swap ", 
	     ((registerBank*)*block1)->getHostCLB()->get_col(), ((registerBank*)*block1)->getHostCLB()->get_row(),
	     ((registerBank*)*block1)->getHostBankNum());
    }
    if(*moveType == MOVELOGICBLOCKCLB){
      printf("entire CLB\n");
    }
    else{
      printf("just registers\n");
    }
#endif
  }
  else if(lb->get_blockType() == LUT || lb->get_blockType() == RAM_BLOCK || 
	  lb->get_blockType() == MULTIPLIER){
    //If we selected any kind of logic block we can do a swap or a retiming move
    row0 = lb->row;
    col0 = lb->col;

    *block0 = (void*) FPGA->getCLBAt(row0, col0)->getLUT(lb->ptc_index);
    *blockType = LB;

#if defined(PIPEPLACEDEBUGHEAVY)
    printf("\n\nSelected block %s at %d, %d to start\n", lb->name.c_str(), ((vprCLB*)*block0)->get_col(), ((vprCLB*)*block0)->get_row());
    
    if(FPGA->getNumberofBlockType(((vprLUT*)*block0)->get_blockType()) == 1){
      // We might get into the situation in which we are unable to generate a move for this
      // netlist - there might be only one legal location for all of the blocks!
      // We should check for this before we get into this part of the code.
      // Can we also use this to help not make windowing make this run forever?
      ERROR("Cannot handle move of only one block!");
    }
#endif
    //Swap this block with some other block of the same type.
    *block1 = *block0;

    //Figure out if we should move just the logic block itself or the whole CLB.
    //Notice that even if we try to move "just the logic block itself", we always
    // keep LUTs together.  What we are determining is if we should bring the
    // FFs in the CLB with too.
    *moveType = MOVELOGICBLOCKCLB;

    if(!clb_level_placement){
      //We are able to do FF or lut-level placement
      if(lut_level_placement){
	//If we are able to do lut-level placement, do it
	*moveType = MOVELOGICBLOCKSELF;

	//The other LUT must have some qualifications:
	// 1) The other LUT bank must not be in the same CLB as me
	// 2) If the other LUT bank has an occupied linked register bank, this one must have
	//     a compatible one too.
	while((((vprLUT*)*block0)->row == ((vprLUT*)*block1)->row &&
	       ((vprLUT*)*block0)->col == ((vprLUT*)*block1)->col) || //(1)
	      (!((vprLUT*)*block1)->isLinkedRegistersEmpty() &&
	       !((vprLUT*)*block0)->matchingLinkedRegisters(((vprLUT*)*block1)))){ //(2)
	  switch(((vprLUT*)*block0)->get_blockType()){
	  case LUT:
	    *block1 = (void*) FPGA->getOtherLUT(row0, col0, rLim);
	    break;
	  case MULTIPLIER:
	    ERROR("Multipliers are not supported yet!");
	    //*block1 = (void*) FPGA->getRandomMult();
	    //*block1 = (void*) FPGA->getOtherMult(row0, col0, rLim);
	    break;
	  case RAM_BLOCK:
	    ERROR("RAM blocks are not supported yet!");
	    //*block1 = (void*) FPGA->getRandomRam();
	    //*block1 = (void*) FPGA->getOtherRam(row0, col0, rLim);
	    break;
	  }
	}
	return;
      }
      else{
	//We are doing FF-level placement, but not LUT-level placement
	if(!lutStickiness && drand48() <= ff_level_placement_probability){
	  //If luts are not sticky, then they could move by themselves as a group.
	  //The FF-level placement die roll succeeded, so only move the luts
	  *moveType = MOVELOGICBLOCKALLLUTS;

	  //Find any LUT that is not in the same CLB.
	  //NOTICE: In theory, we should check that the LUTs for both CLBs are the same.
	  //However, we'll assume that the two sets are compatible.
	  while(((vprLUT*)*block0)->row == ((vprLUT*)*block1)->row &&
		((vprLUT*)*block0)->col == ((vprLUT*)*block1)->col){
	    switch(((vprLUT*)*block0)->get_blockType()){
	    case LUT:
	      *block1 = (void*) FPGA->getOtherLUT(row0, col0, rLim);
	      break;
	    case MULTIPLIER:
	      ERROR("Multipliers are not supported yet!");
	      //*block1 = (void*) FPGA->getRandomMult();
	      //*block1 = (void*) FPGA->getOtherMult(row0, col0, rLim);
	      break;
	    case RAM_BLOCK:
	      ERROR("RAM blocks are not supported yet!");
	      //*block1 = (void*) FPGA->getRandomRam();
	      //*block1 = (void*) FPGA->getOtherRam(row0, col0, rLim);
	      break;
	    }
	  }
	  return;
	}
	else{
	  //The luts are sticky, or the ff-level placement roll did not succeed.
	  //Move these luts together with the FFs.
	}
      }
    }
    else{
      //We are doing CLB-level placement, so we must move together with FFs.
      //Find any other LUT that isn't in the same CLB.
    }
    //We are doing CLB-level placement, so we must move together with FFs.
    //Find any LUT that is not in the same CLB.
    //NOTICE: In theory, we should check that the LUTs and FFs for both CLBs are the same.
    //However, we'll assume that the two sets are compatible.
    while(((vprLUT*)*block0)->row == ((vprLUT*)*block1)->row &&
	  ((vprLUT*)*block0)->col == ((vprLUT*)*block1)->col){
      switch(((vprLUT*)*block0)->get_blockType()){
      case LUT:
	*block1 = (void*) FPGA->getOtherLUT(row0, col0, rLim);
	break;
      case MULTIPLIER:
	      ERROR("Multipliers are not supported yet!");
	      //*block1 = (void*) FPGA->getRandomMult();
	      //*block1 = (void*) FPGA->getOtherMult(row0, col0, rLim);
	      break;
      case RAM_BLOCK:
	ERROR("RAM blocks are not supported yet!");
	//*block1 = (void*) FPGA->getRandomRam();
	      //*block1 = (void*) FPGA->getOtherRam(row0, col0, rLim);
	break;
      }
    }

#if defined(PIPEPLACEDEBUGHEAVY)
    if(((vprLUT*)(*block1))->get_currBlock()){
      printf("Selected block %s at %d, %d to swap ", ((vprLUT*)(*block1))->get_currBlock()->name.c_str(), 
	     ((vprLUT*)*block1)->get_col(), ((vprLUT*)*block1)->get_row());
    }
    else{
      printf("Selected empty block at %d, %d to swap ", ((vprLUT*)*block1)->get_col(), ((vprLUT*)*block1)->get_row());
    }
    if(*moveType == MOVELOGICBLOCKCLB){
      printf("entire CLB\n");
    }
    else if(*moveType == MOVELOGICBLOCKALLLUTS){
      printf("just LUTs\n");
    }
    else{
      printf("just one LUT\n");
    }
#endif
  }
  else{
    ERROR("Unknown block type!");
  }
}

//Perform the ff-level swap homing test followed by, if necessary,
// the ff-level swap test.
void generateFFLevelHomingTest(logicBlock* lb, int *moveType, void** block0, void** block1, int rLim){
  *block1 = (void*) ((registerBank*)*block0)->getBestRegBankInParentHostCLBSwap();
      
  if(*block1){
    //This register bank is not in the same CLB as it's parent and there are
    // eligible register banks there, so perform the homing roll.
    if(drand48() <= ff_level_placement_homing){
      //The homing roll succeeded.
      *moveType = MOVELOGICBLOCKSELF;
      return;
    }
    else{
      //The homing roll didn't succeed, so continue.
      //We'll just grab some other register bank
    }
  }
  else{
    //There are no FFs in the parent's CLB, or it's already in the same CLB, so continue.
    //We'll just grab some other register bank.
  }

  //For whatever reason, we do not want to do a homing move.
  //Just want to grab any other register bank in the array.    
  //We should first figure out if the register is critical enough to warrent FF-level placement
  //This block is critical enough, so we should perform the ff-level placement roll
  *block1 = *block0;
  *moveType = MOVELOGICBLOCKCLB;

  if((lb->maxInputCriticality >= ff_level_placement_crit || lb->maxOutputCriticality >= ff_level_placement_crit) &&
     drand48() <= ff_level_placement_probability){
    //The ff-level placement roll succeeded.
    *moveType = MOVELOGICBLOCKSELF;
	
    //We want to do a ff-level move, so the other register bank has to have some qualifications:
    // 1) Register bank that is not the current one
    // 2) Must be large enough is valid to swap with
    // 3) Must have a current occupancy we can accomadate
    // 4) If it has an occupied linked LUT, this one must have a linked LUT and that 
    //       linked LUT must have the same kind of associated register banks that the
    //       other linked LUT has.
    while(*block0 == *block1 ||  // (1)
	  ((registerBank*)*block0)->getNumMappedRegs() > ((registerBank*)*block1)->getNumPossibleDelay() || // (2)
	  ((registerBank*)*block1)->getNumMappedRegs() > ((registerBank*)*block0)->getNumPossibleDelay() || // (3)
	  (!((registerBank*)*block1)->isLinkedLUTEmpty() && !((registerBank*)*block0)->hasLinkedLUT() &&
	   !((registerBank*)*block0)->linkedLUT->matchingLinkedRegisters(((registerBank*)*block1)->linkedLUT))){ //(4)
      *block1 = (void*) FPGA->getOtherRegBank(((registerBank*)*block0)->getHostCLB()->row, 
					      ((registerBank*)*block0)->getHostCLB()->col, rLim,
					      ((registerBank*)*block0)->getNumMappedRegs(), 
					      ((registerBank*)*block0)->getNumPossibleDelay());
    }
  }
  else{
    //This block is not critical enough to do FF-level placement, or we failed the homing roll, so we'll just do
    // normal clb-level placement.
    //NOTICE: If we have heterogenous CLBs, we should check to see that swapping the two CLBs is possible.
    //That is, the two host CLB should be equivalent.
    while(((registerBank*)*block0)->getHostCLB() == ((registerBank*)*block1)->getHostCLB()){
     *block1 = (void*) FPGA->getOtherRegBank(((registerBank*)*block0)->getHostCLB()->row, 
					      ((registerBank*)*block0)->getHostCLB()->col, rLim);
    }
  }
}

void generatePipelineAwareMovePipeReg(int *moveType, Netlist *netlist, void** block0, void** block1, int *numMove, int rLim){
  long choice;
  Terminal* term;
  registerBank *parentRegBank;
  registerBank *childRegBank;
  
  //Find a pipelining register in the netlist
  term = netlist->getRandomPipeliningRegister();
  
  //Notice that when we select a given register above, we lose this information here.
  // We only point to a given bank in a given block.  This is because moving a single register
  // around has lots of restrictions.  For example, if there are multiple registers in a bank,
  // we cannot move any random register, only the first or last ones as a group.
  *block0 = (void*) term->getRegisterBank();
  
#if defined(PIPEPLACEDEBUGHEAVY)
  printf("\n\nSelected pipelining register %s at %d, %d bank %d to start\n", term->name.c_str(), 
	 ((registerBank*)*block0)->getHostCLB()->get_col(),
	 ((registerBank*)*block0)->getHostCLB()->get_row(),
	 ((registerBank*)*block0)->getHostBankNum());

  if(FPGA->getNumberofBlockType(REGISTER_BANK) == 1){
    // We might get into the situation in which we are unable to generate a move for this
    // netlist - there might be only one legal location for all of the blocks!
    // We should check for this before we get into this part of the code.
    ERROR("Cannot handle move of only one register bank!");
  }
#endif	

  if(clb_level_placement){
    //We are not doing FF-level placement, so swap the entire CLB
    *moveType = MOVEPIPEREGSWAPCLB;

    *block1 = *block0;
    //NOTICE:If we have heterogenous CLBs, we should check to make sure that the two CLBs 
    // are compatible.
    while(((registerBank*)*block0)->getHostCLB() == ((registerBank*)*block1)->getHostCLB()){
      *block1 = (void*) FPGA->getOtherRegBank(((registerBank*)*block0)->getHostCLB()->get_row(),
					      ((registerBank*)*block0)->getHostCLB()->get_col(), rLim);
    }
  }  
  else{
    //We are able to do ff-level moves, so we have to do homing tests.
    //If we are able to do register duplication, the first homing test
    // we should consider is a zipping test.
    if(unzipNets){
      //See if there are any duplicates of this register bank.
      *block1 = (void*) ((registerBank*)*block0)->getRandomDuplicate();
      // If there is a duplicate, perform a homing roll.
      if((*block1) && drand48() <= ff_level_placement_homing){
	// If the homing roll succeeds, suggest a zip and return
	*moveType = MOVEZIPPNET;
	return;
      }
      else{
	// If there are no zipping candidates or the homing roll fails continue to the next homing test
      }
    }
    //If we are able to do register splitting, the next homing test
    // we should consider is a merging test.
    if(splitDelayBlocks){
      //See if this register can be combined with a parent or child
      // If there is merge candidate, perform a homing roll.
      if(((registerBank*)*block0)->getRandomMerge(moveType, numMove, &parentRegBank, &childRegBank)){
	if(drand48() <= ff_level_placement_homing){
	  // If the homing roll succeeds, return
	  *block0 = (void *) parentRegBank;
	  *block1 = (void *) childRegBank;
	  return;
	}
	else{
	  // If the homing roll fails, go onto the next homing test.
	}
      }
      else{
	//There are no merge candidates, so go on to the next homing test
      }
    }

    //The last homing test is the normal ff-level placement homing test.
    //Figure out if this register bank is in the same CLB as it's parent - if not, figure out the "most eligable" register bank
    // in that CLB.
    *block1 = (void*) ((registerBank*)*block0)->getBestRegBankInParentHostCLBSwap();      
    if(*block1){
      //This register bank is not in the same CLB as it's parent and there are
      // eligible register banks there, so perform the homing roll.
      if(drand48() <= ff_level_placement_homing){
	//The homing roll succeeded.
	*moveType = MOVEPIPEREGSWAPSELF;
	return;
      }
      else{
	//The homing roll didn't succeed, so continue.
      }
    }
    else{
      //There are no FFs in the parent's CLB, or it's already in the same CLB, so continue.
    }

    //For whatever reason, we do not want to do a homing move.
    //Just want to grab any other register bank in the array
    *block1 = (void*) FPGA->getOtherRegBank(((registerBank*)*block0)->getHostCLB()->get_row(),
					    ((registerBank*)*block0)->getHostCLB()->get_col(), rLim,
					    ((registerBank*)*block0)->getNumMappedRegs(), 
					    ((registerBank*)*block0)->getNumPossibleDelay());

    //WORKING HERE!
    //Now that we have two particular register banks in mind, we have to figure out what to do with them.  
    if(!splitDelayBlocks){
      //If we are not splitting delay blocks, we can only swap the two
      //We should do some checking to make sure that is legal, and perform
      // the CLB versus FF-level criticality check and probability roll.
      generatePipelineAwareMovePipeRegSwap(moveType, netlist, (registerBank**) block0, 
					   (registerBank**) block1, numMove, rLim);
    }
    else if(!unzipNets){
      //If we are only splitting register banks, we could swap the contents (either
      // only the registers or the whole CLB depending on the criticality checks and
      // probability rolls) or split the registers in the bank
      generatePipelineAwareMovePipeRegSplitOrSwap(moveType, netlist, (registerBank**) block0, 
						  (registerBank**) block1, numMove, rLim);
    }
    else{
      //We are able to swap, split or unzip registers
      generatePipelineAwareMovePipeRegSplitSwapOrUnzip(moveType, netlist, (registerBank**) block0, 
						       (registerBank**) block1, numMove, rLim);
    }
  }

#if defined(PIPEPLACEDEBUGHEAVY)
  printf("\n\nSelected pipelining register at %d, %d bank %d to swap\n", 
	 ((registerBank*)*block1)->getHostCLB()->get_col(),
	 ((registerBank*)*block1)->getHostCLB()->get_row(),
	 ((registerBank*)*block1)->getHostBankNum());
#endif
}

//We want to do something with two register banks, but we can't split them up.
//Therefore, we have to swap them.
//We want to check to make sure this is legal, and determine if we should
// perform a FF or CLB-level swap.
//We should check the criticality of this register, then, if it is critical
// enough to do a FF-level swap, perform the probability roll.
void generatePipelineAwareMovePipeRegSwap(int *moveType, Netlist *netlist, registerBank** regBank0, 
					  registerBank** regBank1, int *numMove, int rLim){
  *regBank1 = *regBank0;
  *moveType = MOVEPIPEREGSWAPCLB;
  
  //We should first figure out if the register bank is critical enough to warrent FF-level placement
  //Check the input criticality of the first register and the output criticality of the last register.
  //Notice, we don't check regBank1.  It might be empty.
  if(((*regBank0)->getFirstSink()->maxInputCriticality >= ff_level_placement_crit || 
      (*regBank0)->getLastSource()->maxOutputCriticality >= ff_level_placement_crit) &&
     drand48() <= ff_level_placement_probability){
    //The ff-level placement roll succeeded.
    *moveType = MOVEPIPEREGSWAPSELF;
	
    //We want to do a ff-level move, so any the other register bank must be:
    // 1) Not the current one
    // 2) Large enough is valid to swap with
    // 3) Have a suitable current occupancy we can accomodate
    // 4) If the other register bank has an occupied linked LUT, this one must
    //     also have a linked LUT, not to mention the two register banks
    //     must be compatible.
    while(*regBank0 == *regBank1 || //(1)
	  (*regBank0)->getNumMappedRegs() > (*regBank1)->getNumPossibleDelay() || //(2)
	  (*regBank1)->getNumMappedRegs() > (*regBank0)->getNumPossibleDelay() || //(3)
	  (!(*regBank1)->isLinkedLUTEmpty() && 
	   (!(*regBank0)->hasLinkedLUT() || !(*regBank0)->hasMatchingLinkedRegisters(*regBank1)))){ //(4)
      *regBank1 = FPGA->getOtherRegBank((*regBank0)->getHostCLB()->row, 
					(*regBank0)->getHostCLB()->col, rLim);
    }
  }
  else{
    //This block is not critical enough to do FF-level placement, or we failed the homing roll, so we'll just do
    // normal clb-level placement.
    //Just grab any register that is not in the same CLB.
    //NOTICE:If we have heterogenous CLBs, we should check to make sure that the two CLBs 
    // are compatible.
    while((*regBank0)->getHostCLB() == (*regBank1)->getHostCLB()){
     *regBank1 = FPGA->getOtherRegBank((*regBank0)->getHostCLB()->row, 
					(*regBank0)->getHostCLB()->col, rLim);
    }
  }

#if defined(PIPEPLACEDEBUGHEAVY)
  if((*regBank1)->isBankAndLinkedLUTEmpty()){
    printf("Selected empty register bank at delay block %d, %d bank %d to swap\n", (*regBank1)->getHostCLB()->get_col(), 
	   (*regBank1)->getHostCLB()->get_row(), (*regBank1)->getHostBankNum());
  }
  else{
    if((*regBank1)->get_currBlock()){
      printf("Selected register bank with mapped logic block %s at delay block %d, %d bank to swap\n", 
	     (*regBank1)->get_currBlock()->name.c_str(), 
	     (*regBank1)->getHostCLB()->get_col(), 
	     (*regBank1)->getHostCLB()->get_row(),
	     (*regBank1)->getHostBankNum());	
    }
    else{
      printf("Selected register bank with mapped pipelining registers at delay block %d, %d bank to swap\n",
	     (*regBank1)->getHostCLB()->get_col(), 
	       (*regBank1)->getHostCLB()->get_row(), (*regBank1)->getHostBankNum());
    }
  }
#endif
}

//We want to do something with two register banks, but we can only split or swap, not duplicate
void generatePipelineAwareMovePipeRegSplitOrSwap(int *moveType, Netlist *netlist, registerBank **regBank0, 
						 registerBank **regBank1, int *numMove, int rLim){
  //First, we should try and perform the normal ff-level placement versus clb-level placement
  // choice.
  generatePipelineAwareMovePipeRegSwap(moveType, netlist, regBank0, regBank1, numMove, rLim);

  //If we chose to perform a ff-level move, we have a few choices to make.
  //The other register bank that we get from the "generatePipelineAwareMovePipeRegSwap" function
  // will always be large enough to perform a swap with.
  //This means that shift registers with more than 2 register mapped to them will
  // not be able to split into a single FF.
  //They will only be allowed to split into another shift register.
  //We handle this by reselecting a new register bank in the "generatePipelineAwareMoveSplitReg"
  // function
  if(*moveType == MOVEPIPEREGSWAPSELF){
    //If the other bank is empty and there is more than one register in this bank, 
    // and the two register banks are not in the same CLB, we might want to split the registers.
    if((*regBank1)->isBankAndLinkedLUTEmpty() && (*regBank0)->getHostCLB() != (*regBank1)->getHostCLB() &&
       (*regBank0)->getNumMappedRegs() > 1){
      //Whether or not we split the register is subject to:
      //   a) the criticality of the input or output nets is greater than or equal to the split threshold
      //   b) the split probability roll succeeds
      if((*regBank0)->getFirstSink()->maxInputCriticality >= splitDelayBlocks_crit ||
	  (*regBank0)->getLastSource()->maxOutputCriticality >= splitDelayBlocks_crit){
	//At least one of the nets passed the criticality threshold test
	if(drand48() <= splitDelayBlocks_probability){
	  //The split roll succeeded, so figure out what we want to do.
	  generatePipelineAwareMoveSplitReg(moveType, regBank0, regBank1, numMove, rLim);
	}
	else{
	  //The split roll did not succeed, so just do a normal ff-level move
	}
      }
      else{
	//The criticality check did not pass, so just do a normal ff-level move
      }
    }
    //If the other bank is not empty, or there is only one register in this bank,
    // let's just swap the two.  We don't care about checking for merging capability here,
    // since we already handle that during the homing check.
    else{
    }    
  }
  //If we chose to perform a clb-level move, we don't want to do anything extra here.
  else{
  }
}

//We have selected two register banks and we are allowed to split, swap or duplicate registers
void generatePipelineAwareMovePipeRegSplitSwapOrUnzip(int *moveType, Netlist *netlist, registerBank **regBank0, 
						      registerBank **regBank1, int *numMove, int rLim){
  int numChoices = 0;
  bool canSplit = false;
  bool canUnzip = false;

  //First, we should try and perform the normal ff-level placement versus clb-level placement
  // choice.
  generatePipelineAwareMovePipeRegSwap(moveType, netlist, regBank0, regBank1, numMove, rLim);

  //If we chose to perform a ff-level move, we have a few choices to make.
  if(*moveType == MOVEPIPEREGSWAPSELF){
    //If the other bank is empty and there is more than one register in this bank, 
    // and the two register banks are not in the same CLB, we might want to split the registers.
    //The other register bank that we get from the "generatePipelineAwareMovePipeRegSwap" function
    // will always be large enough to perform a swap with.
    //This means that shift registers with more than 2 register mapped to them will
    // not be able to split into a single FF.
    //They will only be allowed to split into another shift register.
    //We handle this by reselecting a new register bank in the "generatePipelineAwareMoveSplitReg"
    // function
    if((*regBank1)->isBankAndLinkedLUTEmpty() && (*regBank0)->getHostCLB() != (*regBank1)->getHostCLB() &&
       (*regBank0)->getNumMappedRegs() > 1){
      //Whether or not we split the register is subject to:
      //   a) the criticality of the input or output nets is greater than or equal to the split threshold
      //   b) the split probability roll succeeds
      if((*regBank0)->getFirstSink()->maxInputCriticality >= splitDelayBlocks_crit ||
	 (*regBank0)->getLastSource()->maxOutputCriticality >= splitDelayBlocks_crit){
	//At least one of the nets passed the criticality threshold test
	if(drand48() <= splitDelayBlocks_probability){
	  canSplit = true;
	}
      }
    }
    //If the other bank is empty, the number of registers in this bank could fit into the
    // other bank, we have a fan out greater than 1, and the two banks are not in the same CLB,
    // we could duplicate this entire register bank
    if((*regBank1)->isBankAndLinkedLUTEmpty() && (*regBank0)->getNumMappedRegs() <= (*regBank1)->getNumPossibleDelay() &&
       (*regBank0)->getLastSource()->hasFanOut() && (*regBank0)->getHostCLB() != (*regBank1)->getHostCLB()){
      if((*regBank0)->getFirstSink()->maxInputCriticality >= unzipNets_crit ||
	 (*regBank0)->getLastSource()->maxOutputCriticality >= unzipNets_crit){
	//There is one more check we should make before we say OK to the
	// unzipping: there must be a conflict of interest between two of the sinks.
	//That is, some of the sinks should be closer to regbank0 and some should be closer to
	// regBank1.  If there isn't this split, we shouldn't duplicate the register.
	if((*regBank0)->someSinksCloserTo(*regBank1) && drand48() <= unzipNets_probability){
	  canUnzip = true;
	}
      }
    }

    //If the other bank is not empty, or isn't available to do a split/unzip, just to a regular swap.
    //We don't worry about merging or zipping things back together here, since we already handle that
    // in the homing check.
    //OK, now choose between doing either of the avaiable moves
    if(canSplit && canUnzip){
      //We can either split or unzip, so randomly choose one or the other
      if(lrand48() % 2 == 0){
	//Split the register bank
	generatePipelineAwareMoveSplitReg(moveType, regBank0, regBank1, numMove, rLim);
      }
      else{
	//Duplicate the register bank
	*moveType = MOVEUNZIPPNET;
      }
    }
    else if(canSplit){
      generatePipelineAwareMoveSplitReg(moveType, regBank0, regBank1, numMove, rLim);
    }
    else if(canUnzip){
      *moveType = MOVEUNZIPPNET;
    }
    //We can't or don't want to split or duplicate these registers, so don't do anything else here.
    else{
    }
  }
  //If we chose to perform a clb-level move, we don't want to do anything extra here.
  else{
  }
}
  
//We want to split the registers in bank0 up and put some into the currently empty bank1
void generatePipelineAwareMoveSplitReg(int *moveType, registerBank **regBank0, registerBank **regBank1, int *numMove, int rLim){
  bool canSplitFirst = false;
  bool canSplitLast = false;

  //The register bank that we have chosen will always be large enough to perform a swap with
  // the current register bank.
  //This means that shift registers with more than 2 register mapped to them will
  // not be able to split into a single FF.
  //However, we cannot just select another random register bank, we need to be careful since
  // it must be empty and they cannot be in the same CLB
  *regBank1 = *regBank0;
  while(*regBank0 == *regBank1 || !(*regBank1)->isBankAndLinkedLUTEmpty() ||
	(*regBank0)->getHostCLB() == (*regBank1)->getHostCLB()){
    *regBank1 = FPGA->getOtherRegBank((*regBank0)->getHostCLB()->row, 
					(*regBank0)->getHostCLB()->col, rLim);
  }

  //We know that we want to move split off between 1 and N-1 registers
  //We know that there are at least 2 registers in this bank.
  //*numMove = min((*regBank0)->getNumMappedRegs() - 1, (*regBank1)->getNumFreeRegs());
  *numMove = min((*regBank0)->getNumMappedRegs() - 1, (*regBank1)->getNumPossibleDelay());
  *numMove = (lrand48() % *numMove) + 1;
  
  //Figure out if we can split forward and/or backwards
  if((*regBank0)->getFirstSink()->maxInputCriticality >= splitDelayBlocks_crit){
    canSplitFirst = true;
  }
  if((*regBank0)->getLastSource()->maxOutputCriticality >= splitDelayBlocks_crit){
    canSplitLast = true;
  }

  if(canSplitFirst && canSplitLast){
    //We can either split off the front or the back, so randomly choose one or the other
    if(lrand48() % 2 == 0){
      //Put the first N registers in the bank into bank1.
      *moveType = MOVEREGEMPTYFIRST;
    }
    else{
      //Put the last N registers in the bank into bank1
      *moveType = MOVEREGEMPTYLAST;
    }
  }
  else if(canSplitFirst){
    *moveType = MOVEREGEMPTYFIRST;
  }
  else if(canSplitLast){
    *moveType = MOVEREGEMPTYLAST;
  }
  else{
    ERROR("Something weird happened here!");
  }

#if defined(PIPEPLACEDEBUGHEAVY)
  if(*moveType == MOVEREGEMPTYFIRST){
    printf("Selected empty delay block at %d, %d bank %d to move first %d register into\n", 
	   (*regBank1)->getHostCLB()->get_col(), (*regBank1)->getHostCLB()->get_row(), 
	   (*regBank1)->getHostBankNum(), *numMove);
  }
  else if(*moveType == MOVEREGEMPTYLAST){
    printf("Selected empty delay block at %d, %d bank %d to move last %d register into\n", 
	   (*regBank1)->getHostCLB()->get_col(), (*regBank1)->getHostCLB()->get_row(), 
	   (*regBank1)->getHostBankNum(), *numMove);
  }
#endif
}

//Check to see if it is possible to retime any registers backwards.
//Depending upon how things are connected we might need to grab new register banks
// and the architecture might not have any empty.
//If it is possible at all we will return true and how many we want to move.
//If not, we will return false.
bool generatePipelineAwareMoveRegRetimeBackward(int *moveType, Netlist *netlist, void** block0, enum e_timingMoveBlockType blockType, int *numMove){
  logicBlock *lb;
  int index, i;
  list <int> numMoveList;
  list <int>::iterator intIter;
  int numRegsFreed, numRegsAdded;
  Terminal* tempTerm;

  //Grab the logic block that is currently mapped to this block
  if(blockType == LB){
    lb = ((vprLUT*)*block0)->get_currBlock();
  }
  else if(blockType == REGISTERBANK){
    lb = ((registerBank*)*block0)->get_currBlock();
  }
  else{
    ERROR("Block type not allowed to retime backwards!");
  }

  //Build a list of the number of register we can move back through this logic block - between 1 and all of the ones we can move
  //for(i = 1; i <= lb->numRegCanPushBackward; i++){
  //  numMoveList.push_back(i);
  //}
  numMoveList.push_back(1);

  while (numMoveList.size() != 0){
    //Grab a random number from the list, then delete it from the list
    index = lrand48() % numMoveList.size();
    for(intIter = numMoveList.begin(), i = 0; i < index; intIter++, i++){
    }
    *numMove = *intIter;
    numMoveList.erase(intIter);

    //Figure out how many registers this would free up on the output side.
    //Notice if a register bank is still partially occupied, this will not count them
    numRegsFreed = 0;

    //Go through all of the outputs and see what would happen if we remove this many
    // delays from each
    lb->create_outputListItr();
    for(tempTerm = lb->getnext_outputList(); tempTerm != 0; tempTerm = lb->getnext_outputList()){
      numRegsFreed += tempTerm->numRegsFreedForward(*numMove);
    }
    lb->remove_outputListItr();
    
#ifdef PLACER_DEBUG
    numRegistersEstimatedFreed = numRegsFreed;
#endif
    //cout << "Retiming " << *numMove << " registers backwards would free " << numRegsFreed << " registers " <<endl;

    //Figure out how many registers we would need to add on the input side
    //Notice we try to fill up existing, partially occupied banks if we can.
    numRegsAdded = 0;

    //Go through each of the inputs and figure out how many registers we could push on for free
    // (ones that happen to fill in empty spaces in banks that are already taken)
    // then figure out how many extra we'll need
    lb->create_inputListItr();
    for(tempTerm = lb->getnext_inputList(); tempTerm != 0; tempTerm = lb->getnext_inputList()){
      numRegsAdded += tempTerm->numRegsAddedBackward(*numMove);
    }
    lb->remove_inputListItr();

#ifdef PLACER_DEBUG
    numRegistersEstimatedNeeded = numRegsAdded;
#endif

    //cout << "Retiming " << *numMove << " register backwards would require " << numRegsAdded << " new registers " <<endl;

    //Do we have that many register banks free in the system?
    // If not, try some other number of registers
    if(numRegsFreed + FPGA->numFreeRegisters >= numRegsAdded){
        *moveType = MOVEREGRETIMEBACK;
	return true;
    }
  }

#ifdef PLACER_DEBUG
  numRegistersEstimatedFreed = 0;
  numRegistersEstimatedNeeded = 0;
#endif

  // We cannot move any number of registers back through this block
  return (false);  
}

//Check to see if it is possible to retime any registers forwards.
//Depending upon how things are connected we might need to grab new register banks
// and the architecture might not have any empty.
//If it is possible at all we will return true and how many we want to move.
//If not, we will return false.
bool generatePipelineAwareMoveRegRetimeForward(int *moveType, Netlist *netlist, void** block0, enum e_timingMoveBlockType blockType, int *numMove){
  logicBlock *lb;
  int index, i;
  list <int> numMoveList;
  list <int>::iterator intIter;
  int numRegsFreed, numRegsAdded;
  Terminal* tempTerm;

  //Grab the logic block that is currently mapped to this block
  if(blockType == LB){
    lb = ((vprLUT*)*block0)->get_currBlock();
  }
  else if(blockType == REGISTERBANK){
    lb = ((registerBank*)*block0)->get_currBlock();
  }
  else{
    ERROR("Block type not allowed to retime backwards!");
  }

  //Build a list of the number of register we can move forward through this logic block - between 1 and all of the ones we can move
  //for(i = 1; i <= lb->numRegCanPushForward; i++){
  //  numMoveList.push_back(i);
  //}
  numMoveList.push_back(1);

  while (numMoveList.size() != 0){
    //Grab a random number from the list, then delete it from the list
    index = lrand48() % numMoveList.size();
    for(intIter = numMoveList.begin(), i = 0; i < index; intIter++, i++){
    }
    *numMove = *intIter;
    numMoveList.erase(intIter);

    //Figure out how many registers this would free up on the input side.
    //Notice if a register bank is still partially occupied, this will not count them
    numRegsFreed = 0;

    //Go through all of the inputs and see what would happen if we remove this many
    // delays from each
    lb->create_inputListItr();
    for(tempTerm = lb->getnext_inputList(); tempTerm != 0; tempTerm = lb->getnext_inputList()){
      //While we are doing this, figure out if the number of registers we want to
      // retime forwards is even available as an output.  That is,
      // we will only allow this function to retime to a currently
      // available signals if there is upstream fanout.  If we have a latency 9 
      ///signal in groups of 3 registers, we will only be able to retime backwards 
      // 3, 6 or 9 registers if we have seen any fanout in between.
      numRegsFreed += tempTerm->numRegsFreedBackward(*numMove);
    }
    lb->remove_inputListItr();

#ifdef PLACER_DEBUG
    numRegistersEstimatedFreed = numRegsFreed;
#endif

    //We will return a negative number if we can't retime this signal by this amount due to fan-out
    if(numRegsFreed < 0){
      continue;
    }

    //cout << "Retiming " << *numMove << " registers forwards would free " << numRegsFreed << " registers " <<endl;

    //Figure out how many registers we would need to add on the output side
    //Notice we try to fill up existing, partially occupied banks if we can.
    numRegsAdded = 0;

    //Go through each of the outputs and figure out how many registers we could push on for free
    // (ones that happen to fill in empty spaces in banks that are already taken)
    // then figure out how many extra we'll need
    lb->create_outputListItr();
    for(tempTerm = lb->getnext_outputList(); tempTerm != 0; tempTerm = lb->getnext_outputList()){
      numRegsAdded += tempTerm->numRegsAddedForward(*numMove);
    }
    lb->remove_outputListItr();

#ifdef PLACER_DEBUG
    numRegistersEstimatedNeeded = numRegsAdded;
#endif    

    //cout << "Retiming " << *numMove << " register forwards would require " << numRegsAdded << " new registers " <<endl;

    //Do we have that many register banks free in the system?
    // If not, try some other number of registers
    if(numRegsFreed + FPGA->numFreeRegisters >= numRegsAdded){
        *moveType = MOVEREGRETIMEFORWARD;
	return true;
    }
  }

  // We cannot move any number of registers back through this block
#ifdef PLACER_DEBUG
  numRegistersEstimatedFreed = 0;
  numRegistersEstimatedNeeded = 0;
#endif

  return (false);
}

//Commit a change to the system and calculate new delays and incremental slacks.
// Notice that these slack updates are not quite correct since the critical path might change
//We return true if the move was completed.
bool makePipelineAwareMove(int *moveType, Netlist *netlist, void** block0, void** block1, int index0, int index1, 
			   enum e_timingMoveBlockType blockType, int numMove, double criticalPathDelay, list <PipeliningNet *> *retimeUndoList){
  logicBlock* lb0;
  logicBlock* lb1;
  bool success;
  
  if(*moveType == MOVELOGICBLOCKCLB){
    //Swap the entire CLB associated with these logic blocks
    if(blockType == REGISTERBANK){
      //We focused on a register bank, so use that as a basis to call the swap function
      ((registerBank*)(*block0))->swapEntireStickyMappings((registerBank*)(*block1), criticalPathDelay);
    }
    else if(blockType == LB){
      //We focused on a LUT, so use that as a basis to call the swap function
      ((vprLUT*)(*block0))->swapStickyMapping((vprLUT*)(*block1), criticalPathDelay);
    }
    else{
      ERROR("Unknown block type for MOVELOGICBLOCKCLB!");
    }
  }
  else if(*moveType == MOVELOGICBLOCKALLLUTS){
    //We have selected a LUT and we would like to swap all of the luts in the
    // CLB with the other CLB's luts.
    if(blockType == LB){
      if(lut_level_placement){
	//The swapLUTGroupMapping function relies on netlist-level clb information.
	//Thus, once we start performing lut-level moves we can no longer use that
	// function.
	ERROR("This should not occur!");
      }
      else{
	lb0=((vprLUT*)(*block0))->get_currBlock();
	lb1=((vprLUT*)(*block1))->get_currBlock();      
	((vprLUT*)(*block0))->swapLUTGroupMappingAndAssociatedRegBanks((vprLUT*)(*block1), criticalPathDelay);
      }
    }
    else{
      ERROR("Unknown block type for MOVELOGICBLOCKALLLUTS!");
    }
  }
  else if(*moveType == MOVELOGICBLOCKSELF){
    // Only swap the specific blocks associated with these logic blocks
    if(blockType == IOPIN){
      lb0 = ((vprIOB*)(*block0))->getfrom_currBlockArray(index0);
      lb1 = ((vprIOB*)(*block1))->getfrom_currBlockArray(index1);
      ((vprIOB*)(*block0))->swapMapping(index0, lb0, (vprIOB*)(*block1), index1, lb1, criticalPathDelay);
    }
    else if(blockType == REGISTERBANK){
      ((registerBank*)(*block0))->swapBankAndAssociatedLUT((registerBank *)(*block1), criticalPathDelay);
    }
    else if(blockType == LB){
      lb0=((vprLUT*)(*block0))->get_currBlock();
      lb1=((vprLUT*)(*block1))->get_currBlock();      
      ((vprLUT*)(*block0))->swapSingleLUTMappingAndAssociatedRegBanks(lb0, (vprLUT*)(*block1), lb1, criticalPathDelay);
    }
    else{
      ERROR("Unknown block type for MOVELOGICBLOCKSELF!");
    }
  }

  //We are moving just a register bank
  else if(*moveType == MOVEPIPEREGSWAPSELF){
    ((registerBank *)(*block0))->swapBankAndAssociatedLUT((registerBank *)(*block1), criticalPathDelay);
  }
  //We are moving the entire CLB that a register bank belongs to
  else if(*moveType == MOVEPIPEREGSWAPCLB){
    ((registerBank*)(*block0))->swapEntireStickyMappings((registerBank*)(*block1), criticalPathDelay);
  }

  //We are moving the first N registers from block 0 to an empty block 1
  else if(*moveType == MOVEREGEMPTYFIRST || *moveType == MOVEREGEMPTYLAST ||
	  *moveType == MOVEREGPARENT || *moveType == MOVEREGCHILD){
    if(*moveType == MOVEREGEMPTYFIRST){
      ((registerBank *)(*block0))->migrateFirstNRegs((registerBank *)(*block1), numMove);
    }
    //We are moving the last N register from block 0 to an empty block 1
    else if(*moveType == MOVEREGEMPTYLAST){
      ((registerBank *)(*block0))->migrateLastNRegs((registerBank *)(*block1), numMove);
    }
    //We are moving the first N register from the child block 1 to the parent block 0
    else if(*moveType == MOVEREGPARENT){
      ((registerBank *)(*block1))->migrateFirstNRegs((registerBank *)(*block0), numMove);
    }
    //We are moving the last N register from the parent block 0 to the child block 1
    else if(*moveType == MOVEREGCHILD){
      ((registerBank *)(*block0))->migrateLastNRegs((registerBank *)(*block1), numMove);
    }

    //Update the sink and slack timings for the nets we just moved
    ((registerBank *)(*block0))->updateSinkDelaysAndSlacks();
    ((registerBank *)(*block1))->updateSinkDelaysAndSlacks();
    ((registerBank *)(*block0))->calculateMaxInputOutputCriticalitiesJustRB(criticalPathDelay);
    ((registerBank *)(*block1))->calculateMaxInputOutputCriticalitiesJustRB(criticalPathDelay);
  }
  else if(*moveType == MOVEUNZIPPNET){
    ((registerBank *)(*block0))->duplicateRegs(netlist, (registerBank *)(*block1), criticalPathDelay);

    //Update the sink and slack timings for the nets we just modified
    ((registerBank *)(*block0))->updateSinkDelaysAndSlacks();
    ((registerBank *)(*block1))->updateSinkDelaysAndSlacks();
    ((registerBank *)(*block0))->calculateMaxInputOutputCriticalitiesJustRB(criticalPathDelay);
    ((registerBank *)(*block1))->calculateMaxInputOutputCriticalitiesJustRB(criticalPathDelay);
  }
  else if(*moveType == MOVEZIPPNET){
    ((registerBank *)(*block0))->zipRegs(netlist, (registerBank *)(*block1), criticalPathDelay);

    //Update the sink and slack timings for the nets we just modified.
    //Block 1 will be empty at this time.
    ((registerBank *)(*block0))->updateSinkDelaysAndSlacks();
    ((registerBank *)(*block0))->calculateMaxInputOutputCriticalitiesJustRB(criticalPathDelay);
  }
  else if(*moveType == MOVEREDOUNZIPPNET){
    ((registerBank *)(*block0))->reDuplicateRegs(netlist, (registerBank *)(*block1), criticalPathDelay);
  
    //Update the sink and slack timings for the nets we just modified
    ((registerBank *)(*block0))->updateSinkDelaysAndSlacks();
    ((registerBank *)(*block1))->updateSinkDelaysAndSlacks();
    ((registerBank *)(*block0))->calculateMaxInputOutputCriticalitiesJustRB(criticalPathDelay);
    ((registerBank *)(*block1))->calculateMaxInputOutputCriticalitiesJustRB(criticalPathDelay);
  }
  //We want to move N registers from each output of block 0 to the input of block 0
  else if(*moveType == MOVEREGRETIMEBACK){

    if(blockType == LB){
      if(!((vprLUT*)(*block0))->get_currBlock()->retimeRegistersBackwards(netlist, numMove, criticalPathDelay, retimeUndoList)){
	//We failed to be able to add all of the registers we needed backwards.
	//Undo the partially completed retiming.
	
	((vprLUT*)(*block0))->get_currBlock()->abortRetimeRegistersBackwards(netlist, numMove, criticalPathDelay, retimeUndoList);
	
	//Now we have to treat this, as far as figuring out correct sink delays and updating 
	// net costs, as if we did a regular undo retime backward.
	//This is because we put back a bunch of registers upstream of this logic block.
	((vprLUT*)(*block0))->get_currBlock()->updateSinkDelaysAndSlacksForwards();
	((vprLUT*)(*block0))->get_currBlock()->calculateMaxInputOutputCriticalitiesForwards(criticalPathDelay);
	
	*moveType = MOVEREGRETIMEBACKABORT;
	
#ifdef PLACER_DEBUG
	numRegistersEstimatedFreed = 0;
	numRegistersEstimatedNeeded = 0;
#endif
	return false;
      }
      else{
	//We succeeded in the retiming, so update all of the sinks and slacks for the
	//new nets we created upstream
	((vprLUT*)(*block0))->get_currBlock()->updateSinkDelaysAndSlacksBackwards();
	((vprLUT*)(*block0))->get_currBlock()->calculateMaxInputOutputCriticalitiesBackwards(criticalPathDelay);
      }
    }
    else if(blockType == REGISTERBANK){
      if(!((registerBank*)(*block0))->get_currBlock()->retimeRegistersBackwards(netlist, numMove, criticalPathDelay, retimeUndoList)){
	//We failed to be able to add all of the registers we needed backwards.
	//Undo the partially completed retiming.
	//netlist->printPlacement();
	//netlist->print_netList();
	
	((registerBank*)(*block0))->get_currBlock()->abortRetimeRegistersBackwards(netlist, numMove, criticalPathDelay, retimeUndoList);
	
	//Now we have to treat this, as far as figuring out correct sink delays and updating 
	// net costs, as if we did a regular undo retime backward.
	//This is because we put back a bunch of registers upstream of this logic block.
	((registerBank*)(*block0))->get_currBlock()->updateSinkDelaysAndSlacksForwards();
	((registerBank*)(*block0))->get_currBlock()->calculateMaxInputOutputCriticalitiesForwards(criticalPathDelay);
	
	*moveType = MOVEREGRETIMEBACKABORT;
	
#ifdef PLACER_DEBUG
	numRegistersEstimatedFreed = 0;
	numRegistersEstimatedNeeded = 0;
#endif
	
	return false;
      }
      else{
	//We succeeded in the retiming, so update all of the sinks and slacks for the
	//new nets we created upstream
	((registerBank*)(*block0))->get_currBlock()->updateSinkDelaysAndSlacksBackwards();
	((registerBank*)(*block0))->get_currBlock()->calculateMaxInputOutputCriticalitiesBackwards(criticalPathDelay);
      }
    }
    else{
      ERROR("Block type is not allowed to be retimed!");
    }
  }
  else if(*moveType == MOVEREGRETIMEFORWARD){
    //netlist->printPlacement();
    //netlist->print_netList();
    if(blockType == LB){
      if(!((vprLUT*)(*block0))->get_currBlock()->retimeRegistersForwards(netlist, numMove, criticalPathDelay, retimeUndoList)){
	//We failed to be able to add all of the registers we needed forwards.
	//Undo the partially completed retiming.
	((vprLUT*)(*block0))->get_currBlock()->abortRetimeRegistersForwards(netlist, numMove, criticalPathDelay, retimeUndoList);
	
	//Now we have to treat this, as far as figuring out correct sink delays and updating 
	// net costs, as if we did an undo retime forward.
	//This is because we put back a bunch of registers upstream of this logic block.
	((vprLUT*)(*block0))->get_currBlock()->updateSinkDelaysAndSlacksBackwards();
	((vprLUT*)(*block0))->get_currBlock()->calculateMaxInputOutputCriticalitiesBackwards(criticalPathDelay);	
	*moveType = MOVEREGRETIMEFORWARDABORT;
	
#ifdef PLACER_DEBUG
	numRegistersEstimatedFreed = 0;
	numRegistersEstimatedNeeded = 0;
#endif
	
	return false;
      }
      else{
	//We successfully added registers downstream of this logic block, so update the
	// new sink delays and slacks!
	((vprLUT*)(*block0))->get_currBlock()->updateSinkDelaysAndSlacksForwards();
	((vprLUT*)(*block0))->get_currBlock()->calculateMaxInputOutputCriticalitiesForwards(criticalPathDelay);
      }
    }
    else if(blockType == REGISTERBANK){
      if(!((registerBank*)(*block0))->get_currBlock()->retimeRegistersForwards(netlist, numMove, criticalPathDelay, retimeUndoList)){
	//We failed to be able to add all of the registers we needed forwards.
	//Undo the partially completed retiming.
	((registerBank*)(*block0))->get_currBlock()->abortRetimeRegistersForwards(netlist, numMove, criticalPathDelay, retimeUndoList);
	((registerBank*)(*block0))->get_currBlock()->calculateMaxInputOutputCriticalitiesForwards(criticalPathDelay);
	
	//Now we have to treat this, as far as figuring out correct sink delays and updating 
	// net costs, as if we did an undo retime forward.
	//This is because we put back a bunch of registers upstream of this logic block.
	((registerBank*)(*block0))->get_currBlock()->updateSinkDelaysAndSlacksBackwards();
	((registerBank*)(*block0))->get_currBlock()->calculateMaxInputOutputCriticalitiesBackwards(criticalPathDelay);
	
	*moveType = MOVEREGRETIMEFORWARDABORT;
	
#ifdef PLACER_DEBUG
	numRegistersEstimatedFreed = 0;
	numRegistersEstimatedNeeded = 0;
#endif
	
	return false;
      }
      else{
	//We successfully added registers downstream of this logic block, so update the
	// new sink delays and slacks!
	((registerBank*)(*block0))->get_currBlock()->updateSinkDelaysAndSlacksForwards();
	((registerBank*)(*block0))->get_currBlock()->calculateMaxInputOutputCriticalitiesForwards(criticalPathDelay);
      }
    }
    else{
      ERROR("Block type is not allowed to be retimed!");
    }
  }
  else if(*moveType == MOVEREGUNDOREGRETIMEBACK){
    //netlist->printPlacement();
    //netlist->print_netList();
    if(blockType == LB){
      ((vprLUT*)(*block0))->get_currBlock()->undoRetimeRegistersBackwards(netlist, numMove, criticalPathDelay, retimeUndoList);
      ((vprLUT*)(*block0))->get_currBlock()->updateSinkDelaysAndSlacksForwards();
      ((vprLUT*)(*block0))->get_currBlock()->calculateMaxInputOutputCriticalitiesForwards(criticalPathDelay);
    }
    else if(blockType == REGISTERBANK){
      ((registerBank*)(*block0))->get_currBlock()->undoRetimeRegistersBackwards(netlist, numMove, criticalPathDelay, retimeUndoList);
      ((registerBank*)(*block0))->get_currBlock()->updateSinkDelaysAndSlacksForwards();
      ((registerBank*)(*block0))->get_currBlock()->calculateMaxInputOutputCriticalitiesForwards(criticalPathDelay);
    }
    else{
      ERROR("Block type is not allowed to be retimed!");
    }
  }
  else if(*moveType == MOVEREGUNDOREGRETIMEFORWARD){
    //netlist->printPlacement();
    //netlist->print_netList();
    if(blockType == LB){
      ((vprLUT*)(*block0))->get_currBlock()->undoRetimeRegistersForwards(netlist, numMove, criticalPathDelay, retimeUndoList);
      ((vprLUT*)(*block0))->get_currBlock()->updateSinkDelaysAndSlacksBackwards();
      ((vprLUT*)(*block0))->get_currBlock()->calculateMaxInputOutputCriticalitiesBackwards(criticalPathDelay);
    }
    else if(blockType == REGISTERBANK){
      ((registerBank*)(*block0))->get_currBlock()->undoRetimeRegistersForwards(netlist, numMove, criticalPathDelay, retimeUndoList);
      ((registerBank*)(*block0))->get_currBlock()->updateSinkDelaysAndSlacksBackwards();
      ((registerBank*)(*block0))->get_currBlock()->calculateMaxInputOutputCriticalitiesBackwards(criticalPathDelay);
    }
    else{
      ERROR("Block type is not allowed to be retimed!");
    }
  }
  else if(*moveType == MOVENULL){

  }
  else{
    ERROR("Unknown move type!");
  }
  return true;
}

//We made a change to the system and we would like to undo it.
void undoPipelineAwareMove(int *moveType, Netlist *netlist, void** block0, void** block1, int index0, int index1, enum e_timingMoveBlockType blockType, int numMove,
			     double criticalPathDelay, list <PipeliningNet *> *retimeUndoList){
  logicBlock* lb0;logicBlock* lb1;
  
  // We just did a normal swap operation last time, so undoing it is easy
  // Since before we swapped block 1 might have been empty, now block 0 might be empty, so we put it
  //  as the second argument
  if(*moveType == MOVELOGICBLOCKSELF || *moveType == MOVELOGICBLOCKALLLUTS || *moveType == MOVELOGICBLOCKCLB || 
     *moveType == MOVEPIPEREGSWAPSELF || *moveType == MOVEPIPEREGSWAPCLB){
    makePipelineAwareMove(moveType, netlist, block1, block0, index1, index0, blockType, numMove, criticalPathDelay, NULL);
  }

  //If we split the first N from block 0 into an empty block 1, we should reverse the operation
  // by appending the registers from the parent (block1) back to the child (block0)
  else if(*moveType == MOVEREGEMPTYFIRST){
    *moveType = MOVEREGCHILD;
    makePipelineAwareMove(moveType, netlist, block1, block0, -1, -1, blockType, numMove, criticalPathDelay,  NULL);
  }
  //If we split the last N from block 0 into an empty block 1, we should reverse the operation
  // by appending the registers from the child (block1) back into the parent (block0)
  else if(*moveType == MOVEREGEMPTYLAST){
    *moveType = MOVEREGPARENT;
    makePipelineAwareMove(moveType, netlist, block0, block1, -1, -1, blockType, numMove, criticalPathDelay,  NULL);
  }
  //We moved the first N registers from the child (block1) to the parent (block0) so to reverse the operation
  // we should move the last N registers from the parent to the child
  else if(*moveType == MOVEREGPARENT){
    *moveType = MOVEREGCHILD;
    makePipelineAwareMove(moveType, netlist, block0, block1, -1, -1, blockType, numMove, criticalPathDelay,  NULL);
  }
  //We moved the last N registers from the parent (block0) to the child (block1) so to reverse the operation
  // we should move the first N registers from the child to the parent
  else if(*moveType == MOVEREGCHILD){
    *moveType = MOVEREGPARENT;
    makePipelineAwareMove(moveType, netlist, block0, block1, -1, -1, blockType, numMove, criticalPathDelay,  NULL);
  }
  //If we just duplicated the registers from block 0 into block 1, to reverse this, we just zip them back into block 0
  else if(*moveType == MOVEUNZIPPNET){
    *moveType = MOVEZIPPNET;
    makePipelineAwareMove(moveType, netlist, block0, block1, -1, -1, blockType, numMove, criticalPathDelay,  NULL);
  }
  //If we just zipped the registers from block 1 back into block 0, we can reverse this, but we have to make sure to
  // re-connect the correct sinks (still held by the registers and pnets sitting in block1) back to block 1 for this move to 
  // be accurately reversed.
  else if(*moveType == MOVEZIPPNET){
    *moveType = MOVEREDOUNZIPPNET;
    makePipelineAwareMove(moveType, netlist, block0, block1, -1, -1, blockType, numMove, criticalPathDelay,  NULL);
  }
  else if(*moveType == MOVEREGRETIMEBACK){
    *moveType = MOVEREGUNDOREGRETIMEBACK;
    makePipelineAwareMove(moveType, netlist, block0, block1, -1, -1, blockType, numMove, criticalPathDelay,  retimeUndoList);
  }
  else if(*moveType == MOVEREGRETIMEFORWARD){
    *moveType = MOVEREGUNDOREGRETIMEFORWARD;
    makePipelineAwareMove(moveType, netlist, block0, block1, -1, -1, blockType, numMove, criticalPathDelay,  retimeUndoList);
  }
  else if(*moveType == MOVENULL){

  }
  else{
    ERROR("Unknown move type!");
  }
}

//Evaluate the cost of the given move, but use the "delta delay" method.
//We usually do:
//  timing cost of a source-sink pair = point to point delay * criticality of source-sink pair ^ exp
//  deltaTimingCost = (new delay * new crit)^exp - (old delay * old crit)^exp
//  deltaCost = (alpha * ALL NETS deltaWireCost / ALL NETS oldWireCost) + 
//               [(1-alpha) * ALL NETS deltaTimingCost / ALL NETS oldTimingCost]
//  ALL NETS newTimingCost = ALL NETS oldTimingCost + ALL NETS deltaTimingCost
//
//However, to avoid issues where the delta cost of a net gets weird when we
// update the slack incrementally each time we move a block, we would like to calculate the timing cost
// and delta cost a bit differently.
//
//First, the timing cost is calculated almost the same:
//  timing cost of a source-sink pair = point to point delay * criticality of source-sink pair
//that is, the timing cost is still dependant on the old vs. new criticality (what was potentially making the
//  delta cost weird earlier), but taking out the exponent.
//Thus, the delta timing cost is still basically the same:
// deltaTimingCost = (new delay * new crit) - (old delay * old crit)
//
//The big difference that we make is that we create a second "delta timing cost" equation:
// deltaTimingCost' = (new delay - old delay) * new crit^exp
//
//Then we use this delta timing cost' in our overall delta cost function:
// deltaCost = (alpha * ALL NETS deltaWireCost / ALL NETS oldWireCost) + 
//               [(1-alpha) * ALL NETS deltaTimingCost' / ALL NETS oldTimingCost]
//
//However, the newTimingCost is still calculated the same
// ALL NETS newTimingCost = ALL NETS oldTimingCost + ALL NETS deltaTimingCost
//
double determineDeltaCostDeltaDelay(Netlist *netlist, int moveType, void** block0, void** block1, int index0, int index1, 
			  enum e_timingMoveBlockType blockType, int numMove,
				    bool timing_driven, double criticalPathDelay, list <PipeliningNet *> *retimeUndoList){
  logicBlock* lb0;logicBlock* lb1;
  double deltaWire = 0.0;
  double deltaTiming = 0.0;
  double deltaTimingPrime = 0.0;
  double deltaCost = 0.0;
  double newMaxCut = 0.0;
  double deltaMaxCut = 0.0;
  double newAveCut = 0.0;
  double deltaAveCut = 0.0;

  //Go through the nets associated with the two blocks and record the change in their wirelength and timing.
  //We have to make sure that we do not double-count this change if they are connected to each other!
  if(moveType == MOVELOGICBLOCKCLB){
    //We swap the entire CLB associated with these logic blocks
    if(blockType == REGISTERBANK){
      //We focused on a register bank, so use that as a basis to call the swap function
      if(clb_level_placement){
	((registerBank*)(*block0))->updateCostAndDeltasForEntireStickyCostDeltaDelay((registerBank*)(*block1), timing_driven, criticalPathDelay, 
										     &deltaWire, &deltaTiming, &deltaTimingPrime);
      }
    }
    else if(blockType == LB){
      ((vprLUT*)(*block0))->updateCostAndDeltasForStickyCostDeltaDelay((vprLUT*)(*block1), timing_driven, criticalPathDelay, 
								       &deltaWire, &deltaTiming, &deltaTimingPrime);
    }
    else{
      ERROR("Unknown block type for MOVELOGICBLOCKCLB!");
    }
  }
  else if(moveType == MOVELOGICBLOCKALLLUTS){
    //We have selected a LUT and we have swapped all of the luts in the CLB with the other CLB's luts.
    if(blockType == LB){
      ((vprLUT*)(*block0))->updateCostAndDeltasForLUTGroupAndAssociatedRegBanksCostDeltaDelay((vprLUT*)(*block1), timing_driven, criticalPathDelay, 
											      &deltaWire, &deltaTiming, &deltaTimingPrime);
    }
    else{
      ERROR("Unknown block type for MOVELOGICBLOCKALLLUTS!");
    }
  }
  else if(moveType == MOVELOGICBLOCKSELF){
    // Only swap the specific blocks associated with these logic blocks
    if(blockType == IOPIN){
      lb0 = ((vprIOB*)(*block0))->getfrom_currBlockArray(index0);
      lb1 = ((vprIOB*)(*block1))->getfrom_currBlockArray(index1);
      
      //Since we swapped them, block 0 might be the empty one now!
      //Update the wire and timing costs for the nets that we just moved based on where we moved to and the changes made to the slack vectors
      lb1->updateCostsAndDeltasCostDeltaDelay(timing_driven, criticalPathDelay, &deltaWire, &deltaTiming, &deltaTimingPrime);
      if(lb0)
	lb0->updateCostsAndDeltasCostDeltaDelay(timing_driven, criticalPathDelay, &deltaWire, &deltaTiming, &deltaTimingPrime);    
    }
    else if(blockType == REGISTERBANK){
      ((registerBank*)(*block0))->updateCostAndDeltasAndAssociatedLUTCostDeltaDelay(timing_driven, criticalPathDelay, 
										    &deltaWire, &deltaTiming, &deltaTimingPrime);
      ((registerBank*)(*block1))->updateCostAndDeltasAndAssociatedLUTCostDeltaDelay(timing_driven, criticalPathDelay, &deltaWire, 
										     &deltaTiming, &deltaTimingPrime);
    }
    else if(blockType == LB){
      ((vprLUT*)(*block0))->updateCostAndDeltasCostDeltaDelay(timing_driven, criticalPathDelay, 
							       &deltaWire, &deltaTiming, &deltaTimingPrime);
      ((vprLUT*)(*block1))->updateCostAndDeltasCostDeltaDelay(timing_driven, criticalPathDelay, 
							       &deltaWire, &deltaTiming, &deltaTimingPrime);
    }
    else{
      ERROR("Unknown block type for MOVELOGICBLOCKSELF!");
    }
  }

  //We moved the entire CLBs associated with these register banks
  else if(moveType == MOVEPIPEREGSWAPCLB){
    //We swapped the entire contents of the two CLBs, but did not use any netlist-level clb
    // information for the FFs, only for the LUTs
    ((registerBank*)(*block0))->updateCostAndDeltasForEntireStickyCostDeltaDelay((registerBank*)(*block1), timing_driven, criticalPathDelay, 
										 &deltaWire, &deltaTiming, &deltaTimingPrime);
  }
  //We did something to just the register banks
  else if(moveType == MOVEPIPEREGSWAPSELF || moveType == MOVEREGEMPTYFIRST || moveType == MOVEREGEMPTYLAST ||
	  moveType == MOVEREGPARENT || moveType == MOVEREGCHILD ||
	  moveType == MOVEUNZIPPNET || moveType == MOVEREDOUNZIPPNET){
    //Update the wire and timing costs for the nets we just moved based on where they moved to and the changes made to the slack vectors
    ((registerBank *)(*block0))->updateWireAndTimingCosts(timing_driven, criticalPathDelay);
    ((registerBank *)(*block1))->updateWireAndTimingCosts(timing_driven, criticalPathDelay);
    
    //Figure out what the overall change in wirelength and timing was
    deltaWire += ((registerBank *)(*block0))->determineDeltaWireCost();
    //cout << "Delta Wire after block 0 = " << deltaWire << endl;
    deltaWire += ((registerBank *)(*block1))->determineDeltaWireCost();
    //cout << "Delta Wire after block 1 = " << deltaWire << endl;
    if(timing_driven){
      deltaTiming += ((registerBank *)(*block0))->determineDeltaTimingCost();
      //cout << "Delta Timing after block 0 = " << deltaTiming << endl;
      deltaTimingPrime += ((registerBank *)(*block0))->determineDeltaTimingCostPrime();

      deltaTiming += ((registerBank *)(*block1))->determineDeltaTimingCost();
      //cout << "Delta Timing after block 1 = " << deltaTiming << endl;
      deltaTimingPrime += ((registerBank *)(*block1))->determineDeltaTimingCostPrime();
    }
  }
  else if(moveType == MOVENULL){

  }
  else{
    ERROR("Delta delay moves not compatible with zipping or retiming moves");
  }

  if(timing_driven){
    //If we are timing-driven, we have to consider both the timing impact and wirelength impact of this move
    deltaCost = tdPlacementLambda * (deltaTimingPrime / netlist->getTimingCost()) + ((1.0 - tdPlacementLambda) * (deltaWire / netlist->getWireCost()));

    netlist->setWireCost(deltaWire + netlist->getWireCost());
    netlist->setTimingCost(deltaTiming + netlist->getTimingCost());

    return(deltaCost);
  }
  else{
    ERROR("Delta delays moves should not be used with congestion-driven placement!");
  }
}

//We've made a given move, now we just have to compute the delta cost
double determineDeltaCost(Netlist *netlist, int moveType, void** block0, void** block1, 
			  int index0, int index1, enum e_timingMoveBlockType blockType, int numMove,
			  bool timing_driven, double criticalPathDelay, list <PipeliningNet *> *retimeUndoList){
  logicBlock* lb0;logicBlock* lb1;
  double deltaWire = 0.0;
  double deltaTiming = 0.0;
  double deltaCost = 0.0;
  double newMaxCut = 0.0;
  double deltaMaxCut = 0.0;
  double newAveCut = 0.0;
  double deltaAveCut = 0.0;

  //Go through the nets associated with the two blocks and record the change in their wirelength and timing.
  //We have to make sure that we do not double-count this change if they are connected to each other!
  if(moveType == MOVELOGICBLOCKCLB){
    //We swap the entire CLB associated with these logic blocks
    if(blockType == REGISTERBANK){
      //We focused on a register bank, so use that as a basis to call the swap function
      if(clb_level_placement){
	((registerBank*)(*block0))->updateCostAndDeltasForEntireSticky((registerBank*)(*block1), timing_driven, criticalPathDelay, 
								       &deltaWire, &deltaTiming);
      }
    }
    else if(blockType == LB){
      ((vprLUT*)(*block0))->updateCostAndDeltasForSticky((vprLUT*)(*block1), timing_driven, criticalPathDelay, 
							 &deltaWire, &deltaTiming);
    }
    else{
      ERROR("Unknown block type for MOVELOGICBLOCKCLB!");
    }
  }
 else if(moveType == MOVELOGICBLOCKALLLUTS){
    //We have selected a LUT and we have swapped all of the luts in the CLB with the other CLB's luts.
    if(blockType == LB){
      ((vprLUT*)(*block0))->updateCostAndDeltasForLUTGroupAndAssociatedRegBanks((vprLUT*)(*block1), timing_driven, criticalPathDelay, 
										&deltaWire, &deltaTiming);
    }
    else{
      ERROR("Unknown block type for MOVELOGICBLOCKALLLUTS!");
    }
  }
 else if(moveType == MOVELOGICBLOCKSELF){
    // Only swap the specific blocks associated with these logic blocks
    if(blockType == IOPIN){
      lb0 = ((vprIOB*)(*block0))->getfrom_currBlockArray(index0);
      lb1 = ((vprIOB*)(*block1))->getfrom_currBlockArray(index1);
      
      //Since we swapped them, block 0 might be the empty one now!
      //Update the wire and timing costs for the nets that we just moved based on where we moved to and the changes made to the slack vectors
      lb1->updateCostsAndDeltas(timing_driven, criticalPathDelay, &deltaWire, &deltaTiming);
      if(lb0)
	lb0->updateCostsAndDeltas(timing_driven, criticalPathDelay, &deltaWire, &deltaTiming);    
    }
    else if(blockType == REGISTERBANK){
      ((registerBank*)(*block0))->updateCostAndDeltasAndAssociatedLUT(timing_driven, criticalPathDelay, &deltaWire, &deltaTiming);
      ((registerBank*)(*block1))->updateCostAndDeltasAndAssociatedLUT(timing_driven, criticalPathDelay, &deltaWire, &deltaTiming);
    }
    else if(blockType == LB){
      ((vprLUT*)(*block0))->updateCostAndDeltas(timing_driven, criticalPathDelay, 
						 &deltaWire, &deltaTiming);
      ((vprLUT*)(*block1))->updateCostAndDeltas(timing_driven, criticalPathDelay, 
						 &deltaWire, &deltaTiming);
    }
    else{
      ERROR("Unknown block type for MOVELOGICBLOCKSELF!");
    }
  }

  //We moved the entire CLBs associated with these register banks
  else if(moveType == MOVEPIPEREGSWAPCLB){
    //We swapped the entire contents of the two CLBs, but did not use any netlist-level clb
    // information for the FFs, only for the LUTs
    ((registerBank*)(*block0))->updateCostAndDeltasForEntireSticky((registerBank*)(*block1), timing_driven, criticalPathDelay, 
								   &deltaWire, &deltaTiming);
  }
  //We did something to just the register banks
  else if(moveType == MOVEPIPEREGSWAPSELF || moveType == MOVEREGEMPTYFIRST || moveType == MOVEREGEMPTYLAST ||
	  moveType == MOVEREGPARENT || moveType == MOVEREGCHILD ||
	  moveType == MOVEUNZIPPNET || moveType == MOVEREDOUNZIPPNET){

    ((registerBank*)(*block0))->updateCostAndDeltas(timing_driven, criticalPathDelay, 
						    &deltaWire, &deltaTiming);
    ((registerBank*)(*block1))->updateCostAndDeltas(timing_driven, criticalPathDelay, 
						    &deltaWire, &deltaTiming);

//     if(!clb_level_placement){
//       if(((registerBank*)(*block0))->getFirstSink()){
// 	if(((registerBank*)(*block0))->getFirstSink()->name.compare("i_39__0Reg3") == 0 ||
// 	   ((registerBank*)(*block0))->getFirstSink()->name.compare("i_58__0Reg3") == 0 ||
// 	   ((registerBank*)(*block0))->getFirstSink()->name.compare("i_58__0Reg7") == 0 ||
// 	   ((registerBank*)(*block0))->getFirstSink()->name.compare("i_43__0Reg3") == 0 ||
// 	   ((registerBank*)(*block0))->getFirstSink()->name.compare("i_14__8:ble_2Reg0") == 0 ||
// 	   ((registerBank*)(*block0))->getFirstSink()->name.compare("o_42__2:ble_3Reg0") == 0 ||
// 	   ((registerBank*)(*block0))->getFirstSink()->name.compare("n_n344_1Reg3") == 0 ||
// 	   ((registerBank*)(*block0))->getFirstSink()->name.compare("sink n_n293_4:ble_2Reg0") == 0 ||
// 	   ((registerBank*)(*block0))->getFirstSink()->name.compare("[827]_2:ble_3Reg0") == 0 ||
// 	   ((registerBank*)(*block0))->getFirstSink()->name.compare("[132]_1Reg0") == 0 ||
// 	   ((registerBank*)(*block0))->getFirstSink()->name.compare("i_35__4:ble_2Reg0") == 0 ||
// 	   ((registerBank*)(*block0))->getFirstSink()->name.compare("n_n297_1Reg3") == 0){
// 	  cout << "Moving " << ((registerBank*)(*block0))->getFirstSink()->name << ", deltaWire = " << deltaWire << " deltaTiming = " << deltaTiming <<endl;
// 	  cout << "\tStarting location was (" << ((registerBank*)(*block0))->getHostCLB()->col << "," << ((registerBank*)(*block0))->getHostCLB()->row << ")" << endl;
// 	  cout << "\tEnding location was swapped with ";
// 	  if(((registerBank*)(*block1))->getFirstSink()){
// 	    cout << "register " << ((registerBank*)(*block1))->getFirstSink()->name;
// 	    cout << " at (" << ((registerBank*)(*block1))->getHostCLB()->col << "," << ((registerBank*)(*block1))->getHostCLB()->row << ")" << endl; 
// 	  }
// 	  else{
// 	    cout << "empty register at (" << ((registerBank*)(*block1))->getHostCLB()->col << "," << ((registerBank*)(*block1))->getHostCLB()->row << ")" << endl;
// 	  }
// 	}
//       }
//       if(((registerBank*)(*block1))->getFirstSink()){
// 	if(((registerBank*)(*block1))->getFirstSink()->name.compare("i_39__0Reg3") == 0 ||
// 	   ((registerBank*)(*block1))->getFirstSink()->name.compare("i_58__0Reg3") == 0 ||
// 	   ((registerBank*)(*block1))->getFirstSink()->name.compare("i_58__0Reg7") == 0 ||
// 	   ((registerBank*)(*block1))->getFirstSink()->name.compare("i_43__0Reg3") == 0 ||
// 	   ((registerBank*)(*block1))->getFirstSink()->name.compare("i_14__8:ble_2Reg0") == 0 ||
// 	   ((registerBank*)(*block1))->getFirstSink()->name.compare("o_42__2:ble_3Reg0") == 0 ||
// 	   ((registerBank*)(*block1))->getFirstSink()->name.compare("n_n344_1Reg3") == 0 ||
// 	   ((registerBank*)(*block1))->getFirstSink()->name.compare("sink n_n293_4:ble_2Reg0") == 0 ||
// 	   ((registerBank*)(*block1))->getFirstSink()->name.compare("[827]_2:ble_3Reg0") == 0 ||
// 	   ((registerBank*)(*block1))->getFirstSink()->name.compare("[132]_1Reg0") == 0 ||
// 	   ((registerBank*)(*block1))->getFirstSink()->name.compare("i_35__4:ble_2Reg0") == 0 ||
// 	   ((registerBank*)(*block1))->getFirstSink()->name.compare("n_n297_1Reg3") == 0){
// 	  cout << "Moving " << ((registerBank*)(*block1))->getFirstSink()->name << ", deltaWire = " << deltaWire << " deltaTiming = " << deltaTiming <<endl;
// 	  cout << "\tStarting location was (" << ((registerBank*)(*block1))->getHostCLB()->col << "," << ((registerBank*)(*block1))->getHostCLB()->row << ")" << endl; 
// 	  cout << "\tEnding location was swapped with ";
// 	  if(((registerBank*)(*block0))->getFirstSink()){
// 	    cout << "register " << ((registerBank*)(*block0))->getFirstSink()->name;
// 	    cout << " at (" << ((registerBank*)(*block0))->getHostCLB()->col << "," << ((registerBank*)(*block0))->getHostCLB()->row << ")" << endl; 
// 	  }
// 	  else{
// 	    cout << "empty register at (" << ((registerBank*)(*block0))->getHostCLB()->col << "," << ((registerBank*)(*block0))->getHostCLB()->row << ")" << endl;
// 	  }
// 	}
//       }
//     }
  }

  else if(moveType == MOVEZIPPNET){
    ((registerBank *)(*block0))->updateWireAndTimingCosts(timing_driven, criticalPathDelay);

    //If we are keeping track of the X cut costs, we need to remove the contribution of the nets
    // driven my the registers we are deleting
    if(maxCutCost){
      ((registerBank *)(*block1))->updateDeletedCutCost();
    }

    //Figure out what the overall change in wirelength and timing was
    deltaWire += ((registerBank *)(*block0))->determineDeltaWireCost();
    //cout << "Delta Wire after block 0 = " << deltaWire << endl;

    //We have to remove the cost of the nets we don't need anymore - this is all the nets in block 1
    deltaWire -= ((registerBank *)(*block1))->returnTotalWireCost();
    //cout << "Delta Wire after block 1 = " << deltaWire << endl;

    if(timing_driven){
      deltaTiming += ((registerBank *)(*block0))->determineDeltaTimingCost();
      //cout << "Delta Timing after block 0 = " << deltaTiming << endl;

      //We have to remove the cost of the nets we don't need anymore - this is all the nets in block 1
      deltaTiming -= ((registerBank *)(*block1))->returnTotalTimingCost();
      //cout << "Delta Timing after block 1 = " << deltaTiming << endl;
    }
  }

  else if(moveType == MOVEREGRETIMEBACK || moveType == MOVEREGUNDOREGRETIMEFORWARD){
    if(moveType == MOVEREGUNDOREGRETIMEFORWARD){
      if(blockType == LB){
	lb0=((vprLUT*)(*block1))->get_currBlock();
      }
      else if(blockType == REGISTERBANK){
	lb0=((registerBank*)(*block1))->get_currBlock();
      }
      else{
	ERROR("Block type not allowed to be retimed!");
      }
    }
    else{
      if(blockType == LB){
	lb0=((vprLUT*)(*block0))->get_currBlock();
      }
      else if(blockType == REGISTERBANK){
	lb0=((registerBank*)(*block0))->get_currBlock();
      }
      else{
	ERROR("Block type not allowed to be retimed!");
      }
    }

    //This takes care of the nets we just made on the output of the block
    lb0->updateWireAndTimingCostsBackwards(timing_driven, criticalPathDelay);

   //If we are keeping track of the X cut costs, we need to remove the contribution of the nets
    // driven my the registers we are deleting
    if(maxCutCost){      
      updateDeletedCutCostForwards(retimeUndoList);
    }

    //Figure out what the overall change in wirelength and timing was
    deltaWire += lb0->determineDeltaWireCostBackwards();
    //cout << "Delta Wire after logic block = " << deltaWire << endl;
    
    //We have to remove the cost of the nets we don't need anymore - this is all the nets referred to in the undo list
    deltaWire -= returnTotalWireCostForwards(retimeUndoList);
    //cout << "Delta Wire after undo list = " << deltaWire << endl;
    
    //We might have created pipelined nets for non-pipelined nets.  Take account of these.
    deltaWire -= netlist->wireCostSurplus;
    netlist->wireCostSurplus = 0.0;

    if(timing_driven){
      deltaTiming += lb0->determineDeltaTimingCostBackwards();
      //cout << "Delta Timing after logic block = " << deltaTiming << endl;

      //We have to remove the cost of the nets we don't need anymore - this is all the nets referred to in the undo list
      deltaTiming -= returnTotalTimingCostForwards(retimeUndoList);
      //cout << "Delta Timing after undo list = " << deltaTiming << endl;

      deltaTiming -= netlist->timingCostSurplus;
      netlist->timingCostSurplus = 0.0;
    }
  }
  else if(moveType == MOVEREGRETIMEFORWARD || moveType == MOVEREGUNDOREGRETIMEBACK){
    if(moveType == MOVEREGUNDOREGRETIMEBACK){
      if(blockType == LB){
	lb0=((vprLUT*)(*block1))->get_currBlock();
      }
      else if(blockType == REGISTERBANK){
	lb0=((registerBank*)(*block1))->get_currBlock();
      }
      else{
	ERROR("Block type not allowed to be retimed!");
      }
    }
    else{
      if(blockType == LB){
	lb0=((vprLUT*)(*block0))->get_currBlock();
      }
      else if(blockType == REGISTERBANK){
	lb0=((registerBank*)(*block0))->get_currBlock();
      }
      else{
	ERROR("Block type not allowed to be retimed!");
      }
    }

    //This takes care of the nets we just made on the output of the block
    lb0->updateWireAndTimingCostsForwards(timing_driven, criticalPathDelay);

    //If we are keeping track of the X cut costs, we need to remove the contribution of the nets
    // driven my the registers we are deleting
    if(maxCutCost){      
      updateDeletedCutCostBackwards(retimeUndoList, lb0);
    }

    //Figure out what the overall change in wirelength and timing was
    deltaWire += lb0->determineDeltaWireCostForwards();
    //cout << "Delta Wire after logic block = " << deltaWire << endl;
    
    // We have to remove the cost of the nets we don't need anymore 
    deltaWire -= returnTotalWireCostBackwards(retimeUndoList, timing_driven, criticalPathDelay, lb0);
    //cout << "Delta Wire after undo list = " << deltaWire << endl;

    //We might have had orphaned nets due to fan-out/undoing retiming or we might have created
    // pipelined nets for non-pipelined nets.  Take account of these.
    deltaWire -= netlist->wireCostSurplus;
    netlist->wireCostSurplus = 0.0;
    //cout << "Delta Wire after surplus = " << deltaWire << endl;
    
    if(timing_driven){
      deltaTiming += lb0->determineDeltaTimingCostForwards();
      //cout << "Delta Timing after logic block = " << deltaTiming << endl;

      //We have to remove the cost of the nets we don't need anymore - this is all the nets referred to in the undo list
      deltaTiming -= returnTotalTimingCostBackwards(retimeUndoList, lb0);
      //cout << "Delta Timing after undo list = " << deltaTiming << endl;

      deltaTiming -= netlist->timingCostSurplus;
      netlist->timingCostSurplus = 0.0;
      //cout << "Delta Timing after surplus = " << deltaTiming << endl;
    }
  }

  else if(moveType == MOVEREGRETIMEBACKABORT){
    if(blockType == LB){
      lb0=((vprLUT*)(*block0))->get_currBlock();
    }
    else if(blockType == REGISTERBANK){
      lb0=((registerBank*)(*block0))->get_currBlock();
    }
    else{
      ERROR("Block type not allowed to be retimed!");
    }    

    //This takes care of the nets we just made on the output of the block
    lb0->updateWireAndTimingCostsForwards(timing_driven, criticalPathDelay);

    //Figure out what the overall change in wirelength and timing was,
    // but throw it out.
    lb0->determineDeltaWireCostForwards();
    
    netlist->wireCostSurplus = 0.0;

    if(timing_driven){
      lb0->determineDeltaTimingCostForwards();
      netlist->timingCostSurplus = 0.0;
    }
  }

  else if(moveType == MOVEREGRETIMEFORWARDABORT){
    if(blockType == LB){
      lb0=((vprLUT*)(*block0))->get_currBlock();
    }
    else if(blockType == REGISTERBANK){
      lb0=((registerBank*)(*block0))->get_currBlock();
    }
    else{
      ERROR("Block type not allowed to be retimed!");
    }    
    //This takes care of the nets we just made on the output of the block
    lb0->updateWireAndTimingCostsBackwards(timing_driven, criticalPathDelay);

    //Figure out what the overall change in wirelength and timing was,
    // but throw it out.
    lb0->determineDeltaWireCostBackwards();

    netlist->wireCostSurplus = 0.0;

    if(timing_driven){
      lb0->determineDeltaTimingCostBackwards();
      netlist->timingCostSurplus = 0.0;
    }
  }

  else if(moveType == MOVENULL){
    
  }

  else{
    ERROR("Unknown move type");
  }

  //Figure out the new max cut  
  if(maxCutCost){
    //Go through the cut cost vector to determine the largest value
    for(vector <double>::iterator iter = FPGA->cutCostVector.begin(); iter != FPGA->cutCostVector.end(); iter++){
      //printf("Cut Cost = %d\n", (int) *iter);
      if(aveCutCost){
	newAveCut += *iter;
      }

      if(newMaxCut < *iter){
	//printf("New Max Cut = %f\n", *iter);
	newMaxCut = *iter;
      }
    }
    deltaMaxCut = newMaxCut - netlist->maxCut;

    if(aveCutCost){
      newAveCut /= FPGA->cutCostVector.size();
    }
    deltaAveCut = newAveCut - netlist->aveCut;
  } 

  //cout << "Delta Wire = " << deltaWire << endl;
  //cout << "Delta Timing = " << deltaTiming << endl;
  //if(maxCutCost){
    //cout << "Delta MaxCut = " << deltaMaxCut << endl;
  //}
  //if(aveCutCost){
    //cout << "Delta AveCut = " << deltaAveCut << endl;
  //}

  if(deltaTiming > 0 && moveType != MOVEPIPEREGSWAPCLB){
    deltaTiming = deltaTiming;
    //printPlacement(netlist);
  }
  if(deltaTiming > 0){
    //Moving towards the center space with a positive cost
    if(moveType == MOVEPIPEREGSWAPCLB && 
       (((registerBank*)*block1)->getHostCLB()->get_row() == 13 || ((registerBank*)*block1)->getHostCLB()->get_row() == 14 ||
	((registerBank*)*block1)->getHostCLB()->get_row() == 12) &&
       (((registerBank*)*block1)->getHostCLB()->get_col() == 8 ||  ((registerBank*)*block1)->getHostCLB()->get_col() == 9)){
      deltaTiming = deltaTiming;
      //printPlacement(netlist);
    }
  }

  if(deltaTiming < 0){
    //Moving away from the center space with a cheaper cost
    if(moveType == MOVEPIPEREGSWAPCLB && 
       (((registerBank*)*block0)->getHostCLB()->get_row() == 13 || ((registerBank*)*block0)->getHostCLB()->get_row() == 14 ||
	((registerBank*)*block0)->getHostCLB()->get_row() == 12) &&
       (((registerBank*)*block0)->getHostCLB()->get_col() == 8 || ((registerBank*)*block0)->getHostCLB()->get_col() == 9)){
      deltaTiming = deltaTiming;
      //printPlacement(netlist);
    }
  }

  if(timing_driven){
    if(maxCutCost && !aveCutCost){
      //If we are timing-driven, we have to consider both the timing impact and wirelength impact of this move
      //However, if we are paying attention to max cut, we have to incorporate that too.
      deltaCost = tdPlacementLambda * (deltaTiming / netlist->getTimingCost()) + 
	((1.0 - tdPlacementLambda) * (0.7*(deltaWire / netlist->getWireCost()) + 0.3 * (deltaMaxCut/ netlist->maxCut)));
      netlist->maxCut = newMaxCut;
    }
    else if(maxCutCost && aveCutCost){
      deltaCost = tdPlacementLambda * (deltaTiming / netlist->getTimingCost()) + 
	((1.0 - tdPlacementLambda) * (0.7*(deltaAveCut / netlist->aveCut) + 0.3 * (deltaMaxCut/ netlist->maxCut)));
      netlist->maxCut = newMaxCut;
      netlist->aveCut = newAveCut;
    }
    else{
      //If we are timing-driven, we have to consider both the timing impact and wirelength impact of this move
      deltaCost = tdPlacementLambda * (deltaTiming / netlist->getTimingCost()) + ((1.0 - tdPlacementLambda) * (deltaWire / netlist->getWireCost()));
    }
    netlist->setWireCost(deltaWire + netlist->getWireCost());
    netlist->setTimingCost(deltaTiming + netlist->getTimingCost());
    //cout << "Delta Wire Cost = " << deltaWire << endl;
    //cout << "Wire Cost = " << netlist->getWireCost() << endl;
    //cout << "Delta Timing Cost = " << deltaTiming << endl;
    //cout << "Timing Cost = " << netlist->getTimingCost() << endl;
    //cout << "DeltaCost = " << deltaCost << endl << endl <<endl;

    return(deltaCost);
  }
  else{
    if(maxCutCost && !aveCutCost){
      //If we are not timing-driven, only consider the max cut and wirelength impact
      netlist->setWireCost(deltaWire + netlist->getWireCost());
      netlist->maxCut = newMaxCut;

      deltaCost = (0.7 * deltaWire) + (0.3 * deltaMaxCut);
      return deltaCost;
    }
    else if(maxCutCost && aveCutCost){
      //If we are not timing-driven, only consider the max cut and ave cut
      netlist->setWireCost(deltaWire + netlist->getWireCost());
      netlist->maxCut = newMaxCut;
      netlist->aveCut = newAveCut;
      
      deltaCost = (0.7 * deltaAveCut) + (0.3 * deltaMaxCut);
      return deltaCost;
    }
    else{
      //If we are not timing-driven, only consider the wirelength impact
      netlist->setWireCost(deltaWire + netlist->getWireCost());
      //cout << "DeltaCost = " << deltaWire << endl;
      return deltaWire;
    }
  }
}

//We removed all of these nets up to the nerfed registers (temporarily) 
// so decrement the cut costs associated with these nets
void updateDeletedCutCostForwards(list <PipeliningNet *> *retimeUndoList){
  //Decrement the span of this pipelining net
  for(list <PipeliningNet *>::iterator pnIter = retimeUndoList->begin(); pnIter != retimeUndoList->end(); pnIter++){
    (*pnIter)->decrementNerfedSpanForwards();
  }
}

//We removed all of these nets up to the nerfed registers (temporarily) 
// so decrement the cut costs associated with these nets
void updateDeletedCutCostBackwards(list <PipeliningNet *> *retimeUndoList, logicBlock *lb){
  list <PipeliningNet *>::iterator pnIter;
  list<Terminal*>::iterator sinkIter;

  //Decrement the span of this pipelining net
  for(pnIter = retimeUndoList->begin(), sinkIter = lb->inputList.begin(); 
      pnIter != retimeUndoList->end(); pnIter++, sinkIter++){
    (*pnIter)->decrementNerfedSpanBackwards(*sinkIter);
  }
}

//We removed all of these nets up to the nerfed registers (temporarily)
// so we need to know what their wire cost was
double returnTotalWireCostForwards(list <PipeliningNet *> *retimeUndoList){
  double wireCost = 0.0;
  
  //Go through each of these pipelining nets, up to the nerfed registers,
  // and figure out their wire cost
  for(list <PipeliningNet *>::iterator pnIter = retimeUndoList->begin(); pnIter != retimeUndoList->end(); pnIter++){
    wireCost += (*pnIter)->returnNerfedWireCostForwards();
  }
  return wireCost;
}

//We removed all of these nets up to the nerfed registers (temporarily)
// so we need to know what their wire cost was.
//Since we might have orphaned a netlist, this will also include an evaluation of the last, unnerfed net.
//Since this actually does the reCalculatePlacementCost, it must be used before the returnNerfedTimingCostBackwards function.
double returnTotalWireCostBackwards(list <PipeliningNet *> *retimeUndoList, bool timing_driven, double criticalPathDelay, logicBlock *lb){
  list <PipeliningNet *>::iterator pnIter;
  list<Terminal*>::iterator sinkIter;
  double wireCost = 0.0;
  
  //Go through each of these pipelining nets, up to the nerfed registers,
  // and figure out their wire cost
  for(pnIter = retimeUndoList->begin(), sinkIter = lb->inputList.begin(); 
      pnIter != retimeUndoList->end(); pnIter++, sinkIter++){
    wireCost += (*pnIter)->returnNerfedWireCostBackwards(timing_driven, criticalPathDelay, *sinkIter);
  }
  return wireCost;
}

double returnTotalTimingCostForwards(list <PipeliningNet *> *retimeUndoList){
  double timingCost = 0.0;
  
  //Go through each of these pipelining nets, up to the nerfed registers,
  // and figure out their wire cost
  for(list <PipeliningNet *>::iterator pnIter = retimeUndoList->begin(); pnIter != retimeUndoList->end(); pnIter++){
    timingCost += (*pnIter)->returnNerfedTimingCostForwards();
  }
  return timingCost;
}

double returnTotalTimingCostBackwards(list <PipeliningNet *> *retimeUndoList, logicBlock *lb){
  list <PipeliningNet *>::iterator pnIter;
  list <Terminal *>:: iterator sinkIter;
  double timingCost = 0.0;
  
  //Go through each of these pipelining nets, up to the nerfed registers,
  // and figure out their wire cost
  for(pnIter = retimeUndoList->begin(), sinkIter = lb->inputList.begin(); 
      pnIter != retimeUndoList->end(); pnIter++, sinkIter++){
    timingCost += (*pnIter)->returnNerfedTimingCostBackwards(*sinkIter);
  }
  return timingCost;
}

//We have decided to accept the retiming move, so get rid of the nets
// and nerfed registers.
void clearRetimeUndoListForward(list <PipeliningNet *> *retimeUndoList){
  //Go through each of these pipelining nets, up to the nerfed registers,
  // and start deleting nets and registers
  for(list <PipeliningNet *>::iterator pnIter = retimeUndoList->begin(); pnIter != retimeUndoList->end(); pnIter++){
    (*pnIter)->deleteNerfedNetsAndRegistersForwards();
    //cout << "Freeing net " << (*pnIter)->name << endl;
    delete (*pnIter);
  }

  retimeUndoList->clear();
}

//We have decided to accept the retiming move, so get rid of the nets
// and nerfed registers.
void clearRetimeUndoListBackward(list <PipeliningNet *> *retimeUndoList, void *block, enum e_timingMoveBlockType blockType){
  logicBlock *lb;
  list <PipeliningNet *>::iterator pnIter;
  list <Terminal *>:: iterator sinkIter;

  if(blockType == LB){
    lb = ((vprLUT*)block)->get_currBlock();
  }
  else if(blockType == REGISTERBANK){
    lb = ((registerBank*)block)->get_currBlock();
  }
  else{
    ERROR("Block type not allowed to be retimed!");
  }

  //Go through each of these pipelining nets, up to the nerfed registers,
  // and start deleting nets and registers.
  //Notice that there may be fewer nets in the retime undo list than there are
  // input pins.  That's because we might have aborted a retiming.
  for(pnIter = retimeUndoList->begin(), sinkIter = lb->inputList.begin(); 
      pnIter != retimeUndoList->end(); pnIter++, sinkIter++){
    (*pnIter)->deleteNerfedNetsAndRegistersBackwards(*sinkIter);
  }

  retimeUndoList->clear();
}

//We have decided to accept the retiming move, so un-nerf the registers
// upstream of this logic block
void clearNerfedBlocksBackward(void* block0, enum e_timingMoveBlockType blockType){
  logicBlock *lb;

  if(blockType == LB){
    lb=((vprLUT*)block0)->get_currBlock();
  }
  else if(blockType == REGISTERBANK){
    lb=((registerBank*)block0)->get_currBlock();
  }
  else{
    ERROR("Block type not allowed to be retimed!");
  }
  lb->clearNerfedBlocksBackward();
}

//We have decided to accept the retiming move, so un-nerf the registers
// downstream of this logic block
void clearNerfedBlocksForward(void* block0, enum e_timingMoveBlockType blockType){
  logicBlock *lb;
  if(blockType == LB){
    lb=((vprLUT*)block0)->get_currBlock();
  }
  else if(blockType == REGISTERBANK){
    lb=((registerBank*)block0)->get_currBlock();
  }
  else{
    ERROR("Block type not allowed to be retimed!");
  }
  lb->clearNerfedBlocksForward();
}

// Anneal the netlist
void pipelineAnneal(Netlist *netlist, bool timing_driven, double initialTemperature, float startRLim){
  void* block0;
  void* block1;
  int indexOrBank0, indexOrBank1;
  int moveType;
  enum e_timingMoveBlockType blockType;
  int numMove;
  double dCost;
  float rLim = startRLim;
  int movesRejected;
  int nonZeroMovesAccepted;
  double criticalPathDelay;
  double deltaCostAccept;
  double deltaCostReject;
  double averageDeltaCostAccept;
  double averageDeltaCostNonZeroAccept;
  double averageDeltaCostReject;
  int numMoveLogic;
  int numRegSwap;
  int numRegSplit;
  int numRegMerge;
  int numZip;
  int numUnzip;
  int numMoveLogicReject;
  int numRegSwapReject;
  int numRegSplitReject;
  int numRegMergeReject;
  int numZipReject;
  int numUnzipReject;
  double aveRegSplit;
  double aveRegMerge;

  int numMoveRegRetime;
  double aveRegRetime;
  int numRegRetimeReject;
  double aveCostRegRetime;

  list <PipeliningNet *> retimeUndoList;

  bool moveSucceeded;
  int numMovesFailed;

  double averageOverallCost;
  double stdDevOverallCost;

#if defined(PARANOID_CHECK_PIPEPLACE)
  double oldWireCost;
  double oldTimingCost;
  double oldMaxCut;
  double oldAveCut;
  double reallyOldWireCost;
  double reallyOldTimingCost;
  double reallyOldMaxCut;
  double reallyOldAveCut;  
#endif
  
  double currTemp = initialTemperature;
  bool done = false;

#ifdef BREAKONANNEALNUMBER
  int numberMoves = 0;
#endif

  // The number of moves per temperature is # moveable blocks ^ (4/3)
  int numMovesPerTemp = (int)(placeInnerLoop * pow((double)netlist->calculateNumBlocksAndPipeliningRegisters(), 1.33));
  int numMovesBeforeUpdate = (int)(tdPlacementUpdateInterval * (double) numMovesPerTemp);

  if(numMovesBeforeUpdate <= 0){
    numMovesBeforeUpdate = 1;
  }


  // Calculate everything from scratch before we begin
  criticalPathDelay =  calculateCriticalPathDelay(netlist, true);
  netlist->calculateMaxInputOutputCriticalities(criticalPathDelay);
  netlist->calculatePlacementCost(timing_driven, criticalPathDelay);

  printf("Moves per Temperature iteration = %d\n", numMovesPerTemp);

#if defined(PIPEPLACEDEBUG)
  if(timing_driven){
    if(!splitDelayBlocks && !unzipNets){
      printf("\n\nTemp\t\tMovAcc\t\tNZMoAcc\t\tCPDelay\tWCost\t\tAvgOverallCost\tTimingCost\tAveAccDCost\tAveNZAccDCost\tAveRejDCost\n");
      printf("--------------------------------------------------------------------------------------------------\n");
    }
    else if(splitDelayBlocks && !unzipNets){
      printf("\n\nTemp\t\tMovAcc\t\tNZMoAcc\t\tCP Delay\tWCost\t\tAvgOverallCost\tTimingCost\tAveAccDCost\tAveNZAccDCost\tAveRejDCost\tLogic\tRSwap\tRSplit\tRMerge\tLAcc\tRSwAcc\tRSpAcc\tRMAcc\tAveRSp\tAveRMer\tSpMerR\n");
      printf("------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------\n");
    }
    else if(splitDelayBlocks && unzipNets){
//       printf("\n\nTemp\t\tMovAcc\t\tNZMoAcc\t\tCP Delay\tWCost\t\tTimingCost\tAvgOverallCost\tAveAccDCost\tAveNZAccDCost\tAveRejDCost\tLogic\tRSwap\tRSplit\tRMerge\tZip\tUnzip\tLAcc\tRSwAcc\tRSpAcc\tRMAcc\tZAcc\tUZAcc\tAveRSp\tAveRMer\n");
//       printf("------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------\n");
      printf("\n\nTemp\t\tMovAcc\t\tNZMoAcc\t\tCP Delay\tWCost\t\tTimingCost\tAvgOverallCost\tAveAccDCost\tAveNZAccDCost\tAveRejDCost\tLogic\tRSwap\tRSplit\tRMerge\tZip\tUnzip\tLAcc\tRSwAcc\tRSpAcc\tRMAcc\tZAcc\tUZAcc\n");
      printf("------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------\n");
    }
  }
  else{
    if(!splitDelayBlocks && !unzipNets){
      printf("\n\nTemp\t\tMovAcc\tNZMoAcc\tCP Delay\tWCost\tAveAccDCost\tAveNZAccDCost\tAveRejDCost");
      printf("--------------------------------------------------------------------------------------------------\n");    
    }
    else if(splitDelayBlocks && !unzipNets){
      printf("\n\nTemp\t\tMovAcc\tNZMoAcc\tCP Delay\tWCost\tAveAccDCost\tAveNZAccDCost\tAveRejDCost\tLogic\tRSwap\tRSplit\tRMerge\tLAcc\tRSwAcc\tRSpAcc\tRMAcc\tAveRSp\tAveRMerge\tSpMerR\n");
      printf("--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------\n");
    }
    else if(splitDelayBlocks && unzipNets){
      printf("\n\nTemp\t\tMovAcc\tNZMoAcc\tCP Delay\tWCost\tAveAccDCost\tAveNZAccDCost\tAveRejDCost\tLogic\tRSwap\tRSplit\tRMerge\tZip\tUnzip\tLAcc\tRSwAcc\tRSpAcc\tRMAcc\tZAcc\tUZAcc\tAveRSp\tAveRMerge\tSpMerR\n");
      printf("--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------\n");    
    }
  }
#else
  if(timing_driven ){
    printf("\n\nTemp\t\tMovAcc\t\tNZMoAcc\t\tCP Delay\tWCost\t\tTimingCost\tAvgOverallCost\tAveAccDCost\tAveNZAccDCost\tAveRejDCost\trLim\tTExp\n");
    printf("--------------------------------------------------------------------------------------------------\n");
  }
  else{
    printf("\n\nTemp\t\tMovAcc\t\tNZMoAcc\t\tCP Delay\tWCost\t\tAveAccDCost\tAveNZAccDCost\tAveRejDCost");
    printf("--------------------------------------------------------------------------------------------------\n");    
  }
#endif

  while(!done){
    movesRejected = 0;
    nonZeroMovesAccepted = 0;

    averageDeltaCostAccept = 0.0;
    averageDeltaCostNonZeroAccept = 0.0;
    averageDeltaCostReject = 0.0;

    deltaCostAccept = 0.0;
    deltaCostReject = 0.0;
    
    numMoveLogic = 0; numRegSwap = 0; numRegSplit = 0; numRegMerge = 0; 
    numMoveLogicReject = 0; numRegSwapReject = 0; numRegSplitReject = 0; numRegMergeReject = 0; 
    numZip = 0; numUnzip = 0; numZipReject = 0; numUnzipReject = 0;
    aveRegSplit = 0.0; aveRegMerge = 0.0;
    numMoveRegRetime = 0;
    aveRegRetime = 0.0;
    numRegRetimeReject = 0;
    aveCostRegRetime = 0.0;

    numMovesFailed = 0;

    averageOverallCost = 0.0;
    stdDevOverallCost = 0.0;

    numAttemptedLUTMoves = 0;
    numAttemptedRegMoves = 0;
    numAttemptedIOMoves = 0;
    //averageAttemptedLUTMoveManhattanDistance = 0;
    //averageAttemptedRegMoveManhattanDistance = 0;
    //averageAttemptedIOMoveManhattanDistance = 0;

    if(timing_driven){
      netlist->overallCost = 1.0;
    }
    else{
      netlist->overallCost = netlist->wireCost;
    }

    //If retiming is enabled, we should go through the netlist and try to retime critical nets
    if(placementRetiming){
      attemptRetiming(netlist, currTemp, criticalPathDelay, timing_driven);
      
      //Now that we have retimed a few blocks, recalculate the criticality for everything.
      criticalPathDelay =  calculateCriticalPathDelay(netlist, true);

      //cout << "Critical Path updated to " << criticalPathDelay << endl;
      //Now, the costs for all of the nets might have changed, so calculate the placement cost from scratch
      netlist->calculateMaxInputOutputCriticalities(criticalPathDelay);
      netlist->calculatePlacementCost(timing_driven, criticalPathDelay);      

      cout << "After retiming, wire cost = " <<  netlist->wireCost;
      cout << " timing cost = " <<  netlist->timingCost;
      cout << " critical path delay = " <<  criticalPathDelay <<endl;
    }

    for(int i = 0; i < numMovesPerTemp; i++){
#if defined(PARANOID_CHECK_PIPEPLACE)
      reallyOldWireCost = netlist->wireCost;
      reallyOldTimingCost = netlist->timingCost;
      reallyOldMaxCut = netlist->maxCut;
      reallyOldAveCut = netlist->aveCut;
#endif

#if defined(PIPEPLACEDEBUGHEAVY)
      cout << "Wire Cost before move = " <<  netlist->wireCost <<endl;
      cout << "Timing Cost before move = " <<  netlist->timingCost <<endl;
      if(maxCutCost){
	cout << "Max Cut before move = " <<  netlist->maxCut <<endl;
      }
      if(aveCutCost){
	cout << "Ave Cut before move = " <<  netlist->aveCut <<endl;
      }
      //netlist->print_netList(true);
#endif

#ifdef BREAKONANNEALNUMBER
      numberMoves++;
      
      if(numberMoves == 27041){
 	cout << endl << "Before move!" << endl;
	netlist->printPlacement();
	netlist->print_netList(true);
      }
#endif

//       if(i == 222 && !clb_level_placement){
// 	cout << "STOP HERE!" << endl;
//       }

      //Generate a random move
      moveType = generatePipelineAwareMove(netlist, &block0, &block1, &indexOrBank0, &indexOrBank1, &blockType, &numMove, (int) (rLim + 0.5));
      
      //Make the move and update all the proper sink delays.
      //We are also making incremental changes to the slacks - they are inexact, but we'll see what
      // we can do with them
      //We should keep track of if the move failed or not.
      //If it failed, we should not try to undo it.
      moveSucceeded = makePipelineAwareMove(&moveType, netlist, &block0, &block1, indexOrBank0, indexOrBank1, 
					 blockType, numMove, criticalPathDelay, &retimeUndoList);

#if defined (PARANOID_CHECK_PIPEPLACE)
    //Make sure that the mappings of the netlist logic block and the array block match
    netlist->checkLogicBlockMappings();

    //Check to make sure that the mappings of the array blocks match the logic blocks
    FPGA->checkArrayMappings();

#endif

      //Figure out the cost of the move
      if(deltaDelayCost){
	dCost = determineDeltaCostDeltaDelay(netlist, moveType, &block0, &block1, indexOrBank0, indexOrBank1, blockType, numMove, 
					     timing_driven, criticalPathDelay, &retimeUndoList);
      }
      else{
	dCost = determineDeltaCost(netlist, moveType, &block0, &block1, indexOrBank0, indexOrBank1, blockType, numMove, 
				   timing_driven, criticalPathDelay, &retimeUndoList);
      }

#ifdef BREAKONANNEALNUMBER
      if(numberMoves == 27041){
 	cout << endl<< "After move!" << endl;
 	netlist->printPlacement();
   	netlist->print_netList(true);
      }
#endif

#if defined(PIPEPLACEDEBUG)      
      if(moveType == MOVELOGIC){
	numMoveLogic ++;
      }
      else if(moveType == MOVEREGEMPTYSWAP || moveType == MOVEREGSWAP){
	numRegSwap ++;
      }
      else if(moveType == MOVEREGEMPTYFIRST || moveType == MOVEREGEMPTYLAST){
	numRegSplit ++;
	aveRegSplit += numMove;
      }
      else if(moveType == MOVEREGPARENT || moveType == MOVEREGCHILD){
	//This isn't exactly merging since we might or might not be trying to push all of the registers back together
	numRegMerge ++;
	aveRegMerge += numMove;
      }
      else if(moveType == MOVEZIPPNET){
	numZip ++;
      }
      else if (moveType == MOVEUNZIPPNET){
	numUnzip ++;
      }
      else if (moveType == MOVEREGRETIMEBACK || moveType == MOVEREGRETIMEFORWARD){
	numMoveRegRetime ++;
	aveRegRetime += numMove;
	aveCostRegRetime += dCost;
      }
      //else if (moveType == MOVEREGRETIMEBACKABORT || moveType == MOVEREGRETIMEFORWARDABORT){
      //} 
      if(!moveSucceeded){
	numMovesFailed++;
      }
#endif

#if defined(PARANOID_CHECK_PIPEPLACE)
      oldWireCost = netlist->wireCost;
      oldTimingCost = netlist->timingCost;
      oldMaxCut = netlist->maxCut;
      oldAveCut = netlist->aveCut;
      
      //Recalculate everthing from scratch
      netlist->calculatePlacementCost(timing_driven, criticalPathDelay);
      if(fabs(oldWireCost - netlist->wireCost) / netlist->wireCost > 0.001){
	ERROR("Problem doing incremental cost update");
      }
      if(fabs(oldTimingCost - netlist->timingCost) / netlist->timingCost > 0.001 ){
	ERROR("Problem doing incremental cost update");
      }
      if(oldMaxCut != netlist->maxCut){
	ERROR("Problem doing incremental cost update");
      }
      if(oldAveCut != netlist->aveCut){
	ERROR("Problem doing incremental cost update");
      }
#endif

//       if(dCost == 0.0){
// 	cout << "Zero Cost!" << endl;
//       }
      
      //Figure out whether we want to reject this move
      //This is actually redundant since the dCost will be 0 if the move failed.
      if(dCost > 0.0 && moveSucceeded){
	if(drand48() >= exp(-1.0*(dCost/currTemp))){
	  deltaCostReject += dCost;
	  movesRejected++;

#ifdef PIPEPLACEDEBUG
	  if(moveType == MOVELOGIC){
	    numMoveLogicReject ++;
	  }
	  else if(moveType == MOVEREGEMPTYSWAP || moveType == MOVEREGSWAP){
	    numRegSwapReject ++;
	  }
	  else if(moveType == MOVEREGEMPTYFIRST || moveType == MOVEREGEMPTYLAST){
	    numRegSplitReject ++;
	  }
	  else if(moveType == MOVEREGPARENT || moveType == MOVEREGCHILD){
	    //This isn't exactly merging since we might or might not be trying to push all of the registers back together
	    numRegMergeReject ++;
	  }
	  
	  else if(moveType == MOVEZIPPNET){
	    numZipReject ++;
	  }
	  else if (moveType == MOVEUNZIPPNET){
	    numUnzipReject ++;
	  }	  
	  else if (moveType == MOVEREGRETIMEBACK || moveType == MOVEREGRETIMEFORWARD){
	    numRegRetimeReject++;
	  }
#endif

	  //Undo the move
	  //cout << "Undoing move of type" << moveType << endl;
	  undoPipelineAwareMove(&moveType, netlist, &block0, &block1, indexOrBank0, indexOrBank1, blockType, numMove, 
				criticalPathDelay, &retimeUndoList);
    
	  //Calculate the original cost back again
	  if(deltaDelayCost){
	    dCost = determineDeltaCostDeltaDelay(netlist, moveType, &block1, &block0, indexOrBank1, indexOrBank0, blockType, 
						 numMove, timing_driven, criticalPathDelay, &retimeUndoList);
	  }
	  else{
	    dCost = determineDeltaCost(netlist, moveType, &block1, &block0, indexOrBank1, indexOrBank0, blockType, numMove, 
				       timing_driven, criticalPathDelay, &retimeUndoList);
	  }
	  
#ifdef PLACER_DEBUG
	  numRegistersEstimatedFreed = 0;
	  numRegistersEstimatedNeeded = 0;
#endif

#ifdef BREAKONANNEALNUMBER
 	  if(numberMoves == 4185){
 	    cout << endl << "After undo move!" << endl;
 	    netlist->printPlacement();
 	    netlist->print_netList(true);
 	  }
#endif

	  //If we unzipped a net, then undid it by zipping a net, we've still got info that we don't need anymore
	  // sitting in block 1.
	  // We should go ahead and delete those registers and nets now.
	  if(moveType == MOVEZIPPNET){
	    ((registerBank*)block1)->freeBank();
	  }

	  //If we just retimed registers backwards, then we wanted to undo this move, we created
	  // a new series of "undo" nets and registers.  We should delete the nets and registers
	  // we don't need anymore looking forward and confirm the registers backwards.
	  //This is essentially because we "retimed" register forwards.
	  else if(moveType == MOVEREGUNDOREGRETIMEBACK){
	    clearRetimeUndoListBackward(&retimeUndoList, block0, blockType);
	    clearNerfedBlocksForward(block0, blockType);
	  }
	  
	  //If we just retimed registers forwards, we should delete the nets and registers
	  // we don't need anymore looking backward and confirm the registers forwards
	  else if(moveType == MOVEREGUNDOREGRETIMEFORWARD){
	    clearRetimeUndoListForward(&retimeUndoList);
	    clearNerfedBlocksBackward(block0, blockType);
	  }

#if defined(PARANOID_CHECK_PIPEPLACE)
	  netlist->calculatePlacementCost(timing_driven, criticalPathDelay);
	  if(fabs(reallyOldWireCost - netlist->wireCost) / netlist->wireCost > 0.001){
	    ERROR("Problem doing undo cost update");
	  }
	  if(fabs(reallyOldTimingCost - netlist->timingCost) / netlist->timingCost > 0.001 ){
	    ERROR("Problem doing undo cost update");
	  }
	  if(reallyOldMaxCut != netlist->maxCut){
	    ERROR("Problem doing incremental cost update");
	  }
	  if(reallyOldAveCut != netlist->aveCut){
	    ERROR("Problem doing incremental cost update");
	  }
#endif
	
	}
	else{
	  netlist->overallCost += dCost;
	  averageOverallCost += netlist->overallCost;

	  deltaCostAccept += dCost;
	  //cout << "Accepted move with dCost = " << dCost <<endl;
	  nonZeroMovesAccepted ++;

	  //If we just zipped a net, we should go ahead and delete the registers in block1 now
	  // we kept them around because of incremental cost calculation and just in case
	  // we had to undo the move, but now we can get rid of them
	  if(moveType == MOVEZIPPNET){
	    ((registerBank*)block1)->freeBank();
	  }

	  //If we just retimed registers backwards, we should delete the nets and registers
	  // we don't need anymore looking forward and confirm the registers backwards
	  else if(moveType == MOVEREGRETIMEBACK || moveType == MOVEREGRETIMEFORWARDABORT){
	    clearRetimeUndoListForward(&retimeUndoList);
	    clearNerfedBlocksBackward(block0, blockType);
	  }
	  
	  //If we just retimed registers forwards, we should delete the nets and registers
	  // we don't need anymore looking backward and confirm the registers forwards
	  else if(moveType == MOVEREGRETIMEFORWARD || moveType == MOVEREGRETIMEBACKABORT){
	    clearRetimeUndoListBackward(&retimeUndoList, block0, blockType);
	    clearNerfedBlocksForward(block0, blockType);
	  }
	}
      }
      else{
	if(dCost < 0.0){
	  nonZeroMovesAccepted ++;
	  //cout << "Accepted move with dCost = " << dCost <<endl;
	}

	netlist->overallCost += dCost;
	averageOverallCost += netlist->overallCost;

	deltaCostAccept += dCost;

	//If we just zipped a net, we should go ahead and delete the registers in block1 now
	// we kept them around because of incremental cost calculation and just in case
	// we had to undo the move, but now we can get rid of them
	if(moveType == MOVEZIPPNET){
	  ((registerBank*)block1)->freeBank();
	}

	//If we just retimed registers backwards, we should delete the nets and registers
	// we don't need anymore looking forward and confirm the registers backwards
	else if(moveType == MOVEREGRETIMEBACK || moveType == MOVEREGRETIMEFORWARDABORT){
	  clearRetimeUndoListForward(&retimeUndoList);
	  clearNerfedBlocksBackward(block0, blockType);
	}
	
	//If we just retimed registers forwards, we should delete the nets and registers
	// we don't need anymore looking backward and confirm the registers forwards
	else if(moveType == MOVEREGRETIMEFORWARD || moveType == MOVEREGRETIMEBACKABORT){
	  clearRetimeUndoListBackward(&retimeUndoList, block0, blockType);
	  clearNerfedBlocksForward(block0, blockType);
	}
      }

      //To be completely accurate, we should now recalculate the critical path delay and slacks
      // for the whole system.
      //However, this is too timing consuming to do after each move, so we don't do it all of the time
      // we only do it after every N moves
      if(i % numMovesBeforeUpdate == (numMovesBeforeUpdate - 1)){
	criticalPathDelay =  calculateCriticalPathDelay(netlist, true);
	//cout << "Critical Path updated to " << criticalPathDelay << endl;

	//Now, the costs for all of the nets might have changed, so calculate the placement cost from scratch
	netlist->calculateMaxInputOutputCriticalities(criticalPathDelay);
	netlist->calculatePlacementCost(timing_driven, criticalPathDelay);
      }

#if defined(PIPEPLACEDEBUGHEAVY)
      //netlist->checkNets();

      cout << "Wire Cost after move = " <<  netlist->wireCost <<endl;
      cout << "Timing Cost after move = " <<  netlist->timingCost <<endl;
      if(maxCutCost){
	cout << "Max Cut after move = " <<  netlist->maxCut <<endl;
      }
      if(aveCutCost){
	cout << "Ave Cut after move = " <<  netlist->aveCut <<endl;
      }
      //netlist->print_netList();
      //netlist->printPlacement();
#endif

#if defined(PARANOID_CHECK_PIPEPLACE)
      //FPGA->checkFreeRegisters();
      netlist->checkNets();
      if(numRegistersEstimatedFreed != 0){
	//ERROR("Did not estimate # of freed registers correctly");
      }
      if(numRegistersEstimatedNeeded != 0){
	//ERROR("Did not estimate # of needed registers correctly");
      }
#endif
    }

    //Calculate some statistics about this iteration
    //netlist->printPlacement();
    
    if(movesRejected > 0){
      averageDeltaCostReject = deltaCostReject / (double)(movesRejected);
    }
    if(numMovesPerTemp - movesRejected > 0){
      averageDeltaCostAccept = deltaCostAccept / (double)(numMovesPerTemp - movesRejected);
    }
    if(nonZeroMovesAccepted > 0){
      //cout << "Total delta Accept = " << deltaCostAccept << endl;
      averageDeltaCostNonZeroAccept = deltaCostAccept / (double) nonZeroMovesAccepted;
      //cout << "Accepted " << nonZeroMovesAccepted << " nonzero moves with an average of " << averageDeltaCostNonZeroAccept << endl;
    }
    if(numMoveRegRetime > 0){
      aveRegRetime = aveRegRetime / numMoveRegRetime;  
      aveCostRegRetime = aveCostRegRetime / numMoveRegRetime;
    }

    if(numMovesPerTemp - movesRejected != 0){
      averageOverallCost /= numMovesPerTemp - movesRejected;
    }
    else{
      averageOverallCost = netlist->overallCost;
    }

#if defined (PIPEPLACEDEBUG)
    aveRegSplit = aveRegSplit / numRegSplit;
    aveRegMerge = aveRegMerge / numRegMerge;
    if(timing_driven){
      if((numRegSplit - numRegSplitReject) < 0){
	ERROR("SOMETHING HAPPENED HERE!");
      }
      if(maxCutCost && !aveCutCost){
	printf("%e\t%1.4f\t%1.4f\t%e\t%.1f\t%e\t%e\t%e\t%e\t%e\t%f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f", 
	       currTemp, (float)(numMovesPerTemp - movesRejected) / (float) numMovesPerTemp, 
	       (float)(nonZeroMovesAccepted) / (float) numMovesPerTemp, criticalPathDelay, netlist->getWireCost(), netlist->getTimingCost(), averageOverallCost,
	       averageDeltaCostAccept, averageDeltaCostNonZeroAccept, averageDeltaCostReject, netlist->maxCut, (float) numMoveLogic / (float) numMovesPerTemp,
	       (float)numRegSwap / float(numMovesPerTemp), (float) numRegSplit / (float) numMovesPerTemp, (float) numRegMerge / (float) numMovesPerTemp,
	       (float)(numMoveLogic - numMoveLogicReject) / (float) numMoveLogic, (float)(numRegSwap - numRegSwapReject) / (float) numRegSwap,
	       (float)(numRegSplit - numRegSplitReject) / (float) numRegSplit, (float)(numRegMerge - numRegMergeReject) / (float) numRegMerge,
	       aveRegSplit, aveRegMerge, (float)(numRegSplit - numRegSplitReject)/(float)(numRegMerge - numRegMergeReject));
      }
      
      else if(aveCutCost){
	printf("%e\t%1.4f\t%1.4f\t%e\t%.1f\t%e\t%e\t%e\t%e\t%e\t%f\t%f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f", 
	       currTemp, (float)(numMovesPerTemp - movesRejected) / (float) numMovesPerTemp, 
	       (float)(nonZeroMovesAccepted) / (float) numMovesPerTemp, criticalPathDelay, netlist->getWireCost(), netlist->getTimingCost(), averageOverallCost,
	       averageDeltaCostAccept, averageDeltaCostNonZeroAccept, averageDeltaCostReject, netlist->maxCut, netlist->aveCut, (float) numMoveLogic / (float) numMovesPerTemp,
	       (float)numRegSwap / float(numMovesPerTemp), (float) numRegSplit / (float) numMovesPerTemp, (float) numRegMerge / (float) numMovesPerTemp,
	       (float)(numMoveLogic - numMoveLogicReject) / (float) numMoveLogic, (float)(numRegSwap - numRegSwapReject) / (float) numRegSwap,
	       (float)(numRegSplit - numRegSplitReject) / (float) numRegSplit, (float)(numRegMerge - numRegMergeReject) / (float) numRegMerge,
	       aveRegSplit, aveRegMerge, (float)(numRegSplit - numRegSplitReject)/(float)(numRegMerge - numRegMergeReject));
      }
      else{
// 	printf("%e\t%1.4f\t%1.4f\t%e\t%.1f\t%e\t%e\t%e\t%e\t%e\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%d\t%d\t%d\t%d", 
// 	       currTemp, (float)(numMovesPerTemp - movesRejected) / (float) numMovesPerTemp, 
// 	       (float)(nonZeroMovesAccepted) / (float) numMovesPerTemp, criticalPathDelay, netlist->getWireCost(), netlist->getTimingCost(), averageOverallCost,
// 	       averageDeltaCostAccept, averageDeltaCostNonZeroAccept, averageDeltaCostReject, (float) numMoveLogic / (float) numMovesPerTemp,
// 	       (float)numRegSwap / float(numMovesPerTemp), (float) numRegSplit / (float) numMovesPerTemp, (float) numRegMerge / (float) numMovesPerTemp,
// 	       (float) numZip / (float) numMovesPerTemp,(float) numUnzip / (float) numMovesPerTemp,
// 	       (float)(numMoveLogic - numMoveLogicReject) / (float) numMoveLogic, (float)(numRegSwap - numRegSwapReject) / (float) numRegSwap,
// 	       (float)(numRegSplit - numRegSplitReject) / (float) numRegSplit, (float)(numRegMerge - numRegMergeReject) / (float) numRegMerge,
// 	       (float)(numZip - numZipReject) / (float) numZip, (float)(numUnzip - numUnzipReject) / (float) numUnzip,
// 	       aveRegSplit, aveRegMerge, (float)(numRegSplit - numRegSplitReject)/(float)(numRegMerge - numRegMergeReject),
// 	       numRegSplit, numRegSplitReject, numRegMerge, numRegMergeReject);

	printf("%e\t%1.4f\t%1.4f\t%e\t%.1f\t%e\t%e\t%e\t%e\t%e\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%1.2f\t%1.2f\t%d\t%d\t%1.2f\t%1.4e\t%d", 
	       currTemp, (float)(numMovesPerTemp - movesRejected) / (float) numMovesPerTemp, 
	       (float)(nonZeroMovesAccepted) / (float) numMovesPerTemp, criticalPathDelay, netlist->getWireCost(), netlist->getTimingCost(), averageOverallCost,
	       averageDeltaCostAccept, averageDeltaCostNonZeroAccept, averageDeltaCostReject, numMoveLogic,
	       numRegSwap, numRegSplit, numRegMerge, numZip, numUnzip,(numMoveLogic - numMoveLogicReject), (numRegSwap - numRegSwapReject),
	       (numRegSplit - numRegSplitReject), (numRegMerge - numRegMergeReject),
	       (numZip - numZipReject), (numUnzip - numUnzipReject),
	       aveRegSplit, aveRegMerge,  numMoveRegRetime, numRegRetimeReject, aveRegRetime, aveCostRegRetime, numMovesFailed);
      }
    }
    else{
      if(maxCutCost && !aveCutCost){
	printf("%e\t%1.4f\t%1.4f\t%e\t%.1f\t%e\t%e\t%e\t%f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f", 
	       currTemp, (float)(numMovesPerTemp - movesRejected) / (float) numMovesPerTemp, 
	       (float)(nonZeroMovesAccepted) / (float) numMovesPerTemp, criticalPathDelay, netlist->getWireCost(), averageDeltaCostAccept, 
	       averageDeltaCostNonZeroAccept, averageDeltaCostReject, netlist->maxCut, (float) numMoveLogic / (float) numMovesPerTemp,
	       (float)numRegSwap / float(numMovesPerTemp), (float) numRegSplit / (float) numMovesPerTemp, (float) numRegMerge / (float) numMovesPerTemp,
	       (float)(numMoveLogic - numMoveLogicReject) / (float) numMoveLogic, (float)(numRegSwap - numRegSwapReject) / (float) numRegSwap,
	       (float)(numRegSplit - numRegSplitReject) / (float) numRegSplit, (float)(numRegMerge - numRegMergeReject) / (float) numRegMerge,
	       aveRegSplit, aveRegMerge, (float)(numRegSplit - numRegSplitReject)/(float)(numRegMerge - numRegMergeReject));
      }
      else if(aveCutCost){
	printf("%e\t%1.4f\t%1.4f\t%e\t%.1f\t%e\t%e\t%e\t%f\t%f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f", 
	       currTemp, (float)(numMovesPerTemp - movesRejected) / (float) numMovesPerTemp, 
	       (float)(nonZeroMovesAccepted) / (float) numMovesPerTemp, criticalPathDelay, netlist->getWireCost(), averageDeltaCostAccept, 
	       averageDeltaCostNonZeroAccept, averageDeltaCostReject, netlist->maxCut, netlist->aveCut, (float) numMoveLogic / (float) numMovesPerTemp,
	       (float)numRegSwap / float(numMovesPerTemp), (float) numRegSplit / (float) numMovesPerTemp, (float) numRegMerge / (float) numMovesPerTemp,
	       (float)(numMoveLogic - numMoveLogicReject) / (float) numMoveLogic, (float)(numRegSwap - numRegSwapReject) / (float) numRegSwap,
	       (float)(numRegSplit - numRegSplitReject) / (float) numRegSplit, (float)(numRegMerge - numRegMergeReject) / (float) numRegMerge,
	       aveRegSplit, aveRegMerge, (float)(numRegSplit - numRegSplitReject)/(float)(numRegMerge - numRegMergeReject));
      }
      else{
	printf("%e\t%1.4f\t%1.4f\t%e\t%.1f\t%e\t%e\t%e\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f\t%1.2f", 
	       currTemp, (float)(numMovesPerTemp - movesRejected) / (float) numMovesPerTemp, 
	       (float)(nonZeroMovesAccepted) / (float) numMovesPerTemp, criticalPathDelay, netlist->getWireCost(), averageDeltaCostAccept, 
	       averageDeltaCostNonZeroAccept, averageDeltaCostReject, (float) numMoveLogic / (float) numMovesPerTemp,
	       (float)numRegSwap / float(numMovesPerTemp), (float) numRegSplit / (float) numMovesPerTemp, (float) numRegMerge / (float) numMovesPerTemp,
	       (float)(numMoveLogic - numMoveLogicReject) / (float) numMoveLogic, (float)(numRegSwap - numRegSwapReject) / (float) numRegSwap,
	       (float)(numRegSplit - numRegSplitReject) / (float) numRegSplit, (float)(numRegMerge - numRegMergeReject) / (float) numRegMerge,
	       aveRegSplit, aveRegMerge, (float)(numRegSplit - numRegSplitReject)/(float)(numRegMerge - numRegMergeReject));
      }
   }
#else
    if(timing_driven){
      if(maxCutCost && !aveCutCost){
	printf("%e\t%f\t%f\t%e\t%f\t%e\t%e\t%e\t%e\t%e\t%e\t%f", currTemp, (float)(numMovesPerTemp - movesRejected) / (float) numMovesPerTemp, 
	       (float)(nonZeroMovesAccepted) / (float) numMovesPerTemp, criticalPathDelay, netlist->getWireCost(), netlist->getTimingCost(), averageOverallCost, 
	       averageDeltaCostAccept, averageDeltaCostNonZeroAccept, averageDeltaCostReject, netlist->maxCut);
      }
      
      else if(aveCutCost){
	printf("%e\t%f\t%f\t%e\t%f\t%e\t%e\t%e\t%e\t%e\t%f\t%f", currTemp, (float)(numMovesPerTemp - movesRejected) / (float) numMovesPerTemp, 
	       (float)(nonZeroMovesAccepted) / (float) numMovesPerTemp, criticalPathDelay, netlist->getWireCost(), netlist->getTimingCost(), averageOverallCost,
	       averageDeltaCostAccept, averageDeltaCostNonZeroAccept, averageDeltaCostReject, netlist->maxCut, netlist->aveCut);
      }
      else{
	printf("%e\t%f\t%f\t%e\t%f\t%e\t%e\t%e\t%e\t%e\t%f\t%f", currTemp, (float)(numMovesPerTemp - movesRejected) / (float) numMovesPerTemp, 
	       (float)(nonZeroMovesAccepted) / (float) numMovesPerTemp, criticalPathDelay, netlist->getWireCost(), netlist->getTimingCost(), averageOverallCost,
	       averageDeltaCostAccept, averageDeltaCostNonZeroAccept, averageDeltaCostReject, rLim, tdPlacementCritExponent);
      }
    }
    else{
      if(maxCutCost && !aveCutCost){
	printf("%e\t%f\t%f\t%e\t%f\t%e\t%e\t%e\t%f", currTemp, (float)(numMovesPerTemp - movesRejected) / (float) numMovesPerTemp, 
	       (float)(nonZeroMovesAccepted) / (float) numMovesPerTemp, criticalPathDelay, netlist->getWireCost(), averageDeltaCostAccept, 
	       averageDeltaCostNonZeroAccept, averageDeltaCostReject, netlist->maxCut);
      }
      else if(aveCutCost){
	printf("%e\t%f\t%f\t%e\t%f\t%e\t%e\t%e\t%f\t%f", currTemp, (float)(numMovesPerTemp - movesRejected) / (float) numMovesPerTemp, 
	       (float)(nonZeroMovesAccepted) / (float) numMovesPerTemp, criticalPathDelay, netlist->getWireCost(), averageDeltaCostAccept, 
	       averageDeltaCostNonZeroAccept, averageDeltaCostReject, netlist->maxCut, netlist->aveCut);
      }
      else{
	printf("%e\t%f\t%f\t%e\t%f\t%e\t%e\t%e", currTemp, (float)(numMovesPerTemp - movesRejected) / (float) numMovesPerTemp, 
	       (float)(nonZeroMovesAccepted) / (float) numMovesPerTemp, criticalPathDelay, netlist->getWireCost(), averageDeltaCostAccept, 
	       averageDeltaCostNonZeroAccept, averageDeltaCostReject);
      }
    }
#endif

    cout << endl;

    //Calculate the average, max and min criticalities for all pin-to-pin connections
    //netlist->calculateAvgMaxMinCrits(criticalPathDelay);
    

    //printf("Num Splits = %d, Num Split Attempts = %d, NumSplitSwap = %d, Num Merges = %d, Num Merge Attempts = %d, numMergeSwaps = %d \n", 
    //numSplits, numSplitCalls, numSplitSwapCalls, numMerges, numMergeCalls, numMergeSwapCalls);

    //averageAttemptedIOMoveManhattanDistance /= numAttemptedIOMoves;
    //averageAttemptedLUTMoveManhattanDistance /= numAttemptedLUTMoves;
    //averageAttemptedRegMoveManhattanDistance /= numAttemptedRegMoves;
    //cout << "Average Attempted IO move Manhattan distance = " <<  averageAttemptedIOMoveManhattanDistance << " over " << numAttemptedIOMoves << " number of IO moves" << endl;
    //cout << "Average Attempted LUT move Manhattan distance = " <<  averageAttemptedLUTMoveManhattanDistance << " over " << numAttemptedLUTMoves << " number of LUT moves" << endl;
    //cout << "Average Attempted Reg move Manhattan distance = " <<  averageAttemptedRegMoveManhattanDistance << " over " << numAttemptedRegMoves << " number of Reg moves" << endl;

    //Update the temperature, range limit and timing criticality exponent
    currTemp = updateTemp(currTemp, numMovesPerTemp - movesRejected, numMovesPerTemp);
    rLim = updateRLimAndCritExp(rLim, numMovesPerTemp - movesRejected, numMovesPerTemp);

    //Re-calculate the critical path delay
    criticalPathDelay =  calculateCriticalPathDelay(netlist, true);

    //Update the cost of the current placement due to new critical path delay
    netlist->calculateMaxInputOutputCriticalities(criticalPathDelay);
    netlist->calculatePlacementCost(timing_driven, criticalPathDelay);

    //See if we should terminate the annealing
    done = exitCondition(currTemp, netlist->overallCost, netlist->calculateNumNetsAndPNets());
    //done = true;

    if(lotsTimingInfo){
      netlist->printAllSinkCriticalities(criticalPathDelay, true);
    }
  }
  
  cout << "Final Wire Cost " << netlist->wireCost << endl;
  cout << "Final placement Timing Cost" << netlist->timingCost << endl;
  cout << "Final Critical Path Delay " << criticalPathDelay << endl;
  netlist->printAllSinkCriticalities(criticalPathDelay, true);
  netlist->printCriticalPath(true);
  //netlist->printNetWireCost(true);
}

double updateTemp(double currTemp, int movesAccepted, int movesPerTemp){
  double accPercent = (double)movesAccepted / (double) movesPerTemp;
  if(accPercent > 0.96){
    return (currTemp * 0.5);
  }
  else if(accPercent > 0.8){
    return (currTemp * 0.9);
  }
  if(accPercent > 0.15){
    return (currTemp * 0.95);
  }
  else{
    return (currTemp * 0.8);
  }
}

//Figure out if we should terminate the annealing
bool exitCondition(double currTemp, double cost, int numNets){
  //We are done if the temperature is less than 0.005 * cost of an average net

  if(currTemp < 0.005 * cost / (double) numNets){
    return true;
  }
  return false;    
}

//This updates the Rlim and timing criticality exponent
float updateRLimAndCritExp(float rLim, int movesAccepted, int movesPerTemp){
  float newRLim;
  float accPercent = (float) movesAccepted / (float) movesPerTemp;
  float maxMove;
  
  //This is the range limit mode suggested by VPR
  //As per VPR -
  //<quote>
  /* Update the range limited to keep acceptance prob. near 0.44.  Use *
   * a floating point rlim to allow gradual transitions at low temps.  */
  //</quote>

  maxMove = (float) max(FPGA->get_numRows(), FPGA->get_numCols());  

  if(noWindowing){
    newRLim = maxMove;
  }
  else{
    //Don't let the range get larger than the device
    newRLim = flomin(rLim * (0.56 + accPercent), maxMove);
    
    //Don't let the range get smaller than 1
    newRLim = flomax(newRLim, 1.0);
  }

  if(ff_level_placement_activation >= (float)newRLim/(float)maxMove && clb_level_placement){
    cout << "FF-level placement activated!" << endl;
    clb_level_placement = false;
  }
  if(retime_placement_activation >= (float)newRLim/(float)maxMove && !placementRetiming){
    cout << "Retiming activated!" << endl;
    placementRetiming = true;
  }
  if(splitDelayBlocks_activation >= (float)newRLim/(float)maxMove && !splitDelayBlocks){
    cout << "Splitting of register banks activated!" << endl;
    splitDelayBlocks = true;
  }
  if(unzipNets_activation >= (float)newRLim/(float)maxMove && !unzipNets){
    cout << "Register zipping/unzipping activated!" << endl;
    splitDelayBlocks = true;
    unzipNets = true;
  }

  if(ff_level_placement_activation == 0.0001 && newRLim == 1.0 && clb_level_placement){
    cout << "FF-level placement activated!" << endl;
    clb_level_placement = false;
  }
  if(retime_placement_activation == 0.0001 && newRLim == 1.0 && !placementRetiming){
    cout << "Retiming activated!" << endl;
    placementRetiming = true;
  }
  if(splitDelayBlocks_activation == 0.0001 && newRLim == 1.0 && !splitDelayBlocks){
    cout << "Splitting of register banks activated!" << endl;
    splitDelayBlocks = true;
  }
  if(unzipNets_activation == 0.0001 && newRLim == 1.0 && !unzipNets){
    cout << "Register zipping/unzipping activated!" << endl;
    splitDelayBlocks = true;
    unzipNets = true;
  }

  tdPlacementCritExponent = (1.0 - ((newRLim - 1.0) / (maxMove - 1.0))) * (tdPlacementCritExponentFinal - 1.0) + 1.0;

  return newRLim;
}

//Go through the netlist and find registers that are not in the same CLB as their host LUT.
//If they are not in the same CLB, try and move it there.
//1) If the current CLB has a free register, just put it there.
//2) If it doesn't, swap positions with an unrelated FF in that CLB (there must be at least one)
void findOrphanRegisters(Netlist *netlist, bool timing_driven){
  void* block0;
  void* block1;
  Terminal *outTerm;
  Terminal *sink;
  logicBlock *sourceLB;
  int sourceRow, sourceCol;
  int registerRow, registerCol;
  double criticalPathDelay;
  double dCost;
  int moveType;
  int numRegisters = 0;
  int numPackableRegisters = 0;
  int numOrphanedRegisters = 0;
  int minSinkDistance;

  double oldWireCost;
  double oldTimingCost;
  double oldCriticalPathDelay;

  if(clb_level_placement){
    clb_level_placement = false;
    cout << "FF-level_placement activated" << endl;
  }

  if(timing_driven){
    netlist->overallCost = 1.0;
  }
  else{
    netlist->overallCost = netlist->wireCost;
  }

  criticalPathDelay = calculateCriticalPathDelay(netlist, true);

  for(list<logicBlock*>::iterator lbIter = netlist->lbList.begin(); lbIter != netlist->lbList.end(); lbIter++){
    if((*lbIter)->get_blockType() == REGISTER_BANK){
      numRegisters++;
      registerRow = (*lbIter)->row;
      registerCol = (*lbIter)->col;
      
      //Check to see if it's source is a LUT
      if((*lbIter)->returnDriverType() == LUT){
	cout << "Register " << (*lbIter)->name;
	numPackableRegisters++;

	sourceLB = (*lbIter)->getRegisterSource();

	sourceRow = sourceLB->row;
	sourceCol = sourceLB->col;

	//Show how far this register is away from it's sinks
	//(*lbIter)->printDistanceToAllSinks();
	  
	if(sourceCol != registerCol || sourceRow != registerRow){
	  //This register is not in the same CLB as it's parent LUT
	  numOrphanedRegisters++;
	  //cout << " is orphaned at (" << registerCol << ", " << registerRow;
	  //cout << "), instead of with source LUT at (" << sourceCol << ", " << sourceRow << ")" << endl;	  
	  //cout << "Distance to parent LUT = " << abs(sourceCol - registerCol) + abs(sourceRow - registerRow) <<endl;
	  
	  cout << " orphaned with minSinkDistance = " << (*lbIter)->minDistanceToAllSinks();
	  cout << " and distance to parent LUT = " << abs(sourceCol - registerCol) + abs(sourceRow - registerRow) <<endl;

	  block0 = (void*)(*lbIter)->rb;

	  //Find a register in the host LUT's CLB to swap with
	  block1 = (void *) FPGA->getCLBAt(sourceRow, sourceCol)->getFirstEmptyUnrelatedRegisterBank();
	  
	  if(block1){
	    oldWireCost = netlist->wireCost;
	    oldTimingCost = netlist->timingCost;
	    oldCriticalPathDelay = criticalPathDelay;

	    moveType = MOVELOGICBLOCKSELF;
	    makePipelineAwareMove(&moveType, netlist, &block0, &block1, 0, 0, 
				  REGISTERBANK, 0, criticalPathDelay, NULL);
	    
#if defined (PARANOID_CHECK_PIPEPLACE)
    //Make sure that the mappings of the netlist logic block and the array block match
    netlist->checkLogicBlockMappings();

    //Check to make sure that the mappings of the array blocks match the logic blocks
    FPGA->checkArrayMappings();

#endif

	    //Figure out the cost of the move
	    if(deltaDelayCost){
	      dCost = determineDeltaCostDeltaDelay(netlist, moveType, &block0, &block1, 0, 0, REGISTERBANK, 0, 
						   timing_driven, criticalPathDelay, NULL);
	    }
	    else{
	      dCost = determineDeltaCost(netlist, moveType, &block0, &block1, 0, 0, REGISTERBANK, 0, 
					 timing_driven, criticalPathDelay, NULL);
	    }
	    netlist->overallCost += dCost;

	    criticalPathDelay = calculateCriticalPathDelay(netlist, true);

	    cout << "Move to parent delta cost = " << dCost << endl;
	    cout << "Move to parent delta wire cost = " << netlist->wireCost - oldWireCost << endl;
	    cout << "Move to parent delta timing cost = " << netlist->timingCost - oldTimingCost << endl;
	    cout << "Move to parent delta critical path delay = " << criticalPathDelay - oldCriticalPathDelay << endl;

	    //Undo the move
	    undoPipelineAwareMove(&moveType, netlist, &block0, &block1, 0, 0, REGISTERBANK, 0, 
				  criticalPathDelay, NULL);
	    
	    //Calculate the original cost back again
	    if(deltaDelayCost){
	      dCost = determineDeltaCostDeltaDelay(netlist, moveType, &block1, &block0, 0, 0, REGISTERBANK, 0, 
						   timing_driven, criticalPathDelay, NULL);
	    }
	    else{
	      dCost = determineDeltaCost(netlist, moveType, &block1, &block0, 0, 0, REGISTERBANK, 0, 
					 timing_driven, criticalPathDelay, NULL);
	    }
	    netlist->overallCost += dCost;

	    criticalPathDelay = calculateCriticalPathDelay(netlist, true);
	  }
	}
	else{
	  cout << " united with parent minSinkDistance = " << (*lbIter)->minDistanceToAllSinks() << endl;
	}
      }
    }
  }
  cout << numOrphanedRegisters << " registers out of a total of " << numRegisters << " registers and ";
  cout << numPackableRegisters << " packable registers were orphaned in the regular placement" << endl;
}

//Go through the entire netlist and try to retime the most critical logic blocks
void attemptRetiming(Netlist *netlist, double currTemp, double criticalPathDelay, bool timing_driven){
  void* block0;
  void* block1;
  int indexOrBank0, indexOrBank1;
  int moveType;
  enum e_timingMoveBlockType blockType;
  int select, choices;
  int numMove;
  bool moveSucceeded;
  double dCost;
  list <PipeliningNet *> retimeUndoList;

  //Go through all of the logic blocks we have.
  for(list<logicBlock*>::iterator lbIter = netlist->lbList.begin(); lbIter != netlist->lbList.end(); lbIter++){
    if((*lbIter)->get_blockType() == REGISTER_BANK){
      block0 = (void*) (*lbIter)->rb;
      blockType = REGISTERBANK;
    }
    else if((*lbIter)->get_blockType() == LUT || (*lbIter)->get_blockType() == RAM_BLOCK || (*lbIter)->get_blockType() == MULTIPLIER){
      block0 = (void*) FPGA->getCLBAt((*lbIter)->row, (*lbIter)->col)->getLUT((*lbIter)->ptc_index);
      blockType = LB;
    }
    //We don't care about retiming I/O pins
    else{
      continue;
    }

    choices = 0;
    moveType = MOVENULL;
    
    // See if we could pull a set of registers from the output to the input or from the input to the output.
    //See if we can push a register backwards at all.
    if((*lbIter)->numRegCanPushBackward > 0){
      //If we can, is the max criticality of all the input nets large enough for us to want to?
      if((*lbIter)->maxInputCriticality >= retime_placement_crit){
	//Pushing registers back will be "0" if we are able to do so
	choices ++;
      }
    }
    
    //See if we can push a register forwards at all
    if((*lbIter)->numRegCanPushForward > 0){
      //If we can, is the max criticality of all the output nets large enough for us to want to?
      if((*lbIter)->maxOutputCriticality >= retime_placement_crit){
	//Pushing register forward will be "1" if we are able to do both a forward
	// and backwards push.
	//If we are only able to do a forward push this will be "0", of course
	choices ++;
      }
    }
    
    //We evenly select among the choices we have.
    //See if we have any retiming choices
    if(choices > 0){
      select = lrand48() % choices;
      
      if(select == 0){
	//If we can push a register backwards, this means we want to do a backwards push
	if((*lbIter)->numRegCanPushBackward > 0 && (*lbIter)->maxInputCriticality >= retime_placement_crit){
	  //We would like to make a backwards retiming move - so we should check if we can actually do so, given
	  // the current state of the architecture.
	  if(generatePipelineAwareMoveRegRetimeBackward(&moveType, netlist, &block0, blockType, &numMove)){
	    //If we want to and we can actually do it, we should perform the probability roll
	    if(drand48() > retime_placement_probability){
	      //The probability roll failed.
	      moveType = MOVENULL;
	    }
	  }
	}
	//If we can't, this means we want to do a forward push
	else{
	  //We would like to make a forwards retiming move - so we should check if we can actually do so, given
	  // the current state of the architecture.
	  if(generatePipelineAwareMoveRegRetimeForward(&moveType, netlist, &block0, blockType, &numMove)){
	    //If we want to and we can actually do it, we should perform the probability roll
	    if(drand48() > retime_placement_probability){
	      //The probability roll failed.
	      moveType = MOVENULL;
	    }
	  }
	}
      }
      else{
	//If select is 1, we know we can do both a forward and backwards move and we have elected to do a forwards move.
	//We should check if we can actually do so, given the current state of the architecture.
	if(generatePipelineAwareMoveRegRetimeForward(&moveType, netlist, &block0, blockType, &numMove)){
	  //If we want to and we can actually do it, we should perform the probability roll
	  if(drand48() > retime_placement_probability){
	    //The probability roll failed.
	    moveType = MOVENULL;
	  }
	}
      }
      
      moveSucceeded = makePipelineAwareMove(&moveType, netlist, &block0, NULL, -1, -1, 
					    blockType, numMove, criticalPathDelay, &retimeUndoList);

      //Figure out the cost of the move
      if(deltaDelayCost){
	dCost = determineDeltaCostDeltaDelay(netlist, moveType, &block0, NULL, -1, -1, 
					     blockType, numMove, timing_driven, criticalPathDelay, &retimeUndoList);
      }
      else{
	dCost = determineDeltaCost(netlist, moveType, &block0, NULL, -1, -1, 
				   blockType, numMove, timing_driven, criticalPathDelay, &retimeUndoList);
      }

      //Figure out whether we want to reject this move
      //This is actually redundant since the dCost will be 0 if the move failed.
      if(dCost > 0.0 && moveSucceeded && drand48() >= exp(-1.0*(dCost/currTemp))){
	//Undo the move
	undoPipelineAwareMove(&moveType, netlist, &block0, NULL, -1, -1, 
			      blockType, numMove, criticalPathDelay, &retimeUndoList);
	
	//Calculate the original cost back again
	if(deltaDelayCost){
	  dCost = determineDeltaCostDeltaDelay(netlist, moveType, NULL, &block0, -1, -1,
					       blockType, numMove, timing_driven, criticalPathDelay, &retimeUndoList);
	}
	else{
	  dCost = determineDeltaCost(netlist, moveType, NULL, &block0, -1, -1,
				     blockType, numMove, timing_driven, criticalPathDelay, &retimeUndoList);
	}
	
	//If we just retimed registers backwards, then we wanted to undo this move, we created
	// a new series of "undo" nets and registers.  We should delete the nets and registers
	// we don't need anymore looking forward and confirm the registers backwards.
	//This is essentially because we "retimed" register forwards.
	if(moveType == MOVEREGUNDOREGRETIMEBACK){
	  clearRetimeUndoListBackward(&retimeUndoList, block0, blockType);
	  clearNerfedBlocksForward(block0, blockType);
	}
	
	//If we just retimed registers forwards, we should delete the nets and registers
	// we don't need anymore looking backward and confirm the registers forwards
	else if(moveType == MOVEREGUNDOREGRETIMEFORWARD){
	  clearRetimeUndoListForward(&retimeUndoList);
	  clearNerfedBlocksBackward(block0, blockType);
	}
      }
      else{
	netlist->overallCost += dCost;
	
	//If we just retimed registers backwards, we should delete the nets and registers
	// we don't need anymore looking forward and confirm the registers backwards
	if(moveType == MOVEREGRETIMEBACK || moveType == MOVEREGRETIMEFORWARDABORT){
	  clearRetimeUndoListForward(&retimeUndoList);
	  clearNerfedBlocksBackward(block0, blockType);
	}
	
	//If we just retimed registers forwards, we should delete the nets and registers
	// we don't need anymore looking backward and confirm the registers forwards
	else if(moveType == MOVEREGRETIMEFORWARD || moveType == MOVEREGRETIMEBACKABORT){
	  clearRetimeUndoListBackward(&retimeUndoList, block0, blockType);
	  clearNerfedBlocksForward(block0, blockType);
	}
      }
    }
  }
}
