// Copyright (C) 2006 Ken Eguro, Scott Hauck, Akshay Sharma
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

#include "include.h"

extern vprFPGA* FPGA;
extern long int sharingCost;
extern bool critical_path_headroom;
extern double critical_path_headroom_fac;

#ifdef ROUTER_DEBUG
extern long routerPops;
extern long routerPushes;
#endif

//extern long int wireCost;
extern bool lockRegPlacement;

// Main entry point for all routing in the system
// Includes main routing loop
void route(Netlist *netlist, bool pipeline_aware, bool timing_driven){
  int itr;
  double criticalPathDelay;
  double lastCriticalPathDelay;
  double lastSharingCost;
  double minCriticalPathDelay;
  int minSharingCost;
  int minSharingCostIter;
  time_t start_time,end_time;

  minCriticalPathDelay = DBL_MAX;
  minSharingCost = INT_MAX;
  minSharingCostIter = -1;

  FPGA->updateHistoryCosts(1);
  //Since we don't know the critical path delay yet, just fill in a really small number to prevent a divide by 0

#ifdef ROUTER_DEBUG
  routerPushes = 0;
  routerPops = 0;
#endif

  start_time=time(0);


  // Perform the first routing iteration
  //The slacks of all nets are estimates from either the regular placement process or from
  // the estimated timing analysis done after sucking in the placement.
  //Thus, we should put in a very large critical path delay here so that
  // each route assumes they have very little slack.
  //I realize this is counter-intuitive, since a critical path of 0
  // would make everyone's slack infinite, but necessary since we do not re-calculate slacks
  // until we re-calculate critical path delays.
  if(timing_driven){
    routeNetsTimingDriven(netlist, pipeline_aware, 1, DBL_MAX);
  }
  else{
    routeNetsCongestionDriven(netlist, pipeline_aware, 1);
  }
  criticalPathDelay = calculateCriticalPathDelay(netlist, pipeline_aware);

  minSharingCost = sharingCost;
  minSharingCostIter = 1;
  
  if(sharingCost == 0){
    minCriticalPathDelay = criticalPathDelay;
  }

  lastSharingCost = sharingCost;
  lastCriticalPathDelay = criticalPathDelay;
  
  //print out all of the nets' information
  //netlist->print_netList();  

  //We have completed the first iteration of routing.  Now we might have to resort.
  //If we not pipeline aware, we only re-sort if we are timing-driven.  We do this
  //   based on most critical to least critical
  //If we are pipeline aware and not timing-driven, we don't have to re-sort either.
  //However, if we are pipeline aware and timing-driven, we should re-sort.  We still sort sinks in
  //   non-decreasing order, but within a given latency we sort based on most critial
  //   to least critical.  Criticality is obvious for non-pipelined signals, and the
  //   criticality of a pipelined signal is determined by the maximum criticality of
  //   all its links
  //If we have locked all of the register assignments during placement, we should also
  // re-sort the order of the sinks for pipelined nets.  This is different from the 
  // sorting we do for regular Armada routing since we can do the sorting on each individual
  // pipelined net by itself.  
  if(lockRegPlacement && timing_driven){
    netlist->sort_netPipelineSinks_LockedRegs(timing_driven);
  }
  else{
    if(pipeline_aware && timing_driven){
      netlist->sort_netPipelineSinks(timing_driven);
    }
    else if(timing_driven){
      netlist->sort_netSinksTiming();
    }
  }

  //wireCost=calcWireCost(netlist);
  cout<<endl<<endl<<"<<<<<<<<<<<<<<<<<<<<<< BEGIN ROUTING PHASE >>>>>>>>>>>>>>>>>>>>>>>>" <<endl;
  cout<<"ITERATION 1\tSHARING = "<<sharingCost<< "\tDELAY = " << criticalPathDelay;

#ifdef ROUTER_DEBUG
  cout << "\tRPush = " << routerPushes << "\tRPop = " << routerPops;
#endif

  cout.flush();

  //netlist->printCriticalPath(pipeline_aware);
  
  //netlist->print_netList();

  if(criticalPathDelay == 0.0){
    ERROR("Critical Path reached 0.0");
  }

  //We might have resolved sharing in the first iteration
  //however, this is done without real timing information so we have to do at least one
  //more iteration
  for(itr=2; itr<=200; itr++){
    FPGA->updateHistoryCosts(itr);
    if(timing_driven){
      routeNetsTimingDriven(netlist, pipeline_aware, itr, criticalPathDelay);
    }
    else{
      routeNetsCongestionDriven(netlist, pipeline_aware, itr);
    }

    criticalPathDelay = calculateCriticalPathDelay(netlist, pipeline_aware);

    if(lockRegPlacement && timing_driven){
      netlist->sort_netPipelineSinks_LockedRegs(timing_driven);
    }
    else{
      if(pipeline_aware && timing_driven){
	netlist->sort_netPipelineSinks(timing_driven);
      }
      else if(timing_driven){
	netlist->sort_netSinksTiming();
      }
    }
    
    //print out all of the nets' information
    //netlist->print_netList();

    cout<<endl<<"ITERATION "<<itr<<"\tSHARING = "<<sharingCost << "\tDELAY = " << criticalPathDelay;

#ifdef ROUTER_DEBUG
    cout << "\tRPush = " << routerPushes << "\tRPop = " << routerPops;
#endif

    //netlist->printCriticalPath(pipeline_aware);

    //netlist->print_netList();
    
    if(criticalPathDelay == 0.0){
      ERROR("Critical Path reached 0.0");
    }
    
    //printNetsAndRecalcCosts();

    if(minSharingCost > sharingCost){
      minSharingCost = sharingCost;
      minSharingCostIter = itr;
    }

    if(sharingCost == 0 && minCriticalPathDelay > criticalPathDelay){
      minCriticalPathDelay = criticalPathDelay;
    }
	      
    if(sharingCost==0){
      //printNetsAndRecalcCosts();
      cout<<endl<<endl<<"SUCCESFULLY PLACED AND ROUTED THIS BLOKE!!";
      cout<< " with critical path " << criticalPathDelay << endl;
      //cout << " and min critical path delay " << minCriticalPathDelay << endl;

      netlist->printCriticalPath(pipeline_aware);
      break;
    }
    
    lastSharingCost = sharingCost;
    lastCriticalPathDelay = criticalPathDelay;
    cout.flush();
  }

  if(itr==201 && sharingCost != 0){
    /*       printNetsAndRecalcCosts(); */
    cout << endl << "Min sharing cost = " << minSharingCost << " first reached in iteration " << minSharingCostIter;
    cout<<endl<<endl<<"<<<<<<<<<<<<<<<<<<<<<< END ROUTING PHASE >>>>>>>>>>>>>>>>>>>>>>>>";
    cout<<endl<<endl;
  }
  
  /* start measuring the routing runtime */
  end_time=time(0);
  
  cout<<endl<<endl<<"TOTAL ROUTING RUNTIME IN SECONDS : "<<end_time-start_time;
  
  cout<<endl<<endl;
  cout<<endl;
  
} 


void routeNetsTimingDriven(Netlist *netlist, bool pipeline_aware, int iter, double criticalPathDelay){
  int net_ctr=0;
  long int oldSharing;
  long int newSharing;

  netlist->create_netListItr();
  for(Net* net=netlist->getnext_netList();net!=0;net=netlist->getnext_netList()){

    if(iter==1){
      if(!net->pipelined || !pipeline_aware){
	routeNetTimingDriven(iter, net, criticalPathDelay);
      }
      else{
	//If the placement of the registers is fixed, we have to do pipeline aware routing, but not QuickRoute
	// or Armada - just a version of PathFinder that goes between the correct terminals of the pipelined nets
	if(lockRegPlacement){
	  routeNetPathFinderLockedRegistersTimingDriven(iter, net, criticalPathDelay);
	}
	else{
	  routeNetQuickRoute_D(iter, net, criticalPathDelay);
	}
      }
    }
    else{
      if(!net->pipelined || !pipeline_aware){
	net->rip_rTree();
	//On subsequent iterations use the net's last known delay to determine criticality
	routeNetTimingDriven(iter, net, criticalPathDelay);
      }
      else{
	//If the placement of the registers is fixed, we have to do pipeline aware routing, but not QuickRoute
	// or Armada - just a version of PathFinder that goes between the correct terminals of the pipelined nets
	if(lockRegPlacement){
	  routeNetPathFinderLockedRegistersTimingDriven(iter, net, criticalPathDelay);
	}
	else{
	  //For each net, make the source's external delay be it's arrival/departure time
	  net->source->externalDelay = net->source->departureTime;
	  
	  //Make each sink's external delay be the critical path minus it's required time
	  for(vector<Terminal*>::iterator sinkTerm = net->sinkVector.begin(); sinkTerm != net->sinkVector.end(); sinkTerm++){
	    (*sinkTerm)->externalDelay = criticalPathDelay - (*sinkTerm)->requiredTime;
	  }

	  //Pipelined nets have to get treated differently
	  //Since the pipelined nets have already been routed at least once, we should re-route using pipeRoute-D
	  net->ripUpPipelineInformation();
	  if(critical_path_headroom_fac){
	    routeNetQuickRoute_D(iter, net, criticalPathDelay + (critical_path_headroom_fac * criticalPathDelay));
	  }
	  else{
	    routeNetQuickRoute_D(iter, net, criticalPathDelay);
	  }
	}
      }
    }

#ifdef ROUTER_DEBUG
    net->print_Info(false);
    net->print_rTree();
#endif
    //Go through this net and make sure that all of the sinks have a real route to them at the correct latency
    if(!net->followSinks(pipeline_aware)){
      ERROR("Net did not route to all sinks!");
    }
  }
  netlist->remove_netListItr();

  oldSharing = sharingCost;
  
  FPGA->zeroOccupancy();

#ifdef PARANOID_CHECK_ROUTE
  netlist->reCalculateOccupancy(pipeline_aware);
  
  newSharing = FPGA->countNewSharing(0);
  if(oldSharing != newSharing){
    cout << "Old sharing = " << oldSharing << " new sharing = " << newSharing << endl;
    ERROR("Sharing Cost is different!");
  }
#endif
}

void routeNetsCongestionDriven(Netlist *netlist, bool pipeline_aware, int iter){
  int net_ctr=0;
  long int oldSharing;
  long int newSharing;

  netlist->create_netListItr();
  for(Net* net=netlist->getnext_netList();net!=0;net=netlist->getnext_netList()){

    if(iter==1){
      if(!net->pipelined || !pipeline_aware){
	routeNetCongestionDriven(iter, net);
      }
      else{
	//If the placement of the registers is fixed, we have to do pipeline aware routing, but not QuickRoute
	// or Armada - just a version of PathFinder that goes between the correct terminals of the pipelined nets
	if(lockRegPlacement){
	  routeNetPathFinderLockedRegistersCongestionDriven(iter, net);
	}
	else{
	  //Pipelined nets have to get treated differently
	  //If we are timing-driven, this will be done as if all nets are timing-critical  
	  routeNetQuickRoute(iter, net);
	}
      }
    }
    else{
      if(!net->pipelined || !pipeline_aware){
	net->rip_rTree();
	//On subsequent iterations use the net's last known delay to determine criticality
	//routeNet(iter,net,pathfinder_pshare, timing_driven, criticalPathDelay);
	routeNetCongestionDriven(iter, net);
      }
      else{
	//If the placement of the registers is fixed, we have to do pipeline aware routing, but not QuickRoute
	// or Armada - just a version of PathFinder that goes between the correct terminals of the pipelined nets
	if(lockRegPlacement){
	  routeNetPathFinderLockedRegistersCongestionDriven(iter, net);
	}
	else{
	  //Pipelined nets have to get treated differently
	  //For the congestion-driven case, all we have to do is rip-up the existing routes, delete the current information
	  //and call routeNetInitialPipeRoute again.
	  net->ripUpPipelineInformation();
	  routeNetQuickRoute(iter, net);
	}
      }
    }

#ifdef ROUTER_DEBUG
    net->print_Info(false);
    net->print_rTree();
#endif
    //Go through this net and make sure that all of the sinks have a real route to them at the correct latency
    if(!net->followSinks(pipeline_aware)){
      ERROR("Net did not route to all sinks!");
    }
  }
  netlist->remove_netListItr();

  oldSharing = sharingCost;
  
  FPGA->zeroOccupancy();

#ifdef PARANOID_CHECK_ROUTE
  netlist->reCalculateOccupancy(pipeline_aware);
  
  newSharing = FPGA->countNewSharing(0);
  if(oldSharing != newSharing){
    cout << "Old sharing = " << oldSharing << " new sharing = " << newSharing << endl;
    ERROR("Sharing Cost is different!");
  }
#endif
}


/***
 * This function finds a route for the net pointed to by 'net'. It assumes that all
 * routingNode and Net objects have been cleaned up and initialized for routing.
 * This cleanup MUST be separately handled by another routine. Specifically, the 
 * previous routing tree should have been ripped up, and routing node members (isExplored,
 * occupancy,presentSharing,pathCost,predecessor and inrTree) should have been 
 * accordingly changed. 
 ***/
void routeNetTimingDriven(int iter, Net* net, double criticalPathDelay){
  routingNode* rnode;routingNode* nb_rnode;
  routingEdge* nb_edge;
  double path_cost,nb_path_cost,new_path_cost,heur_estimate;
  double net_delay, new_net_delay;
  Terminal* terminal;
  priority_queue<pqNode>* PQ;
  treeNode* tree_node;
  list<routingNode*> cleanupList;
  double linkCriticality;
  int i;

  list<routingNode*>::iterator rNodeIter;

#ifdef PARANOID_CHECK_ROUTE
  //Go through the entire routing graph and make sure that all nodes are unexplored and not in a routing tree
  for(rNodeIter = FPGA->rnodeList.begin(); rNodeIter != FPGA->rnodeList.end(); rNodeIter++){
    if((*rNodeIter)->inrTree == true){
      ERROR("Found routing node inside routing tree before routing has begun!");
    }
    else if((*rNodeIter)->isExplored == true){
      ERROR("Found explored routing node before routing has begun!");
    }
  }
#endif

  /* new root tree node. initialize it to source routing node */
  net->init_rTreeRoot();
  PQ=new priority_queue<pqNode>;


  //Get the source routing node
  terminal = net->get_source();
  rnode = terminal->get_rNode();

  //Add it to the PQ
  //We should, really, add the neighbors of the source to the PQ,
  // and not the source itself, just in case the source is also
  // the sink.
  //If we do that, we force ourselves to go through at least 1 edge.
  //However, if we are doing TD routing with multiple sinks this still might not
  // be enough.
  //When we "update the PQ" with the existing routing tree we will
  // run into the same problem.
  //If we really want to solve that problem we will put in a flag
  // to gate the first time we see the source.
  //Push the source into the PQ at a cost of criticality of link * netDelay (Aij*dj)at this node
  //If we are here we are certainly going to route to sink 0
  PQ->push(pqNode(rnode, calculateCrit(net->slackVector.at(0),criticalPathDelay) * rnode->intrinsicDelay, 0.0, 0, rnode->intrinsicDelay));
#ifdef ROUTER_DEBUG
  //cout << "Push source on with cost " << calculateCrit(net->slackVector.at(0),criticalPathDelay) * rnode->intrinsicDelay << " and delay " << rnode->intrinsicDelay << endl;
  routerPushes++;
#endif

  rnode->set_inrTree(true);

  //If the source is an embedded pipelining net, don't increase it's occupancy.
  //  It's just a single node so if both the source and destination
  //  nets both increment the occupancy it will look like it's overused when it's not
  if(rnode->get_type() != DELAY){
    rnode->set_occupancy(rnode->get_occupancy()+1);
    if(rnode->get_occupancy() > rnode->get_capacity()){
      sharingCost+=1;
      //cout << "Node " << rnode->name << " is shared" << endl;
    }
  }

#ifdef ROUTER_DEBUG
  cout<<endl<<endl<<"<<<<<<<<<============ BEGIN ROUTER DEBUG FOR "<<net->get_name();
  cout<<"============>>>>>>>>>";
  cout<<endl<<"SOURCE LOGIC BLOCK OF THIS NET IS "<<terminal->get_lBlock()->get_name();
  cout<<" AND ITS OUTPUT IS MAPPED TO "<<rnode->get_name();

  cout << "Routing net " << net->name << endl;
#endif

  /* start exploration */
  routingNode* sink_rnode;
  int sink_count=0;

  while(1){
    //If we are doing timing-driven routing, the sinks of the net will be sorted
    //in order of decreasing criticality (except for iteration 1)
    sink_rnode=net->sinkVector.at(sink_count)->rNode;
    linkCriticality = calculateCrit(net->slackVector.at(sink_count),criticalPathDelay);
    
    //Loop until we have an unexplored routing node
    while(1){
      if(PQ->empty())
	ERROR("priority queue is empty before net routing completed...");

      rnode=(PQ->top()).rNode;
      //We only want to consider nodes that have not been explored
      if(rnode->isExplored){
#ifdef ROUTER_DEBUG
	cout << "Popped " << (PQ->top()).rNode->get_name() <<" predecessor ";
	if((PQ->top()).predNode)
	  cout<<(PQ->top()).predNode->get_name();
	cout << " but node has been explored already!" <<endl;
	routerPops++;
#endif
	PQ->pop();
	continue;
      }
      else{
	//Mark the node explored, set its pathcost/delay and predecessor routing node
	//Notice that the node's pathcost/delay does not get written until it is actually popped off the PQ
	rnode->isExplored=true;
	rnode->pathCost=(PQ->top()).pathCost;
	rnode->netDelay = (PQ->top()).netDelay;
	rnode->predecessor=(PQ->top()).predNode;
	
	cleanupList.push_back(rnode);  //this list holds routing nodes that need a clean-up

#ifdef ROUTER_DEBUG
	cout<<endl<<"POPPED "<<rnode->get_name()<<" AT PATH COST "<<rnode->get_pathCost();
	cout<<" and delay " << rnode->netDelay << " AND PREDECESSOR NODE ";
	if(rnode->get_predecessor())
	  cout<<rnode->get_predecessor()->get_name();
	routerPops++;
#endif
	PQ->pop();
	break;
      }
    }

    //Check to see if we have found a new sink of the net
    //NOTICE, THIS IS ENFORCING A SPECIFIC ORDER TO THE ROUTING.
    if(rnode==sink_rnode){
#ifdef ROUTER_DEBUG
      cout << endl << "Found sink " <<  sink_rnode->name << " # " << sink_count << " of " << net->sinkVector.size() << " at cost " << rnode->pathCost;
      cout<<endl<<endl<<"++++++++++++ FOUND A SINK; WILL NOW ENTER BACKTRACE ++++++++++++";
#endif

      //Record the link delay from source to this sink
      //since the rnode may be shared with several nets
      net->linkDelayVector.at(sink_count) = rnode->netDelay;
      
      tree_node=backTrace(net,rnode,iter);

      if(sink_count<(net->sinkVector.size()-1)){
	//If there are unrouted sinks in this net we have to push the entire routing tree
	//into a new PQ
	delete PQ;
	PQ=new priority_queue<pqNode>;
	
#ifdef ROUTER_DEBUG
	cout << endl << "Routing tree for net " << net->name << " after adding sink " << sink_count;
	net->print_rTree();
#endif
	//When we find a sink, clean up the pathCost of the everything not in the routing tree
	while(!cleanupList.empty()){
	  if(!cleanupList.front()->inrTree){
	    cleanupList.front()->isExplored=false;
	    cleanupList.front()->predecessor=0;
	    cleanupList.front()->pathCost=DBL_MAX;
	    cleanupList.front()->netDelay=DBL_MAX;
	    //cleanupList.front()->isRouted=false;
	  }
	  cleanupList.pop_front();
	}

	//now reinit the PQ to the current routing tree ala VPR's directed search
	//If we are doing timing-driven routing we have to reinit the PQ to the current routing tree
	//but each link is added at a cost of (unrouted sink's link criticality * path delay to node)
	//Notice that since the slackVector is initially set to small numbers, 
	// and iteration 1 is given a very large critical path delay, all links are considered critical
#ifdef ROUTER_DEBUG
	cout << "Update PQ for net " << net->name << " node " << sink_count+1 << " of " << net->sinkVector.size() << " criticality "; 
	cout << (1 - (net->slackVector.at(sink_count+1)/criticalPathDelay)) << " slack= " << net->slackVector.at(sink_count+1);
	cout << " criticalPathDelay = " << criticalPathDelay << endl;
#endif
	updatePQTimingDriven(PQ, (net->rTree).root, net->sinkVector.at(sink_count+1)->rNode, 
		 calculateCrit(net->slackVector.at(sink_count+1), criticalPathDelay));
	sink_count++;
      }
      else{
	//We have just routed the last sink of the net
#ifdef ROUTER_DEBUG
	cout << endl << "Routing tree for net " << net->name << " after adding last sink " << sink_count<< endl;
	net->print_rTree();
	cout << "Cleaning up all nodes touched by search!" <<endl;
#endif
	
	//We have no more sinks to route so we should delete the PQ, clean up all of the nets we have touched
	//and end the search
	delete PQ;
	/* clean up all routing nodes that have been affected by the exploration */
	while(!cleanupList.empty()){
	  cleanupList.front()->isExplored = false;
	  cleanupList.front()->inrTree = false;
	  cleanupList.front()->predecessor = 0;
	  cleanupList.front()->pathCost = DBL_MAX;
	  cleanupList.front()->netDelay = DBL_MAX;
	  //cleanupList.front()->isRouted = false;
	  cleanupList.pop_front();
	}

	//Also clean up all of the nodes in the routing tree
	(net->get_rTree())->clean_rNodes(net->get_rTree()->get_root());

	//We have found all of the sinks so we are done!
	break;
      }
    }

    else{
      //This is not a sink so we should look at it's neighbors
      rnode->create_fanoutListItr();
      for(nb_edge=rnode->getnext_fanoutList(); nb_edge!=0; nb_edge=rnode->getnext_fanoutList()){
	nb_rnode=nb_edge->destNode;

	//Check to see if this node's neighbor is a delay node that is not the sink we're looking for.
	//We only use this routine to route "unpipelined" signals, so we shouldn't be using delay nodes
	// to route.
	if(nb_rnode->type == DELAY && nb_rnode != sink_rnode){
	  continue;
	}

	/* A* algorithm */
	//heur_estimate=ALPHA*cost_to_target(nb_rnode,sink_rnode);
	heur_estimate = 0;
	
#ifdef ROUTER_DEBUG
	cout<<endl<<"LOOKING AT NEIGHBOR NODE "<<nb_rnode->get_name();
#endif
	
	//Only add unexplored neighbors that are not already in the routing tree to the queue
	//Also, make sure they are within the routing bounding box
	if(!nb_rnode->isExplored && !nb_rnode->inrTree && net->checkNodeRoutingBoundingBox(nb_rnode)){
	  path_cost=rnode->pathCost;
	  net_delay = rnode->netDelay;
	  
	  //New timing driven cost
	  //Notice that since the slackVector is initially set to small numbers, 
	  // and iteration 1 is given a very large critical path delay, all links are considered critical
	  new_path_cost = path_cost + nb_rnode->returnTimingCost(iter, linkCriticality, nb_edge->intrinsicDelay);
	  
	  new_net_delay = net_delay + nb_rnode->intrinsicDelay + nb_edge->intrinsicDelay;
	  
	  PQ->push(pqNode(nb_rnode,new_path_cost,heur_estimate,rnode, new_net_delay));
#ifdef ROUTER_DEBUG
	  routerPushes++;
	  cout<<endl<<"\tADDED NODE "<<nb_rnode->get_name()<<" AT PATH COST "<<
	    new_path_cost<< " Delay " << new_net_delay<< " AND PREDECESSOR "<<rnode->get_name();
#endif
	}
      }
      rnode->remove_fanoutListItr();
    }
  }
}

/***
 * This function finds a route for the net pointed to by 'net'. It assumes that all
 * routingNode and Net objects have been cleaned up and initialized for routing.
 * This cleanup MUST be separately handled by another routine. Specifically, the 
 * previous routing tree should have been ripped up, and routing node members (isExplored,
 * occupancy,presentSharing,pathCost,predecessor and inrTree) should have been 
 * accordingly changed. 
 ***/
void routeNetCongestionDriven(int iter, Net* net){
  routingNode* rnode;routingNode* nb_rnode;
  routingEdge* nb_edge;
  double path_cost,nb_path_cost,new_path_cost,heur_estimate;
  double net_delay, new_net_delay;
  Terminal* terminal;
  priority_queue<pqNode>* PQ;
  treeNode* tree_node;
  list<routingNode*> cleanupList;
  int i;

  list<routingNode*>::iterator rNodeIter;

#ifdef PARANOID_CHECK_ROUTE
  //Go through the entire routing graph and make sure that all nodes are unexplored and not in a routing tree
  for(rNodeIter = FPGA->rnodeList.begin(); rNodeIter != FPGA->rnodeList.end(); rNodeIter++){
    if((*rNodeIter)->inrTree == true){
      ERROR("Found routing node inside routing tree before routing has begun!");
    }
    else if((*rNodeIter)->isExplored == true){
      ERROR("Found explored routing node before routing has begun!");
    }
  }
#endif

  /* new root tree node. initialize it to source routing node */
  net->init_rTreeRoot();
  PQ=new priority_queue<pqNode>;

  //Get the source routing node
  terminal = net->get_source();
  rnode = terminal->get_rNode();

  //Add it to the PQ
  //We should, really, add the neighbors of the source to the PQ,
  // and not the source itself, just in case the source is also
  // the sink.
  //If we do that, we force ourselves to go through at least 1 edge.
  //However, if we are doing TD routing with multiple sinks this still might not
  // be enough.
  //When we "update the PQ" with the existing routing tree we will
  // run into the same problem.
  //If we really want to solve that problem we will put in a flag
  // to gate the first time we see the source.
  //If we are doing congestion-driven routing we need to add the source to the PQ at cost 0
  //Even if we are not in timing mode we should still report timing numbers
  PQ->push(pqNode(rnode,0.0,0.0,0, rnode->intrinsicDelay));
#ifdef ROUTER_DEBUG
  routerPushes++;
#endif

  rnode->set_inrTree(true);

  //If the source is an embedded pipelining net, don't increase it's occupancy.
  //  It's just a single node so if both the source and destination
  //  nets both increment the occupancy it will look like it's overused when it's not
  if(rnode->get_type() != DELAY){
    rnode->set_occupancy(rnode->get_occupancy()+1);
    if(rnode->get_occupancy() > rnode->get_capacity()){
      sharingCost+=1;
      //cout << "Node " << rnode->name << " is shared" << endl;
    }
  }

#ifdef ROUTER_DEBUG
  cout<<endl<<endl<<"<<<<<<<<<============ BEGIN ROUTER DEBUG FOR "<<net->get_name();
  cout<<"============>>>>>>>>>";
  cout<<endl<<"SOURCE LOGIC BLOCK OF THIS NET IS "<<terminal->get_lBlock()->get_name();
  cout<<" AND ITS OUTPUT IS MAPPED TO "<<rnode->get_name();

  cout << "Routing net " << net->name << endl;
#endif

  /* start exploration */
  int sink_count=0;

  while(1){
    //Loop until we have an unexplored routing node
    while(1){
      if(PQ->empty())
	ERROR("priority queue is empty before net routing completed...");

      rnode=(PQ->top()).rNode;
      //We only want to consider nodes that have not been explored
      if(rnode->isExplored){
#ifdef ROUTER_DEBUG
	cout << "Popped " << (PQ->top()).rNode->get_name() <<" predecessor ";
	if((PQ->top()).predNode)
	  cout<<(PQ->top()).predNode->get_name();
	cout << " but node has been explored already!" <<endl;
	routerPops++;
#endif
	PQ->pop();
	continue;
      }
      else{
	//Mark the node explored, set its pathcost/delay and predecessor routing node
	//Notice that the node's pathcost/delay does not get written until it is actually popped off the PQ
	rnode->isExplored=true;
	rnode->pathCost=(PQ->top()).pathCost;
	rnode->netDelay = (PQ->top()).netDelay;
	rnode->predecessor=(PQ->top()).predNode;
	
	cleanupList.push_back(rnode);  //this list holds routing nodes that need a clean-up

#ifdef ROUTER_DEBUG
	cout<<endl<<"POPPED "<<rnode->get_name()<<" AT PATH COST "<<rnode->get_pathCost();
	cout<<" and delay " << rnode->netDelay << " AND PREDECESSOR NODE ";
	if(rnode->get_predecessor())
	  cout<<rnode->get_predecessor()->get_name();
	routerPops++;
#endif
	PQ->pop();
	break;
      }
    }

    //Check to see if we have found any new sink of the net 
    if(net->foundNewSink(rnode)){
#ifdef ROUTER_DEBUG
      cout << endl << "Found sink " <<  rnode->name << " # " << sink_count << " of " << net->sinkVector.size() << " at cost " << rnode->pathCost;
      cout<<endl<<endl<<"++++++++++++ FOUND A SINK; WILL NOW ENTER BACKTRACE ++++++++++++";
#endif

      //Record the link delay from source to this sink
      //since the rnode may be shared with several nets
      net->linkDelayVector.at(sink_count) = rnode->netDelay;
      
      tree_node=backTrace(net, rnode, iter);

      if(sink_count<(net->sinkVector.size()-1)){
	//If there are unrouted sinks in this net we have to push the entire routing tree
	//into a new PQ
	delete PQ;
	PQ=new priority_queue<pqNode>;
	
#ifdef ROUTER_DEBUG
	cout << endl << "Routing tree for net " << net->name << " after adding sink " << sink_count;
	net->print_rTree();
#endif
	//When we find a sink, clean up the pathCost of the everything not in the routing tree
	while(!cleanupList.empty()){
	  if(!cleanupList.front()->inrTree){
	    cleanupList.front()->isExplored=false;
	    cleanupList.front()->predecessor=0;
	    cleanupList.front()->pathCost=DBL_MAX;
	    cleanupList.front()->netDelay=DBL_MAX;
	    cleanupList.front()->isRouted=false;
	  }
	  cleanupList.pop_front();
	}

	//now reinit the PQ to the current routing tree ala VPR's directed search
	//In congestion-driven routing we add the entire existing routing tree to the PQ at a cost of zero	   
	updatePQCongestionDriven(PQ,(net->rTree).root,net->sinkVector.at(sink_count+1)->rNode);

	rnode->isRouted=true;
	
	sink_count++;
      }
      else{
	//We have just routed the last sink of the net
#ifdef ROUTER_DEBUG
	cout << endl << "Routing tree for net " << net->name << " after adding last sink " << sink_count<< endl;
	net->print_rTree();
	cout << "Cleaning up all nodes touched by search!" <<endl;
#endif
	
	//We have no more sinks to route so we should delete the PQ, clean up all of the nets we have touched
	//and end the search
	delete PQ;
	/* clean up all routing nodes that have been affected by the exploration */
	while(!cleanupList.empty()){
	  cleanupList.front()->isExplored = false;
	  cleanupList.front()->inrTree = false;
	  cleanupList.front()->predecessor = 0;
	  cleanupList.front()->pathCost = DBL_MAX;
	  cleanupList.front()->netDelay = DBL_MAX;
	  cleanupList.front()->isRouted = false;
	  cleanupList.pop_front();
	}

	//Also clean up all of the nodes in the routing tree
	(net->get_rTree())->clean_rNodes(net->get_rTree()->get_root());

	//We have found all of the sinks so we are done!
	break;
      }
    }

    else{
      //If this node was a non-sink register, then we should not explore the
      // neighbors of this node.
      //We are doing unpipelined routing, so thye should not be used.
      if(rnode->type == DELAY){
	continue;
      }

      //This is not a sink so we should look at it's neighbors
      rnode->create_fanoutListItr();
      for(nb_edge=rnode->getnext_fanoutList(); nb_edge!=0; nb_edge=rnode->getnext_fanoutList()){
	nb_rnode=nb_edge->destNode;

	/* A* algorithm */
	//heur_estimate=ALPHA*cost_to_target(nb_rnode, rnode);
	heur_estimate = 0;
	
#ifdef ROUTER_DEBUG
	cout<<endl<<"LOOKING AT NEIGHBOR NODE "<<nb_rnode->get_name();
#endif
	
	//Only add unexplored neighbors that are not already in the routing tree to the queue
	//Also, make sure they are within the routing bounding box
	if(!nb_rnode->isExplored && !nb_rnode->inrTree && net->checkNodeRoutingBoundingBox(nb_rnode)){
	  path_cost=rnode->pathCost;
	  net_delay = rnode->netDelay;
	  
	  //New cost
	  new_path_cost=path_cost+nb_rnode->returnCongestionCost(iter);
	  
	  new_net_delay = net_delay + nb_rnode->intrinsicDelay + nb_edge->intrinsicDelay;
	  
	  PQ->push(pqNode(nb_rnode,new_path_cost,heur_estimate,rnode, new_net_delay));
#ifdef ROUTER_DEBUG
	  routerPushes++;
	  cout<<endl<<"\tADDED NODE "<<nb_rnode->get_name()<<" AT PATH COST "<<
	    new_path_cost<< " Delay " << new_net_delay<< " AND PREDECESSOR "<<rnode->get_name();
#endif
	}
      }
      rnode->remove_fanoutListItr();
    }
  }
}

// /***
//  * This function finds a route to sink_rnode from net's route tree. It is assumed here
//  * that there is an existing route tree that connects all sinks EXCEPT sink_rnode. Routing 
//  * node members (isExplored,occupancy,presentSharing,pathCost,predecessor and inrTree) 
//  * should be appropriately initialized.  
//  ***/
// void routeSink(int iter,Net* net,routingNode* sink_rnode,int change_costs, bool timing_driven){
//   routingNode* rnode;routingNode* nb_rnode;
//   routingEdge* nb_edge;
//   double path_cost,nb_path_cost,new_path_cost,heur_estimate;
//   Terminal* terminal;
//   priority_queue<pqNode> PQ;
//   treeNode* tree_node;
//   list<routingNode*> cleanupList;

//   /* init the PQ to the existing route tree of net */
//   if(timing_driven){
//     ERROR("Timing-driven placement not implemented yet!");
//   }
//   else{
//     initPQ(PQ,(net->get_rTree())->get_root(),sink_rnode);
//   }

//   /* start exploration */
//   while(1){

//     /* loop until an unexplored routing node is obtained */
//     while(1){
//       if(PQ.empty())
// 	ERROR("priority queue is empty before net routing completed...");

//       rnode=PQ.top().rNode;
//       if(rnode->isExplored==false){
// 	/* found unexplored routing node...mark it explored,set its pathcost and */
// 	/* predecessor routing node and break */
// 	rnode->isExplored=true;
// 	rnode->pathCost=PQ.top().pathCost;
// 	rnode->predecessor=PQ.top().predNode;
// 	cleanupList.push_back(rnode);  //this list holds routing nodes that need a clean-up
// 	PQ.pop();
// #ifdef ROUTER_DEBUG
// 	routerPops++;
// #endif
// 	break;
//       }

//       PQ.pop();
// #ifdef ROUTER_DEBUG
//       routerPops++;
// #endif
//     }

// #ifdef ROUTER_DEBUG
//     cout<<endl<<"POPPED "<<rnode->get_name()<<" AT PATH COST "<<rnode->get_pathCost();
//     cout<<" AND PREDECESSOR NODE ";
//     if(rnode->get_predecessor())
//       cout<<rnode->get_predecessor()->get_name();
// #endif

//     //if(rnode->type==SINK){
//     /* found the sink terminal that we are looking for */
//     if(rnode==sink_rnode){
// #ifdef ROUTER_DEBUG
//       cout<<endl<<endl<<"++++++++++++ FOUND SINK; WILL NOW ENTER BACKTRACE ++++++++++++";
// #endif

//       backTrace(net,rnode,iter);

//       /* clean up all routing nodes that have been affected by the exploration */
//       while(!cleanupList.empty()){
// 	cleanupList.front()->isExplored=false;
// 	cleanupList.front()->inrTree = false;
// 	cleanupList.front()->predecessor=0;
// 	cleanupList.front()->pathCost=DBL_MAX;
// 	cleanupList.front()->netDelay=DBL_MAX;
// 	//cleanupList.front()->isRouted=false;
// 	cleanupList.pop_front();
//       }

//       /* now clean up all the routing nodes that are in the new route tree */
//       (net->get_rTree())->clean_rNodes(net->get_rTree()->get_root());

//       /* exploration ends...return */
//       return;

//     }
//     //}

//     /* loop thru the fanout neighbor nodes of rnode */ 
//     rnode->create_fanoutListItr();
//     for(nb_edge=rnode->getnext_fanoutList();nb_edge!=0;nb_edge=rnode->getnext_fanoutList()){
//       nb_rnode=nb_edge->destNode;

//       /* A* algorithm */
//       //heur_estimate=ALPHA*cost_to_target(nb_rnode,sink_rnode);
//       heur_estimate = 0;

// #ifdef ROUTER_DEBUG
//       cout<<endl<<"LOOKING AT NEIGHBOR NODE "<<nb_rnode->get_name();
// #endif

//       if(nb_rnode->isExplored==false){
// 	path_cost=rnode->pathCost;
// 	nb_path_cost=nb_rnode->pathCost + nb_edge->intrinsicCost;
// 	new_path_cost=path_cost+nb_rnode->returnCongestionCost(iter);
// 	PQ.push(pqNode(nb_rnode,new_path_cost,heur_estimate,rnode));

// #ifdef ROUTER_DEBUG
// 	routerPushes++;
// 	cout<<endl<<"\tADDED NODE "<<nb_rnode->get_name()<<" AT PATH COST "<<
// 	  new_path_cost<<" AND PREDECESSOR "<<rnode->get_name();
// #endif
//       }

//     }
//     rnode->remove_fanoutListItr();

//   }
// }

//Rip up all of the routing associated with this netlist
void removeRouting(Netlist *netlist){
  netlist->create_netListItr();
  for(Net* net=netlist->getnext_netList();net!=0;net=netlist->getnext_netList()){
    //if(net->sinkVector==0)
    //continue;
    net->rip_rTree();
  }
  netlist->remove_netListItr();
  
}

double cost_to_target(routingNode* rnode,routingNode* sink_rnode){
  double estimate1=((double)(abs(rnode->row_low-sink_rnode->row_low)+abs(rnode->col_low-sink_rnode->col_low)))/4.0;
  double estimate2=((double)(abs(rnode->row_high-sink_rnode->row_low)+abs(rnode->col_high-sink_rnode->col_low)))/4.0;
  return doubmin(estimate1,estimate2);
}

// int calcWireCost(Netlist *netlist){
//   int wire_cost=0;
//   list<PipeliningNet *>::iterator pNetIter;

//   netlist->create_netListItr();
//   for(Net* net=netlist->getnext_netList();net!=0;net=netlist->getnext_netList()){
//     if(net->pipelined){
//       for(pNetIter = net->pipelineNetList.begin(); pNetIter != net->pipelineNetList.end(); pNetIter++){
// 	wire_cost+=(*pNetIter)->get_rNodeCount();
//       }
//     }
//     else{
//       wire_cost+=net->get_rNodeCount();
//     }
//   }
//   netlist->remove_netListItr();

//   return wire_cost;

// }

// void printNetsAndRecalcCosts(Netlist *netlist, bool pipeline_aware){
//   int wire_cost=0;int sharing_cost=0;

//   netlist->create_netListItr();
//   for(Net* net=netlist->getnext_netList();net!=0;net=netlist->getnext_netList()){
//     net->print_Info(pipeline_aware);
//     net->print_rTree();
//     //wire_cost+=net->get_rNodeCount();
//     //cout<<endl<<"WIRE COST = "<<wire_cost;
//   }
//   netlist->remove_netListItr();

//   sharing_cost=FPGA->countSharing(0);

//   cout<<endl<<endl<<"^^^^^^^^^^^ FINAL WIRE COST = "<<wire_cost<<" ^^^^^^^^^^^^";
//   cout<<endl<<"^^^^^^^^^^^ FINAL SHARING COST = "<<sharing_cost<<" ^^^^^^^^^^^^";
//   cout<<endl<<endl;
// }

//This performs regular PathFinder given the register assignement found during placement
void routeNetPathFinderLockedRegistersTimingDriven(int iter, Net *net, double criticalPathDelay){
  //Go through all of the pipelined nets and route them
  for(list<PipeliningNet *>::iterator pNetIter = net->pipelineNetList.begin(); pNetIter != net->pipelineNetList.end(); pNetIter++){
    if(iter != 1){
      //If this is not the first iteration, we should rip up the old routing tree and start again.
      (*pNetIter)->rip_rTree();
    }
    routePipelineNetPathFinderLockedRegistersTimingDriven(iter, *pNetIter, criticalPathDelay);
  }
}

//This performs regular PathFinder given the register assignement found during placement
void routeNetPathFinderLockedRegistersCongestionDriven(int iter, Net *net){
  //Go through all of the pipelined nets and route them
  for(list<PipeliningNet *>::iterator pNetIter = net->pipelineNetList.begin(); pNetIter != net->pipelineNetList.end(); pNetIter++){
    if(iter != 1){
      //If this is not the first iteration, we should rip up the old routing tree and start again.
      (*pNetIter)->rip_rTree();
    }
    routePipelineNetPathFinderLockedRegistersCongestionDriven(iter, *pNetIter);
  }
}

//For this pipelined net, just find a path between the source and all of the sinks.
void routePipelineNetPathFinderLockedRegistersTimingDriven(int iter, PipeliningNet *pNet, double criticalPathDelay){
  routingNode* rnode;routingNode* nb_rnode;
  routingEdge* nb_edge;
  double path_cost,nb_path_cost,new_path_cost,heur_estimate;
  double net_delay, new_net_delay;
  Terminal* terminal;
  priority_queue<pqNode>* PQ;
  treeNode* tree_node;
  list<routingNode*> cleanupList;
  list<routingNode*>::iterator clean_itr;
  double linkCriticality;
  int i;

  list<routingNode*>::iterator rNodeIter;

#ifdef PARANOID_CHECK_ROUTE
  //Go through the entire routing graph and make sure that all nodes are unexplored and not in a routing tree
  for(rNodeIter = FPGA->rnodeList.begin(); rNodeIter != FPGA->rnodeList.end(); rNodeIter++){
    if((*rNodeIter)->inrTree == true){
      ERROR("Found routing node inside routing tree before routing has begun!");
    }
    else if((*rNodeIter)->isExplored == true){
      ERROR("Found explored routing node before routing has begun!");
    }
  }
#endif

  /* new root tree node. initialize it to source routing node */
  pNet->init_rTreeRoot();
  PQ=new priority_queue<pqNode>;


  //Get the source routing node
  terminal = pNet->source;
  rnode=terminal->get_rNode();

#ifdef ROUTER_DEBUG
  cout<<endl<<endl<<"<<<<<<<<<============ BEGIN ROUTER DEBUG FOR "<<pNet->name;
  cout<<"============>>>>>>>>>";
  cout<<endl<<"SOURCE OF THIS NET IS MAPPED TO "<<rnode->get_name();
#endif

  //Set the node as explored and add it to the PQ
  //Push the source into the PQ at a cost of criticality of link * netDelay (Aij*dj)at this node
  //If we are here we are certainly going to route to sink 0
  PQ->push(pqNode(rnode, calculateCrit(pNet->slackVector.at(0),criticalPathDelay) * rnode->intrinsicDelay, 0.0, 0, rnode->intrinsicDelay));
  //cout << "Push source on with cost " << calculateCrit(pNet->slackVector.at(0),criticalPathDelay) * rnode->intrinsicDelay << " and delay " << rnode->intrinsicDelay << endl;
#ifdef ROUTER_DEBUG
  routerPushes++;
#endif

  rnode->set_inrTree(true);
  //If the source is an embedded pipelining net, don't increase it's occupancy.
  //  It's just a single node so if both the source and destination
  //  nets both increment the occupancy it will look like it's overused when it's not
  if(rnode->get_type() != DELAY){
    rnode->set_occupancy(rnode->get_occupancy()+1);
    if(rnode->get_occupancy() > rnode->get_capacity()){
      sharingCost+=1;
      //cout << "Node " << rnode->name << " is shared" << endl;
    }
  }

  /* start exploration */
  routingNode* sink_rnode;
  int sink_count=0;
  
#ifdef ROUTER_DEBUG
  cout << "Routing net " << pNet->name << endl;
#endif

  while(1){
    //If we are doing timing-driven routing, the sinks of the net will be sorted
    //in order of decreasing criticality (except for iteration 1)
    sink_rnode=pNet->sinkVector.at(sink_count)->rNode;
    linkCriticality = calculateCrit(pNet->slackVector.at(sink_count),criticalPathDelay);
    
    //Loop until we have an unexplored routing node
    while(1){
      if(PQ->empty())
	ERROR("priority queue is empty before net routing completed...");

      rnode=(PQ->top()).rNode;
      //We only want to consider nodes that have not been explored
      if(rnode->isExplored == true){
#ifdef ROUTER_DEBUG
	cout << "Popped " << (PQ->top()).rNode->get_name() <<" predecessor ";
	if((PQ->top()).predNode)
	  cout<<(PQ->top()).predNode->get_name();
	cout << " but node has been explored already!" <<endl;
	routerPops++;
#endif
	PQ->pop();
	continue;
      }
      else{
	//Mark the node explored, set its pathcost/delay and predecessor routing node
	//Notice that the node's pathcost/delay does not get written until it is actually popped off the PQ
	rnode->isExplored=true;
	rnode->pathCost=(PQ->top()).pathCost;
	rnode->netDelay = (PQ->top()).netDelay;
	rnode->predecessor=(PQ->top()).predNode;
	
	cleanupList.push_back(rnode);  //this list holds routing nodes that need a clean-up
	
#ifdef ROUTER_DEBUG
	cout<<endl<<"POPPED "<<rnode->get_name()<<" AT PATH COST "<<rnode->get_pathCost();
	cout<<" and delay " << rnode->netDelay << " AND PREDECESSOR NODE ";
	if(rnode->get_predecessor())
	  cout<<rnode->get_predecessor()->get_name();
	routerPops++;
#endif
	PQ->pop();
	break;
      }
    }

    //Check to see if we have found a new sink of the net
    if(rnode==sink_rnode){
#ifdef ROUTER_DEBUG
      cout << endl << "Found sink " <<  sink_rnode->name << " # " << sink_count << " of " << pNet->sinkVector.size() << " at cost " << rnode->pathCost;
      cout << " at delay " << rnode->netDelay;
      cout<<endl<<endl<<"++++++++++++ FOUND A SINK; WILL NOW ENTER BACKTRACE ++++++++++++";
#endif
      
      //Record the link delay from source to this sink
      //since the rnode may be shared with several nets
      pNet->linkDelayVector.at(sink_count) = rnode->netDelay;
      
      tree_node=backTrace(pNet,rnode,iter);

      if(sink_count<(pNet->sinkVector.size()-1)){
	//If there are unrouted sinks in this net we have to push the entire routing tree
	//into a new PQ
	delete PQ;
	PQ=new priority_queue<pqNode>;
	
#ifdef ROUTER_DEBUG
	cout << endl << "Routing tree for net " << pNet->name << " after adding sink " << sink_count;
	pNet->print_rTree();
#endif
	
	//When we find a sink, clean up the pathCost of the everything not in the routing tree
	while(!cleanupList.empty()){
	  if(cleanupList.front()->inrTree==false){
	    cleanupList.front()->isExplored=false;
	    cleanupList.front()->predecessor=0;
	    cleanupList.front()->pathCost=DBL_MAX;
	    cleanupList.front()->netDelay=DBL_MAX;
	  }
	  cleanupList.pop_front();
	}

	//now reinit the PQ to the current routing tree ala VPR's directed search
	//If we are doing timing-driven routing we have to reinit the PQ to the current routing tree
	//but each link is added at a cost of (unrouted sink's link criticality * path delay to node)
	//Notice that since the slackVector is initially set to all 0's, iteration 1 considers all links to be critical
#ifdef ROUTER_DEBUG
	cout << "Update PQ for net " << pNet->name << " node " << sink_count+1 << " of " << pNet->sinkVector.size() << " criticality "; 
	cout << (1 - (pNet->slackVector.at(sink_count+1)/criticalPathDelay)) << " slack= " << pNet->slackVector.at(sink_count+1);
	cout << " criticalPathDelay = " << criticalPathDelay << endl;
#endif
	updatePQTimingDriven(PQ, (pNet->rTree).root, pNet->sinkVector.at(sink_count+1)->rNode, 
			     calculateCrit(pNet->slackVector.at(sink_count+1), criticalPathDelay));
	sink_count++;
      }
      else{
	//We have just routed the last sink of the net
#ifdef ROUTER_DEBUG
	cout << endl << "Routing tree for net " << pNet->name << " after adding last sink " << sink_count<< endl;
	pNet->print_rTree();
	cout << "Cleaning up all nodes touched by search!" <<endl;
#endif
	
	//We have no more sinks to route so we should delete the PQ, clean up all of the nets we have touched
	//and end the search
	delete PQ;
	/* clean up all routing nodes that have been affected by the exploration */
	while(!cleanupList.empty()){
	  cleanupList.front()->isExplored=false;
	  cleanupList.front()->inrTree = false;
	  cleanupList.front()->predecessor=0;
	  cleanupList.front()->pathCost=DBL_MAX;
	  cleanupList.front()->netDelay=DBL_MAX;
	  cleanupList.pop_front();
	}

	//Also cleanup all nodes in the routing tree
	(pNet->get_rTree())->clean_rNodes(pNet->get_rTree()->get_root());

	//We have found all of the sinks so we are done!
	break;
      }
    }

    else{
      //This is not a sink so we should look at it's neighbors
      rnode->create_fanoutListItr();
      for(nb_edge=rnode->getnext_fanoutList();nb_edge!=0;nb_edge=rnode->getnext_fanoutList()){
	nb_rnode=nb_edge->destNode;

	//Check to see if this node's neighbor is a delay node that is not the sink we're looking for.
	//We only use this routine to route "unpipelined" signals, so we shouldn't be using delay nodes
	// to route.
	if(nb_rnode->type == DELAY && nb_rnode != sink_rnode){
	  continue;
	}

	/* A* algorithm */
	//heur_estimate=ALPHA*cost_to_target(nb_rnode,sink_rnode);
	heur_estimate = 0;
	
#ifdef ROUTER_DEBUG
	cout<<endl<<"LOOKING AT NEIGHBOR NODE "<<nb_rnode->get_name();
#endif
	
	//Only add unexplored neighbors that are not already in the routing tree to the queue
	//Also make sure they are within the routing bounding box
	if(nb_rnode->isExplored==false && nb_rnode->inrTree == false && pNet->checkNodeRoutingBoundingBox(nb_rnode)){
	  path_cost=rnode->pathCost;
	  net_delay = rnode->netDelay;
	  
	  //New timing driven cost
	  //Since the slackVector is initialized with all 0.0, the first iteration considers all links to be critical
	  new_path_cost = path_cost + nb_rnode->returnTimingCost(iter, linkCriticality, nb_edge->intrinsicDelay);
	  new_net_delay = net_delay + nb_rnode->intrinsicDelay + nb_edge->intrinsicDelay;
	  
	  PQ->push(pqNode(nb_rnode,new_path_cost,heur_estimate,rnode, new_net_delay));
#ifdef ROUTER_DEBUG
	  routerPushes++;
	  cout<<endl<<"\tADDED NODE "<<nb_rnode->get_name()<<" AT PATH COST "<<
	    new_path_cost<< " Delay " << new_net_delay<< " AND PREDECESSOR "<<rnode->get_name();
#endif
	}
      }
      rnode->remove_fanoutListItr();
    }
  }  
}

//For this pipelined net, just find a path between the source and all of the sinks.
void routePipelineNetPathFinderLockedRegistersCongestionDriven(int iter, PipeliningNet *pNet){
  routingNode* rnode;routingNode* nb_rnode;
  routingEdge* nb_edge;
  double path_cost,nb_path_cost,new_path_cost,heur_estimate;
  double net_delay, new_net_delay;
  Terminal* terminal;
  priority_queue<pqNode>* PQ;
  treeNode* tree_node;
  list<routingNode*> cleanupList;
  list<routingNode*>::iterator clean_itr;
  int i;

  list<routingNode*>::iterator rNodeIter;

#ifdef PARANOID_CHECK_ROUTE
  //Go through the entire routing graph and make sure that all nodes are unexplored and not in a routing tree
  for(rNodeIter = FPGA->rnodeList.begin(); rNodeIter != FPGA->rnodeList.end(); rNodeIter++){
    if((*rNodeIter)->inrTree == true){
      ERROR("Found routing node inside routing tree before routing has begun!");
    }
    else if((*rNodeIter)->isExplored == true){
      ERROR("Found explored routing node before routing has begun!");
    }
  }
#endif

  /* new root tree node. initialize it to source routing node */
  pNet->init_rTreeRoot();
  PQ=new priority_queue<pqNode>;


  //Get the source routing node
  terminal = pNet->source;
  rnode=terminal->get_rNode();

#ifdef ROUTER_DEBUG
  cout<<endl<<endl<<"<<<<<<<<<============ BEGIN ROUTER DEBUG FOR "<<pNet->name;
  cout<<"============>>>>>>>>>";
  cout<<endl<<"SOURCE OF THIS NET IS MAPPED TO "<<rnode->get_name();
#endif

  //Set the node as explored and add it to the PQ
  //If we are doing congestion-driven routing we need to add the source to the PQ at cost 0
  //Even if we are not in timing mode we should still report timing numbers
  PQ->push(pqNode(rnode,0.0,0.0,0, rnode->intrinsicDelay));
#ifdef ROUTER_DEBUG
  routerPushes++;
#endif

  rnode->set_inrTree(true);
  //If the source is an embedded pipelining net, don't increase it's occupancy.
  //  It's just a single node so if both the source and destination
  //  nets both increment the occupancy it will look like it's overused when it's not
  if(rnode->get_type() != DELAY){
    rnode->set_occupancy(rnode->get_occupancy()+1);
    if(rnode->get_occupancy() > rnode->get_capacity()){
      sharingCost+=1;
      //cout << "Node " << rnode->name << " is shared" << endl;
    }
  }

  /* start exploration */
  int sink_count=0;
  
#ifdef ROUTER_DEBUG
  cout << "Routing net " << pNet->name << endl;
#endif

  while(1){   
    //Loop until we have an unexplored routing node
    while(1){
      if(PQ->empty())
	ERROR("priority queue is empty before net routing completed...");

      rnode=(PQ->top()).rNode;
      //We only want to consider nodes that have not been explored
      if(rnode->isExplored == true){
#ifdef ROUTER_DEBUG
	cout << "Popped " << (PQ->top()).rNode->get_name() <<" predecessor ";
	if((PQ->top()).predNode)
	  cout<<(PQ->top()).predNode->get_name();
	cout << " but node has been explored already!" <<endl;
	routerPops++;
#endif
	PQ->pop();
	continue;
      }
      else{
	//Mark the node explored, set its pathcost/delay and predecessor routing node
	//Notice that the node's pathcost/delay does not get written until it is actually popped off the PQ
	rnode->isExplored=true;
	rnode->pathCost=(PQ->top()).pathCost;
	rnode->netDelay = (PQ->top()).netDelay;
	rnode->predecessor=(PQ->top()).predNode;
	
	cleanupList.push_back(rnode);  //this list holds routing nodes that need a clean-up
	
#ifdef ROUTER_DEBUG
	cout<<endl<<"POPPED "<<rnode->get_name()<<" AT PATH COST "<<rnode->get_pathCost();
	cout<<" and delay " << rnode->netDelay << " AND PREDECESSOR NODE ";
	if(rnode->get_predecessor())
	  cout<<rnode->get_predecessor()->get_name();
	routerPops++;
#endif
	PQ->pop();
	break;
      }
    }

    //Check to see if we have found any new sink of the net 
    if(pNet->foundNewSink(rnode)){
#ifdef ROUTER_DEBUG
      cout << endl << "Found sink " <<  rnode->name << " # " << sink_count << " of " << pNet->sinkVector.size() << " at cost " << rnode->pathCost;
      cout << " at delay " << rnode->netDelay;
      cout<<endl<<endl<<"++++++++++++ FOUND A SINK; WILL NOW ENTER BACKTRACE ++++++++++++";
#endif
      
      //Record the link delay from source to this sink
      //since the rnode may be shared with several nets
      pNet->linkDelayVector.at(sink_count) = rnode->netDelay;
      
      tree_node=backTrace(pNet,rnode,iter);

      if(sink_count<(pNet->sinkVector.size()-1)){
	//If there are unrouted sinks in this net we have to push the entire routing tree
	//into a new PQ
	delete PQ;
	PQ=new priority_queue<pqNode>;
	
#ifdef ROUTER_DEBUG
	cout << endl << "Routing tree for net " << pNet->name << " after adding sink " << sink_count;
	pNet->print_rTree();
#endif
	rnode->isRouted=true;
	
	//When we find a sink, clean up the pathCost of the everything not in the routing tree
	while(!cleanupList.empty()){
	  if(cleanupList.front()->inrTree==false){
	    cleanupList.front()->isExplored=false;
	    cleanupList.front()->predecessor=0;
	    cleanupList.front()->pathCost=DBL_MAX;
	    cleanupList.front()->netDelay=DBL_MAX;
	    cleanupList.front()->isRouted=false;
	  }
	  cleanupList.pop_front();
	}

	//now reinit the PQ to the current routing tree ala VPR's directed search
	//If we are doing timing-driven routing we have to reinit the PQ to the current routing tree
	//but each link is added at a cost of (unrouted sink's link criticality * path delay to node)
	//In congestion-driven routing we add the entire existing routing tree to the PQ at a cost of zero	   
	updatePQCongestionDriven(PQ,(pNet->rTree).root, pNet->sinkVector.at(sink_count+1)->rNode);
	sink_count++;
      }
      else{
	//We have just routed the last sink of the net
#ifdef ROUTER_DEBUG
	cout << endl << "Routing tree for net " << pNet->name << " after adding last sink " << sink_count<< endl;
	pNet->print_rTree();
	cout << "Cleaning up all nodes touched by search!" <<endl;
#endif
	
	//We have no more sinks to route so we should delete the PQ, clean up all of the nets we have touched
	//and end the search
	delete PQ;
	/* clean up all routing nodes that have been affected by the exploration */
	while(!cleanupList.empty()){
	  cleanupList.front()->isExplored=false;
	  cleanupList.front()->inrTree = false;
	  cleanupList.front()->predecessor=0;
	  cleanupList.front()->pathCost=DBL_MAX;
	  cleanupList.front()->netDelay=DBL_MAX;
	  cleanupList.front()->isRouted=false;
	  cleanupList.pop_front();
	}

	//Also cleanup all nodes in the routing tree
	(pNet->get_rTree())->clean_rNodes(pNet->get_rTree()->get_root());

	//We have found all of the sinks so we are done!
	break;
      }
    }

    else{
      //This is not a sink so we should look at it's neighbors
      //If this node was a non-sink register, then we should not explore the
      // neighbors of this node.
      //We are doing unpipelined routing, so they should not be used.
      if(rnode->type == DELAY){
	continue;
      }      

      rnode->create_fanoutListItr();
      for(nb_edge=rnode->getnext_fanoutList();nb_edge!=0;nb_edge=rnode->getnext_fanoutList()){
	nb_rnode=nb_edge->destNode;

	/* A* algorithm */
	//heur_estimate=ALPHA*cost_to_target(nb_rnode, rnode);
	heur_estimate = 0;
	
#ifdef ROUTER_DEBUG
	cout<<endl<<"LOOKING AT NEIGHBOR NODE "<<nb_rnode->get_name();
#endif
	
	//Only add unexplored neighbors that are not already in the routing tree to the queue
	//Also make sure they are within the routing bounding box
	if(nb_rnode->isExplored==false && nb_rnode->inrTree == false && pNet->checkNodeRoutingBoundingBox(nb_rnode)){
	  path_cost=rnode->pathCost;
	  net_delay = rnode->netDelay;
	  
	  //New cost
	  new_path_cost=path_cost+nb_rnode->returnCongestionCost(iter);
	  new_net_delay = net_delay + nb_rnode->intrinsicDelay + nb_edge->intrinsicDelay;
	    
	  PQ->push(pqNode(nb_rnode,new_path_cost,heur_estimate,rnode, new_net_delay));
#ifdef ROUTER_DEBUG
	  routerPushes++;
	  cout<<endl<<"\tADDED NODE "<<nb_rnode->get_name()<<" AT PATH COST "<<
	    new_path_cost<< " Delay " << new_net_delay<< " AND PREDECESSOR "<<rnode->get_name();
#endif
	}
      }
      rnode->remove_fanoutListItr();
    }
  }  
}
