// Copyright (C) 2006 Ken Eguro, Scott Hauck, Akshay Sharma
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

#include "include.h"

#ifdef ROUTER_DEBUG
extern long routerPushes;
#endif

extern long int sharingCost;
//extern long int wireCost;

double calculateCrit(double slack, double criticalPathDelay){
  //Should we dis-allow criticalities larger than 0.99?
  //This doesn't do it if we have negative slack
  return(doubmax(doubmin((0.99 - (slack / criticalPathDelay)), 0.99), 0));
}


//From scratch, calculate the arrival/departure time of all terminals in the circuit to determine
//the critical path.  Then update the pathDelay for all nets.
double calculateCriticalPathDelay(Netlist *netlist, bool pipeline_aware){
  double criticalPath;

  //First, initialize all arrival and departure times to big numbers
  netlist->initTimingGraph(pipeline_aware);

  //Fill in all arrival and departure times
  netlist->calculateArrivalDepartureTimes(pipeline_aware);

  criticalPath = netlist->returnCriticalPathDelay(pipeline_aware);

  netlist->calculateRequiredTimes(criticalPath, pipeline_aware);

  netlist->calculateSlack(criticalPath, pipeline_aware);

  return(criticalPath);
}

void initPQ(priority_queue<pqNode>& PQ,treeNode* tree_node,routingNode* new_sink_rnode){
  treeNode* t=tree_node;
  treeNode* child;
  double heur_estimate;

  t->get_rNode()->set_inrTree(true);

  //heur_estimate=ALPHA*cost_to_target(t->get_rNode(),new_sink_rnode);
  heur_estimate = 0;
  //if(t->get_rNode()->get_type()!=SINK && t->get_rNode()->get_type()!=INPUT_PIN){
  if(t->get_rNode()->get_type()!=SINK){
    if(t->get_parent()){
      PQ.push(pqNode(t->get_rNode(),0.,heur_estimate,t->get_parent()->get_rNode()));
#ifdef ROUTER_DEBUG
      routerPushes++;
#endif
    }
    else{
      PQ.push(pqNode(t->get_rNode(),0.,heur_estimate,0));
#ifdef ROUTER_DEBUG
      routerPushes++;
#endif
    }
    t->get_rNode()->set_pathCost(0.);
  }

  t->create_childListItr();
  for(child=t->getnext_childList();child!=0;child=t->getnext_childList())
    initPQ(PQ,child,new_sink_rnode);
  t->remove_childListItr();

}

//Examine the current tree node and put the it in the PQ.
//If we are doing timing-driven routing we have to reinit the PQ to the current routing tree
//but each link is added at a cost of (criticality * path delay to node)
//If we are doing congestion-driven routing everything is pushed in at a cost of 0.0
//Lastly, recursively call this on nodes the current node might be connected to
void updatePQTimingDriven(priority_queue<pqNode>* PQ,treeNode* tree_node,routingNode* new_sink_rnode, double criticality){
  treeNode* t=tree_node;
  treeNode* child;
  double heur_estimate;
  
  if(t->get_rNode()==0)
    ERROR("something fishy with the backTrace operation here...");
  if(t->get_parent()){
    if(t->get_parent()->get_rNode()!=t->get_rNode()->get_predecessor()){
      cout << "Node " << t->get_rNode()->name << endl;
      ERROR("parent mismatch here...");
    }
  }
  //cout << "Looking at node " << t->get_rNode()->name << endl; 
  
  //Add everthing in the routing tree to the PQ except for sinks
  //if(t->get_rNode()->get_type()!=SINK){
  /* A* algorithm */
  
  //heur_estimate=ALPHA*cost_to_target(t->get_rNode(),new_sink_rnode);
  heur_estimate = 0;
  PQ->push(pqNode(t->get_rNode(),criticality * t->get_rNode()->netDelay, heur_estimate, 
		  t->get_rNode()->get_predecessor(), t->get_rNode()->netDelay));
#ifdef ROUTER_DEBUG
  routerPushes++;
#endif
  
#if defined(ROUTER_DEBUG)
  cout <<endl<< "Criticality = " << criticality;
  cout<<endl<<"PUSHED "<<t->get_rNode()->get_name()<<" AT PATH COST " << criticality * t->get_rNode()->netDelay;
  cout<<" and delay " << t->get_rNode()->netDelay;
  if(t->get_rNode()->get_predecessor()){
    cout << " AND PREDECESSOR NODE " <<  t->get_rNode()->get_predecessor()->get_name() << endl;
  }
  else{
    cout << " with no predecessor " << endl;
  }
#endif      
  t->get_rNode()->set_pathCost(criticality * t->get_rNode()->netDelay);
  t->get_rNode()->set_inrTree(true);
  //Why is this false?
  //This is false so that we can check a node when we pop it
  t->get_rNode()->set_isExplored(false);
  
  t->create_childListItr();
  for(child=t->getnext_childList();child!=0;child=t->getnext_childList())
    updatePQTimingDriven(PQ,child,new_sink_rnode, criticality);
  t->remove_childListItr();
  
}

//Examine the current tree node and put the it in the PQ.
//If we are doing timing-driven routing we have to reinit the PQ to the current routing tree
//but each link is added at a cost of (criticality * path delay to node)
//If we are doing congestion-driven routing everything is pushed in at a cost of 0.0
//Lastly, recursively call this on nodes the current node might be connected to
void updatePQCongestionDriven(priority_queue<pqNode>* PQ,treeNode* tree_node,routingNode* new_sink_rnode){
  treeNode* t=tree_node;
  treeNode* child;
  double heur_estimate;
  
  if(t->get_rNode()==0)
    ERROR("something fishy with the backTrace operation here...");
  if(t->get_parent()){
    if(t->get_parent()->get_rNode()!=t->get_rNode()->get_predecessor()){
      cout << "Node " << t->get_rNode()->name << endl;
      ERROR("parent mismatch here...");
    }
  }
  //cout << "Looking at node " << t->get_rNode()->name << endl; 
  
  //Add everthing in the routing tree to the PQ except for sinks
  //if(t->get_rNode()->get_type()!=SINK){
  /* A* algorithm */
  
  //heur_estimate=ALPHA*cost_to_target(t->get_rNode(),new_sink_rnode);
  heur_estimate = 0;

  //Even if we are not in timing mode we should still report timing numbers
  PQ->push(pqNode(t->get_rNode(),0.,heur_estimate,t->get_rNode()->get_predecessor(), t->get_rNode()->netDelay));
#ifdef ROUTER_DEBUG
  routerPushes++;
#endif
    
#if defined(ROUTER_DEBUG)
  cout<<endl<<"PUSHED "<<t->get_rNode()->get_name()<<" AT PATH COST " << 0;
  cout<<" and delay " << t->get_rNode()->netDelay;
#endif
  
  t->get_rNode()->set_pathCost(0.);
  t->get_rNode()->set_inrTree(true);
  t->get_rNode()->set_isExplored(false);
  
  t->create_childListItr();
  for(child=t->getnext_childList();child!=0;child=t->getnext_childList())
    updatePQCongestionDriven(PQ,child,new_sink_rnode);
  t->remove_childListItr();
  
}

treeNode* backTrace(Net* net,routingNode* sink_rnode,int iter){
  treeNode* curr_tree_node;
  treeNode* prev_tree_node=0;
  routingNode* curr_node;

  /* start backtracing from the newly found sink to a node in the routing tree */
  for(curr_node=sink_rnode;curr_node->get_inrTree()==false;
      curr_node=curr_node->get_predecessor()){

    /* create a new tree node for the current routing node on the backtraced path */
    curr_tree_node=new treeNode(curr_node);

    //Set the routing node that this tree node points to be in the routing tree
    curr_tree_node->rNode->inrTree = true;


#if defined(ROUTER_DEBUG)
    cout<<endl<<"\tADDED "<<curr_node->get_name()<<" TO ROUTING TREE with predecessor " << curr_node->get_predecessor()->name;
#endif

    if(prev_tree_node!=0){
      /* hook up parent and child tree nodes */
      curr_tree_node->addto_childList(prev_tree_node);
      prev_tree_node->set_parent(curr_tree_node);
    }

    curr_node->set_occupancy(curr_node->get_occupancy()+1);
    prev_tree_node=curr_tree_node;

    /* change the sharing and wire costs to reflect the addition of this node */
    if(curr_node->get_occupancy()>curr_node->get_capacity()){
      //sharingCost += curr_node->get_occupancy() - curr_node->get_capacity();
      sharingCost+=1;
      
      // 	cout << "Node " << curr_node->name << " has a sharing of " << curr_node->get_occupancy() - curr_node->get_capacity() << endl;
      
      // 	if(curr_node->name.compare("2178") == 0){
      // 	  cout << "FOUND 2178!" << endl;
      // 	}
      //cout << "Increasing sharing cost due to node " << curr_node->name << " occupancy = " << curr_node->get_occupancy() << endl;
    }
//     if(curr_node->get_type()==BUS_OR_WIRE){
//       //wireCost+=1;
//     }
  }

  /* curr_node now points to a routing node that is present in the route tree. */
  /*   cout<<endl<<endl<<"NEED TO FIND "<<curr_node->get_name()<<" IN ROUTE TREE"; */

  /* find the tree node corresponding to curr_node */
  curr_tree_node=net->find_treeNode(curr_node);

  /* connect the new tree segment to the existing route tree */
  if(prev_tree_node!=0){
    curr_tree_node->addto_childList(prev_tree_node);
    prev_tree_node->set_parent(curr_tree_node);
  }

#if defined(ROUTER_DEBUG)
  cout<<endl<<endl<<"LEAVING BACKTRACE...";
#endif

  /*   cout<<endl<<endl<<"FOUND TREE NODE "<<curr_tree_node->get_rNode()->get_name(); */

  /* return the tree node that joins the new tree segment to the existing route tree */
  return prev_tree_node; 

}

treeNode* backTrace(PipeliningNet* pNet,routingNode* sink_rnode,int iter){
  treeNode* curr_tree_node;
  treeNode* prev_tree_node=0;
  routingNode* curr_node;

  /* start backtracing from the newly found sink to a node in the routing tree */
  for(curr_node=sink_rnode;curr_node->get_inrTree()==false;
      curr_node=curr_node->get_predecessor()){

    /* create a new tree node for the current routing node on the backtraced path */
    curr_tree_node=new treeNode(curr_node);

    //Set the routing node that this tree node points to be in the routing tree
    curr_tree_node->rNode->inrTree = true;


#if defined(ROUTER_DEBUG)
    cout<<endl<<"\tADDED "<<curr_node->get_name()<<" TO ROUTING TREE with predecessor " << curr_node->get_predecessor()->name;
#endif

    if(prev_tree_node!=0){
      /* hook up parent and child tree nodes */
      curr_tree_node->addto_childList(prev_tree_node);
      prev_tree_node->set_parent(curr_tree_node);
    }

    curr_node->set_occupancy(curr_node->get_occupancy()+1);
    prev_tree_node=curr_tree_node;

    /* change the sharing and wire costs to reflect the addition of this node */
    if(curr_node->get_occupancy()>curr_node->get_capacity()){
       sharingCost+=1;

       //if(curr_node->name.compare("385") == 0){
       // cout << "FOUND 385!" << endl;
       //}

       //cout << "Increasing sharing cost due to node " << curr_node->name << " occupancy = " << curr_node->get_occupancy() << endl;
       //sharingCost += curr_node->get_occupancy() - curr_node->get_capacity();
    }
    if(curr_node->get_type()==BUS_OR_WIRE){
      //wireCost+=1;
    }
  }

  /* curr_node now points to a routing node that is present in the route tree. */
  /*   cout<<endl<<endl<<"NEED TO FIND "<<curr_node->get_name()<<" IN ROUTE TREE"; */

  /* find the tree node corresponding to curr_node */
  curr_tree_node=pNet->find_treeNode(curr_node);

  /* connect the new tree segment to the existing route tree */
  if(prev_tree_node!=0){
    curr_tree_node->addto_childList(prev_tree_node);
    prev_tree_node->set_parent(curr_tree_node);
  }

#if defined(ROUTER_DEBUG)
  cout<<endl<<endl<<"LEAVING BACKTRACE...";
#endif

  /*   cout<<endl<<endl<<"FOUND TREE NODE "<<curr_tree_node->get_rNode()->get_name(); */

  /* return the tree node that joins the new tree segment to the existing route tree */
  return prev_tree_node; 

}
