// Copyright (C) 2006 Ken Eguro, Scott Hauck, Akshay Sharma
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

#include "include.h"
extern long int sharingCost;
extern int MAXID;

#ifdef ROUTER_DEBUG
extern long routerPops;
extern long routerPushes;
#endif

//This is the primary entry point for the timing-driven pipelined router.
void routeNetQuickRoute_D(int iter, Net* net, double criticalPathDelay){
  PipeliningNet *pNet;
  //char tempStr [MAX_STR_LEN];
  string tempStr;

  list<PipeliningNet *>::iterator pNetIter;
  priority_queue<pqNode>* PQ;
  list<pExplore> cleanupList;
  vector <unsigned int> *rootPathVector;

  vector<Terminal*>::iterator sinkIter;
  vector<int>::iterator latencyIter;

  int failedLatency;
  int critIndex;

  bool success;

  //Check to make sure that all of the routing nodes are un-explored/etc before we start
#ifdef PARANOID_CHECK_ROUTE_PIPELINE
  checkEmptyRRGraph(true);
#endif
  
  //First, we should create a pipelined net for the source and put it in the list
  //strcpy(tempStr, net->name);
  //sprintf(tempStr, "%sPipelineNet%d", tempStr, net->pipelineNetList.size());
  tempStr = net->name + "PipelineNet" + itoa(net->pipelineNetList.size());
  pNet = new PipeliningNet(tempStr, net->source, 0, net);
  net->pipelineNetList.push_back(pNet);

  //Make the source terminal point to this new pipelined net
  pNet->source->destPipeliningNet = pNet;

  //Put the source node into the routing tree of the first pipelined net
  //and increment the occupancy
  pNet->init_rTreeRoot();

  //Increment occupancy and put it into the tree at latency 0
  if(pNet->source->rNode->get_type() != DELAY){
    pNet->source->rNode->set_occupancy(pNet->source->rNode->occupancy + 1);
    if(pNet->source->rNode->get_occupancy() > pNet->source->rNode->get_capacity()){
      sharingCost+=1;
    }
  }
  pNet->source->rNode->set_inPRTree(0);

  //Make a new PQ
  PQ=new priority_queue<pqNode>();
      
  //Timing-driven routing finds sinks in order
  for(sinkIter = net->sinkVector.begin(), latencyIter = net->latencyVector.begin(); 
      sinkIter != net->sinkVector.end(), latencyIter != net->latencyVector.end(); 
      sinkIter++, latencyIter++){
    //cout << "Attempting to route to sink " << (*sinkIter)->name << endl;
    
    //Check to make sure that all of the routing nodes are un-explored (although not the inPRTree)
#ifdef PARANOID_CHECK_ROUTE_PIPELINE
    checkEmptyRRGraph(false);
#endif

    //Make sure that the PQ is empty
    if(!PQ->empty()){
      ERROR("PQ not empty!");
    }

    //Add all of the existing routing trees of an equal or smaller latency than the target into the PQ
    //cout << "Adding existing paths to PQ" << endl;
    for(pNetIter = net->pipelineNetList.begin(); pNetIter != net->pipelineNetList.end(); pNetIter++){
      if((*pNetIter)->latency <= *latencyIter){
	//cout << "\tCreating base pathvector for pipelining net " << (*pNetIter)->name << endl;
	rootPathVector = createPathVectorRoot(net, *pNetIter, &cleanupList);
	//cout << "\tAdding nodes of pipelining net " << (*pNetIter)->name << " to PQ" << endl;
	for(critIndex = 0; critIndex < NUMASSUMEDCRIT; critIndex++){
	  addRoutingNodes2PQ_D(PQ, (*pNetIter), (*pNetIter)->rTree.root, critIndex, 
			     (*pNetIter)->latency, rootPathVector, &cleanupList, criticalPathDelay);
	}
	delete rootPathVector;
      }
    }
      
    //Find a sink at the appropriate latency and backtrace/add sinks to appropriate pipelined nets
    //cout << "Begin routing to sink " << (*sinkIter)->name << " at latency " << (*latencyIter) << endl;
    if(!findRouteQuickRouteSink_D(iter, net, *sinkIter, *latencyIter, PQ, &cleanupList, criticalPathDelay)){
      //OK, the router got stuck, so try to route to this sink
      cout << "Primary router failed, invoking secondary router" << endl;

      //Try to route to this sink from decreasing latency existing routes
      //At the very worse, we can always try routing again from the 0 latency route
      //This will almost always allow for a N latency route if one exists...
      //cout << "Attempting to route to " << (*sinkIter)->name << " at latency " << *latencyIter << endl;

      failedLatency = *latencyIter;
      success = false;
      while(!success){
 	//Make sure that the PQ is empty
 	if(!PQ->empty()){
 	  ERROR("PQ not empty!");
 	}
	      
 	cout << "Attempting to begin from " << failedLatency << " latency nets" << endl;

 	//Add only the N latency nets to the PQ
	for(pNetIter = net->pipelineNetList.begin(); pNetIter != net->pipelineNetList.end(); pNetIter++){
	  if((*pNetIter)->latency == failedLatency){ 
	    //cout << "\tAdding nodes of pipelining net " << (*pNetIter)->name << " to PQ" << endl;
	    rootPathVector = createPathVectorRoot(net, *pNetIter, &cleanupList);	
	    for(critIndex = 0; critIndex < NUMASSUMEDCRIT; critIndex++){
	      addRoutingNodes2PQ_D(PQ, (*pNetIter), (*pNetIter)->rTree.root, critIndex, 
				   (*pNetIter)->latency, rootPathVector, &cleanupList, criticalPathDelay);
	    }
	    delete rootPathVector;
	  }
	}

	//Try to find this particular sink
	success = findRouteQuickRouteSink_D(iter, net, *sinkIter, *latencyIter, PQ, &cleanupList, criticalPathDelay);	      

	if(success){
	  cout << "Succeeded!" << endl;
	}

	else{
	  cout << "Failed!" << endl;
	  if(failedLatency > 0){
	    failedLatency--;
	  }
	  else{
	    cout << "Failed from all latencies!  I'm out of suggestions!" << endl;
	    ERROR("Failed routing!");
	  }
	}
      }
    }
  }

  //We are done routing so we should reset all of the sinks to unrouted
  for(sinkIter = net->sinkVector.begin(); sinkIter != net->sinkVector.end(); sinkIter++){
    (*sinkIter)->isRouted = false;
  }
  
  //Go through all of the pipelining nets and un-set the inRoutingTree variables
  for(pNetIter = net->pipelineNetList.begin(); pNetIter != net->pipelineNetList.end(); pNetIter++){
    (*pNetIter)->rTree.clearInPRTree((*pNetIter)->rTree.root);
  }

  delete PQ;
  
  //cout << "Net completed routing!" << endl;
}

//Find a timing-driven route from the given PQ to a particular sink.
bool findRouteQuickRouteSink_D(int iter, Net* net, Terminal *sinkTerminal, int latency, priority_queue<pqNode>* PQ, 
			       list<pExplore> *cleanupList, double criticalPathDelay){
   PipeliningNet* pNet;

  routingNode* currentRNode;
  int currentLatency;
  int currentExploreIndex;
  int currentACIndex;
  double currentCongestion;
  double currentExternalDelay;
  double currentUpstreamCost;
  vector <unsigned int> *currentPathVector;

  vector<Terminal*>::iterator sinkIter;
  vector<int>::iterator latencyIter;
  routingNode* nb_rnode;
  routingEdge* nb_edge;
  vector <unsigned int> *nb_pathVector;  

  double path_cost;
  double delay;
  double new_path_cost;
  double new_delay;
  double actual_delay;
  double new_congestion;
  //double crit;
  int critIndex;

  bool foundAtLatency;
  
  pExplore *prev;

  //Try to route to the sink
  while(1){
    //Grab something usable off of the PQ
    while(1){
      if(PQ->empty()){
	//We have emptied the PQ before we found the sink
	cout << "Could not find route for net " << net->name << endl;

	//Clean up all nodes that have been touched
	while(!cleanupList->empty()){
	  cleanupList->front().rNode->pPredecessor.at(cleanupList->front().assumed_crit_index).at(cleanupList->front().latency).at(cleanupList->front().exploreIndex).clear();
	  cleanupList->front().rNode->pPathCost.at(cleanupList->front().assumed_crit_index).at(cleanupList->front().latency).at(cleanupList->front().exploreIndex) = DBL_MAX;
	  cleanupList->front().rNode->pNetDelay.at(cleanupList->front().assumed_crit_index).at(cleanupList->front().latency).at(cleanupList->front().exploreIndex) = DBL_MAX;
	  cleanupList->front().rNode->pIsExplored.at(cleanupList->front().assumed_crit_index).at(cleanupList->front().latency) = 0;
	  cleanupList->front().rNode->nodeID = 0;
	  cleanupList->pop_front();
	}
	return false;
      }
      
      currentRNode=(PQ->top()).rNode;
      currentLatency = PQ->top().latency;
      currentACIndex = PQ->top().assumed_crit_index;
      currentExploreIndex = currentRNode->pIsExplored.at(currentACIndex).at(currentLatency);
      
      //If this is a delay node, we have to see if we are the first one to reach it at this latency, at any assumed criticality
      if(currentRNode->type == DELAY ){
	if(currentRNode->check_exploredAtLatency(currentLatency)){
#if defined(ROUTER_DEBUG)
	  cout << "\tPopped delay node " << currentRNode->get_name() <<" predecessor ";
	  if((PQ->top()).pred.rNode){
	    cout<<(PQ->top()).pred.rNode->get_name();
	    cout << " AC index " << (PQ->top()).pred.assumed_crit_index << " latency " << (PQ->top()).pred.latency;
	    cout << "and e_Index " << (PQ->top()).pred.exploreIndex;
	  }
	  cout << " cong " << (PQ->top()).netCongestion;
	  cout << " exDelay " << (PQ->top()).externalDelay;
	  cout << " upCost " << (PQ->top()).upstreamCost;
	  cout << " but node has been explored at latency " << currentLatency << endl;
#endif
	  delete PQ->top().pathVector;
	  PQ->pop();
#ifdef ROUTER_DEBUG
	  routerPops++;
#endif
	  continue;
	}
      }

      //Determine if we have explored the current node K times already at the given latency and ACIndex
      if(currentExploreIndex >= KFUNCTION(currentLatency)){
#if defined(ROUTER_DEBUG)
	cout << "\tPopped " << currentRNode->get_name() <<" predecessor ";
	if((PQ->top()).pred.rNode){
	  cout<<(PQ->top()).pred.rNode->get_name();
	  cout << " AC index " << (PQ->top()).pred.assumed_crit_index << " latency " << (PQ->top()).pred.latency;
	  cout << " and e_Index " << (PQ->top()).pred.exploreIndex; 
	}
	cout << " cong " << (PQ->top()).netCongestion;
	cout << " exDelay " << (PQ->top()).externalDelay;
	cout << " upCost " << (PQ->top()).upstreamCost;
	cout << " but node has been explored at latency " << currentLatency << " ";
	cout << currentExploreIndex << " times by AC index " << currentACIndex << endl;
#endif
	delete PQ->top().pathVector;
	PQ->pop();
#ifdef ROUTER_DEBUG
	routerPops++;
#endif
	continue;
      }
      else{
	//This node is free to be explored
	currentPathVector = PQ->top().pathVector;  

	//Mark node explored once more at current latency and set pathcost/delay/predecessor
	currentRNode->pIsExplored.at(currentACIndex).at(currentLatency)++;
	currentRNode->pPathCost.at(currentACIndex).at(currentLatency).at(currentExploreIndex) = (PQ->top()).pathCost;
	currentRNode->pNetDelay.at(currentACIndex).at(currentLatency).at(currentExploreIndex) = (PQ->top()).netDelay;
	currentRNode->pPredecessor.at(currentACIndex).at(currentLatency).at(currentExploreIndex) = (PQ->top()).pred;
	addNode2PathVector(currentRNode, currentPathVector, cleanupList);
	currentCongestion = (PQ->top()).netCongestion;
	currentExternalDelay = (PQ->top()).externalDelay;
	currentUpstreamCost = (PQ->top()).upstreamCost;

	cleanupList->push_back(pExplore(currentRNode, currentACIndex, currentLatency, currentExploreIndex));
		
#if defined(ROUTER_DEBUG)
	cout <<"\tPopped "<<currentRNode->get_name();
	cout << " at path cost "<< currentRNode->pPathCost.at(currentACIndex).at(currentLatency).at(currentExploreIndex);
	cout << " has been explored at AC index " << currentACIndex << ", latency " << currentLatency << " " << currentExploreIndex << " times has delay ";
	cout << currentRNode->pNetDelay.at(currentACIndex).at(currentLatency).at(currentExploreIndex) << " and predecessor node ";
	
	if(currentRNode->pPredecessor.at(currentACIndex).at(currentLatency).at(currentExploreIndex).rNode){
	  cout << currentRNode->pPredecessor.at(currentACIndex).at(currentLatency).at(currentExploreIndex).rNode->get_name();
	  cout << " AC index " << currentRNode->pPredecessor.at(currentACIndex).at(currentLatency).at(currentExploreIndex).assumed_crit_index;
	  cout << " latency " << currentRNode->pPredecessor.at(currentACIndex).at(currentLatency).at(currentExploreIndex).latency;
	  cout << " and e_Index " << currentRNode->pPredecessor.at(currentACIndex).at(currentLatency).at(currentExploreIndex).exploreIndex;
	}
	cout << " cong " << currentCongestion;
	cout << " exDelay " << currentExternalDelay;
	cout << " upCost " << currentUpstreamCost;
	cout << endl;
#endif

	pNet = PQ->top().sourcePNet;

	PQ->pop();	
#ifdef ROUTER_DEBUG
	routerPops++;
#endif
	break;
      }
    }

    //We've got something usable off of the PQ, so now see if we have hit the sink at the correct latency
    if(currentRNode == sinkTerminal->rNode && currentLatency == latency){
#if defined(ROUTER_DEBUG)
	cout << "\tFOUNDSINK " <<  currentRNode->name << " at latency " << currentLatency;
	cout << " at cost " << currentRNode->pPathCost.at(currentACIndex).at(currentLatency).at(currentExploreIndex);
	cout << " and delay " << currentRNode->pNetDelay.at(currentACIndex).at(currentLatency).at(currentExploreIndex) << endl <<endl;
#endif

	//Mark sink as routed
	sinkTerminal->isRouted = true;

	//Backtrace and create the appropriate pipelining nets, etc.
	backTracePipelinedRoute_D(net, pNet, sinkTerminal, currentACIndex, currentLatency, currentExploreIndex, criticalPathDelay);

	delete currentPathVector;
	
	//Clean up all nodes that have been touched
	while(!cleanupList->empty()){
	  cleanupList->front().rNode->pPredecessor.at(cleanupList->front().assumed_crit_index).at(cleanupList->front().latency).at(cleanupList->front().exploreIndex).clear();
	  cleanupList->front().rNode->pPathCost.at(cleanupList->front().assumed_crit_index).at(cleanupList->front().latency).at(cleanupList->front().exploreIndex) = DBL_MAX;
	  cleanupList->front().rNode->pNetDelay.at(cleanupList->front().assumed_crit_index).at(cleanupList->front().latency).at(cleanupList->front().exploreIndex) = DBL_MAX;
	  cleanupList->front().rNode->pIsExplored.at(cleanupList->front().assumed_crit_index).at(cleanupList->front().latency) = 0;
	  cleanupList->front().rNode->nodeID = 0;
	  cleanupList->pop_front();
	}
	MAXID = 0;

	//We need to free the rest of the path vectors that we pushed onto the queue
	while(!PQ->empty()){
	  delete PQ->top().pathVector;
	  PQ->pop();
	}

	//Exit
	return(true);
    }

    //We didn't hit the sink at the correct latency, so we should add neighbors to PQ
    currentRNode->create_fanoutListItr();
    for(nb_edge=currentRNode->getnext_fanoutList(); nb_edge!=0; nb_edge=currentRNode->getnext_fanoutList()){
      nb_rnode=nb_edge->destNode;
      
#if defined(ROUTER_DEBUG)
      cout<<"\t\tLooking at neighbor node "<<nb_rnode->get_name() << endl;
#endif
      
      //Only add explore-able neighbors to the queue
      //Don't add the neighbor if-
      //1) Neighbor is in explored path to current node
      if(checkNodeID(currentPathVector, nb_rnode->nodeID)){
	//cout << "\t\t\tNeigbor not added since already in current path" << endl;
        continue;
      }

      //2) Neighbor has already been explored K times at latency or once already at latency 0
      //   Notice that we do not do this check if we are at a delay node right now since we launch a new
      //   series of ACindexes at delay nodes.
      if(currentRNode->type != DELAY){
	if(nb_rnode->pIsExplored.at(currentACIndex).at(currentLatency) >= KFUNCTION(currentLatency)){
	  //cout << "\t\t\tNeighbor not added since already explored " << nb_rnode->pIsExplored.at(currentACIndex).at(currentLatency) << " times" << endl;
	  continue;
	}
      }

      //3) Neighbor is a delay node and the sink that we are looking for does not require a latency that high
      if(nb_rnode->type == DELAY && currentLatency == latency){
	//cout << "\t\t\tNeighbor not added since we do not need to search above latency " << currentLatency << endl;
	continue;
      }
      
      //4) Neighbor is already in tree at given latency
      if((nb_rnode->type != DELAY && nb_rnode->check_inPRTree(currentLatency)) ||
	 (nb_rnode->type == DELAY && nb_rnode->check_inPRTree(currentLatency + 1))){
	if(nb_rnode->type == DELAY){
	  //cout << "\t\t\tNeighbor node not added since it is already in the routing tree at latency " << currentLatency + 1 << endl;
	}
	else{
	  //cout << "\t\t\tNeighbor node not added since it is already in the routing tree at latency " << currentLatency << endl;
	}
	continue;
      }

      //5) Neighbor is a delay node and it has already been explored at the next latency
      if(nb_rnode->type == DELAY && nb_rnode->check_exploredAtLatency(currentLatency + 1)){
	//cout << "\t\t\tNeighbor delay node not added since already explored at latency " << currentLatency + 1<< endl;
	continue;
      }

      //Add the neighbor to the PQ
      path_cost = currentRNode->pPathCost.at(currentACIndex).at(currentLatency).at(currentExploreIndex);
      
      if(currentRNode->type == DELAY ){
	delay = currentRNode->intrinsicDelay;
      }
      else{
	delay = currentRNode->pNetDelay.at(currentACIndex).at(currentLatency).at(currentExploreIndex);
      }

      if(nb_rnode->type == DELAY || (nb_rnode == sinkTerminal->rNode && currentLatency == latency)){
	//Delay nodes get explored at latency L + 1 and do not add in their own intrinsic delay
	if(nb_rnode->type == DELAY){
	  new_delay = delay + nb_edge->intrinsicDelay;
	}
	//Sinks get explored at latency L and do include their own intrinsic delay
	else{
	  new_delay = delay + nb_rnode->intrinsicDelay + nb_edge->intrinsicDelay;
	}

	//If our neighbor is a delay node or a sink it doesn't matter if we are currently at a delay node or not.
	//This is because we add the neighbor in at it's measured criticality.	
	//In this case, the cost of the path to a delay node is:
	//[MC/(1-MC)]*delay since/to last reg + congestion since last reg + combined delay/congestion cost of any upstream routes

	//Notice that if we are at latency 0 or latency L, we might have a non-zero external delay.
	//Latency 0 sources have to see the upstream delay from the nearest upstream register and latency L sinks
	//need to see the downstream delay from the nearest downstream register.
	//Normally, we don't have to do this check since we usually incrementally add to the cost instead of recalculating.
	//That means that only the latency 0 and L waves will have a non-zero "currentExternalDelay" value".
	actual_delay = new_delay + currentExternalDelay;
	new_congestion = currentCongestion +  nb_rnode->returnCongestionCost(iter);

	//For the first iteration, the criticalPathDelay is set to DBL_MIN so as to ensure that all the nets are considered to be critical
	//crit = doubmin((delay / criticalPathDelay), 0.99);
	//crit = crit / (1 - crit);
	//new_path_cost = (crit * actual_delay) + new_congestion + currentUpstreamCost;
	new_path_cost = calculateNewPathCost(actual_delay, criticalPathDelay, new_congestion, currentUpstreamCost);
	
	//Make a new path vector for the neighbor
	nb_pathVector = new vector <unsigned int> (*currentPathVector);

	if(nb_rnode->type == DELAY){
	  if(currentLatency + 1 == latency){
	    //Latency L nets need to see the extra delay past the sink so we need to push these searches onto the PQ with the correct external delay loading
	    PQ->push(pqNode(pNet, nb_rnode, new_path_cost, new_delay, currentACIndex, currentLatency + 1, 
			    pExplore(currentRNode, currentACIndex, currentLatency, currentExploreIndex), nb_pathVector,
			    0.0, sinkTerminal->externalDelay, new_path_cost));

#ifdef ROUTER_DEBUG
	    routerPushes++;
	    cout <<"\t\t\tPushed Delay Node " << nb_rnode->get_name() <<" at Path Cost " << new_path_cost<< " Delay " << new_delay;
	    cout << " AC index " << currentACIndex << " latency " << currentLatency + 1;
	    cout << " and predecessor "<< currentRNode->get_name();
	    cout << " AC index " << currentACIndex;
	    cout << " latency " << currentLatency;
	    cout << " e_Index " << currentExploreIndex;
	    cout << " cong " << 0.0;
	    cout << " exDelay " << sinkTerminal->externalDelay;
	    cout << " and upCost " << new_path_cost << endl;
#endif	  
	  }

	  else{
	    //Non-latency L delay nodes should re-set the congestion cost, external delay and upstream cost.
	    PQ->push(pqNode(pNet, nb_rnode, new_path_cost, new_delay, currentACIndex, currentLatency + 1, 
			    pExplore(currentRNode, currentACIndex, currentLatency, currentExploreIndex), nb_pathVector,
			    0.0, 0.0, new_path_cost));
#ifdef ROUTER_DEBUG
	    routerPushes++;
	    cout <<"\t\t\tPushed Delay Node " << nb_rnode->get_name() <<" at Path Cost " << new_path_cost<< " Delay " << new_delay;
	    cout << " AC index " << currentACIndex << " latency " << currentLatency + 1;
	    cout << " and predecessor "<< currentRNode->get_name();
	    cout << " AC index " << currentACIndex;
	    cout << " latency " << currentLatency;
	    cout << " e_Index " << currentExploreIndex;
	    cout << " cong " << 0.0;
	    cout << " exDelay " << 0.0;
	    cout << " and upCost " << new_path_cost << endl;
#endif	  
	  }	  
	}
	else{
	  //Notice that this is exactly the same procedure for both delay node and sink except that we push a sink in at the current latency instead of + 1
	  PQ->push(pqNode(pNet, nb_rnode, new_path_cost, new_delay, currentACIndex, currentLatency, 
			  pExplore(currentRNode, currentACIndex, currentLatency, currentExploreIndex), nb_pathVector,
			  0.0, sinkTerminal->externalDelay, new_path_cost));
#ifdef ROUTER_DEBUG
	  routerPushes++;
	  cout <<"\t\t\tPushed sink " << nb_rnode->get_name() <<" at Path Cost " << new_path_cost<< " Delay " << new_delay;
	  cout << " AC index " << currentACIndex << " latency " << currentLatency;
	  cout << " and predecessor "<< currentRNode->get_name();
	  cout << " AC index " << currentACIndex;
	  cout << " latency " << currentLatency;
	  cout << " e_Index " << currentExploreIndex;
	  cout << " cong " << 0.0;
	  cout << " exDelay " << sinkTerminal->externalDelay;
	  cout << " and upCost " << new_path_cost << endl;
#endif	  	  
	}
      }
      else{
	//The cost of non-delay, non-sink nodes is:
	//combined delay/congestion cost of any upstream routes +
	//        [AC/(1-AC)]*(unregistered upstream/downstream delay) + [AC/(1-AC)]*(delay within routing tree) + 
	//        [AC/(1-AC)]*(delay up to node from this exploration) + 
	//        (congestion since last register)
	//The first three terms have already been accounted for.
	//The first term is accounted for when we pushed the post-delay nodes into the PQ.
	//The second and third terms are accounted for either by previous explorations
	//or when we added to the PQ from the routing tree.
	//Thus all we have to add in is the incremental delay contribution and the congestion contibution
	new_delay = delay + nb_rnode->intrinsicDelay + nb_edge->intrinsicDelay;
	new_congestion = currentCongestion +  nb_rnode->returnCongestionCost(iter);

	if(currentRNode->type == DELAY ){
	  //If we are currently at a delay node we should spawn a new series of assumed criticalities
	  for(critIndex = 0; critIndex < NUMASSUMEDCRIT; critIndex++){
	    if(nb_rnode->pIsExplored.at(critIndex).at(currentLatency) >= KFUNCTION(currentLatency)){
	      //cout << "\t\t\tNeighbor not added since already explored " << nb_rnode->pIsExplored.at(critIndex).at(currentLatency) << " times" << endl;
	      continue;
	    }
	    
	    //crit = convertACrit(critIndex);
	    //crit = crit / (1 - crit);
	    //new_path_cost = path_cost + (crit * (nb_rnode->intrinsicDelay + nb_edge->intrinsicDelay)) + nb_rnode->returnCongestionCost(iter);
	    new_path_cost = path_cost + (calculateNewPathCostAC(critIndex, nb_rnode->intrinsicDelay + nb_edge->intrinsicDelay,
								nb_rnode->returnCongestionCost(iter), 0.0));			

	    //Make a new path vector for the neighbor
	    nb_pathVector = new vector <unsigned int> (*currentPathVector);
	    PQ->push(pqNode(pNet, nb_rnode, new_path_cost, new_delay, critIndex, currentLatency, 
			    pExplore(currentRNode, currentACIndex, currentLatency, currentExploreIndex), nb_pathVector,
			    new_congestion, currentExternalDelay, currentUpstreamCost));

#ifdef ROUTER_DEBUG
	    routerPushes++;
	    cout <<"\t\t\tPushed Node " << nb_rnode->get_name() <<" at Path Cost " << new_path_cost<< " Delay " << new_delay;
	    cout << " AC index " << critIndex;
	    cout << " latency " << currentLatency;
	    cout << " and predecessor "<< currentRNode->get_name();
	    cout << " AC index " << currentACIndex;
	    cout << " latency " << currentLatency;
	    cout << " and e_Index " << currentExploreIndex;
	    cout << " cong " << new_congestion;
	    cout << " exDelay " << currentExternalDelay;
	    cout << " and upCost " << currentUpstreamCost << endl;
#endif	  	  
	  }
	}
	else{
	  //If we are not currently at a delay node we should just continue the same assumed criticality that we have now.
	  //However, don't add this neighbor to the PQ if my neighbor's delay would make the total net's delay equal to or 
	  //greater than the difference between the current and next assumed criticality indexes.
	  //We do not prune the largest assumed criticality, though
	  if(currentACIndex != (NUMASSUMEDCRIT - 1)){
	    if(new_delay >= (convertACrit(currentACIndex + 1) * criticalPathDelay)){
	      //cout << "\t\t\tNot adding node " << nb_rnode->name << " at ACIndex " << currentACIndex;
	      //cout << " since it will be better covered by another AC Index" << endl;
	      continue;
	    }
	  }
      
	  //crit = convertACrit(currentACIndex);
	  //crit = crit / (1 - crit);
	  //new_path_cost = path_cost + (crit * (nb_rnode->intrinsicDelay + nb_edge->intrinsicDelay)) + nb_rnode->returnCongestionCost(iter) + nb_edge->intrinsicCost;
	  new_path_cost = path_cost + (calculateNewPathCostAC(currentACIndex, nb_rnode->intrinsicDelay + nb_edge->intrinsicDelay,
							      nb_rnode->returnCongestionCost(iter), 0.0));		  


	  //Make a new path vector for the neighbor
	  nb_pathVector = new vector <unsigned int> (*currentPathVector);
	  PQ->push(pqNode(pNet, nb_rnode, new_path_cost, new_delay, currentACIndex, currentLatency, 
			  pExplore(currentRNode, currentACIndex, currentLatency, currentExploreIndex), nb_pathVector,
			  new_congestion, currentExternalDelay, currentUpstreamCost));
#ifdef ROUTER_DEBUG
	  routerPushes++;
	  cout <<"\t\t\tPushed Node " << nb_rnode->get_name() <<" at Path Cost " << new_path_cost<< " Delay " << new_delay;
	  cout << " AC index " << currentACIndex << " latency " << currentLatency;
	  cout << " and predecessor "<< currentRNode->get_name();
	  cout << " AC index " << currentACIndex;
	  cout << " latency " << currentLatency;
	  cout << " and e_Index " << currentExploreIndex;
	  cout << " cong " << new_congestion;
	  cout << " exDelay " << currentExternalDelay;
	  cout << " and upCost " << currentUpstreamCost << endl;
#endif	  	  
	}	
      }
    }
    currentRNode->remove_fanoutListItr();;

    delete currentPathVector;
  }
}

//This adds all of the routing nodes in the pipelining net's routing tree to the PQ
//This is my version of updatePQ
//One big difference is that this function also needs to fill in the pathvector.
//This is the timing-driven version so we add the nodes with their respective timing cost
//Notice that we trim additions that 
void addRoutingNodes2PQ_D(priority_queue<pqNode>* PQ, PipeliningNet* pNet, treeNode* tree_node, 
			int assumed_crit_index, int latency, vector <unsigned int> *basePathVector, list<pExplore> *cleanupList, double criticalPathDelay){
  treeNode* t=tree_node;
  treeNode* child;
  routingNode* parentRNode;
  vector <unsigned int> *pathVector;
  //double crit;
  double cost;
  
  //This net may have nothing in it's routing tree
  if(!t){
    return;
  }

  if(t->get_rNode()==0)
    ERROR("something fishy with the backTrace operation here...");

  if(t->get_parent()){
    parentRNode = t->get_parent()->get_rNode();
  }
  else{
    parentRNode = 0;
  }

  //Add everything in the routing tree to the PQ except for non-source delay nodes
  //Sink delay nodes get added to the PQ by their destination nets
  if(t->get_rNode()->get_type() != DELAY || pNet->source->rNode == t->get_rNode()){
    //cout << "\t\tAdding node " << t->get_rNode()->name << " to PQ" << endl;

    //Make sure that I am in the routing tree at the proper latency
    if(!t->rNode->check_inPRTree(latency)){
      ERROR("Node is not in PRouting tree!");
    }

    //If we are routing in timing-driven mode, push the entire tree at the given latency and 
    //a cost of the (assumed criticality * the intrinsic delay) + upward stream delay cost
    //As far as the predecessor goes, set it to null
    //Notice that source delay nodes have to be treated differently.  Their
    //linkDelay is actually for the path TO, the route, not the new one, since this is
    //obviously 0.0
    
    //Do not add in node if current delay is already more than difference between current
    //and next assumed criticality index.
    //E.g. the current critical path delay is 10 ns and I'm exploring a node at an assumed criticality of 0.3.
    //When the delay along this path gets to be 4ns or greater I will stop adding children.
    //An exeception is made for explorations of the largest assumed criticality index, (NUMASSUMEDCRIT - 1)
    //Since we are working on routing graphs here we will assume that there are no negative edges.
    if(assumed_crit_index != (NUMASSUMEDCRIT - 1)){
      if((t->linkDelay + pNet->source->externalDelay) >= (convertACrit(assumed_crit_index + 1) * criticalPathDelay)){
	//cout << "\t\t\tNot adding node " << t->get_rNode()->get_name() << " at ACIndex " << assumed_crit_index;
	//cout << " since it will be better covered by another AC Index" << endl;
	return;
      }
    }
    
    //Make a copy of the base pathvector
    pathVector = new vector <unsigned int> (*basePathVector);
    
    //Add myself into the pathvector
    addNode2PathVector(t->rNode, pathVector, cleanupList);
    
    //Notice that only the latency 0 net will add in a non-zero external delay
    //and only the non-latency 0 nets will add in a non-zero upstreamCost
    //We add these in with no congestion cost
    //crit = convertACrit(assumed_crit_index);
      //cost = ((crit / (1 - crit)) * (t->linkDelay + pNet->source->externalDelay)) + pNet->source->upstreamCost;
    cost = calculateNewPathCostAC(assumed_crit_index, t->linkDelay + pNet->source->externalDelay, 0.0, pNet->source->upstreamCost);
    
    PQ->push(pqNode(pNet, t->get_rNode(), cost, t->linkDelay, assumed_crit_index, latency, pExplore(), pathVector, 0.0, pNet->source->externalDelay, pNet->source->upstreamCost));
#ifdef ROUTER_DEBUG
    routerPushes++;
    cout<<"\t\tPushed "<< t->get_rNode()->get_name() <<" at Path Cost " << cost;
    cout<<", delay " << t->linkDelay << " ACIndex " << assumed_crit_index << " latency " << latency;
    cout << " upstreamDelay " << pNet->source->externalDelay << " upstreamCost " <<  pNet->source->upstreamCost<< endl; 
#endif

    //Add any children to the PQ
    t->create_childListItr();
    for(child=t->getnext_childList();child!=0;child=t->getnext_childList()){
      addRoutingNodes2PQ_D(PQ, pNet, child, assumed_crit_index, latency, pathVector, cleanupList, criticalPathDelay);
    }
    t->remove_childListItr();
  }
}

void backTracePipelinedRoute_D(Net* net, PipeliningNet* sourcePNet, Terminal* sinkTerminal, 
			       int assumed_crit_index, int lat, int e_index, double criticalPathDelay){
  treeNode* curr_tree_node;
  treeNode* prev_tree_node=0; 
  pExplore* current;
  pExplore* end;

  PipeliningNet *currPNet;
  PipeliningNet *nextPNet;
  Terminal* currSink;
  //char tempStr [MAX_STR_LEN];
  string tempStr;

  double actual_delay;
  double crit;
  list <Terminal*> upstreamCostList;

  Terminal* currTerm;
  Terminal* nextTerm;

  //cout << "Backtrace" << endl;  
  //cout << "\tPath back to source or source net" << endl;
  end = new pExplore(sinkTerminal->rNode, assumed_crit_index, lat, e_index);
  current = end;
  //btpr(current);

  //Figure out what pipelining net should be the host for the sink terminal
  if(lat != sourcePNet->latency){
    //strcpy(tempStr, net->name);
    //sprintf(tempStr, "%sPipelineNet%d", tempStr, net->pipelineNetList.size());
    tempStr = net->name + "PipelineNet" + itoa(net->pipelineNetList.size());
    currPNet = new PipeliningNet(tempStr, lat, net);
    net->pipelineNetList.push_back(currPNet);
  }
  else{
    currPNet = sourcePNet;
  }

  //Add the sink terminal to the current pipelining net
  currPNet->sinkVector.push_back(sinkTerminal);
  currPNet->linkDelayVector.push_back(current->rNode->pNetDelay.at(current->assumed_crit_index).at(current->latency).at(current->exploreIndex));
  currPNet->slackVector.push_back(0.0);
  sinkTerminal->sourcePipeliningNet = currPNet;
  
  while(1){
    //See if we are done
    if(!current->rNode->pPredecessor.at(current->assumed_crit_index).at(current->latency).at(current->exploreIndex).rNode){
      //Reattach this node to the source net
      //Find this node in the source net, then add the correct children/parent relationship
      //If a net has the same sink listed twice there will not be a previous tree node...
      if(prev_tree_node){
	curr_tree_node = currPNet->find_treeNode(current->rNode);
	curr_tree_node->addto_childList(prev_tree_node);
	prev_tree_node->set_parent(curr_tree_node);
      }

      //Go through the list of terminals to set the proper upstream costs
      currTerm = currPNet->source;
      while(!upstreamCostList.empty()){
	nextTerm = upstreamCostList.front();
	upstreamCostList.pop_front();
	
	nextTerm->upstreamCost += currTerm->upstreamCost;
	//cout << "Terminal " << nextTerm->name << " upstream cost set to " << nextTerm->upstreamCost << endl;
	currTerm = nextTerm;
      }

      //Release the beginning pExplore
      delete end;
      break;
    }
    
    if(current->rNode->type == DELAY){
      //This is the root of the current tree.
      //Notice that we do NOT use the pNetDelay value, this is the delay to the node from the source side, not the
      //delay starting from the node to the sink side.
      curr_tree_node = new treeNode(current->rNode, current->rNode->intrinsicDelay);
      
      //cout << "\t\tMaking new treeNode for ";
      //cout << "Node " << current->rNode->name << " delay " << current->rNode->intrinsicDelay << endl;
    }
    else{
      curr_tree_node = new treeNode(current->rNode, current->rNode->pNetDelay.at(current->assumed_crit_index).at(current->latency).at(current->exploreIndex));

      //cout << "\t\tMaking new treeNode for ";
      //cout << "Node " << current->rNode->name << " delay " << current->rNode->pNetDelay.at(current->assumed_crit_index).at(current->latency).at(current->exploreIndex) << endl;
    }
    
    if(prev_tree_node != 0){
      curr_tree_node->addto_childList(prev_tree_node);
      prev_tree_node->set_parent(curr_tree_node);
    }

    //Increment occupancy
    current->rNode->set_occupancy(current->rNode->get_occupancy()+1);
    if(current->rNode->get_occupancy()>current->rNode->get_capacity()){
      sharingCost+=1;
      //cout << "Increasing sharing cost due to node " << current->rNode->name << endl;
    }
    if(current->rNode->get_type()==BUS_OR_WIRE){
      //wireCost+=1;
    }
    //Mark myself as in the PRouting tree
    if(current->rNode->check_inPRTree(current->latency)){
      ERROR("Node already in PRouting tree!");
    }
    else{
      current->rNode->set_inPRTree(current->latency);
    }

    //If we have hit a delay node so we need to do a few things
    if(current->rNode->type == DELAY){
      //1) Make it the root of the current pipelining net
      currPNet->rTree.set_root(curr_tree_node);

      //2) Make a new tree node for this delay node
      curr_tree_node = new treeNode(current->rNode, current->rNode->pNetDelay.at(current->assumed_crit_index).at(current->latency).at(current->exploreIndex));

      //3) Evaluate the upstream pipelining net
      if(current->latency - 1 != sourcePNet->latency){
	//We need to create a new pipelining net and add it to the list
	//strcpy(tempStr, net->name);
	//sprintf(tempStr, "%sPipelineNet%d", tempStr, net->pipelineNetList.size());
	tempStr = net->name + "PipelineNet" + itoa(net->pipelineNetList.size());
	nextPNet = new PipeliningNet(tempStr, current->latency - 1, net);
	net->pipelineNetList.push_back(nextPNet);
      }
      else{
	nextPNet = sourcePNet;
      }

      //4) Make a new terminal node for this delay node and add it to the main net's list
      //strcpy(tempStr, net->name);
      //sprintf(tempStr, "%sReg%d", tempStr, net->pipelineRegisterList.size());
      tempStr = net->name + "Reg" + itoa( net->pipelineRegisterList.size());
      currSink = new Terminal(net, current->rNode, nextPNet, currPNet, tempStr);
      net->pipelineRegisterList.push_back(currSink);

      //5) Make this terminal to be the source of the old net and add it to the sink list of the new one
      currPNet->source = currSink;
      nextPNet->sinkVector.push_back(currSink);
      nextPNet->linkDelayVector.push_back(curr_tree_node->linkDelay);
      nextPNet->slackVector.push_back(0.0);

      //6) Set the partial upstream cost for this terminal
      actual_delay = current->rNode->pNetDelay.at(current->assumed_crit_index).at(current->latency).at(current->exploreIndex) + currPNet->source->externalDelay;
      crit = doubmin((actual_delay / criticalPathDelay), 0.99);
      //crit = actual_delay / criticalPathDelay;
      crit = crit / (1 - crit);
      currSink->upstreamCost = crit * actual_delay;
      //cout << "Set partial upstream cost of Terminal " << currSink->name << " to " << currSink->upstreamCost << endl;
      upstreamCostList.push_front(currSink);

      //7) Make the new net the current net since we a completely done with the previous one
      currPNet = nextPNet;
    }

    prev_tree_node=curr_tree_node;      
    current = &(current->rNode->pPredecessor.at(current->assumed_crit_index).at(current->latency).at(current->exploreIndex));
  }
}
