// Copyright (C) 2006 Ken Eguro, Scott Hauck, Akshay Sharma
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

#include "include.h"
extern vprFPGA *FPGA;
extern bool maxCutCost;
extern bool carefulHoming;

#ifdef PLACER_DEBUG
extern int numRegistersEstimatedFreed;
extern int numRegistersEstimatedNeeded;
#endif
extern bool lut_level_placement;

// We assume that registers are added in the correct order if that is important
// That is, in a particular bank, register 0 is added before register 1,
// register 1 is added before register 2, etc.if they are in a
// shift-chain orientation.
void registerBank::addRoutingNode(routingNode *r){
  routingNodeVector.push_back(r);
}

//Return the first sink in this bank
Terminal* registerBank::getFirstSink(){
  if(currSinkList.empty()){
    return NULL;
  }
  else{
    return currSinkList.front();
  }
}

Terminal* registerBank::getLastSource(){
  if(currSourceList.empty()){
    return NULL;
  }
  else{
    return currSourceList.back();
  }
}

//Return the first routing node of this register bank
routingNode* registerBank::getFirstRoutingNode(){
  if(routingNodeVector.empty()){
    ERROR("Register bank has no routing nodes assigned!");
  }
  else{
    return routingNodeVector.front();
  }
}

bool registerBank::isBankAndLinkedLUTEmpty(){
  if(currSourceList.size() != currSinkList.size()){
    ERROR("Bank has a source/sink mismatch!");
  }
  if(linkedLUT){
    if(!linkedLUT->isJustLUTEmpty()){
      return false;
    }
  }
  
  return(currSourceList.empty());
}

bool registerBank::isJustBankEmpty(){
  if(currSourceList.size() != currSinkList.size()){
    ERROR("Bank has a source/sink mismatch!");
  }
  return(currSourceList.empty());
}

bool registerBank::isLinkedLUTEmpty(){
  if(linkedLUT){
    return(linkedLUT->isJustLUTEmpty());
  }
  return true;
}

//How many registers are currently mapped to this bank?
int registerBank::getNumMappedRegs(){
  if(currSourceList.size() != currSinkList.size()){
    ERROR("Bank has a source/sink mismatch!");
  }
  return(currSourceList.size());
}

//How many registers could be mapped to this bank?
int registerBank::getNumPossibleDelay(){
  //if(routingNodeVector.size() == 0){
  //  ERROR("Bank has no routing nodes assigned!");
  //}
  return(routingNodeVector.size());
}

//How many registers are currently vacent in this bank?
int registerBank::getNumFreeRegs(){
  if(currSourceList.size() != currSinkList.size()){
    ERROR("Bank has a source/sink mismatch!");
  }

  if(linkedLUT){
    if(!linkedLUT->isJustLUTEmpty()){
      return 0;
    }
  }
  return(routingNodeVector.size() - currSourceList.size());
}


//Is this bank at capacity?
bool registerBank::bankFull(){
  if(routingNodeVector.size() == 0){
    ERROR("Bank has no routing nodes assigned!");
  }

  if(routingNodeVector.size() == currSourceList.size()){
    return true;
  }
  else{
    return false;
  }
}

//Put this register into this bank
void registerBank::placeTerminal(Terminal *T){
  // See if this bank is already full
  if(bankFull()){
    ERROR("placeTerminal: Attempted to map terminal to full bank!");
  }

  if(linkedLUT){
    if(linkedLUT->currBlock){
      ERROR("A linked LUT should be empty!");
    }
  }

  if(currSourceList.size() != currSinkList.size()){
    ERROR("Somehow the source and sink vectors got out of sync!");
  }

  if(isJustBankEmpty()){
    FPGA->occupyRegBank(this);
  }

  //Since this is only used by pipelined nets, there is only
  // one terminal - it is both a source and a sink
  currSinkList.push_back(T);
  currSinkList.back()->set_rNode(routingNodeVector.at(currSinkList.size() - 1));
  currSourceList.push_back(T);
  currSourceList.back()->set_rNode(routingNodeVector.at(currSourceList.size() - 1));

  //Make sure this terminal also knows who it belongs to
  T->setRegisterBank(this);
}

//Put this register into the front of this bank instead of the end
//Notice, this does not set any routing nodes since things get
// screwed up every time we push in a new register.
//We must call updateRoutingNodes after this function is done.
void registerBank::placeTerminalFront(Terminal *T){
  // See if this bank is already full
  if(bankFull()){
    ERROR("placeTerminal: Attempted to map terminal to full bank!");
  }

  if(linkedLUT){
    if(linkedLUT->currBlock){
      ERROR("A linked LUT should be empty!");
    }
  }

  if(currSourceList.size() != currSinkList.size()){
    ERROR("Somehow the source and sink vectors got out of sync!");
  }

  if(isJustBankEmpty()){
    FPGA->occupyRegBank(this);
  }

  //Since this is only used by pipelined nets, there is only
  // one terminal - it is both a source and a sink
  currSinkList.push_front(T);
  currSourceList.push_front(T);

  //Make sure this terminal also knows who it belongs to
  T->setRegisterBank(this);
}

void registerBank::placeLogicBlock(logicBlock *lb){
  if(!lb){
    ERROR("Cannot place null logic block!");
  }

  if(currSourceList.size() != currSinkList.size()){
    ERROR("Somehow the source and sink vectors got out of sync!");
  }

  if(!isJustBankEmpty()){
    ERROR("Attempted to map logic block to occupied bank!");
  }

  if(get_currBlock()){
    ERROR("Attempted to map logic block to bank already allocated to another logic block!!");
  }

  FPGA->occupyRegBank(this);
  
  //Make sure that the logic block knows who it is mapped to
  lb->set_registerBank(this);
  lb->set_ptc_index(hostBankNum);
  currBlock = lb;
  
  //Map the input net of this lb to the first sink node of the RB
  if(lb->inputList.size() > 1){
    ERROR("Register has more than one input!");
  }
  currSinkList.push_back(lb->inputList.front());
  currSinkList.back()->set_rNode(routingNodeVector.at(currSinkList.size() - 1));

  //Map the output net of this lb to the first source node the RB
  if(lb->outputList.size() > 1){
    ERROR("Register has more than one output!");
  }
  currSourceList.push_back(lb->outputList.front());
  currSourceList.back()->set_rNode(routingNodeVector.at(currSourceList.size() - 1));
}

void registerBank::unPlaceLogicBlock(logicBlock *lb){
  if(!lb){
    ERROR("Cannot unplace null logic block!");
  }

  if(currSourceList.size() != currSinkList.size()){
    ERROR("Somehow the source and sink vectors got out of sync!");
  }

  if(isJustBankEmpty()){
    ERROR("Attempted to unmap empty bank!");
  }

  if(get_currBlock() != lb){
    ERROR("Attempted to unmap logic block that was not previously mapped!");
  }

  FPGA->vacateRegBank(this);
  
  //Make sure that the logic block knows who it is mapped to
  lb->unset_registerBank(this);
  lb->set_ptc_index(-1);
  currBlock = 0;
  
  if(currSinkList.size() != 1){
    ERROR("Sink list should have exactly one terminal!");
  }
  
  currSinkList.front()->clearRNode();
  currSinkList.clear();

  if(currSourceList.size() != 1){
    ERROR("Source list should have exactly one terminal!");
  }
  
  currSourceList.front()->clearRNode();
  currSourceList.clear();
}

//Put this register into the bank after the parent.
//The parent might not be mapped to this register bank, though.
void registerBank::insertRegisterForwards(Terminal *T, Terminal *parent){
  list <Terminal *>::iterator sinkIter;
  list <Terminal *>::iterator sourceIter;
  int i;

  // See if this bank is already full
  if(bankFull()){
    ERROR("Attempted to map terminal to full bank!");
  }
  
  if(currSourceList.size() != currSinkList.size()){
    ERROR("Somehow the source and sink vectors got out of sync!");
  }

  if(isJustBankEmpty()){
    FPGA->occupyRegBank(this);
    currSinkList.push_back(T);
    currSinkList.back()->set_rNode(routingNodeVector.at(currSinkList.size() - 1));
    currSourceList.push_back(T);
    currSourceList.back()->set_rNode(routingNodeVector.at(currSourceList.size() - 1));    
  }
  else{
    //Find the parent in the sink/source list
    for(sinkIter = currSinkList.begin(), sourceIter = currSourceList.begin(), i = 0; 
	sinkIter != currSinkList.end() && sourceIter != currSourceList.end();
	sinkIter ++, sourceIter++, i++){
      if(*sinkIter == parent){
	break;
      }
    }
    
    if(sinkIter == currSinkList.end()){
      //We didn't find the parent in the list, so just put it in the front of the bank
      // and then update the routing node assignments of everything behind it
      currSinkList.push_front(T);
      currSourceList.push_front(T);
      for(sinkIter = currSinkList.begin(), i = 0; sinkIter != currSinkList.end(); sinkIter++, i++){
	(*sinkIter)->set_rNode(routingNodeVector.at(i));
      }
    }
    else{
      //We did find the parent in the list, so insert it behind the parent, then update 
      // the routing node assignment of everyone behind the parent
      sinkIter++;
      sourceIter++;
      i++;
      currSinkList.insert(sinkIter, T);
      currSourceList.insert(sourceIter, T);
      sinkIter --;
      sourceIter --;
      for( ; sinkIter != currSinkList.end(); sinkIter++, i++){
	(*sinkIter)->set_rNode(routingNodeVector.at(i));
      }
    }
  }
}

//Put this register into the bank after the parent.
//The parent might not be mapped to this register bank, though.
void registerBank::insertRegisterBackwards(Terminal *T, Terminal *sink){
  list <Terminal *>::iterator sinkIter;
  list <Terminal *>::iterator sourceIter;
  int i;

  // See if this bank is already full
  if(bankFull()){
    ERROR("Attempted to map terminal to full bank!");
  }
  
  if(currSourceList.size() != currSinkList.size()){
    ERROR("Somehow the source and sink vectors got out of sync!");
  }

  if(isJustBankEmpty()){
    FPGA->occupyRegBank(this);
    currSinkList.push_back(T);
    currSinkList.back()->set_rNode(routingNodeVector.at(currSinkList.size() - 1));
    currSourceList.push_back(T);
    currSourceList.back()->set_rNode(routingNodeVector.at(currSourceList.size() - 1));    
  }
  else{
    //Find the sink in the sink/source list
    for(sinkIter = currSinkList.begin(), sourceIter = currSourceList.begin(), i = 0; 
	sinkIter != currSinkList.end() && sourceIter != currSourceList.end();
	sinkIter ++, sourceIter++, i++){
      if(*sinkIter == sink){
	break;
      }
    }
    
    if(sinkIter == currSinkList.end()){
      //We didn't find the sink in the list, so just put it in the back of the bank
      // and then update the routing node assignment
      currSinkList.push_back(T);
      currSourceList.push_back(T);
      currSinkList.back()->set_rNode(routingNodeVector.at(currSourceList.size() - 1));
    }
    else{
      //We did find the sink in the list, so insert it in front of the sink, then update 
      // the routing node assignment of everyone behind the parent
      currSinkList.insert(sinkIter, T);
      currSourceList.insert(sourceIter, T);
      sinkIter--;
      sourceIter--;
      for( ; 	sinkIter != currSinkList.end(); sinkIter++, i++){
	(*sinkIter)->set_rNode(routingNodeVector.at(i));
      }
    }
  }
}

//Take the entire contents of this register bank and swap it with
// the contests of the other one.
//This is only called from pipelining registers, which means we are not
// doing clb-level placement.
void registerBank::swapBank(registerBank *otherBank){
  list <Terminal*> tempTermList;
  list <Terminal*>::iterator iter;
  int i;
  logicBlock *tempLb;

  //The other block may be empty, so we will make it full and empty this one
  if(otherBank->isJustBankEmpty()){
    FPGA->occupyRegBank(otherBank);
    FPGA->vacateRegBank(this);
  }

  //Make sure that the two banks are still in sync
  if(currSinkList.size() != currSourceList.size() || 
     otherBank->currSinkList.size() != otherBank->currSourceList.size()){
    ERROR("Bank is out of sync!");
  }

  //Copy all of sinks from this bank
  for(iter = currSinkList.begin(); iter != currSinkList.end(); iter++){
    tempTermList.push_back(*iter);
  }
  //Clear the sink list from this bank
  currSinkList.clear();  

  //Copy the sinks from the other bank and update what register bank they are pointed to,
  //along with the routing node they are mapped to
  for(iter = otherBank->currSinkList.begin(), i = 0;
      iter != otherBank->currSinkList.end(); iter++, i++){
    currSinkList.push_back(*iter);
    (*iter)->setRegisterBank(this);
    (*iter)->set_rNode(routingNodeVector.at(i));
  }
  //Clear the sink list from the other block
  otherBank->currSinkList.clear();

  //Copy the sinks that used to be in this block into the other block
  for(iter = tempTermList.begin(), i = 0; iter != tempTermList.end(); iter++, i++){
    otherBank->currSinkList.push_back(*iter);
    (*iter)->setRegisterBank(otherBank);
    (*iter)->set_rNode(otherBank->routingNodeVector.at(i));
  }
  //Clear the temp sink list
  tempTermList.clear();

  //Copy all of sources from this bank
  for(iter = currSourceList.begin(); iter != currSourceList.end(); iter++){
    tempTermList.push_back(*iter);
  }
  //Clear the source list from this bank
  currSourceList.clear();  

  //Copy the sources from the other bank
  for(iter = otherBank->currSourceList.begin(), i = 0;
      iter != otherBank->currSourceList.end(); iter++, i++){
    currSourceList.push_back(*iter);
    (*iter)->setRegisterBank(this);
    (*iter)->set_rNode(routingNodeVector.at(i));
  }
  //Clear the source list from the other bank
  otherBank->currSourceList.clear();

  //Copy the sources that used to be in this bank into the other bank
  for(iter = tempTermList.begin(), i = 0; iter != tempTermList.end(); iter++, i ++){
    otherBank->currSourceList.push_back(*iter);
    (*iter)->setRegisterBank(otherBank);
    (*iter)->set_rNode(otherBank->routingNodeVector.at(i));
  }
  
  //Find out if either bank was associated with a logic block
  //If so, we should update the logic block to point to the new delay block and bank
  tempLb = currBlock;
  currBlock = otherBank->currBlock;
  otherBank->currBlock = tempLb;
  if(currBlock){
    currBlock->unset_registerBank(otherBank);
    currBlock->set_registerBank(this);
  }

  if(otherBank->currBlock){
    otherBank->currBlock->unset_registerBank(this);
    otherBank->currBlock->set_registerBank(otherBank);
  }
}

//Take the entire contents of this register bank and swap it with
// the contests of the other one.
//This register bank is occupied, but the other one may be empty or
// have an occupied linked LUT location.
//Lastly, update the timing information.
void registerBank::swapBankAndAssociatedLUT(registerBank *otherBank, double criticalPathDelay){
  list <Terminal*> tempTermList;
  list <Terminal*>::iterator iter;
  int i;
  logicBlock *tempLb;

  //The other block may be empty, so we will make it full and empty this one
  if(otherBank->isBankAndLinkedLUTEmpty()){
    FPGA->occupyRegBank(otherBank);
    FPGA->vacateRegBank(this);
  }
  else{
    //See if the other bank has an occupied linked LUT.
    //If so, migrate it over.
    if(otherBank->hasLinkedLUT() && otherBank->linkedLUT->currBlock){
      //Make sure we have a linked LUT.
      if(!linkedLUT){
	ERROR("Register bank does not have linked LUT!");
      }
      
      linkedLUT->swapLUTAndAssociatedRegBanks(otherBank->linkedLUT);

      //Update the sink and slack timings for the lut and registers we just moved
      //Our linked LUT is occupied.
      linkedLUT->updateSinkDelaysAndSlacks();
      //The other linked LUT is not occupied, but we have to update all registers
      // that are linked to the linked LUT.
      otherBank->linkedLUT->updateSinkDelaysAndSlacks();
      
      //Update the input and output criticalities for the registers we just moved
      //Our linked LUT is occupied.
      linkedLUT->currBlock->calculateMaxInputOutputCriticality(criticalPathDelay);
      //The other linked LUT is not occupied, but we have to update all registers
      // that are linked to the linked LUT.
      otherBank->linkedLUT->calculateMaxInputOutputCriticality(criticalPathDelay);
      return;
    }
  }
  
  //If we made it this far, we only want to transfer the 
  // contents of the two specified register banks.
  //Make sure that the two banks are still in sync
  if(currSinkList.size() != currSourceList.size() || 
     otherBank->currSinkList.size() != otherBank->currSourceList.size()){
    ERROR("Bank is out of sync!");
  }

  //Copy all of sinks from this bank
  for(iter = currSinkList.begin(); iter != currSinkList.end(); iter++){
    tempTermList.push_back(*iter);
  }
  //Clear the sink list from this bank
  currSinkList.clear();  

  //Copy the sinks from the other bank and update what register bank they are pointed to,
  //along with the routing node they are mapped to
  for(iter = otherBank->currSinkList.begin(), i = 0;
      iter != otherBank->currSinkList.end(); iter++, i++){
    currSinkList.push_back(*iter);
    (*iter)->setRegisterBank(this);
    (*iter)->set_rNode(routingNodeVector.at(i));
  }
  //Clear the sink list from the other block
  otherBank->currSinkList.clear();

  //Copy the sinks that used to be in this block into the other block
  for(iter = tempTermList.begin(), i = 0; iter != tempTermList.end(); iter++, i++){
    otherBank->currSinkList.push_back(*iter);
    (*iter)->setRegisterBank(otherBank);
    (*iter)->set_rNode(otherBank->routingNodeVector.at(i));
  }
  //Clear the temp sink list
  tempTermList.clear();

  //Copy all of sources from this bank
  for(iter = currSourceList.begin(); iter != currSourceList.end(); iter++){
    tempTermList.push_back(*iter);
  }
  //Clear the source list from this bank
  currSourceList.clear();  

  //Copy the sources from the other bank
  for(iter = otherBank->currSourceList.begin(), i = 0;
      iter != otherBank->currSourceList.end(); iter++, i++){
    currSourceList.push_back(*iter);
    (*iter)->setRegisterBank(this);
    (*iter)->set_rNode(routingNodeVector.at(i));
  }
  //Clear the source list from the other bank
  otherBank->currSourceList.clear();

  //Copy the sources that used to be in this bank into the other bank
  for(iter = tempTermList.begin(), i = 0; iter != tempTermList.end(); iter++, i ++){
    otherBank->currSourceList.push_back(*iter);
    (*iter)->setRegisterBank(otherBank);
    (*iter)->set_rNode(otherBank->routingNodeVector.at(i));
  }
  
  //Find out if either bank was associated with a logic block
  //If so, we should update the logic block to point to the new delay block and bank
  tempLb = currBlock;
  currBlock = otherBank->currBlock;
  otherBank->currBlock = tempLb;
  if(currBlock){
    currBlock->unset_registerBank(otherBank);
    currBlock->set_registerBank(this);
  }

  if(otherBank->currBlock){
    otherBank->currBlock->unset_registerBank(this);
    otherBank->currBlock->set_registerBank(otherBank);
  }

  //Update the sink and slack timings for the registers we just moved
  updateSinkDelaysAndSlacks();
  otherBank->updateSinkDelaysAndSlacks();
  
  //Update the input and output criticalities for the registers we just moved
  calculateMaxInputOutputCriticalitiesJustRB(criticalPathDelay);
  otherBank->calculateMaxInputOutputCriticalitiesJustRB(criticalPathDelay);
}

//Copy all of the sources and sinks to the other bank, then clean this bank out
void registerBank::copyContentsToAndEmpty(registerBank *otherBank){
  list <Terminal*>::iterator iter;
  int i;

  //The other block may be empty, so we will make it full and empty this one
  if(!otherBank->isBankAndLinkedLUTEmpty()){
    ERROR("The other bank should be empty!");
  }

  FPGA->occupyRegBank(otherBank);
  FPGA->vacateRegBank(this);

  //Make sure that the two banks are still in sync
  if(currSinkList.size() != currSourceList.size()){
    ERROR("Bank is out of sync!");
  }

  //Copy all of sinks from this bank
  for(iter = currSinkList.begin(), i = 0; iter != currSinkList.end(); iter++, i++){
    otherBank->currSinkList.push_back(*iter);
    (*iter)->setRegisterBank(otherBank);
    (*iter)->set_rNode(otherBank->routingNodeVector.at(i));
  }
  //Clear the sink list from this bank
  currSinkList.clear();  

  //Copy all of sources from this bank
  for(iter = currSourceList.begin(), i = 0; iter != currSourceList.end(); iter++, i++){
    otherBank->currSourceList.push_back(*iter);
    (*iter)->setRegisterBank(otherBank);
    (*iter)->set_rNode(otherBank->routingNodeVector.at(i));
  }
  //Clear the source list from this bank
  currSourceList.clear();  
}

void registerBank::updateSinkDelaysAndSlacks(){
  list <Terminal*>::iterator iter;

  if(currBlock){
    currBlock->updateSinkDelaysAndSlacks();
  }

  //Make sure that the bank is still in sync
  if(currSinkList.size() != currSourceList.size()){
    ERROR("Bank is out of sync!");
  }

  // Update the sink timing for the nets attached to this block
  for(iter = currSinkList.begin(); iter != currSinkList.end(); iter++){
    //This terminal is the sink of a net.
    //Update the sink timing for this net
    (*iter)->updateSinkDelayAndSlack();
  }
  
  // Recalculate the placement cost for the nets attached to the sources of this logic block
  for(iter = currSourceList.begin(); iter != currSourceList.end(); iter++){
    //This terminal is the source of a net.
    //Update the sink timings for all the sinks of this net
    (*iter)->updateSourceDelaysAndSlacks();
  }
}

void registerBank::updateSinkDelaysAndSlacksAndAssociatedLUT(){
  list <Terminal*>::iterator iter;

  //Make sure that the bank is still in sync
  if(currSinkList.size() != currSourceList.size()){
    ERROR("Bank is out of sync!");
  }

  // Update the sink timing for the nets attached to this block
  for(iter = currSinkList.begin(); iter != currSinkList.end(); iter++){
    //This terminal is the sink of a net.
    //Update the sink timing for this net
    (*iter)->updateSinkDelayAndSlack();
  }
  
  // Recalculate the placement cost for the nets attached to the sources of this logic block
  for(iter = currSourceList.begin(); iter != currSourceList.end(); iter++){
    //This terminal is the source of a net.
    //Update the sink timings for all the sinks of this net
    (*iter)->updateSourceDelaysAndSlacks();
  }

  //See if there is a linked LUT we should update too
  if(linkedLUT && linkedLUT->currBlock){
    linkedLUT->currBlock->updateSinkDelaysAndSlacks();
  }
}

void registerBank::updateWireAndTimingCosts(bool timing_driven, double criticalPathDelay){
  list<Terminal*>::iterator iter;
  
  //Make sure that the bank is still in sync
  if(currSinkList.size() != currSourceList.size()){
    ERROR("Bank is out of sync!");
  }

  // Recalculate the placement cost for the net attached to the sinks
  for(iter = currSinkList.begin(); iter != currSinkList.end(); iter++){
    (*iter)->updateSinkWireAndTimingCost(timing_driven, criticalPathDelay);
  }
  
  // Recalculate the placement cost for the nets attached to the sources
  for(iter = currSourceList.begin(); iter != currSourceList.end(); iter++){
    (*iter)->updateSourceWireAndTimingCosts(timing_driven, criticalPathDelay);
  }
}

double registerBank::determineDeltaWireCost(){
  //Go through the all of the nets this block is attached to and figure out the total increase
  // or decrease in wirelength.
  list<Terminal*>::iterator iter;
  double deltaWire = 0.0;

  //Make sure that this bank is still in sync
  if(currSinkList.size() != currSourceList.size()){
    ERROR("Bank is out of sync!");
  }
  
  // Recalculate the placement cost for the nets attached to the sinks
  for(iter = currSinkList.begin(); iter != currSinkList.end(); iter++){
    deltaWire += (*iter)->sinkDeltaWireCost();
  }
  
  // Recalculate the placement cost for the nets attached to the sources 
  for(iter = currSourceList.begin(); iter != currSourceList.end(); iter++){
    deltaWire += (*iter)->sourceDeltaWireCost();
  }
  return deltaWire;
}

double registerBank::determineDeltaTimingCost(){
  //Go through the all of the nets this block is attached to and figure out the total increase
  // or decrease in timing cost.
  list<Terminal*>::iterator iter;
  double deltaTiming = 0.0;
  
  //Make sure that this bank is still in sync
  if(currSinkList.size() != currSourceList.size()){
    ERROR("Bank is out of sync!");
  }

  // Recalculate the placement cost for the nets attached to the sinks
  for(iter = currSinkList.begin(); iter != currSinkList.end(); iter++){
    deltaTiming += (*iter)->sinkDeltaTimingCost();
  }
  
  // Recalculate the placement cost for the nets attached to the sources
  for(iter = currSourceList.begin(); iter != currSourceList.end(); iter++){
    deltaTiming += (*iter)->sourceDeltaTimingCost();
  }
  return deltaTiming;
}

double registerBank::determineDeltaTimingCostPrime(){
  //Go through the all of the nets this block is attached to and figure out the total increase
  // or decrease in timing cost.
  list<Terminal*>::iterator iter;
  double deltaTimingPrime = 0.0;
  
  //Make sure that this bank is still in sync
  if(currSinkList.size() != currSourceList.size()){
    ERROR("Bank is out of sync!");
  }

  // Recalculate the placement cost for the nets attached to the sinks
  for(iter = currSinkList.begin(); iter != currSinkList.end(); iter++){
    deltaTimingPrime += (*iter)->sinkDeltaTimingCostPrime();
  }
  
  // Recalculate the placement cost for the nets attached to the sources
  for(iter = currSourceList.begin(); iter != currSourceList.end(); iter++){
    deltaTimingPrime += (*iter)->sourceDeltaTimingCostPrime();
  }
  return deltaTimingPrime;
}

//Move the first N register in this bank to the end of the other bank
void registerBank::migrateFirstNRegs(registerBank *otherBank, int numRegs){
  list <Terminal*>::iterator iter;
  int i;

  //Make sure that the two banks are still in sync
  if(currSinkList.size() != currSourceList.size() || 
     otherBank->currSinkList.size() != otherBank->currSourceList.size()){
    ERROR("Bank is out of sync!");
  }

  //Make sure that this many registers are even mapped to this block
  if(getNumMappedRegs() < numRegs){
    ERROR("Trying to migrate more registers than are mapped to this block!");
  }
  
  //Make sure that this many registers can fit in the other block
  if(otherBank->getNumFreeRegs() < numRegs){
    ERROR("Trying to migrate more registers than can fit in this block!");
  }

  //Is the other block empty?  If so, we will fill it.
  if(otherBank->isBankAndLinkedLUTEmpty()){
    FPGA->occupyRegBank(otherBank);
  }

  //Push the first N sinks from this block into the end of the other block and update what delay
  // block/routing node they are pointed to.
  for(iter = currSinkList.begin(), i = 0;
      iter != currSinkList.end() && i < numRegs; iter++, i++){
    otherBank->currSinkList.push_back(*iter);
    (*iter)->setRegisterBank(otherBank);
    (*iter)->set_rNode(otherBank->routingNodeVector.at(otherBank->currSinkList.size() - 1));
  }

  currSinkList.erase(currSinkList.begin(), iter);

  //Copy the first N sources from this block into the end of the other block and update what delay
  // block/routing node they are pointed to.
  for(iter = currSourceList.begin(), i = 0;
      iter != currSourceList.end() && i < numRegs; iter++, i++){
    otherBank->currSourceList.push_back(*iter);
    (*iter)->setRegisterBank(otherBank);
    (*iter)->set_rNode(otherBank->routingNodeVector.at(otherBank->currSourceList.size() - 1));
  }
  
  //Remove these sources from this block
  //currSourceList.erase(currSourceList.begin(), iter--);
  currSourceList.erase(currSourceList.begin(), iter);
  //CHECK THE ABOVE STATEMENT!
  
  //Did we merge all of the registers from this bank into the other one?
  if(isBankAndLinkedLUTEmpty()){
    FPGA->vacateRegBank(this);
  }

  //Find out if either bank was associated with a logic block.
  //This is an error because we cannot break up monolithic register banks, only swap them
  if(currSourceList.size() != 0){
    if(currSourceList.front()->lBlock){
      ERROR("Breaking up delay block!");
    }
  }
  if(otherBank->currSourceList.size() != 0){
    if(otherBank->currSourceList.front()->lBlock){
      ERROR("Breaking up delay block!");
    }
  }
}

void registerBank::migrateLastNRegs(registerBank *otherBank, int numRegs){
  list <Terminal*>::iterator iter;
  int i;

  //Make sure that the two banks are still in sync
  if(currSinkList.size() != currSourceList.size() || 
     otherBank->currSinkList.size() != otherBank->currSourceList.size()){
    ERROR("Bank is out of sync!");
  }

  //Make sure that this many registers are even mapped to this block
  if(getNumMappedRegs() < numRegs){
    ERROR("Trying to migrate more registers than are mapped to this block!");
  }
  
  //Make sure that this many registers can fit in the other block
  if(otherBank->getNumFreeRegs() < numRegs){
    ERROR("Trying to migrate more registers than can fit in this block!");
  }

  //Is the other block empty?  If so, we will fill it.
  if(otherBank->isBankAndLinkedLUTEmpty()){
    FPGA->occupyRegBank(otherBank);
  }

  //Push the last N sinks from this block into the front of the other block
  for(iter = currSinkList.begin(), i = 0; i < currSinkList.size() - numRegs; iter++, i++){
    
  }
  otherBank->currSinkList.insert(otherBank->currSinkList.begin(), 
				      iter, currSinkList.end());

  //Remove these sinks from this block
  currSinkList.erase(iter, currSinkList.end());
  
  //Go through all of the sinks in the other block and update what delay block/routing node they are pointed to.
  //At least the routing node portion is absolutely necessary for everyone since all the blocks slide over
  for(iter = otherBank->currSinkList.begin(), i = 0;
      iter != otherBank->currSinkList.end(); iter++, i++){
    (*iter)->setRegisterBank(otherBank);
    (*iter)->set_rNode(otherBank->routingNodeVector.at(i));
  }

  //Push the last N sources from this block into the front of the other block
  for(iter = currSourceList.begin(), i = 0; i < currSourceList.size() - numRegs; iter++, i++){
    
  }
  otherBank->currSourceList.insert(otherBank->currSourceList.begin(), 
					iter, currSourceList.end());

  //Remove these sources from this block
  currSourceList.erase(iter, currSourceList.end());
  
  //Go through all of the sources in the other block and update what delay block/routing node they are pointed to.
  //At least the routing node portion is absolutely necessary for everyone since all the blocks slide over
  for(iter = otherBank->currSourceList.begin(), i = 0;
      iter != otherBank->currSourceList.end(); iter++, i++){
    (*iter)->setRegisterBank(otherBank);
    (*iter)->set_rNode(otherBank->routingNodeVector.at(i));
  }

  //Did we merge all of the registers from this bank into the other one?
  if(isBankAndLinkedLUTEmpty()){
    FPGA->vacateRegBank(this);
  }

  //Find out if either bank was associated with a logic block.
  //This is an error because we cannot break up static delay blocks, only swap them
  if(currSourceList.size() != 0){
    if(currSourceList.front()->lBlock){
      ERROR("Breaking up delay block!");
    }
  }
  if(otherBank->currSourceList.size() != 0){
    if(otherBank->currSourceList.front()->lBlock){
      ERROR("Breaking up delay block!");
    }
  }
}

//Take the registers in this bank and duplicate them into the other bank.  
//Assign the sinks to whichever register bank is closer.
//We have already determined, from earlier tests, that at least one sink
// will be attached to each register bank.
void registerBank::duplicateRegs(Netlist *netlist, registerBank *otherBank, double criticalPathDelay){
  Net* parentNet;
  PipeliningNet* currPNet;
  PipeliningNet* nextPNet;
  Terminal* regTerm;
  //char tempStr [MAX_STR_LEN];
  string tempStr;
  int i;

  int myDistance;
  int otherDistance;
  Terminal* currSink;

  //Make sure that the two banks are still in sync
  if(currSinkList.size() != currSourceList.size() || 
     otherBank->currSinkList.size() != otherBank->currSourceList.size()){
    ERROR("Delay Blocks are out of sync!");
  }

  //We know that the other block used to be empty.
  //FPGA->occupyRegBank(otherBank);

  //Find the net that drives this register bank (the source pnet of the first register)
  currPNet = getFirstSink()->sourcePipeliningNet;
  parentNet = currPNet->parentNet;

  //Until we have duplicated all of the registers in the block, do this
  for(i = 0; i < currSinkList.size(); i++){
    //1) Create a new pipeline net at a latency one greater than the current one
    //   and add it to the main net's list
    //strcpy(tempStr, parentNet->name);
    //sprintf(tempStr, "%sPipelineNet%d", tempStr, parentNet->registerID + 1);
    tempStr = parentNet->name + "PipelineNet" + itoa(parentNet->registerID + 1);
    nextPNet = new PipeliningNet(tempStr, currPNet->latency + 1, parentNet);
    parentNet->pipelineNetList.push_back(nextPNet);
    
    //2) Create a new pipelining register and add it to the main net's list of registers
    //   along with the netlist's list of registers
    //strcpy(tempStr, parentNet->name);
    //sprintf(tempStr, "%sReg%d", tempStr, parentNet->registerID);
    tempStr = parentNet->name + "Reg" + itoa(parentNet->registerID);
    parentNet->registerID ++;
    regTerm = new Terminal(parentNet, currPNet, nextPNet, tempStr);
    parentNet->pipelineRegisterList.push_back(regTerm);
    netlist->placementPipelineRegisterVector.push_back(regTerm);
    regTerm->departureTime = 0.0;
    regTerm->requiredTime = criticalPathDelay;

    //3) Make this register a sink of the current net and the source of the next net
    //The slack should be the sink required time minus the source departure time
    //  minus the delay of the wire between the two.
    //That is, CPD - source DT - 0.
    currPNet->pushSinkPlusData(regTerm, 0.0, regTerm->requiredTime - currPNet->source->departureTime);
    nextPNet->source = regTerm;
    
    //4) Place the newly created register into the other register bank
    otherBank->placeTerminal(regTerm);

    currPNet = nextPNet;
  }

  //Now that we have duplicated the registers, we have to split up the sinks
  // of the original register bank.
  currPNet = getLastSource()->destPipeliningNet;

  //Disconnect any sinks from the original net that are closer to the newly created register.
  for(i = 0; i < currPNet->sinkVector.size(); i++){
    currSink = currPNet->sinkVector.at(i);

    //Figure out their distance to this register bank
    myDistance = abs(currSink->getLogicBlockCol() - hostCLB->col) +
      abs(currSink->getLogicBlockRow() - hostCLB->row); 
    
    //Figure out their distance to the other register bank
    otherDistance = abs(currSink->getLogicBlockCol() - otherBank->hostCLB->col) +
      abs(currSink->getLogicBlockRow() - otherBank->hostCLB->row); 
    
    //If the distance to the newly created register from the sink is smaller than to the
    // original one, move it to the
    if(otherDistance < myDistance){
      //The slack should be the sink required time minus the source departure time
      //  minus the delay of the wire between the two.
      //Note that the departure time of a register should always be 0
      //That is, Sink RT - 0 - 0
      nextPNet->pushSinkPlusData(currSink, 0.0, 
				 currSink->requiredTime - nextPNet->source->departureTime);
      //Update the sink's source net
      currSink->sourcePipeliningNet = nextPNet;
      currPNet->removeSinkPlusData(currSink);
      i--;
    }
  }

  //**********SOMETHING TO LOOK INTO
  //Find out if this bank was associated with a logic block.
  //If so, we should update the netlist to show that we duplicated a monolithic
  // register bank
  if(currSourceList.size() != 0){
    if(currSourceList.front()->lBlock){
      ERROR("Should not be duplicating grouped register banks!");
    }
  }
}

//The other bank has duplicate registers to this one.  We want to transfer the
// nets that use that one to take from this one.  However, only notify the nets
// and sinks, don't disturb the information sitting in the other bank yet.  We
// might need that information if we want to undo this operation.
void registerBank::zipRegs(Netlist *netlist, registerBank *otherBank, double criticalPathDelay){
  PipeliningNet* recPNet;
  PipeliningNet* sendPNet;
  Net* parentNet;
  vector<Terminal*>::iterator tvIter;
  list<Terminal*>::iterator tlIter;
  int select;

  //Make sure that neither bank is associated with a logic block
  if(currSourceList.size() != 0){
    if(currSourceList.front()->lBlock){
      ERROR("Should not be merging grouped register banks!");
    }
  }
  if(otherBank->currSourceList.size() != 0){
    if(otherBank->currSourceList.front()->lBlock){
      ERROR("Should not be merging grouped register banks!");
    }
  }

  //Make sure that the two banks are still in sync
  if(currSinkList.size() != currSourceList.size() || 
     otherBank->currSinkList.size() != otherBank->currSourceList.size()){
    ERROR("Delay Blocks are out of sync!");
  }

  //Find the net that drives both of these register banks (the source pnet of the first register)
  recPNet = otherBank->getFirstSink()->sourcePipeliningNet;
  parentNet = recPNet->parentNet;

  //Double check that this bank has the same source net
  if(recPNet != getFirstSink()->sourcePipeliningNet){
    ERROR("Trying to combine two nets that do not have the same source!");
  }

  //Take that drive net and remove the reference to the first sink in the other bank
  recPNet->removeSinkPlusData(otherBank->getFirstSink());

  //Now detach all of the sinks from the other bank and attach them to the end of this one
  //Since we might have to undo this move, we'll leave the registers in the other bank alone
  // for the moment. That is, the registers in the other bank will know who they used to 
  //  connect to, but the sinks themselves will think they are attached to this one
  recPNet = getLastSource()->destPipeliningNet;
  
  sendPNet = otherBank->getLastSource()->destPipeliningNet;
  for(tvIter = sendPNet->sinkVector.begin(); tvIter != sendPNet->sinkVector.end(); tvIter++){
    //The slack should be the sink required time minus the source departure time
    //  minus the delay of the wire between the two.
    //Notice, this isn't perfect since the required time for the source might change,
    // but we can't update this without doing a full STA.
    //Thus, we will use CPD - source DT - 0
    recPNet->pushSinkPlusData(*tvIter, 0.0, (*tvIter)->requiredTime - recPNet->source->departureTime);
    (*tvIter)->sourcePipeliningNet = recPNet;    
  }

  //Go through each register in the other bank and delete it from the 
  // parent conventional net's list of registers / main
  // netlist's list of registers.
  //Also, find the destination pnets for each of these registers
  // and remove it from the parent conventional net's list of pnets.
  //Notice, we are not deleting the register itself yet - that is
  // still hanging around in the block, but the netlist should not
  // know anything about these registers any more
  for(tlIter = otherBank->currSinkList.begin(); 
      tlIter != otherBank->currSinkList.end(); tlIter++){
    parentNet->removePipelineRegFromPRList(*tlIter);
    netlist->removePipelineRegisterFromPlacementPRVector(*tlIter);
    parentNet->removePipelineNetFromPNList((*tlIter)->destPipeliningNet);
  }

  //Although we haven't actually done it yet, assume that we will accept this
  // change and vacate the other duplicate bank in terms of free register counting
  //FPGA->vacateRegBank(otherBank);
}

//We zipped the registers from the other bank into this one, but now we want to undo that.
//The other bank still should have all of its registers and internal connections intact, 
// but we have to reconnect the ends.
//We also have to put the references to these registers and pipelining net back into the
// parent nets.
void registerBank::reDuplicateRegs(Netlist *netlist, registerBank *otherBank, double criticalPathDelay){
  Net* parentNet;
  PipeliningNet* currPNet;
  PipeliningNet* nextPNet;
  int select;

  vector<Terminal*>::iterator tIter;

  //Make sure that this bank is not associated with a monolithic register bank
  if(currSourceList.size() != 0){
    if(currSourceList.front()->lBlock){
      ERROR("Should not be duplicating grouped register banks!");
    }
  }

  //Make sure that the two banks are still in sync
  if(currSinkList.size() != currSourceList.size() || 
     otherBank->currSinkList.size() != otherBank->currSourceList.size()){
    ERROR("Register banks are out of sync!");
  }

  //We have to undo the zip that we did earlier.  As far as empty register bookkeeping
  // goes, we assumed that we were going to accept the zip from earlier.  Thus, to keep
  // things straight, we have to re-occupy the other bank
  //FPGA->occupyRegBank(otherBank);

  //Find the net that drives this register bank (the source pnet of the first register)
  currPNet = getFirstSink()->sourcePipeliningNet;
  parentNet = currPNet->parentNet;

  //First, the registers and their pNets still exist in block 1.
  //We knew we might have to undo this, so we didn't delete them yet. 
  //Thus, we don't have to create them, but we do have to:
  // 1) hook up the first one
  //  The slack should be the sink required time minus the source departure time
  //   minus the delay of the wire between the two.
  //  That is, CPD - source DT - 0.
  currPNet->pushSinkPlusData(otherBank->getFirstSink(), 0.0, 
			     criticalPathDelay - currPNet->source->departureTime);

  // 2) put all of the registers back into the parent net's list/the netlist's list
  //    along with the pnets back into the parent net's list
  for(list <Terminal*>::iterator iter = otherBank->currSinkList.begin(); 
      iter != otherBank->currSinkList.end(); iter++){
    parentNet->pipelineRegisterList.push_back(*iter);
    netlist->placementPipelineRegisterVector.push_back(*iter);
    parentNet->pipelineNetList.push_back((*iter)->destPipeliningNet);    
  }

  //Now that we have duplicated the registers, we have to split up the sinks
  // of the original register bank.  
  //The last pipelining net in bank 1 knows who it wants to connect to,
  // we just have to:
  //1) remove the sink from the last net in bank 0
  //2) tell the sink itself who it's parent is
  currPNet = getLastSource()->destPipeliningNet;
  nextPNet = otherBank->getLastSource()->destPipeliningNet;

  for(tIter = nextPNet->sinkVector.begin(); tIter != nextPNet->sinkVector.end(); tIter++){
    //Find this sink in the bank 0 net and delete it
    currPNet->removeSinkPlusData(*tIter);

    //Now inform the sink itself who it's new source is
    (*tIter)->sourcePipeliningNet = nextPNet;
  }
}

//This terminal should be the first sink mapped to this register.
//We would like to remove this terminal from this register bank.
//However, note that we are NOT remapping the routing nodes.
//That will have to be taken care of later.
void registerBank::removeStartReference(Terminal *T){
  if(currSinkList.front() != T || currSourceList.front() != T){
    ERROR("Trying to remove non-initial register from bank!");
  }
  else{
    currSinkList.pop_front();
    currSourceList.pop_front();
    //currSinkList.erase(currSinkList.begin());
    //currSourceList.erase(currSourceList.begin());

    //If this removes the last mapped sink from the bank, this bank has
    // been vacated.
    if(currSinkList.size() == 0){
      FPGA->vacateRegBank(this);
#ifdef PLACER_DEBUG
      numRegistersEstimatedFreed -= getNumPossibleDelay();
#endif
    }
  }
}

//This terminal should be the last sink mapped to this register.
//We would like to remove this terminal from this register bank.
void registerBank::removeEndReference(Terminal *T){
  if(currSinkList.back() != T || currSourceList.back() != T){
    ERROR("Trying to remove non-final register from bank!");
  }
  else{
    currSinkList.pop_back();
    currSourceList.pop_back();

    //If this removes the last mapped sink from the bank, this bank has
    // been vacated.
    if(currSinkList.size() == 0){
      FPGA->vacateRegBank(this);
#ifdef PLACER_DEBUG
      numRegistersEstimatedFreed -= getNumPossibleDelay();
#endif
    }
  }
}

// When we zipped the registers from this bank into another, we knew
//  we might have to undo the move.  Thus, we left the registers inside
//  this bank intact.  Clear them now.
void registerBank::freeBank(){
  list<Terminal*>::iterator tIter;

  for(tIter = currSinkList.begin(); tIter != currSinkList.end(); tIter++){
    delete ((*tIter)->destPipeliningNet);
    delete *tIter;
  }
  currSinkList.clear();
  currSourceList.clear();

  FPGA->vacateRegBank(this);
}

//Return the total wire cost of the nets driven by the registers in this bank
double registerBank::returnTotalWireCost(){
  list<Terminal*>::iterator tIter;
  double wireCost = 0.0;
  
  for(tIter = currSinkList.begin(); tIter != currSinkList.end(); tIter++){
    wireCost += ((*tIter)->destPipeliningNet->wireCost);
  }
  return wireCost;
}

//Return the total timing cost of the nets driven by the registers in this bank
double registerBank::returnTotalTimingCost(){
  list<Terminal*>::iterator tIter;
  double timingCost = 0.0;
  
  for(tIter = currSinkList.begin(); tIter != currSinkList.end(); tIter++){
    timingCost += ((*tIter)->destPipeliningNet->timingCost);
  }
  return timingCost;
}

//We are going to get rid of the registers in this bank so update the cut cost appropriately
void registerBank::updateDeletedCutCost(){
  list<Terminal*>::iterator tIter;
  int i;

  for(tIter = currSinkList.begin(); tIter != currSinkList.end(); tIter++){
    //Decrement the span of the pipelining net driven by this register
    for(i = (*tIter)->destPipeliningNet->minCol; i < (*tIter)->destPipeliningNet->maxCol; i++){
      FPGA->cutCostVector.at(i) += -1.0;
    }
  }
}

void registerBank::print_info(){
  cout << "Location = " << getHostCLB()->col  << ", " << getHostCLB()->row
       << " bank " << getHostBankNum() << endl;
  cout << "\tBank has " << getNumPossibleDelay() << " registers" << endl;
}

//We moved around some of the register mapped to this register bank, so
// make sure everyone is mapped to the correct location.
void registerBank::updateRoutingNodes(){
  list <Terminal *>::iterator sinkIter, sourceIter;
  int i;

  //Make sure that the two banks are still in sync
  if(currSinkList.size() != currSourceList.size()){
    ERROR("Register bank are out of sync!");
  }  
  
  for(sinkIter = currSinkList.begin(), sourceIter = currSourceList.begin(), i = 0;
      sinkIter != currSinkList.end(), sourceIter != currSourceList.end(); 
      sinkIter ++, sourceIter++, i++){
    if(*sinkIter != *sourceIter){
      ERROR("Sink and source mismatch!");
    }
    (*sinkIter)->set_rNode(routingNodeVector.at(i));
    (*sourceIter)->set_rNode(routingNodeVector.at(i));
  }
}

//Are the LUTs in the clb at the same location empty?
//How about the other register banks in the same delay block?
bool registerBank::emptyCLB(){
  if(!FPGA->getCLBAt(getHostCLB()->row, getHostCLB()->col)->areLUTsEmpty() || 
     !FPGA->getCLBAt(getHostCLB()->row, getHostCLB()->col)->areRegistersEmpty())
    return false;
  return true;
}

//Swap the contents of the two parent clbs, but don't rely on netlist-level
// clb information for anything.
void registerBank::swapEntireStickyMappings(registerBank* otherBank, double criticalPathDelay){
  vprCLB* myArrayCLB = hostCLB;
  vprCLB* otherArrayCLB = otherBank->hostCLB;

  //Swap any LUTs and FFs that might be mapped to the host physical clbs
  // for these two register banks.
  myArrayCLB->swapAllLUTsAndFFs(otherArrayCLB);
  
  //Update the sink timings and slack for the nets we just moved
  // The slack calculations are not exact since the critical path might have changed
  myArrayCLB->updateSinkDelaysAndSlacksForAllLUTsAndAssociatedFFs();
  myArrayCLB->updateSinkDelaysAndSlacksForAllNonAssociatedFFs();
  otherArrayCLB->updateSinkDelaysAndSlacksForAllLUTsAndAssociatedFFs();
  otherArrayCLB->updateSinkDelaysAndSlacksForAllNonAssociatedFFs();
  
  //Now update the max input and output criticalities
  myArrayCLB->calculateMaxInputOutputCriticalitiesForAllLUTsAndAssociatedFFs(criticalPathDelay);
  myArrayCLB->calculateMaxInputOutputCriticalitiesForAllNonAssociatedFFs(criticalPathDelay);
  otherArrayCLB->calculateMaxInputOutputCriticalitiesForAllLUTsAndAssociatedFFs(criticalPathDelay);
  otherArrayCLB->calculateMaxInputOutputCriticalitiesForAllNonAssociatedFFs(criticalPathDelay);
}

void registerBank::updateCostAndDeltas(bool timing_driven, double criticalPathDelay, double *deltaWire, 
			 double *deltaTiming){
  updateWireAndTimingCosts(timing_driven, criticalPathDelay);
  
  *deltaWire += determineDeltaWireCost();
  
  if(timing_driven){
    *deltaTiming += determineDeltaTimingCost();
  }
}

void registerBank::updateCostAndDeltasCostDeltaDelay(bool timing_driven, double criticalPathDelay, double *deltaWire, 
						     double *deltaTiming, double *deltaTimingPrime){
  updateWireAndTimingCosts(timing_driven, criticalPathDelay);
  
  *deltaWire += determineDeltaWireCost();
  
  if(timing_driven){
    *deltaTiming += determineDeltaTimingCost();
    *deltaTimingPrime += determineDeltaTimingCostPrime();
  }
}

void registerBank::updateCostAndDeltasAndAssociatedLUT(bool timing_driven, double criticalPathDelay, double *deltaWire, 
						       double *deltaTiming){
  if(linkedLUT && !linkedLUT->isLUTAndLinkedRegistersEmpty()){
    linkedLUT->updateCostAndDeltas(timing_driven, criticalPathDelay, deltaWire, deltaTiming);
  }
  else{
    updateCostAndDeltas(timing_driven, criticalPathDelay, deltaWire, deltaTiming);
  }
}

void registerBank::updateCostAndDeltasAndAssociatedLUTCostDeltaDelay(bool timing_driven, double criticalPathDelay, double *deltaWire, 
								     double *deltaTiming, double *deltaTimingPrime){
  if(linkedLUT && !linkedLUT->isLUTAndLinkedRegistersEmpty()){
    linkedLUT->updateCostAndDeltasCostDeltaDelay(timing_driven, criticalPathDelay, deltaWire, deltaTiming, deltaTimingPrime);
  }
  else{
    updateCostAndDeltasCostDeltaDelay(timing_driven, criticalPathDelay, deltaWire, deltaTiming, deltaTimingPrime);
  }
}

void registerBank::updateCostAndDeltasForEntireSticky(registerBank* otherRegisterBank, bool timing_driven, 
						      double criticalPathDelay, double *deltaWire, double *deltaTiming){
  vprCLB* myArrayCLB = hostCLB;
  vprCLB* otherArrayCLB = otherRegisterBank->hostCLB;
  
  myArrayCLB->updateCostAndDeltasForAllLUTsAndAssociatedFFs(timing_driven, criticalPathDelay, 
							    deltaWire, deltaTiming);
  myArrayCLB->updateCostAndDeltasForAllNonAssociatedFFs(timing_driven, criticalPathDelay, 
							deltaWire, deltaTiming);
  otherArrayCLB->updateCostAndDeltasForAllLUTsAndAssociatedFFs(timing_driven, criticalPathDelay, 
							       deltaWire, deltaTiming);
  otherArrayCLB->updateCostAndDeltasForAllNonAssociatedFFs(timing_driven, criticalPathDelay, 
							   deltaWire, deltaTiming);
}

//This function updates the delta costs for all of the LUTs and FFs in the parent CLBs
// of these two register banks.
void registerBank::updateCostAndDeltasForEntireStickyCostDeltaDelay(registerBank* otherRegisterBank, bool timing_driven, 
								    double criticalPathDelay, double *deltaWire, 
								    double *deltaTiming, double *deltaTimingPrime){
  vprCLB* myArrayCLB = hostCLB;
  vprCLB* otherArrayCLB = otherRegisterBank->hostCLB;
  
  myArrayCLB->updateCostAndDeltasForAllLUTsAndAssociatedFFsCostDeltaDelay(timing_driven, criticalPathDelay, 
									  deltaWire, deltaTiming, deltaTimingPrime);
  myArrayCLB->updateCostAndDeltasForAllNonAssociatedFFsCostDeltaDelay(timing_driven, criticalPathDelay, 
								      deltaWire, deltaTiming, deltaTimingPrime);
  otherArrayCLB->updateCostAndDeltasForAllLUTsAndAssociatedFFsCostDeltaDelay(timing_driven, criticalPathDelay, 
									     deltaWire, deltaTiming, deltaTimingPrime);
  otherArrayCLB->updateCostAndDeltasForAllNonAssociatedFFsCostDeltaDelay(timing_driven, criticalPathDelay, 
									 deltaWire, deltaTiming, deltaTimingPrime);
}

//Go through all of the register banks and LUTs associated with the vprCLB at this location.
//If there are any logic blocks mapped to these locations, return the CLB associated with them
c_clb* registerBank::returnMappedCLB(){
  c_clb* mappedCLB = NULL;
  vprCLB *myCLB = FPGA->getCLBAt(hostCLB->row, hostCLB->col);

  //Try to extract a CLB from the FFs at this locations
  for(vector <registerBank*>::iterator rbIter = myCLB->registerBankVector.begin(); 
      rbIter != myCLB->registerBankVector.end(); rbIter++){
    if((*rbIter)->get_currBlock()){
      if(mappedCLB){
	if(mappedCLB != (*rbIter)->get_currBlock()->parentCLB){
	  ERROR("Register bank is mapped to different logical clb!");
	}
      }
      else{
	mappedCLB = (*rbIter)->get_currBlock()->parentCLB;
      }
    }
  }
  
  //Try to extract a CLB from the LUTs at this locations
  for(vector <vprLUT *>::iterator lutIter = myCLB->lutVector.begin(); 
      lutIter != myCLB->lutVector.end(); lutIter++){
    if((*lutIter)->get_currBlock()){
      if(mappedCLB){
	if(mappedCLB != (*lutIter)->get_currBlock()->parentCLB){
	  ERROR("LUT is mapped to different logical clb!");
	}
      }
      else{
	mappedCLB = (*lutIter)->get_currBlock()->parentCLB;
      }
    }
  }
  return mappedCLB;
}

//Figure out what drives this register bank and try to return the best-choice register bank
// to do a swap with.
//If this register bank is already in the same CLB as its parent, we'll return NULL.
//Otherwise, we'll try to return the "most eligable" register bank in the parent's CLB.
registerBank* registerBank::getBestRegBankInParentHostCLBSwap(){
  int sourceRow, sourceCol;
  int myRow = hostCLB->get_row();
  int myCol = hostCLB->get_col();
  Terminal *sink;
  Terminal *source;
  int sourceIndex;
  int returnIndex;
  vprCLB* sourceCLB;
  
  //Regardless of whether or not this register block is currently mapped to
  // a logic block or a string of individual pipelining registers, we can find out
  // what drives this by looking at the first sink.
  sink = getFirstSink();

  //This sink might be driven by a pipelining net or a regular net
  if(sink->sourcePipeliningNet){
    source = sink->sourcePipeliningNet->source;
  }
  else{
    source = sink->net->source;
  }

  sourceRow = source->getLogicBlockRow();
  sourceCol = source->getLogicBlockCol();

  if(myRow == sourceRow && myCol == sourceCol){
    return NULL;
  }
  
  if(source->lBlock){
    //If an I/O pad is a parent:
    // Return NULL
    if(source->lBlock->blockType == INPUTPAD){
      return NULL;
    }
    else if(source->lBlock->blockType == OUTPUTPAD){
      ERROR("Source is an output pad!");
    }
    else if(source->lBlock->blockType == LUT){
      sourceCLB = FPGA->getCLBAt(sourceRow, sourceCol);

      //If a LUT is the parent:
      // I) Search for the smallest unoccupied reg bank in the parent's CLB that will fit this 
      //      register bank. Check to see how many adequately-sized empty register banks are 
      //      in the parent CLB
      if(sourceCLB->getNumEmptyRegBank(getNumMappedRegs()) > 0){
	return(sourceCLB->getRandomEmptyRegBank(getNumMappedRegs()));
      }
      // II) If (I) is impossible, search for one that will fit this register bank that isn't 
      //         related to the LUTs in this CLB (there might not be one, even though the 
      //         architecture has a 1/1 ratio or more of FF/LUTs due to register unzipping).
      else if(sourceCLB->getNumUnrelatedRegBank(getNumMappedRegs(), getNumPossibleDelay(), linkedLUT) > 0){
	return(sourceCLB->getRandomUnrelatedRegBank(getNumMappedRegs(), getNumPossibleDelay(), linkedLUT));
      }
      // III) If (II) is impossible, just pick a random one that will fit this register bank in 
      //         the parent's CLB.
      else if(sourceCLB->getNumMinSizeRegBank(getNumMappedRegs(), getNumPossibleDelay(), linkedLUT) > 0){
	return(sourceCLB->getRandomMinSizeRegBank(getNumMappedRegs(), getNumPossibleDelay(), linkedLUT));
      }
      // IV) If III fails (there may not be any registers in the parent's CLB at all) return NULL
      else{
	return NULL;
      }
    }
    else if(source->lBlock->blockType == REGISTER_BANK){
      sourceCLB = FPGA->getCLBAt(sourceRow, sourceCol);
      //Do we want to be careful about not choosing the parent itself?
      if(carefulHoming){
	// I) Search for the smallest unoccupied reg bank in the parent's CLB that will fit this 
	//      register bank. Check to see how many adaquately-sized empty register banks are 
	//      in the parent CLB.
	if(sourceCLB->getNumEmptyRegBank(getNumMappedRegs()) > 0){
	  return(sourceCLB->getRandomEmptyRegBank(getNumMappedRegs()));
	}
	// II) If (I) is impossible, search for one that will fit this register bank that isn't 
	//      related to the LUTs in this CLB (there might not be one, since we are driven by a
	//      a register bank), but is NOT the source.
	else if(sourceCLB->getNumUnrelatedRegBank(getNumMappedRegs(), getNumPossibleDelay(), source, linkedLUT) > 0){
	  return(sourceCLB->getRandomUnrelatedRegBank(getNumMappedRegs(), getNumPossibleDelay(), source, linkedLUT));
	}
	//If this fails, do not try to kick some else out, just so we can come home.
	// III) If (II) is impossible, just pick a random one that will fit this register bank in 
	//       the parent's CLB that is NOT the source
	//else if(sourceCLB->getNumMinSizeRegBank(getNumMappedRegs(), getNumPossibleDelay(), source) > 0){
	//  return(sourceCLB->getRandomMinSizeRegBank(getNumMappedRegs(), getNumPossibleDelay(), source));
	//}
	// IV) If III fails (there may not be any registers in the parent's CLB at all) return NULL
	else{
	  return NULL;
	}
      }
      //Don't worry about not choosing the parent itself, just do the same thing you do for LUTs.
      else{
	// I) Search for the smallest unoccupied reg bank in the parent's CLB that will fit this 
	//      register bank.  Check to see how many adaquately-sized empty register banks are 
	//      in the parent CLB
	if(sourceCLB->getNumEmptyRegBank(getNumMappedRegs()) > 0){
	  return(sourceCLB->getRandomEmptyRegBank(getNumMappedRegs()));
	}
	// II) If (I) is impossible, search for one that will fit this register bank that isn't 
	//      related to the LUTs in this CLB (there might not be one, since we are driven by a
	//      a register bank).
	else if(sourceCLB->getNumUnrelatedRegBank(getNumMappedRegs(), getNumPossibleDelay(), linkedLUT) > 0){
	  return(sourceCLB->getRandomUnrelatedRegBank(getNumMappedRegs(), getNumPossibleDelay(), linkedLUT));
	}
	//If this fails, do not try to kick some else out, just so we can come home.
	// III) If (II) is impossible, just pick a random one that will fit this register bank 
	//      in the parent's CLB 
	//else if(sourceCLB->getNumMinSizeRegBank(getNumMappedRegs(), getNumPossibleDelay()) > 0){
	//  return(sourceCLB->getRandomMinSizeRegBank(getNumMappedRegs(), getNumPossibleDelay()));
	//}
	// IV) If III fails (there may not be any registers in the parent's CLB at all) return NULL
	else{
	  return NULL;
	}
      }
    }
    else{
      ERROR("Something is wrong!");
    }
  }
  else{
    //The source is a pipelining register
    sourceCLB = FPGA->getCLBAt(sourceRow, sourceCol);
    //Do we want to be careful about not choosing the parent itself?
    if(carefulHoming){
      // I) Search for the smallest unoccupied reg bank in the parent's CLB that will fit 
      //      this register bank.  Check to see how many adaquately-sized empty register banks
      //      are in the parent CLB
      if(sourceCLB->getNumEmptyRegBank(getNumMappedRegs()) > 0){
	return(sourceCLB->getRandomEmptyRegBank(getNumMappedRegs()));
      }
      // II) If (I) is impossible, search for one that will fit this register bank that isn't 
      //      related to the LUTs in this CLB (there might not be one, since we are driven by a
      //      a register bank) that is NOT the source. 
      else if(sourceCLB->getNumUnrelatedRegBank(getNumMappedRegs(), getNumPossibleDelay(), source, linkedLUT) > 0){
	return(sourceCLB->getRandomUnrelatedRegBank(getNumMappedRegs(), getNumPossibleDelay(), source, linkedLUT));
      }
      //If this fails, do not try to kick some else out, just so we can come home.
      // III) If (II) is impossible, just pick a random one that will fit this register bank 
      //      in the parent's CLB that is NOT the source
      //else if(sourceCLB->getNumMinSizeRegBank(getNumMappedRegs(), getNumPossibleDelay(), source) > 0){
      //return(sourceCLB->getRandomMinSizeRegBank(getNumMappedRegs(), getNumPossibleDelay(), source));
      //}
      // IV) If III fails (there may not be any registers in the parent's CLB at all) return NULL
      else{
	return NULL;
      }
    }
    //Don't worry about not choosing the parent itself, just do the same thing you do for LUTs.
    else{
      // I) Search for the smallest unoccupied reg bank in the parent's CLB that will fit 
      //      this register bank.  Check to see how many adaquately-sized empty register banks
      //      are in the parent CLB
      if(sourceCLB->getNumEmptyRegBank(getNumMappedRegs()) > 0){
	return(sourceCLB->getRandomEmptyRegBank(getNumMappedRegs()));
      }
      // II) If (I) is impossible, search for one that will fit this register bank that isn't 
      //      related to the LUTs in this CLB (there might not be one, since we are driven by a
      //      a register bank).
      else if(sourceCLB->getNumUnrelatedRegBank(getNumMappedRegs(), getNumPossibleDelay(), linkedLUT) > 0){
	return(sourceCLB->getRandomUnrelatedRegBank(getNumMappedRegs(), getNumPossibleDelay(), linkedLUT));
      }
      //If this fails, do not try to kick some else out, just so we can come home.
      // III) If (II) is impossible, just pick a random one that will fit this register bank 
      //      in the parent's CLB 
      //else if(sourceCLB->getNumMinSizeRegBank(getNumMappedRegs(), getNumPossibleDelay(), linkedLUT) > 0){
      //return(sourceCLB->getRandomMinSizeRegBank(getNumMappedRegs(), getNumPossibleDelay(), linkedLUT));
      //}
      // IV) If III fails (there may not be any registers in the parent's CLB at all) return NULL
      else{
	return NULL;
      }
    }
  }
}

//Go through all of the FF's in this bank and update the max input and output criticalities 
// of any that are occupied.
void registerBank::calculateMaxInputOutputCriticalitiesJustRB(double criticalPathDelay){
  list <Terminal*>::iterator iter;

  if(currBlock){
    //There is a logic block associated with this register bank
    currBlock->calculateMaxInputOutputCriticality(criticalPathDelay);
  }
  else if(!isJustBankEmpty()){
    //Just because there isn't a logic block associated with
    // this register bank does not mean that we don't have
    // something mapped.  It could be individual pipelining registers.
    //Make sure that the bank is still in sync
    if(currSinkList.size() != currSourceList.size()){
      ERROR("Bank is out of sync!");
    }
    // Update the sink timing for the nets attached to this block
    for(iter = currSinkList.begin(); iter != currSinkList.end(); iter++){
      (*iter)->calculateMaxInputOutputCriticality(criticalPathDelay);
    }
  }
}

//Go through all of the FF's in this bank and update the max input and output criticalities 
// of any that are occupied.
void registerBank::calculateMaxInputOutputCriticalitiesAndAssociatedLUT(double criticalPathDelay){
  list <Terminal*>::iterator iter;

  if(currBlock){
    //There is a logic block associated with this register bank
    currBlock->calculateMaxInputOutputCriticality(criticalPathDelay);
  }
  else if(!isBankAndLinkedLUTEmpty()){
    //Just because there isn't a logic block associated with
    // this register bank does not mean that we don't have
    // something mapped.  It could be individual pipelining registers
    // or the associated LUT could be occupied.
    if(linkedLUT && linkedLUT->currBlock){
      linkedLUT->currBlock->calculateMaxInputOutputCriticality(criticalPathDelay);
    }
    else{
      //Make sure that the bank is still in sync
      if(currSinkList.size() != currSourceList.size()){
	ERROR("Bank is out of sync!");
      }
      // Update the sink timing for the nets attached to this block
      for(iter = currSinkList.begin(); iter != currSinkList.end(); iter++){
	(*iter)->calculateMaxInputOutputCriticality(criticalPathDelay);
      }
    }
  }
}


//See if there are any duplicates of this register bank that have the same output latency.
registerBank* registerBank::getRandomDuplicate(){
  Terminal *sink;
  vector<registerBank*> duplicateVector;

  //Get the net that drives this register bank and go through all of its sinks.
  //Look for register banks that are of the same size as this one.
  sink = getFirstSink();

  if(sink->sourcePipeliningNet){
    for(vector<Terminal*>::iterator sinkIter = sink->sourcePipeliningNet->sinkVector.begin();
	sinkIter != sink->sourcePipeliningNet->sinkVector.end(); sinkIter++){
      if((*sinkIter)->getRegisterBank()){
	if((*sinkIter)->getRegisterBank() != this && 
	   (*sinkIter)->getRegisterBank()->getNumMappedRegs() == getNumMappedRegs()){
	  duplicateVector.push_back((*sinkIter)->getRegisterBank());
	}
      }
      else{
	// The other sinks of this pipelining net might just be a regular LUT or output pin.
	//ERROR("Pipelining Register is not placed!");
      }
    }
  }
  else{
    ERROR("Why is this pipelining register driven by a regular nets?");
//     for(vector<Terminal*>::iterator sinkIter = sink->net->sinkVector.begin();
// 	sinkIter != sink->net->sinkVector.end(); sinkIter++){
//       if((*sinkIter)->getRegisterBank()){
// 	if((*sinkIter)->getRegisterBank() != this && 
// 	   (*sinkIter)->getRegisterBank()->getNumMappedRegs() == getNumMappedRegs()){
// 	  duplicateVector.push_back((*sinkIter)->getRegisterBank());
// 	}
//       }
//       else{
// 	ERROR("Pipelining Register is not placed!");
//       }
//     }
  }

  if(duplicateVector.size() != 0){
    return(duplicateVector.at(lrand48() % duplicateVector.size()));
  }
  else{
    return NULL;
  }
}

//Go through the parents and children of this register bank to look for ones
// we can merge with.
bool registerBank::getRandomMerge(int *moveType, int *numMove, registerBank **parentRegBank, registerBank **childRegBank){
  Terminal *sink;
  Terminal *source;
  vector<registerBank*> parentMergeVector;
  vector<registerBank*> childMergeVector;
  vector<int> moveTypeVector;
  int choice;
  
  //Get the node that is the source of the net that drives this register bank.
  sink = getFirstSink();

  if(sink->sourcePipeliningNet){
    source = sink->sourcePipeliningNet->source;

    //Check to see if it is not a monolithic register bank (logic block)
    // and not a LUT
    if(!source->lBlock){
      //If it is not, see if the parent has a fanout of one.
      if(sink->sourcePipeliningNet->sinkVector.size() == 1){
	//If it is, see if there is room in the parent register bank for the child register bank (this one).
	if(source->rb->getNumFreeRegs() >= getNumMappedRegs()){
	  //If there is room in the parent register bank for the child register bank (this one), 
	  //   add the merge to the parent to the mergeVector.
	  parentMergeVector.push_back(source->rb);
	  childMergeVector.push_back(this);
	  moveTypeVector.push_back(MOVEREGPARENT);
	}
	if(getNumFreeRegs() >= source->rb->getNumMappedRegs()){
	  //Alternatively, see if there is room in the child register bank (this one) for the parent.
	  //If so, add the merge to the child to the mergeVector.
	  parentMergeVector.push_back(source->rb);
	  childMergeVector.push_back(this);
	  moveTypeVector.push_back(MOVEREGCHILD);
	}
      }
    }
  }
  else {
    source = sink->net->source;
    
    //The register bank is driven by a regular net.  Thus, it must not be a pipelining register.
    // We won't have any merging going on.
  }

  //Get the node that is the source of the net this register bank drives.
  source = getLastSource();

  //Check to see if there is a fanout of one.
  if(source->destPipeliningNet->sinkVector.size() == 1){
    //Check to see if this sink is a register bank that is not a monolithic register bank (logic block)
    // or a LUT
    sink = source->destPipeliningNet->sinkVector.front();
    if(!sink->lBlock){
      //If it is, see if there is room in the parent register bank (this one) for the child register bank
      if(getNumFreeRegs() >= sink->rb->getNumMappedRegs()){
	//If there is room in the parent register bank for the child register bank, add the
	// merge to the parent to the mergeVector.
	parentMergeVector.push_back(this);
	childMergeVector.push_back(sink->rb);
	moveTypeVector.push_back(MOVEREGPARENT);
      }
      if(sink->rb->getNumFreeRegs() >= getNumMappedRegs()){
	//Alternatively, see if there is room in the child register bank for the parent (this one).
	//If so, add the merge to the child to the mergeVector.
	parentMergeVector.push_back(this);
	childMergeVector.push_back(sink->rb);
	moveTypeVector.push_back(MOVEREGCHILD);
      }      
    }
  }

  if(parentMergeVector.size() == 0){
    return false;
  }
  else{
    choice = lrand48() % parentMergeVector.size();
    *moveType = moveTypeVector.at(choice);
    *parentRegBank = parentMergeVector.at(choice);
    *childRegBank = childMergeVector.at(choice);
    return true;
  }
}

//Find out if there are any sinks of the net that this register bank
// drives that are closer to the other register bank.
bool registerBank::someSinksCloserTo(registerBank *otherRegBank){
  PipeliningNet* currPNet = getLastSource()->destPipeliningNet;
  int myDistance;
  int otherDistance;
  int closerToMe = 0;
  int closerToOther = 0;
  
  //Go through all of the sinks of this net
  for(vector<Terminal*>::iterator sinkIter = currPNet->sinkVector.begin(); sinkIter != currPNet->sinkVector.end(); sinkIter++){
    //Figure out their distance to this register bank
    myDistance = abs((*sinkIter)->getLogicBlockCol() - hostCLB->col) +
      abs((*sinkIter)->getLogicBlockRow() - hostCLB->row); 

    //Figure out their distance to the other register bank
    otherDistance = abs((*sinkIter)->getLogicBlockCol() - otherRegBank->hostCLB->col) +
      abs((*sinkIter)->getLogicBlockRow() - otherRegBank->hostCLB->row); 

    if(otherDistance < myDistance){
      closerToOther++;
    }
    else{
      closerToMe++;
    }
  }

  if(closerToOther == 0 || closerToMe == 0){
    return false;
  }
  else{
    return true;
  }
}

//Set the lut assiociation of this register bank and vice-versa
void registerBank::setLUTAssociation(int lutAssociation){
  vprLUT *associatedLUT;;

  //See if this lut actually exists (it should since we ordered all sources/sinks 
  //  before any register chains)
  if(hostCLB->getNumLUTs() <= lutAssociation){
    ERROR("Invalid lut index!");
  }
  else{
    associatedLUT = hostCLB->getLUT(lutAssociation);
  }
  
  //The lut association has been set already, so verify it.
  if(linkedLUT){
    if(linkedLUT != associatedLUT){
      ERROR("Trying to associate shift register bank with 2 different luts!");
    }
  }
  else{
    linkedLUT = associatedLUT;
  }

  associatedLUT->setRegisterAssociation(this);
}

//See if this register bank is linked, via a linked LUT,
// to the same kind of registers.
bool registerBank::hasMatchingLinkedRegisters(registerBank* otherRegisterBank){
  list <registerBank*>::iterator myRBIter;
  list <registerBank*>::iterator otherRBIter;

  if(!hasLinkedLUT() || !otherRegisterBank->hasLinkedLUT()){
    return false;
  }
  
  if(linkedLUT->linkedRegisterBanks.size() != otherRegisterBank->linkedLUT->linkedRegisterBanks.size()){
    return false;
  }
  for(myRBIter = linkedLUT->linkedRegisterBanks.begin(), 
	otherRBIter = otherRegisterBank->linkedLUT->linkedRegisterBanks.begin();
      myRBIter != linkedLUT->linkedRegisterBanks.end() && 
	otherRBIter != otherRegisterBank->linkedLUT->linkedRegisterBanks.end();
      myRBIter++, otherRBIter++){
    if((*myRBIter)->getNumPossibleDelay() != (*otherRBIter)->getNumPossibleDelay()){
      return false;
    }
  }
  return true;
}

//Return true only if the last source in this register bank
// drives more than 1 sink
bool registerBank::lastSourceHasFanout(){
  Terminal* lastSource;

  lastSource = getLastSource();
  
  if(!lastSource){
    return false;
  }
  else{
    return (lastSource->hasFanOut());
  }
}

vprBaseBlock::vprBaseBlock(int r, int c){
  if(r<0 || c<0)
    ERROR("row/col coords and channel width cannot be -ve...");
  row=r;col=c;
  magicNum=21111;
}

vprBaseBlock::~vprBaseBlock(){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  else
    magicNum=0;
}

int vprBaseBlock::get_row(){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  return row;
}

int vprBaseBlock::get_col(){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  return col;
}

void vprBaseBlock::print_info(){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  cout<<endl<<endl<<"FPGA LOCATION AT x = "<<col<<" AND y = "<<row;
}

/******************************************************************/

vprLUT::vprLUT(int r, int c, int ptc_index) : vprBaseBlock(r,c){
  derived_magicNum=21112;
  sourceNode=sinkNode=clock_sinkNode=0;
  //ipin_itr=opin_itr=0;
  currBlock=0;
  index = ptc_index;
}

vprLUT::~vprLUT(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  else
    derived_magicNum=0;
}

vprCLB* vprLUT::getHostCLB(){
  return (FPGA->getCLBAt(row, col));
}

void vprLUT::set_sourceNode(routingNode* r){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  if(!r)
    ERROR("source routing node cannot be NULL...");
  sourceNode=r;
}

void vprLUT::set_sinkNode(routingNode* r){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  if(!r)
    ERROR("sink routing node cannot be NULL...");
  sinkNode=r;
}

void vprLUT::set_currBlock(logicBlock* l){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  currBlock=l;
}

logicBlock* vprLUT::get_currBlock(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  return currBlock;
}

routingNode* vprLUT::get_sourceNode(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  return sourceNode;
}

routingNode* vprLUT::get_sinkNode(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  return sinkNode;
}

void vprLUT::set_blockType(enum e_genericBlockType t){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  if(t != LUT)
    ERROR("Invalid block type");
  blockType = t;
}

enum e_genericBlockType vprLUT::get_blockType(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  return blockType;
}

void vprLUT::print_info(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");

  vprBaseBlock::print_info();

  if(currBlock)
    cout<<endl<<"LOGIC BLOCK : "<<currBlock->get_name();
  else
    cout<<endl<<"LOGIC BLOCK : ----";

  cout<<endl<<"SOURCE NODE : "<<sourceNode->get_name()<<", SINK NODE : "<<sinkNode->get_name();
}

void vprLUT::clearCurrBlock(){
  currBlock=0;  
}

//Are the LUTs and registers in the delay block at the same location empty?
bool vprLUT::emptyCLB(){
  if(!FPGA->getCLBAt(row, col)->areLUTsEmpty() || !FPGA->getCLBAt(row, col)->areRegistersEmpty())
    return false;
  return true;
}

//Are all of the LUTs in the CLB empty?
bool vprLUT::emptyCLBLUTS(){
  if(!FPGA->getCLBAt(row, col)->areLUTsEmpty())
    return false;
  return true;
}

//Figure out if the two luts have the same linked register bank structure.
bool vprLUT::matchingLinkedRegisters(vprLUT* otherLUT){
  list <registerBank*>::iterator myRBIter;
  list <registerBank*>::iterator otherRBIter;

  if(linkedRegisterBanks.size() != otherLUT->linkedRegisterBanks.size()){
    return false;
  }
  for(myRBIter = linkedRegisterBanks.begin(), otherRBIter = otherLUT->linkedRegisterBanks.begin();
      myRBIter != linkedRegisterBanks.end() && otherRBIter != otherLUT->linkedRegisterBanks.end();
      myRBIter++, otherRBIter++){
    if((*myRBIter)->getNumPossibleDelay() != (*otherRBIter)->getNumPossibleDelay()){
      return false;
    }
  }
  return true;
}
    
//Check the contents of these two luts.  If either is occupied,
// swap their contents.  Notice, this not only considers the LUT itself,
// but any associated register banks too.
void vprLUT::swapLUTAndAssociatedRegBanks(vprLUT* otherLUT){
  logicBlock *myLogicBlock = currBlock;
  logicBlock *otherLogicBlock = otherLUT->currBlock;
  list <registerBank*>::iterator myRBIter;
  list <registerBank*>::iterator otherRBIter;

  if(myLogicBlock){
    unPlaceLUT(currBlock, this);
  }
  if(otherLogicBlock){
    unPlaceLUT(otherLogicBlock, otherLUT);
  }
  if(!myLogicBlock || !otherLogicBlock){
    //Just because there are no logic blocks mapped to these luts does not
    // mean they are necessarily empty.
    //Find out if either had any occupied linked register banks
    transferAllLinkedRegisters(otherLUT);
  }
    
  if(myLogicBlock){
    placeLUT(myLogicBlock, otherLUT);
  }
  if(otherLogicBlock){
    placeLUT(otherLogicBlock, this);
  }
}

//Swap just this lut with the other one, then update the timing
void vprLUT::swapSingleLUTMappingAndAssociatedRegBanks(logicBlock* myLB, vprLUT* otherLUT, logicBlock* otherLB, double criticalPathDelay){
  list <registerBank*>::iterator myRBIter;
  list <registerBank*>::iterator otherRBIter;

  //We are not maintaining the LUT packing, so just swap the two selected LUTs
  swapLUTAndAssociatedRegBanks(otherLUT);

  myLB->updateSinkDelaysAndSlacks();
  if(otherLB){
    otherLB->updateSinkDelaysAndSlacks();
  }
  else{
    //The other LUT may not be have a logic block mapped to it,
    // but if it has shift-register capabilities it may have 
    // something mapped to its companion register bank(s)
    //Those registers would now be in this lut's companion register banks
    otherLUT->updateSinkDelaysAndSlacks();
  }
  //Now update the max input and output criticalities
  myLB->calculateMaxInputOutputCriticality(criticalPathDelay);
  if(otherLB){
    otherLB->calculateMaxInputOutputCriticality(criticalPathDelay);
  }
  else{
    //The other LUT may not be have a logic block mapped to it,
    // but if it has shift-register capabilities it may have 
    // something mapped to its companion register bank(s)
    //Those registers would now be in this lut's companion register banks
    otherLUT->calculateMaxInputOutputCriticality(criticalPathDelay);
  }
}

//Swap all of the luts in this CLB with the luts in the other CLB.
//Because we may need to make space, also consider associated register banks.
void vprLUT::swapLUTGroupMappingAndAssociatedRegBanks(vprLUT* otherLUT, double criticalPathDelay){
  vprCLB* myArrayCLB = FPGA->getCLBAt(get_row(), get_col());
  vprCLB* otherArrayCLB = FPGA->getCLBAt(otherLUT->get_row(), otherLUT->get_col());
  
  //Swap any LUTs and FFs that might be mapped to the host physical clbs
  // for these two register banks.
  myArrayCLB->swapAllLUTsAndAssociatedFFs(otherArrayCLB);
  
  //Update the sink timings and slack for the nets we just moved
  // The slack calculations are not exact since the critical path might have changed
  myArrayCLB->updateSinkDelaysAndSlacksForAllLUTsAndAssociatedFFs();
  otherArrayCLB->updateSinkDelaysAndSlacksForAllLUTsAndAssociatedFFs();
  
  //Now update the max input and output criticalities
  myArrayCLB->calculateMaxInputOutputCriticalitiesForAllLUTsAndAssociatedFFs(criticalPathDelay);
  otherArrayCLB->calculateMaxInputOutputCriticalitiesForAllLUTsAndAssociatedFFs(criticalPathDelay);
}

//Swap the entire contents of the two physical CLBs these luts belong to.
//We don't rely on netlist-level clb information for anything.
void vprLUT::swapStickyMapping(vprLUT* otherLUT, double criticalPathDelay){
  vprCLB* myArrayCLB = FPGA->getCLBAt(get_row(), get_col());
  vprCLB* otherArrayCLB = FPGA->getCLBAt(otherLUT->get_row(), otherLUT->get_col());

  //Swap any LUTs and FFs that might be mapped to the host physical clbs
  // for these two register banks.
  myArrayCLB->swapAllLUTsAndFFs(otherArrayCLB);
  
  //Update the sink timings and slack for the nets we just moved
  // The slack calculations are not exact since the critical path might have changed
  myArrayCLB->updateSinkDelaysAndSlacksForAllLUTsAndAssociatedFFs();
  myArrayCLB->updateSinkDelaysAndSlacksForAllNonAssociatedFFs();
  otherArrayCLB->updateSinkDelaysAndSlacksForAllLUTsAndAssociatedFFs();
  otherArrayCLB->updateSinkDelaysAndSlacksForAllNonAssociatedFFs();
  
  //Now update the max input and output criticalities
  myArrayCLB->calculateMaxInputOutputCriticalitiesForAllLUTsAndAssociatedFFs(criticalPathDelay);
  myArrayCLB->calculateMaxInputOutputCriticalitiesForAllNonAssociatedFFs(criticalPathDelay);
  otherArrayCLB->calculateMaxInputOutputCriticalitiesForAllLUTsAndAssociatedFFs(criticalPathDelay);
  otherArrayCLB->calculateMaxInputOutputCriticalitiesForAllNonAssociatedFFs(criticalPathDelay);
}

void vprLUT::updateSinkDelaysAndSlacks(){
  if(currBlock){
    currBlock->updateSinkDelaysAndSlacks();
  }
  else{
    //If there isn't a logic block mapped here, there may be some register banks.
    for(list <registerBank*>::iterator rbIter = linkedRegisterBanks.begin();
	rbIter != linkedRegisterBanks.end(); rbIter++){
      (*rbIter)->updateSinkDelaysAndSlacks();
    }
  }
}

void vprLUT::updateWireAndTimingCosts(bool timing_driven, double criticalPathDelay){
  if(currBlock){
    currBlock->updateWireAndTimingCosts(timing_driven, criticalPathDelay);
  }
  else{
    //If there isn't a logic block mapped here, there may be some register banks.
    for(list <registerBank*>::iterator rbIter = linkedRegisterBanks.begin();
	rbIter != linkedRegisterBanks.end(); rbIter++){
      (*rbIter)->updateWireAndTimingCosts(timing_driven, criticalPathDelay);
    }
  }
}

double vprLUT::determineDeltaWireCost(){
  double deltaWire = 0.0;
  if(currBlock){
    return (currBlock->determineDeltaWireCost());
  }
  else{
    //If there isn't a logic block mapped here, there may be some register banks.
    for(list <registerBank*>::iterator rbIter = linkedRegisterBanks.begin();
	rbIter != linkedRegisterBanks.end(); rbIter++){
      deltaWire += (*rbIter)->determineDeltaWireCost();
    }
    return deltaWire;
  }
}

double vprLUT::determineDeltaTimingCost(){
  double deltaTiming = 0.0;
  if(currBlock){
    return (currBlock->determineDeltaTimingCost());
  }
  else{
    //If there isn't a logic block mapped here, there may be some register banks.
    for(list <registerBank*>::iterator rbIter = linkedRegisterBanks.begin();
	rbIter != linkedRegisterBanks.end(); rbIter++){
      deltaTiming += (*rbIter)->determineDeltaTimingCost();
    }
    return deltaTiming;
  }
}

double vprLUT::determineDeltaTimingCostPrime(){
  double deltaTiming = 0.0;
  if(currBlock){
    return (currBlock->determineDeltaTimingCostPrime());
  }
  else{
    //If there isn't a logic block mapped here, there may be some register banks.
    for(list <registerBank*>::iterator rbIter = linkedRegisterBanks.begin();
	rbIter != linkedRegisterBanks.end(); rbIter++){
      deltaTiming += (*rbIter)->determineDeltaTimingCostPrime();
    }
    return deltaTiming;
  }
}

void vprLUT::updateCostAndDeltas(bool timing_driven, double criticalPathDelay, double *deltaWire, 
				  double *deltaTiming){
  updateWireAndTimingCosts(timing_driven, criticalPathDelay);
  *deltaWire += determineDeltaWireCost();
  if(timing_driven){
    *deltaTiming += determineDeltaTimingCost();
  }
}

void vprLUT::updateCostAndDeltasCostDeltaDelay(bool timing_driven, double criticalPathDelay, double *deltaWire, 
						double *deltaTiming, double *deltaTimingPrime){
  updateWireAndTimingCosts(timing_driven, criticalPathDelay);
  *deltaWire += determineDeltaWireCost();
  if(timing_driven){
    *deltaTiming += determineDeltaTimingCost();
    *deltaTimingPrime += determineDeltaTimingCostPrime();
  }
}


void vprLUT::updateCostAndDeltasForSticky(vprLUT* otherLUT, bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming){
  vprCLB* myArrayCLB = FPGA->getCLBAt(get_row(), get_col());
  vprCLB* otherArrayCLB = FPGA->getCLBAt(otherLUT->get_row(), otherLUT->get_col());
  
  //Update the costs and deltas for all the LUTs we moved.
  myArrayCLB->updateCostAndDeltasForAllLUTsAndAssociatedFFs(timing_driven, criticalPathDelay, 
							    deltaWire, deltaTiming);
  myArrayCLB->updateCostAndDeltasForAllNonAssociatedFFs(timing_driven, criticalPathDelay, 
							deltaWire, deltaTiming);
  
  otherArrayCLB->updateCostAndDeltasForAllLUTsAndAssociatedFFs(timing_driven, criticalPathDelay, 
							       deltaWire, deltaTiming);
  otherArrayCLB->updateCostAndDeltasForAllNonAssociatedFFs(timing_driven, criticalPathDelay, 
							   deltaWire, deltaTiming);
}

void vprLUT::updateCostAndDeltasForStickyCostDeltaDelay(vprLUT* otherLUT, bool timing_driven, double criticalPathDelay, double *deltaWire, 
							double *deltaTiming, double *deltaTimingPrime){
  vprCLB* myArrayCLB = FPGA->getCLBAt(get_row(), get_col());
  vprCLB* otherArrayCLB = FPGA->getCLBAt(otherLUT->get_row(), otherLUT->get_col());
  
  //Update the costs and deltas for all the LUTs we moved.
  myArrayCLB->updateCostAndDeltasForAllLUTsAndAssociatedFFsCostDeltaDelay(timing_driven, criticalPathDelay, 
									  deltaWire, deltaTiming, deltaTimingPrime);
  myArrayCLB->updateCostAndDeltasForAllNonAssociatedFFsCostDeltaDelay(timing_driven, criticalPathDelay, 
								      deltaWire, deltaTiming, deltaTimingPrime);
  
  otherArrayCLB->updateCostAndDeltasForAllLUTsAndAssociatedFFsCostDeltaDelay(timing_driven, criticalPathDelay, 
									     deltaWire, deltaTiming, deltaTimingPrime);
  otherArrayCLB->updateCostAndDeltasForAllNonAssociatedFFsCostDeltaDelay(timing_driven, criticalPathDelay, 
									 deltaWire, deltaTiming, deltaTimingPrime);
}

void vprLUT::updateCostAndDeltasForLUTGroupAndAssociatedRegBanks(vprLUT* otherLUT, bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming){
  vprCLB* myArrayCLB = FPGA->getCLBAt(get_row(), get_col());
  vprCLB* otherArrayCLB = FPGA->getCLBAt(otherLUT->get_row(), otherLUT->get_col());
  
  //Update the costs and deltas for all the LUTs we moved.
  myArrayCLB->updateCostAndDeltasForAllLUTsAndAssociatedFFs(timing_driven, criticalPathDelay, 
							    deltaWire, deltaTiming);
  otherArrayCLB->updateCostAndDeltasForAllLUTsAndAssociatedFFs(timing_driven, criticalPathDelay, 
							       deltaWire, deltaTiming);
}

void vprLUT::updateCostAndDeltasForLUTGroupAndAssociatedRegBanksCostDeltaDelay(vprLUT* otherLUT, bool timing_driven, double criticalPathDelay, 
									       double *deltaWire, double *deltaTiming, double *deltaTimingPrime){
  vprCLB* myArrayCLB = FPGA->getCLBAt(get_row(), get_col());
  vprCLB* otherArrayCLB = FPGA->getCLBAt(otherLUT->get_row(), otherLUT->get_col());
  
  //Update the costs and deltas for all the LUTs we moved.
  myArrayCLB->updateCostAndDeltasForAllLUTsAndAssociatedFFsCostDeltaDelay(timing_driven, criticalPathDelay, 
									  deltaWire, deltaTiming, deltaTimingPrime);
  otherArrayCLB->updateCostAndDeltasForAllLUTsAndAssociatedFFsCostDeltaDelay(timing_driven, criticalPathDelay, 
									     deltaWire, deltaTiming, deltaTimingPrime);
}

//Go through all of the register banks and LUTs associated with the vprCLB at this location.
//If there are any logic blocks mapped to these locations, return the CLB associated with them
c_clb* vprLUT::returnMappedCLBFromLUTsAndFF(){
  c_clb* mappedCLB = NULL;
  vprCLB *myCLB = FPGA->getCLBAt(row, col);

  //Try to extract a CLB from the FFs at this locations
  for(vector <registerBank*>::iterator rbIter = myCLB->registerBankVector.begin(); 
      rbIter != myCLB->registerBankVector.end(); rbIter++){
    if((*rbIter)->get_currBlock()){
      if(mappedCLB){
	if(mappedCLB != (*rbIter)->get_currBlock()->parentCLB){
	  ERROR("Register bank is mapped to different logical clb!");
	}
      }
      else{
	mappedCLB = (*rbIter)->get_currBlock()->parentCLB;
      }
    }
  }
  
  //Try to extract a CLB from the LUTs at this locations
  for(vector <vprLUT *>::iterator lutIter = myCLB->lutVector.begin(); 
      lutIter != myCLB->lutVector.end(); lutIter++){
    if((*lutIter)->get_currBlock()){
      if(mappedCLB){
	if(mappedCLB != (*lutIter)->get_currBlock()->parentCLB){
	  ERROR("LUT is mapped to different logical clb!");
	}
      }
      else{
	mappedCLB = (*lutIter)->get_currBlock()->parentCLB;
      }
    }
  }
  return mappedCLB;
}

//Go through all of the register banks and LUTs associated with the vprCLB at this location.
//If there are any logic blocks mapped to these locations, return the CLB associated with them
c_clb* vprLUT::returnMappedCLBFromLUTs(){
  c_clb* mappedCLB = NULL;
  vprCLB *myCLB = FPGA->getCLBAt(row, col);

  //Try to extract a CLB from the LUTs at this locations
  for(vector <vprLUT *>::iterator lutIter = myCLB->lutVector.begin(); 
      lutIter != myCLB->lutVector.end(); lutIter++){
    if((*lutIter)->get_currBlock()){
      if(mappedCLB){
	if(mappedCLB != (*lutIter)->get_currBlock()->parentCLB){
	  ERROR("Register bank is mapped to different logical clb!");
	}
      }
      else{
	mappedCLB = (*lutIter)->get_currBlock()->parentCLB;
      }
    }
  }
  return mappedCLB;
}

//Associate a register chain with this lut
void vprLUT::setRegisterAssociation(registerBank* associatedRegisterBank){
  if(!associatedRegisterBank){
    ERROR("Trying to associated LUT with unknown register bank!");
  }

  //See if this register bank is already associated with this LUT
  for(list <registerBank*>::iterator rbIter = linkedRegisterBanks.begin(); rbIter != linkedRegisterBanks.end(); rbIter++){
    if(*rbIter == associatedRegisterBank){
      return;
    }
  }
  linkedRegisterBanks.push_back(associatedRegisterBank);
}

//Go through all of the register banks linked to this lut and remove them from the "free register" list
void vprLUT::occupyAllLinkedRegisterBanks(){
 for(list <registerBank*>::iterator rbIter = linkedRegisterBanks.begin(); rbIter != linkedRegisterBanks.end(); rbIter++){
    FPGA->occupyRegBank(*rbIter);
  }
}

//Go through all of the register banks linked to this lut and add them from the "free register" list
void vprLUT::vacateAllLinkedRegisterBanks(){
  for(list <registerBank*>::iterator rbIter = linkedRegisterBanks.begin(); rbIter != linkedRegisterBanks.end(); rbIter++){
    FPGA->vacateRegBank(*rbIter);
  }
}

//Find out if this lut has a linked register bank,
// and if so, if any of them are occupied.
bool vprLUT::hasOccupiedLinkedRegisterBank(){
  if(linkedRegisterBanks.size() == 0){
    return false;
  }
  else{
    for(list <registerBank*>::iterator rbIter = linkedRegisterBanks.begin();
	rbIter != linkedRegisterBanks.end(); rbIter++){
      if(!(*rbIter)->isJustBankEmpty()){
	return true;
      }
    }
    return false;
  }
}

//Transfer the contents of all of the linked registers associated
// with this LUT to the linked registers associated with the other LUT
void vprLUT::transferAllLinkedRegisters(vprLUT * otherLUT){
  list <registerBank*>::iterator myRBIter;
  list <registerBank*>::iterator otherRBIter;

  if(hasOccupiedLinkedRegisterBank() || otherLUT->hasOccupiedLinkedRegisterBank()){
    //Make sure that the two luts have similar companion register banks
    if(otherLUT->linkedRegisterBanks.size() != linkedRegisterBanks.size()){
      ERROR("These two LUTs have different companion register bank setups!");
    }
    for(myRBIter = linkedRegisterBanks.begin(), otherRBIter = otherLUT->linkedRegisterBanks.begin();
	myRBIter != linkedRegisterBanks.end() && otherRBIter != otherLUT->linkedRegisterBanks.end();
	myRBIter++, otherRBIter++){
      if((*myRBIter)->getNumPossibleDelay() != (*otherRBIter)->getNumPossibleDelay()){
	ERROR("These two register banks are not the same!");
      }
      
      if(!(*myRBIter)->isJustBankEmpty() || !(*otherRBIter)->isJustBankEmpty()){
	//One of the two isn't empty so we should do a swap.
	if(!(*myRBIter)->isJustBankEmpty()){
	  (*myRBIter)->swapBank(*otherRBIter);
	}
	else{
	  (*otherRBIter)->swapBank(*myRBIter);
	}
      }
    }
  }
}

void vprLUT::calculateMaxInputOutputCriticality(double criticalPathDelay){
  if(currBlock){
    currBlock->calculateMaxInputOutputCriticality(criticalPathDelay);
  }
  else{
    for(list <registerBank*>::iterator rbIter = linkedRegisterBanks.begin();
	rbIter != linkedRegisterBanks.end(); rbIter++){
      (*rbIter)->calculateMaxInputOutputCriticalitiesJustRB(criticalPathDelay);
    }
  }
}

vprCLB::vprCLB(int r,int c) : vprBaseBlock(r,c){
  derived_magicNum = 21119;
}

void vprCLB::addAllRegBanksToRegisterBankVector(){
  for(vector<registerBank *>::iterator iter = registerBankVector.begin();
      iter != registerBankVector.end(); iter++){
    FPGA->registerBankVector.push_back(*iter);
  }
}

void vprCLB::addAllRegBanksToFreeRegisterBankList(){
  for(vector<registerBank *>::iterator iter = registerBankVector.begin();
      iter != registerBankVector.end(); iter++){
    FPGA->freeRegisterBankList.push_back(*iter);
  }
}

// Correspond a routing node in the RR graph with this block
// We assume that registers are added in the correct order if that is important
// That is, in a particular bank, register 0 is added before register 1,
// register 1 is added before register 2, etc.if they are in a
// shift-chain orientation.
void vprCLB::addRoutingNodeToRegisterBank(routingNode* r, int bank){
  // We might not have that many banks yet
  if(registerBankVector.size() < bank + 1){
    registerBankVector.resize(bank + 1);
  }

  //Alternatively, that bank might not be allocated yet
  if(!registerBankVector.at(bank)){
    registerBankVector.at(bank) = new registerBank();
    registerBankVector.at(bank)->setHostCLB(this, bank);
  }
  
  registerBankVector.at(bank)->addRoutingNode(r);
}

// Correspond a routing node in the RR graph with this block
// We assume that registers are added in the correct order if that is important
// That is, in a particular bank, register 0 is added before register 1,
// register 1 is added before register 2, etc.if they are in a
// shift-chain orientation.
void vprCLB::addRoutingNodeToRegisterBank(routingNode* r, int bank, int regNum, int lutAssociation){
  // We might not have that many banks yet
  if(registerBankVector.size() < bank + 1){
    registerBankVector.resize(bank + 1);
  }

  //Alternatively, that bank might not be allocated yet
  if(!registerBankVector.at(bank)){
    registerBankVector.at(bank) = new registerBank();
    registerBankVector.at(bank)->setHostCLB(this, bank);
  }
  
  //Make sure the registers are being added in the correct order.
  if(registerBankVector.at(bank)->getNumPossibleDelay() != regNum){
    ERROR("Adding register within shift chain out of sequence!");
  }
  registerBankVector.at(bank)->addRoutingNode(r);

  //Associate this register bank with it's corresponding LUT and vice-versa
  registerBankVector.at(bank)->setLUTAssociation(lutAssociation);
}

//Return the number of empty luts in this CLB
int vprCLB::getNumEmptyLUTs(){
  int temp = 0;

  for(vector<vprLUT *>::iterator lutIter = lutVector.begin(); lutIter != lutVector.end(); lutIter++){
    if((*lutIter)->isLUTAndLinkedRegistersEmpty()){
      temp++;
    }
  }
  return temp;
}

int vprCLB::firstEmptyRegisterIndex(){
  int temp = 0;
  
  for(vector<registerBank *>::iterator rbIter = registerBankVector.begin(); 
      rbIter != registerBankVector.end(); rbIter++, temp++){
    if((*rbIter)->isBankAndLinkedLUTEmpty()){
      return (temp);
    }
  }
  return NULL;
}

//Count the number of empty register banks that are at least of a certain size in this CLB
int vprCLB::getNumEmptyRegBank(int minSize){
  int temp = 0;

  for(vector<registerBank *>::iterator rbIter = registerBankVector.begin(); 
      rbIter != registerBankVector.end(); rbIter++){
    if((*rbIter)->isBankAndLinkedLUTEmpty() && (*rbIter)->getNumPossibleDelay() >= minSize){
      temp++;
    }
  }
  return temp;
}
  
//Return a random empty register bank that is at least of a certain size in this CLB
registerBank* vprCLB::getRandomEmptyRegBank(int minSize){
  int returnIndex;
  while(1){
    returnIndex = lrand48() % getNumRegBanks();
    if(getRegBank(returnIndex)->isBankAndLinkedLUTEmpty() && 
       getRegBank(returnIndex)->getNumPossibleDelay() >= minSize){
      return(getRegBank(returnIndex));
    }
  }
}

//Return the number of register banks in this CLB that are at least of a certain size
//  and are not related to anything else in this CLB.  Unrelated means one of a few cases:
//  1) Register bank is not occupied && linked LUT is not occupied (not possible)
//  2) Register bank is occupied and it is either:
//      a) driven by something outside this CLB
//      b) driven by another register
//  3) Linked LUT is occupied, but it does not drive anything inside this CLB,
//      the calling register bank has a linked LUT too and both linked LUTs have
//      matching linked registers
int vprCLB::getNumUnrelatedRegBank(int minSize, int maxCurrOccupancy, vprLUT *linkedLUT){
  int temp = 0;
  Terminal *tempSink;
  Terminal *tempSource;
  int tempSourceRow;
  int tempSourceCol;

  for(vector<registerBank *>::iterator rbIter = registerBankVector.begin(); rbIter != registerBankVector.end(); rbIter++){
    if((*rbIter)->getNumPossibleDelay() >= minSize && (*rbIter)->getNumMappedRegs() <= maxCurrOccupancy){
      //Obviously, if the register bank is empty, it is not related to anything in the same CLB
      if((*rbIter)->isBankAndLinkedLUTEmpty()){
	//  1) Register bank is not occupied && linked LUT is not occupied (not possible)
	ERROR("This should not be possible!");
	//temp++;
      }
      else{
	if((*rbIter)->isLinkedLUTEmpty()){
	  //Either this register bank does not have a linked LUT or it does have a linked LUT but it is not occupied.
	  //  2) Register bank is occupied and it is either:
	  //      a) driven by something outside this CLB
	  //      b) driven by another register
	  tempSink = (*rbIter)->getFirstSink();
	  
	  //This sink might be driven by a pipelining net or a regular net
	  if(tempSink->sourcePipeliningNet){
	    tempSource = tempSink->sourcePipeliningNet->source;
	  }
	  else{
	    tempSource = tempSink->net->source;
	  }

	  tempSourceRow = tempSource->getLogicBlockRow();
	  tempSourceCol = tempSource->getLogicBlockCol();
	  
	  if(tempSource->rb){
	    temp++;
	  }
	  else if(row != tempSourceRow && col != tempSourceCol){
	    temp++;
	  }
	}
	else{
	  //  3) Linked LUT is occupied, but it does not drive anything inside this CLB,
	  //      the calling register bank has a linked LUT too and both linked LUTs have
	  //      matching linked registers
	  if(!linkedLUT){
	    continue;
	  }
	  else if(!linkedLUT->matchingLinkedRegisters((*rbIter)->linkedLUT)){
	    continue;
	  }
	  //Figure out if this LUT drives anything inside this CLB.
	  else if(!(*rbIter)->linkedLUT->get_currBlock()->drivesSomethingIn(row, col)){
	    temp++;
	  }
	}
      }
    }
  }
  return temp;
}

//Return the number of register banks in this CLB that are at least of a certain size,
//  are not related to anything else in this CLB, and do not include the specified sink.  
//Unrelated means one of a few cases:
//  1) Register bank is not occupied && linked LUT is not occupied (not possible)
//  2) Register bank is occupied and it is either:
//      a) driven by something outside this CLB
//      b) driven by another register
//  3) Linked LUT is occupied, but it does not drive anything inside this CLB,
//      the calling register bank has a linked LUT too and both linked LUTs have
//      matching linked registers
int vprCLB::getNumUnrelatedRegBank(int minSize, int maxCurrOccupancy, Terminal *driver, vprLUT *linkedLUT){
  int temp = 0;
  Terminal *tempSink;
  Terminal *tempSource;
  int tempSourceRow;
  int tempSourceCol;

  for(vector<registerBank *>::iterator rbIter = registerBankVector.begin(); rbIter != registerBankVector.end(); rbIter++){
    if((*rbIter)->getNumPossibleDelay() >= minSize && (*rbIter)->getNumMappedRegs() <= maxCurrOccupancy){
      //Obviously, if the register bank is empty, it is not related to anything in the same CLB
      // nor does it contain the forbidden driver.
      if((*rbIter)->isBankAndLinkedLUTEmpty()){
	//  1) Register bank is not occupied && linked LUT is not occupied (not possible)
	ERROR("This should not be possible!");
	//temp++;
      }
      else{
	if((*rbIter)->isLinkedLUTEmpty()){
	  //Either this register bank does not have a linked LUT or it does have a linked LUT but it is not occupied.
	  if(driver != (*rbIter)->getLastSource()){
	    //This register bank is occupied and does not contain the forbidden sink
	    //  2) Register bank is occupied and it is either:
	    //      a) driven by something outside this CLB
	    //      b) driven by another register
	    tempSink = (*rbIter)->getFirstSink();
	    
	    //This sink might be driven by a pipelining net or a regular net
	    if(tempSink->sourcePipeliningNet){
	      tempSource = tempSink->sourcePipeliningNet->source;
	    }
	    else{
	      tempSource = tempSink->net->source;
	    }
	  
	    tempSourceRow = tempSource->getLogicBlockRow();
	    tempSourceCol = tempSource->getLogicBlockCol();
	    
	    if(tempSource->rb){
	      temp++;
	    }
	    else if(row != tempSourceRow && col != tempSourceCol){
	      temp++;
	    }
	  }
	}
	else{
	  //  3) Linked LUT is occupied, but it does not drive anything inside this CLB,
	  //      the calling register bank has a linked LUT too and both linked LUTs have
	  //      matching linked registers
	  if(!linkedLUT){
	    continue;
	  }
	  else if(!linkedLUT->matchingLinkedRegisters((*rbIter)->linkedLUT)){
	    continue;
	  }
	  //Figure out if this LUT drives anything inside this CLB.
	  else if(!(*rbIter)->linkedLUT->get_currBlock()->drivesSomethingIn(row, col)){
	    temp++;
	  }
	}
      }
    }
  }
  return temp;
}

//Return a random register bank in this CLB that is at least of a certain size
//  and are not related to anything else in this CLB.  Unrelated means one of a few cases:
//  1) Register bank is not occupied && linked LUT is not occupied (not possible)
//  2) Register bank is occupied and it is either:
//      a) driven by something outside this CLB
//      b) driven by another register
//  3) Linked LUT is occupied, but it does not drive anything inside this CLB,
//      the calling register bank has a linked LUT too and both linked LUTs have
//      matching linked registers
registerBank* vprCLB::getRandomUnrelatedRegBank(int minSize, int maxCurrOccupancy, vprLUT *linkedLUT){
  int returnIndex;
  registerBank *tempRegBank;
  Terminal *tempSink;
  Terminal *tempSource;
  int tempSourceRow;
  int tempSourceCol;

  while(1){
    returnIndex = lrand48() % getNumRegBanks();
    tempRegBank = getRegBank(returnIndex);

    if(tempRegBank->getNumPossibleDelay() < minSize || tempRegBank->getNumMappedRegs() > maxCurrOccupancy){
      //This bank is not of sufficent size 
      continue;
    }

    //Obviously, if the register bank is empty, it is not related to anything in the same CLB
    if(tempRegBank->isBankAndLinkedLUTEmpty()){
      //  1) Register bank is not occupied && linked LUT is not occupied (not possible)
      ERROR("This should not be possible!");
      //return(tempRegBank);
    }
    else{
      if(tempRegBank->isLinkedLUTEmpty()){
	//Either this register bank does not have a linked LUT or it does have a linked LUT but it is not occupied.
	//  2) Register bank is occupied and it is either:
	//      a) driven by something outside this CLB
	//      b) driven by another register
	tempSink = tempRegBank->getFirstSink();
	
	//This sink might be driven by a pipelining net or a regular net
	if(tempSink->sourcePipeliningNet){
	  tempSource = tempSink->sourcePipeliningNet->source;
	}
	else{
	  tempSource = tempSink->net->source;
	}
	
	tempSourceRow = tempSource->getLogicBlockRow();
	tempSourceCol = tempSource->getLogicBlockCol();
	  
	if(tempSource->rb){
	  return(tempRegBank);
	}
	else if(row != tempSourceRow && col != tempSourceCol){
	  return(tempRegBank);
	}	  
      }
      else{
	//  3) Linked LUT is occupied, but it does not drive anything inside this CLB,
	//      the calling register bank has a linked LUT too and both linked LUTs have
	//      matching linked registers
	if(!linkedLUT){
	  continue;
	}
	else if(!linkedLUT->matchingLinkedRegisters(tempRegBank->linkedLUT)){
	  continue;
	}
	//Figure out if this LUT drives anything inside this CLB.
	else if(!tempRegBank->linkedLUT->get_currBlock()->drivesSomethingIn(row, col)){
	  return(tempRegBank);
	}
      }
    }
  }
}

//Return a random register bank that is at least of a certain size,
//  are not related to anything else in this CLB, and do not include the specified sink.  
//Unrelated means one of a few cases:
//  1) Register bank is not occupied && linked LUT is not occupied (not possible)
//  2) Register bank is occupied and it is either:
//      a) driven by something outside this CLB
//      b) driven by another register
//  3) Linked LUT is occupied, but it does not drive anything inside this CLB,
//      the calling register bank has a linked LUT too and both linked LUTs have
//      matching linked registers
registerBank* vprCLB::getRandomUnrelatedRegBank(int minSize, int maxCurrOccupancy, Terminal *driver, vprLUT *linkedLUT){
  int returnIndex;
  registerBank *tempRegBank;
  Terminal *tempSink;
  Terminal *tempSource;
  int tempSourceRow;
  int tempSourceCol;

  while(1){
    returnIndex = lrand48() % getNumRegBanks();
    tempRegBank = getRegBank(returnIndex);

    if(tempRegBank->getNumPossibleDelay() < minSize || tempRegBank->getNumMappedRegs() > maxCurrOccupancy){
      //This bank is not of sufficent size
      continue;
    }

    //Obviously, if the register bank is empty, it is not related to anything in the same CLB
    // nor does it contain the forbidden driver.
    if(tempRegBank->isBankAndLinkedLUTEmpty()){
      //  1) Register bank is not occupied && linked LUT is not occupied (not possible)
      ERROR("This should not be possible!");
      //return(tempRegBank);
    }
    else{
      if(tempRegBank->isLinkedLUTEmpty()){
	//Either this register bank does not have a linked LUT or it does have a linked LUT but it is not occupied.
	if(driver != tempRegBank->getLastSource()){
	  //This register bank is occupied and does not contain the forbidden sink
	  //  2) Register bank is occupied and it is either:
	  //      a) driven by something outside this CLB
	  //      b) driven by another register
	  tempSink = tempRegBank->getFirstSink();
	  
	  //This sink might be driven by a pipelining net or a regular net
	  if(tempSink->sourcePipeliningNet){
	    tempSource = tempSink->sourcePipeliningNet->source;
	  }
	  else{
	    tempSource = tempSink->net->source;
	  }
	  
	  tempSourceRow = tempSource->getLogicBlockRow();
	  tempSourceCol = tempSource->getLogicBlockCol();
	  
	  if(tempSource->rb){
	    return(tempRegBank);
	  }
	  else if(row != tempSourceRow && col != tempSourceCol){
	    return(tempRegBank);
	  }
	}
	else{
	  //  3) Linked LUT is occupied, but it does not drive anything inside this CLB,
	  //      the calling register bank has a linked LUT too and both linked LUTs have
	  //      matching linked registers
	  if(!linkedLUT){
	    continue;
	  }
	  else if(!linkedLUT->matchingLinkedRegisters(tempRegBank->linkedLUT)){
	    continue;
	  }
	  //Figure out if this LUT drives anything inside this CLB.
	  else if(!tempRegBank->linkedLUT->get_currBlock()->drivesSomethingIn(row, col)){
	    return(tempRegBank);
	  }
	}
      }
    }
  }
}

//Return the number of register banks that are at least of a certain size in this CLB, and:
// 1) Register bank is not occupied && linked LUT is not occupied (not possible)
// 2) Register bank is occupied
// 3) Linked LUT is occupied, the calling register bank has a linked LUT too and both linked LUTs have
//      matching linked registers
int vprCLB::getNumMinSizeRegBank(int minSize, int maxCurrOccupancy, vprLUT *linkedLUT){
  int temp = 0;

  for(vector<registerBank *>::iterator rbIter = registerBankVector.begin(); 
      rbIter != registerBankVector.end(); rbIter++){
    if((*rbIter)->getNumPossibleDelay() >= minSize && (*rbIter)->getNumMappedRegs() <= maxCurrOccupancy){
      if((*rbIter)->isBankAndLinkedLUTEmpty()){
	//  1) Register bank is not occupied && linked LUT is not occupied (not possible)
	ERROR("This should not be possible!");
	//temp++;
      }
      else{
	if((*rbIter)->isLinkedLUTEmpty()){
	  //Either this register bank does not have a linked LUT or it does have a linked LUT but it is not occupied.
	  temp++;
	}
	else if(linkedLUT && linkedLUT->matchingLinkedRegisters((*rbIter)->linkedLUT)){
	  //  3) Linked LUT is occupied, the calling register bank has a linked LUT too and both linked LUTs have
	  //      matching linked registers
	  temp++;
	}
      }
    }
  }
  return temp;
}

//Return the number of register banks that are at least of a certain size in this CLB, 
// do not contain a specific driver and:
// 1) Register bank is not occupied && linked LUT is not occupied (not possible)
// 2) Register bank is occupied
// 3) Linked LUT is occupied, the calling register bank has a linked LUT too and both linked LUTs have
//      matching linked registers
int vprCLB::getNumMinSizeRegBank(int minSize, int maxCurrOccupancy, Terminal *driver, vprLUT *linkedLUT){
  int temp = 0;

  for(vector<registerBank *>::iterator rbIter = registerBankVector.begin(); rbIter != registerBankVector.end(); rbIter++){
    if((*rbIter)->getNumPossibleDelay() >= minSize && (*rbIter)->getNumMappedRegs() <= maxCurrOccupancy){
      if((*rbIter)->isBankAndLinkedLUTEmpty()){
	//  1) Register bank is not occupied && linked LUT is not occupied (not possible)
	ERROR("This should not be possible!");
	//temp++;
      }
      else{
	if((*rbIter)->isLinkedLUTEmpty()){
	  //Either this register bank does not have a linked LUT or it does have a linked LUT but it is not occupied.
	  //This means that the register bank must be occupied.
	  if(driver != (*rbIter)->getLastSource()){	
	    temp++;
	  }
	}
	else if(linkedLUT && linkedLUT->matchingLinkedRegisters((*rbIter)->linkedLUT)){
	  // 3) Linked LUT is occupied, the calling register bank has a linked LUT too and both linked LUTs have
	  //      matching linked registers
	  temp++;
	}
      }
    }
  }
  return temp;
}

//Return a random register bank that is at least of a certain size in this CLB, and:
// 1) Register bank is not occupied && linked LUT is not occupied (not possible)
// 2) Register bank is occupied
// 3) Linked LUT is occupied, the calling register bank has a linked LUT too and both linked LUTs have
//      matching linked registers
registerBank* vprCLB::getRandomMinSizeRegBank(int minSize, int maxCurrOccupancy, vprLUT *linkedLUT){
  int returnIndex;
  registerBank *tempRegBank;

  while(1){
    returnIndex = lrand48() % getNumRegBanks();
    tempRegBank = getRegBank(returnIndex);

    if(tempRegBank->getNumPossibleDelay() < minSize || tempRegBank->getNumMappedRegs() > maxCurrOccupancy){
      continue;
    }

    if(tempRegBank->isBankAndLinkedLUTEmpty()){
      //  1) Register bank is not occupied && linked LUT is not occupied (not possible)
      ERROR("This should not be possible!");
      //return(tempRegBank);
    }
    else{
      if((tempRegBank)->isLinkedLUTEmpty()){
	//Either this register bank does not have a linked LUT or it does have a linked LUT but it is not occupied.
	return(tempRegBank);
      }
      else if(linkedLUT && linkedLUT->matchingLinkedRegisters(tempRegBank->linkedLUT)){
	//  3) Linked LUT is occupied, the calling register bank has a linked LUT too and both linked LUTs have
	//      matching linked registers
	return(tempRegBank);
      }
    }
  }
} 

//Return a random register bank that is at least of a certain size in this CLB, 
// do not contain a specific driver and:
// 1) Register bank is not occupied && linked LUT is not occupied (not possible)
// 2) Register bank is occupied
// 3) Linked LUT is occupied, the calling register bank has a linked LUT too and both linked LUTs have
//      matching linked registers
registerBank* vprCLB::getRandomMinSizeRegBank(int minSize, int maxCurrOccupancy, Terminal *driver, vprLUT *linkedLUT){
  int returnIndex;
  registerBank *tempRegBank;
    
  while(1){
    returnIndex = lrand48() % getNumRegBanks();
    tempRegBank = getRegBank(returnIndex);

    if(tempRegBank->getNumPossibleDelay() < minSize || tempRegBank->getNumMappedRegs() > maxCurrOccupancy){
      continue;
    }

    if(tempRegBank->isBankAndLinkedLUTEmpty()){
      //  1) Register bank is not occupied && linked LUT is not occupied (not possible)
      ERROR("This should not be possible!");
      //return(tempRegBank);
    }
    else{
      if(tempRegBank->isLinkedLUTEmpty()){
	//Either this register bank does not have a linked LUT or it does have a linked LUT but it is not occupied.
	//This means that the register bank must be occupied.
	if(driver != tempRegBank->getLastSource()){	
	  return(tempRegBank);
	}
      }
      else if(linkedLUT && linkedLUT->matchingLinkedRegisters(tempRegBank->linkedLUT)){
	//  3) Linked LUT is occupied, the calling register bank has a linked LUT too and both linked LUTs have
	//      matching linked registers
	return(tempRegBank);
      }
    }
  }
}

vprLUT* vprCLB::getOrMakeLUT(int i){
  //Our lut vector may not be large enough
  if(lutVector.size() <= i){
    lutVector.resize(i + 1);
  } 
  
  if(!lutVector.at(i)){
    lutVector.at(i) = new vprLUT(row, col, i);
  }

  return lutVector.at(i);
}

bool vprCLB::areRegistersEmpty(){
  vector<registerBank *>::iterator rbIter;

  for(rbIter = registerBankVector.begin(); rbIter != registerBankVector.end(); rbIter++){
    if(!(*rbIter)->isBankAndLinkedLUTEmpty()){
      return false;
    }
  }
  return true;
}

bool vprCLB::areLUTsEmpty(){
  vector<vprLUT *>::iterator lutIter;

  for(lutIter = lutVector.begin(); lutIter != lutVector.end(); lutIter++){
    if(!(*lutIter)->isLUTAndLinkedRegistersEmpty()){
      return false;
    }
  }
  return true;
}

//Try to find an empty register bank in the given CLB
// if there isn't one, return NULL
//This assumes that single register banks come before shift registers in the
// bank vector
registerBank* vprCLB::getFirstEmptySingleOrShiftRegisterBank(){
  //Go through all of the register banks in this clb and return the first empty one
  for(vector<registerBank *>::iterator rbIter = registerBankVector.begin(); 
      rbIter != registerBankVector.end(); rbIter++){
    //if((*rbIter)->isJustBankEmpty() && (*rbIter)->getNumPossibleDelay() == 1){
    if((*rbIter)->isBankAndLinkedLUTEmpty()){
      //cout << "Found empty register bank!" << endl;
      return (*rbIter);
    }
  }
  return NULL;
}

//Try to find an empty register bank in the given CLB
//if there isn't one, return NULL
registerBank* vprCLB::getFirstEmptyShiftRegisterBank(int capacity){
  //Go through all of the register banks in this clb and return the first empty one
  for(vector<registerBank *>::iterator rbIter = registerBankVector.begin(); 
      rbIter != registerBankVector.end(); rbIter++){
    if((*rbIter)->isBankAndLinkedLUTEmpty() && (*rbIter)->getNumPossibleDelay() >= capacity){
      //cout << "Found empty register bank!" << endl;
      return (*rbIter);
    }
  }
  return NULL;
}

//Try to find the first empty register bank in the given CLB.
//Failing that, find the first register bank not related to the LUTs in this CLB
registerBank* vprCLB::getFirstEmptyUnrelatedRegisterBank(){
  registerBank *emptyRegBank;
  Terminal *source;
  int sourceRow, sourceCol;
  
  //See if there is an empty register bank in this CLB
  emptyRegBank = getFirstEmptySingleOrShiftRegisterBank();
  
  if(emptyRegBank){
    return emptyRegBank;
  }

  //cout << "All register banks are occupied!  Looking for an unrelated FF" << endl;
  //Failing that, find the first register not related to the LUTs in the host CLB
  for(vector<registerBank *>::iterator rbIter = registerBankVector.begin(); 
      rbIter != registerBankVector.end(); rbIter++){
    //This register bank might be driven by a pipelining net or a regular net
    if((*rbIter)->getFirstSink()->sourcePipeliningNet){
      source = (*rbIter)->getFirstSink()->sourcePipeliningNet->source;
    }
    else{
      source = (*rbIter)->getFirstSink()->net->source;
    }

    //See if this register bank is occupied by a self-feeding register bank in the netlist (constant generator)
    if(source->lBlock && (*rbIter)->getFirstSink()->lBlock){
      if(source->lBlock == (*rbIter)->getFirstSink()->lBlock){
	//cout << "Found a constant generator register bank!" << endl;
	return(*rbIter);
      }
    }


    sourceRow = source->getLogicBlockRow();
    sourceCol = source->getLogicBlockCol();
    
    if(sourceCol != (*rbIter)->getHostCLB()->get_col() || sourceRow != (*rbIter)->getHostCLB()->get_row()){
      //cout << "Found a register bank not driven by something in this CLB!" << endl;
      return(*rbIter);
    }
  }

  ERROR("Unable to find a suitable register bank in this CLB");
  return NULL;
}

//Swap all of the LUTs in this CLB with the ones in the other CLB.
//This function relies on the netlist-level CLB information for these
// luts to be intact.  Thus, while we can use this function when we
// have done FF-level placement, we can't use it if we have done lut-level placement
void vprCLB::swapAllLUTs(vprCLB *otherArrayCLB){
  c_clb* myCLB = 0;
  c_clb* otherCLB = 0;
  vprLUT* myLUT = lutVector.front();
  vprLUT* otherLUT = otherArrayCLB->lutVector.front();

  if(lut_level_placement){
    ERROR("We don't support lut-level placement with grouped LUTs yet!");
  }
  
  myCLB = myLUT->returnMappedCLBFromLUTs();
  otherCLB = otherLUT->returnMappedCLBFromLUTs();
  
  //Remove any logic blocks from the old locations
  if(myCLB)
    myCLB->unPlaceLogicBlocksFromLUTs(myLUT);
  if(otherCLB)
    otherCLB->unPlaceLogicBlocksFromLUTs(otherLUT);

  //Put this into the new locations
  if(myCLB)
    myCLB->placeLogicBlocksToLUTs(otherLUT);
  if(otherCLB)
    otherCLB->placeLogicBlocksToLUTs(myLUT);
}

//Swap all of the LUTs and FFs in this CLB with the ones in the other CLB.
//Don't count on any netlist CLB information.
void vprCLB::swapAllLUTsAndFFs(vprCLB *otherArrayCLB){
  vector<vprLUT *>::iterator myLUTIter;
  vector<vprLUT *>::iterator otherLUTIter;
  int counter;
  vector<registerBank *>::iterator myRBIter;
  vector<registerBank *>::iterator otherRBIter;

  if(lutVector.size() != otherArrayCLB->lutVector.size()){
    ERROR("Incompatible number of luts!");
  }
  if(registerBankVector.size() != otherArrayCLB->registerBankVector.size()){
    ERROR("Incompatible number of register banks!");
  }

  for(myLUTIter = lutVector.begin(), otherLUTIter = otherArrayCLB->lutVector.begin();
      myLUTIter != lutVector.end() && otherLUTIter != otherArrayCLB->lutVector.end();
      myLUTIter++, otherLUTIter++){
    (*myLUTIter)->swapLUTAndAssociatedRegBanks(*otherLUTIter);
  }
  
  for(myRBIter = registerBankVector.begin(), otherRBIter = otherArrayCLB->registerBankVector.begin();
      myRBIter != registerBankVector.end() && otherRBIter != otherArrayCLB->registerBankVector.end();
      myRBIter++, otherRBIter++){
    //Only swap non-associated register banks.  Associated ones have already been
    // taken care of.
    if(!(*myRBIter)->linkedLUT && !(*otherRBIter)->linkedLUT){
      if(!(*myRBIter)->isJustBankEmpty() || !(*otherRBIter)->isJustBankEmpty()){
	//One of the two isn't empty so we should do a swap.
	//This empty check works because these are non-associated register banks.
	if(!(*myRBIter)->isJustBankEmpty()){
	  (*myRBIter)->swapBank(*otherRBIter);
	}
	else{
	  (*otherRBIter)->swapBank(*myRBIter);
	}
      }
    }
  }
}

//Swap all of the LUTs and linked FFs in this CLB with the ones in the other CLB.
//Don't count on any netlist CLB information.
void vprCLB::swapAllLUTsAndAssociatedFFs(vprCLB *otherArrayCLB){
  vector<vprLUT *>::iterator myLUTIter;
  vector<vprLUT *>::iterator otherLUTIter;
  int counter;
  if(lutVector.size() != otherArrayCLB->lutVector.size()){
    ERROR("Incompatible number of luts!");
  }

  for(myLUTIter = lutVector.begin(), otherLUTIter = otherArrayCLB->lutVector.begin();
      myLUTIter != lutVector.end() && otherLUTIter != otherArrayCLB->lutVector.end();
      myLUTIter++, otherLUTIter++){
    (*myLUTIter)->swapLUTAndAssociatedRegBanks(*otherLUTIter);
  }
}

//Update the sink delays and slacks for all of the luts in this physical CLB
void vprCLB::updateSinkDelaysAndSlacksForAllLUTsAndAssociatedFFs(){
  vector<vprLUT *>::iterator lutIter;

  for(lutIter = lutVector.begin(); lutIter != lutVector.end(); lutIter++){
    if(!(*lutIter)->isLUTAndLinkedRegistersEmpty()){
      (*lutIter)->updateSinkDelaysAndSlacks();
    }
  }
}

//Update the sink delays and slacks for all of the FFs in this physical CLB
void vprCLB::updateSinkDelaysAndSlacksForAllNonAssociatedFFs(){
  vector<registerBank *>::iterator rbIter;

  for(rbIter = registerBankVector.begin(); rbIter != registerBankVector.end(); rbIter++){
    if(!(*rbIter)->hasLinkedLUT() && !(*rbIter)->isJustBankEmpty()){
      (*rbIter)->updateSinkDelaysAndSlacks();
    }
  }
}

//Update the costs and deltas for all of the luts in this clb
void vprCLB::updateCostAndDeltasForAllLUTsAndAssociatedFFs(bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming){
  vector<vprLUT *>::iterator lutIter;

  for(lutIter = lutVector.begin(); lutIter != lutVector.end(); lutIter++){
    if(!(*lutIter)->isLUTAndLinkedRegistersEmpty()){
      (*lutIter)->updateCostAndDeltas(timing_driven, criticalPathDelay, deltaWire, 
					 deltaTiming);
    }
  }
}

//Update the costs and deltas for all of the luts in this clb
void vprCLB::updateCostAndDeltasForAllLUTsAndAssociatedFFsCostDeltaDelay(bool timing_driven, double criticalPathDelay, double *deltaWire, 
									 double *deltaTiming, double *deltaTimingPrime){
  vector<vprLUT *>::iterator lutIter;

  for(lutIter = lutVector.begin(); lutIter != lutVector.end(); lutIter++){
    if(!(*lutIter)->isLUTAndLinkedRegistersEmpty()){
      (*lutIter)->updateCostAndDeltasCostDeltaDelay(timing_driven, criticalPathDelay, deltaWire, 
						     deltaTiming, deltaTimingPrime);
    }
  }
}


void vprCLB::updateCostAndDeltasForAllNonAssociatedFFs(bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming){
  vector<registerBank *>::iterator rbIter;

  for(rbIter = registerBankVector.begin(); rbIter != registerBankVector.end(); rbIter++){
    if(!(*rbIter)->linkedLUT && !(*rbIter)->isJustBankEmpty()){
      (*rbIter)->updateCostAndDeltas(timing_driven, criticalPathDelay, deltaWire, deltaTiming);
    }
  }
}

void vprCLB::updateCostAndDeltasForAllNonAssociatedFFsCostDeltaDelay(bool timing_driven, double criticalPathDelay, double *deltaWire, 
								     double *deltaTiming, double *deltaTimingPrime){
  vector<registerBank *>::iterator rbIter;

  for(rbIter = registerBankVector.begin(); rbIter != registerBankVector.end(); rbIter++){
    if(!(*rbIter)->linkedLUT && !(*rbIter)->isJustBankEmpty()){
      (*rbIter)->updateCostAndDeltasCostDeltaDelay(timing_driven, criticalPathDelay, deltaWire, deltaTiming, deltaTimingPrime);
    }
  }
}

void vprCLB::calculateMaxInputOutputCriticalitiesForAllLUTsAndAssociatedFFs(double criticalPathDelay){
  vector<vprLUT *>::iterator lutIter;

  for(lutIter = lutVector.begin(); lutIter != lutVector.end(); lutIter++){
    if(!(*lutIter)->isLUTAndLinkedRegistersEmpty()){
      (*lutIter)->calculateMaxInputOutputCriticality(criticalPathDelay);
    }
  }
}

void vprCLB::calculateMaxInputOutputCriticalitiesForAllNonAssociatedFFs(double criticalPathDelay){
  vector<registerBank *>::iterator rbIter;

  for(rbIter = registerBankVector.begin(); rbIter != registerBankVector.end(); rbIter++){
    if(!(*rbIter)->hasLinkedLUT() && !(*rbIter)->isJustBankEmpty()){
      (*rbIter)->calculateMaxInputOutputCriticalitiesJustRB(criticalPathDelay);
    }
  }
}

/******************************************************************/

vprIOB::vprIOB(int r,int c) : vprBaseBlock(r,c){
  derived_magicNum=21113;
  sourceArray=sinkArray=0;
  currBlockArray=0;
  numIOs=0;
}

vprIOB::vprIOB(int r,int c,int n) : vprBaseBlock(r,c) {
  derived_magicNum=21113;
  numIOs=n;
  sourceArray=new vector<routingNode*>(n);
  sinkArray=new vector<routingNode*>(n);
  currBlockArray=new vector<logicBlock*>(n);

  for(int i=0;i<n;i++){
    sourceArray->at(i)=0;
    sinkArray->at(i)=0;
    currBlockArray->at(i)=0;
  }
}

vprIOB::~vprIOB(){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  else
    derived_magicNum=0;
}

void vprIOB::addto_sourceArray(routingNode* r,int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(!r || i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  sourceArray->at(i)=r;
}

void vprIOB::addto_sinkArray(routingNode* r,int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(!r || i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  sinkArray->at(i)=r;
}

void vprIOB::addto_currBlockArray(logicBlock* l,int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  currBlockArray->at(i)=l;
}

int vprIOB::get_numIOs(){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  return numIOs;
}

routingNode* vprIOB::getfrom_sourceArray(int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  return sourceArray->at(i);
}

routingNode* vprIOB::getfrom_sinkArray(int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  return sinkArray->at(i);
}

logicBlock* vprIOB::getfrom_currBlockArray(int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  return currBlockArray->at(i);
}

void vprIOB::print_info(){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");

  vprBaseBlock::print_info();
}

void vprIOB::lineup_IOs(){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");

  logicBlock* lb;routingNode* rnode;Terminal* term;int index=0;
  for(int i=0;i<numIOs;i++){
    if(lb=currBlockArray->at(i)){
      currBlockArray->at(index)=lb;
      lb->ptc_index=index;
      if(lb->get_blockType() == OUTPUTPAD){
	/* We have an output pad */
	rnode=sinkArray->at(index);
	lb->create_inputListItr();
	term=lb->getnext_inputList();
	//rnode->addto_terminalList(term);
	term->set_rNode(rnode);
	lb->remove_inputListItr();
      }
      else{
	/* We have an input pad */
	rnode=sourceArray->at(index);
	lb->create_outputListItr();
	term=lb->getnext_outputList();
	//rnode->addto_terminalList(term);
	term->set_rNode(rnode);
	lb->remove_outputListItr();
      }
      index++;
    }
  }

}

void vprIOB::clearCurrBlockArray(int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  currBlockArray->at(i)=0;
}

void vprIOB::swapMapping(int myIndex, logicBlock* myLB, vprIOB* otherIOB, int otherIndex, logicBlock* otherLB, double criticalPathDelay){
  //Remove the io pins from the old locations
  unPlaceBlock(myLB, this, IOPIN, myIndex);
  if(otherLB)
    unPlaceBlock(otherLB, otherIOB, IOPIN, otherIndex);	
  
  //Put them in the new locations
  placeBlock(myLB, otherIOB, IOPIN, otherIndex);
  if(otherLB)
    placeBlock(otherLB, this, IOPIN, myIndex);

  //Update the sink timings and slack for the nets we just moved
  // The slack calculations are not exact since the critical path might have changed
  myLB->updateSinkDelaysAndSlacks();
  if(otherLB)
    otherLB->updateSinkDelaysAndSlacks();

  //Now update the max input and output criticalities
  myLB->calculateMaxInputOutputCriticality(criticalPathDelay);
  if(otherLB){
    otherLB->calculateMaxInputOutputCriticality(criticalPathDelay);
  }
}

/******************************************************************/
vprFPGA::vprFPGA(int num_rows,int num_cols,int num_ios){
  int i, j;

  numRows = num_rows; 
  numCols = num_cols; 
  numIOs_perIOB = num_ios;
  numCLBs = numRows * numCols; 
  numIOs = 2 * numIOs_perIOB * (numRows + numCols);
  magicNum = 23111;
  numFreeRegisters = 0;
}

vprFPGA::~vprFPGA(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  else
    magicNum=0;

  //Delete the array
  
 //Make sure we are asking for a CLB location
  for(int row = 0; row <= numRows + 1; row++){
    for(int col = 0; col <= numCols + 1; col++){
      if(row < 1 || row > numRows || col < 1 || col > numCols){
	if(array.at(row).at(col)){
	  delete (vprIOB*) array.at(row).at(col);
	}
      }
      else{
	delete (vprCLB*) array.at(row).at(col);
      }
    }
  }

  //Delete all of the routing nodes
  for(list<routingNode*>::iterator rnodeIter = rnodeList.begin(); rnodeIter != rnodeList.end(); rnodeIter++){
    delete *rnodeIter;
  } 
}

int vprFPGA::get_numCLBs(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  return numCLBs;
}

int vprFPGA::get_numIOs(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  return numIOs;
}

int vprFPGA::countLUT(){
  int r, c, count = 0;
  int i;
  vprCLB* clb;
  vprLUT* lut;
  vector<vprLUT *>::iterator lutIter;

  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  for(r = 0; r < get_numRows(); r++){
    for(c = 0; c < get_numCols(); c++){
      clb = getCLBAt(r + 1, c + 1);
      for(i = 0; i < clb->getNumLUTs(); i++){
	lut = clb->getLUT(i);
	if(lut->get_blockType() == LUT){
	  count ++;
	}
      }
    }
  }
  return (count);
}

int vprFPGA::countMult(){
  int r, c, count = 0;
  int i;
  vprCLB* clb;
  vprLUT* lut;
  vector<vprLUT *>::iterator lutIter;

  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  for(r = 0; r < get_numRows(); r++){
    for(c = 0; c < get_numCols(); c++){
      clb = getCLBAt(r + 1, c + 1);
      for(i = 0; i < clb->getNumLUTs(); i++){
	lut = clb->getLUT(i);
	if(lut->get_blockType() == MULTIPLIER){
	  count ++;
	}
      }
    }
  }
  return (count);
}

int vprFPGA::countRegBanks(){
  int r, c, count = 0;
  vprCLB* clb;
  vprLUT* lut;
  vector<vprLUT *>::iterator lutIter;

  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  for(r = 0; r < get_numRows(); r++){
    for(c = 0; c < get_numCols(); c++){
      count += getCLBAt(r + 1, c + 1)->getNumRegBanks();
    }
  }
  return (count);
}

int vprFPGA::countRAM(){
  int r, c, count = 0;
  int i;
  vprCLB* clb;
  vprLUT* lut;
  vector<vprLUT *>::iterator lutIter;

  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  for(r = 0; r < get_numRows(); r++){
    for(c = 0; c < get_numCols(); c++){
      clb = getCLBAt(r + 1, c + 1);
      for(i = 0; i < clb->getNumLUTs(); i++){
	lut = clb->getLUT(i);
	if(lut->get_blockType() == RAM_BLOCK){
	  count ++;
	}
      }
    }
  }
  return (count);
}

void vprFPGA::addto_rnodeList(routingNode* r){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  if(!r)
    ERROR("trying to add NULL rnode...");
  rnodeList.push_back(r);
}

void vprFPGA::create_rnodeListItr(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  if(rnode_itr)
    ERROR("an iterator already exists...");
  rnode_itr=new list<routingNode*>::iterator;
  *rnode_itr=rnodeList.begin();
}

void vprFPGA::remove_rnodeListItr(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  if(!rnode_itr)
    ERROR("there is no iterator to remove...");
  delete rnode_itr;
  rnode_itr=0;
}

routingNode* vprFPGA::getnext_rnodeList(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  if(!rnode_itr)
    ERROR("trying to access NULL iterator here...");

  if(*rnode_itr == rnodeList.end())
    return 0;
  else{
    routingNode* ret_ptr = **rnode_itr;
    (*rnode_itr)++;
    return ret_ptr;
  }

}

long int vprFPGA::countCongestedNodeOccupancy(int opt){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
}

void vprFPGA::updateHistoryCosts(int iter){
  list<routingNode*>::iterator rnodelist_itr;

  if(magicNum!=23111)
    ERROR("trying to double delete node here...");

  for(rnodelist_itr = rnodeList.begin(); rnodelist_itr != rnodeList.end(); rnodelist_itr++){
    (*rnodelist_itr)->update_historyCost(iter);
  }
}

void vprFPGA::print_info(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");

  list<routingNode*>::iterator rn_itr=rnodeList.begin();
  for(;rn_itr!=rnodeList.end();rn_itr++){
    if((*rn_itr)->type == BUS_OR_WIRE){
      cout << "Wire node " << (*rn_itr)->name << " has delay " << (*rn_itr)->intrinsicDelay << endl;
    }
  }
  
  printf("IOB list has %d blocks\n", iobList.size());
  for(int i = 0; i < iobList.size(); i++){
    printf("\tBlock %d at %d, %d\n", i, (iobList.at(i))->get_row(), (iobList.at(i))->get_col());
  }

  printf("LUT list has %d blocks\n", lutList.size());
  for(int i = 0; i < lutList.size(); i++){
    printf("\tBlock %d at %d, %d\n", i, (lutList.at(i))->get_row(), (lutList.at(i))->get_col());
  }
  printf("MULT list has %d blocks\n", multList.size());
  for(int i = 0; i < multList.size(); i++){
    printf("\tBlock %d at %d, %d\n", i, (multList.at(i))->get_row(), (multList.at(i))->get_col());
  }
  printf("RAM list has %d blocks\n", ramList.size());
  for(int i = 0; i < ramList.size(); i++){
    printf("\tBlock %d at %d, %d\n", i, (ramList.at(i))->get_row(), (ramList.at(i))->get_col());
  }

  cout << "Array has " << registerBankVector.size() << " routing delay register banks" << endl;
  for(vector<registerBank*>::iterator itr = registerBankVector.begin(); itr != registerBankVector.end(); itr++){
    (*itr)->print_info();
  }
}

void vprFPGA::lineup_IOBs(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");

  for(int i=1;i<=numCols;i++){
    getIOBAt(0, i)->lineup_IOs();
    getIOBAt(numRows + 1, i)->lineup_IOs();
  }

  for(int j=1;j<=numRows;j++){
    getIOBAt(j, 0)->lineup_IOs();
    getIOBAt(j, numCols + 1)->lineup_IOs();
  }

}

//Go through all routing nodes in the architecure and determine the
//average intrinsic delay
double vprFPGA::calculateAverageIntrinsicDelay(){
  long numNodes;
  double intrinsicDelaySum;
  list<routingNode*>::iterator rnodeIter;

  numNodes = 0;
  intrinsicDelaySum = 0.0;
  for(rnodeIter = rnodeList.begin(); rnodeIter != rnodeList.end(); rnodeIter++){
    intrinsicDelaySum += (*rnodeIter)->intrinsicDelay;
    numNodes++;
  }

  return (intrinsicDelaySum / (double) numNodes);
}

// Go through all routing nodes in the architecture and scale their base cost
// by the average intrinsic delay of all nodes
// Thus, since the base cost of a input pin is normally set to 0.95 and a output pin
// has a base cost of 1.0, all nodes will maintain their relative proportion to each other
void vprFPGA::scaleBaseCosts(double averageIntrinsicDelay){
  list<routingNode*>::iterator rnodeIter;

  for(rnodeIter = rnodeList.begin(); rnodeIter != rnodeList.end(); rnodeIter++){
    (*rnodeIter)->baseCost = (*rnodeIter)->baseCost * averageIntrinsicDelay;
  }
}

//Initialize some of the placement timing data structures
void vprFPGA::initPlacementDS(bool testInitialPlacementOnly){
  // First, build lists of various block types
  // This is relatively Rapid-specific
  create_block_type_lists();
  
  // Now create the timing table we'll use to estimate delay during placement.
  // We do this even if we are not doing timing-driven placement so that
  //  we can give an estimate of the critical path delay.
  if(!testInitialPlacementOnly){
    createTimingTable();
      
    if(maxCutCost){
      createCutCostVector();
    }
  }
}

//Count the number of delay nodes in the architecture
int vprFPGA::countDelayReg(){
  list<routingNode*>::iterator rnodeIter;
  int count = 0;

  for(rnodeIter = rnodeList.begin(); rnodeIter != rnodeList.end(); rnodeIter++){
    if((*rnodeIter)->type == DELAY){
      count++;
    }
  }

  numFreeRegisters = count;
  return (count);
}

//Go through and make all the occupancy of all of the routing nodes 0
void vprFPGA::zeroOccupancy(){
  list<routingNode*>::iterator rnodeIter;

  for(rnodeIter = rnodeList.begin(); rnodeIter != rnodeList.end(); rnodeIter++){
    (*rnodeIter)->newOccupancy = 0;
  }
}

long int vprFPGA::countNewSharing(int opt){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");

  int sharing=0;long int totalSharing=0;
  list<routingNode*>::iterator rnode_itr;

  for(rnode_itr = rnodeList.begin(); rnode_itr != rnodeList.end(); rnode_itr++){
    if((*rnode_itr)->newOccupancy != (*rnode_itr)->occupancy){
      cout << "Node " << (*rnode_itr)->name << " has old occupancy " << (*rnode_itr)->occupancy << " and new occupancy " << (*rnode_itr)->newOccupancy << endl;
    }

    sharing=(*rnode_itr)->newOccupancy - (*rnode_itr)->get_capacity();
    if(sharing <= 0){
      sharing=0;
    }
    else{
      cout << "Node " << (*rnode_itr)->name << " is shared by " << sharing << endl;
    }
    totalSharing+=sharing;

    if(opt==1)
      cout<<endl<<(*rnode_itr)->get_name()<<"  "<<sharing;
    else if(opt==2){
      if(sharing>0)
	cout<<endl<<(*rnode_itr)->get_name()<<"  "<<(*rnode_itr)->get_occupancy();
    }
  }

  return totalSharing;

}

// The block-type lists let you know where the 5th ALU or 2nd Multiplier is in the system
void vprFPGA::create_block_type_lists(){
  vprCLB* clb;
  vector<vprLUT *>::iterator lutIter;

  for(int r = 0; r < get_numRows(); r++){
    for(int c = 0; c < get_numCols(); c++){
      clb = getCLBAt(r + 1, c + 1);
      
      //Put all of the LUTs into the list
      for(lutIter = clb->lutVector.begin(); lutIter != clb->lutVector.end(); lutIter++){
	if((*lutIter)->get_blockType() == LUT)
	  lutList.push_back(*lutIter);
	else if((*lutIter)->get_blockType() == MULTIPLIER)
	  multList.push_back(*lutIter);
	else if((*lutIter)->get_blockType() == RAM_BLOCK)
	  ramList.push_back(*lutIter);
	else
	  ERROR("Unknown block type in array!");
      }
      
      // Go through all the registers blocks in the system and put
      //  each register bank into a list
      clb->addAllRegBanksToRegisterBankVector();
      clb->addAllRegBanksToFreeRegisterBankList();
    }
  }

  //Put all of the IO pins into the IOB list
  for(int r = 1; r <= get_numRows(); r++){
    //Left side
    iobList.push_back(getIOBAt(r, 0));
    //Right side
    iobList.push_back(getIOBAt(r, get_numCols()+ 1));
  }

  for(int c = 1; c <= get_numCols(); c++){
    //Top
    iobList.push_back(getIOBAt(0, c));
    //Bottom
    iobList.push_back(getIOBAt(get_numRows()+ 1, c));
  }
}

void vprFPGA::createTimingTable(){
  //Initialize the timing table itself
  initializeTimingTable();
  
  //Create a table to keep track of how many times we find a given type of sink
  // at a given distance - we need this to calculate average time
  initializeTimingNumHitTable();
 
  //Launch searches from all of the input pins
  sweepTimingIOPins();
  
  //These aren't necessary since we assume that all blocks have the same
  // connectivity regardless of type.
  //This is fine until we introduce interconnect registers.
  //Launch searches from all of the logic blocks
  sweepTimingLogic();

  //Launch searches from all of the delay registers
  sweepTimingDelay();

  //Calculate average time given how many times we found each kind of sink
  calculateTimingTable();

  emptyTimingNumHitTable();
}

void vprFPGA::createCutCostVector(){
  //The cut cost vector should have N-1 entries since there are N-1 gaps
  // Num Cols only accounts for logic blocks, but we have I/O pins too
  cutCostVector.resize(get_numCols() + 1, 0.0);
}

void vprFPGA::clearCutCostVector(){
  for(vector<double>::iterator iter = cutCostVector.begin(); iter != cutCostVector.end(); iter++){
    *iter = 0.0;
  }
}

//Make enough space for all the values we'll need in the timing table
//The first dimension is the number of different kinds of sources we have
//The second dimension is the number of different source banks we might have
//The third dimension is the number of different kinds of sinks we have
//The fourth dimension is the number of different sink banks we might have
//The fifth is the col separation between the source and the sink
//The sixth is the row separation between the source and the sink
void vprFPGA::initializeTimingTable(){
  int maxBank;

  //Go through and figure out what the maximum delay block bank is.
  //maxBank = 0;
//   for(vector<vector <delayBlock *> >::iterator outerIter = delayBlockArray.begin(); 
//       outerIter != delayBlockArray.end(); outerIter++){
//     for(vector <delayBlock *>::iterator innerIter = outerIter->begin(); 
// 	innerIter != outerIter->end(); innerIter++){
//       if(*innerIter){
// 	if(maxBank < (*innerIter)->getNumRegBanks()){
// 	  maxBank = (*innerIter)->getNumRegBanks();
// 	}
//       }
//     }
//   }
  //Right now, we assume that all banks have the same connectivity
  //This is fine until we introduce interconnect registers.
  maxBank = 1;

  timingTable = new vector <vector <vector <vector <vector <vector <double> > > > > >;
  
  //We have three kinds of sources: logic blocks, input pins and register banks
  timingTable->resize(3);
  ////Right now, we assume that all kinds of blocks have the same connectivity
  //timingTable->resize(1);
  for(vector <vector <vector <vector <vector <vector <double> > > > > >::iterator iter = 
	timingTable->begin(); iter != timingTable->end(); iter++){
    //Each of these can have up to N banks
    (*iter).resize(maxBank);
    for(vector <vector <vector <vector <vector <double> > > > >::iterator iter2 =
	  (*iter).begin(); iter2 != (*iter).end(); iter2++){
      //Each of these three sources can drive one of three kinds of sinks:
      //a logic block, an output pin or a delay block
      (*iter2).resize(3);
      ////Right now, we assume that all kinds of blocks have the same connectivity
      //(*iter2).resize(1);
      for(vector <vector <vector <vector <double> > > >::iterator iter3 = 
	    (*iter2).begin(); iter3 != (*iter2).end(); iter3++){
	//Each of these can have up to N banks
	(*iter3).resize(maxBank);
	for(vector <vector <vector <double> > >::iterator iter4 = 
	      (*iter3).begin(); iter4 != (*iter3).end(); iter4++){
	  //Each of these source-sink pairs can have an col separation of
	  // between negative (num columns + 1) (input pin on right
	  // drives output pin on left
	  // and (num columns + 1) (input pin to output pin).
	  //Thus, if we make each of the third dimension vectors
	  // of size (num columns + 1) * 2 we can accomodate everyone
	  (*iter4).resize(((numCols + 1) * 2) + 1);
	  for(vector <vector <double> >::iterator iter5 = 
		(*iter4).begin(); iter5 != (*iter4).end(); iter5++){
	    //Initialize these values to be really big
	    (*iter5).resize(((numRows + 1) * 2) + 1, 1.0E20);
	  }
	}
      }
    }
  }
}

//Make enough space for all the values we'll need in the num hit table
//Basically, do exactly the same thing as above, only for the number of times we hit each type of sink at a given distance
void vprFPGA::initializeTimingNumHitTable(){
  int maxBank;

  //Go through and figure out what the maximum delay block bank is.
  //This takes out the idea that register banks have different connectivity.
  //I'll have to put it back in when I put registers into the interconnect.
  maxBank = 1;
  // maxBank = 0;
  //   for(vector<vector <delayBlock *> >::iterator outerIter = delayBlockArray.begin(); 
  //       outerIter != delayBlockArray.end(); outerIter++){
  //     for(vector <delayBlock *>::iterator innerIter = outerIter->begin(); 
  // 	innerIter != outerIter->end(); innerIter++){
  //       if(*innerIter){
  // 	if(maxBank < (*innerIter)->getNumRegBanks()){
  // 	  maxBank = (*innerIter)->getNumRegBanks();
  // 	}
  //       }
  //     }
  //   }

  timingNumHitTable = new vector <vector <vector <vector <vector <vector <int> > > > > >;
  
  //We have three kinds of sources: logic blocks, input pins and delay blocks
  timingNumHitTable->resize(3);
  ////Right now, we assume that all kinds of blocks have the same connectivity
  //timingNumHitTable->resize(1);

  for(vector <vector <vector <vector <vector <vector <int> > > > > >::iterator iter = 
	timingNumHitTable->begin(); iter != timingNumHitTable->end(); iter++){
    //Each of these can have up to N banks
    (*iter).resize(maxBank);
    for(vector <vector <vector <vector <vector <int> > > > >::iterator iter2 =
	  (*iter).begin(); iter2 != (*iter).end(); iter2++){
      //Each of these three sources can drive one of three kinds of sinks:
      //a logic block, an output pin or a delay block
      (*iter2).resize(3);
      ////Right now, we assume that all kinds of blocks have the same connectivity
      //(*iter2).resize(1);
      for(vector <vector <vector <vector <int> > > >::iterator iter3 = 
	    (*iter2).begin(); iter3 != (*iter2).end(); iter3++){
	//Each of these can have up to N banks
	(*iter3).resize(maxBank);
	for(vector <vector <vector <int> > >::iterator iter4 = 
	      (*iter3).begin(); iter4 != (*iter3).end(); iter4++){
	  //Each of these source-sink pairs can have an col separation of
	  // between negative (num columns + 1) (input pin on right
	  // drives output pin on left
	  // and (num columns + 1) (input pin to output pin).
	  //Thus, if we make each of the third dimension vectors
	  // of size (num columns + 1 * 2 we can accomodate everyone
	  (*iter4).resize(((numCols + 1) * 2) + 1);
	  for(vector <vector <int> >::iterator iter5 = 
		(*iter4).begin(); iter5 != (*iter4).end(); iter5++){
	    //Initialize these values to be 0 since we haven't found any sinks yet
	    (*iter5).resize(((numRows + 1) * 2) + 1, 0);
	  }
	}
      }
    }
  }
}

//Calculate the average timing for each source/dest type/bank
void vprFPGA::calculateTimingTable(){
  int A, B, C, D, E, F;
  vector <vector <vector <vector <vector <vector <int> > > > > >::iterator iter;
  vector <vector <vector <vector <vector <int> > > > >::iterator iter2;
  vector <vector <vector <vector <int> > > >::iterator iter3;
  vector <vector <vector <int> > >::iterator iter4;
  vector <vector <int> >::iterator iter5;
  vector <int>::iterator iter6;
  
  for(iter = timingNumHitTable->begin(), A = 0; iter != timingNumHitTable->end(); iter++, A++){
    for(iter2 = (*iter).begin(), B = 0; iter2 != (*iter).end(); iter2++, B++){
      for(iter3 = (*iter2).begin(), C = 0; iter3 != (*iter2).end(); iter3++, C++){
	for(iter4 = (*iter3).begin(), D = 0; iter4 != (*iter3).end(); iter4++, D++){
	  for(iter5 = (*iter4).begin(), E = 0; iter5 != (*iter4).end(); iter5++, E++){
	    for(iter6 = (*iter5).begin(), F = 0; iter6 != (*iter5).end(); iter6++, F++){
	      if(*iter6 > 0){
		timingTable->at(A).at(B).at(C).at(D).at(E).at(F) /= (double)(*iter6);
		//cout << "Timing Table[" << A << "][" << B << "][" << C <<"]["  << D;
		//cout << "]["<< E << "][" << F << "] = " << timingTable->at(A).at(B).at(C).at(D).at(E).at(F) << endl;
	      }
	    }
	  }
	}
      }
    }
  }
}

void vprFPGA::emptyTimingTable(){
  delete timingTable;
}

void vprFPGA::emptyTimingNumHitTable(){
  delete timingNumHitTable;
}

//Start a to-all search from some representational IO pins
void vprFPGA::sweepTimingIOPins(){
//   //Start from all ipins
//   for(int r = 1; r <= get_numRows(); r++){
//     //Left side
//     findAllSinksFrom((void*) getIOBAt(r, 0), IOPIN, 0, r, 0);
//     //Right side
//     findAllSinksFrom((void*) getIOBAt(r, get_numCols()+ 1), IOPIN, 0, r, get_numCols()+ 1);
//   }
//   for(int c = 1; c <= get_numCols(); c++){
//     //Top
//     findAllSinksFrom((void*) getIOBAt(0, c), IOPIN, 0, 0, c);
//     //Bottom
//     findAllSinksFrom((void*) getIOBAt(get_numRows()+ 1, c), IOPIN, 0, get_numRows()+ 1, c);
//   }
  //We should launch a search from each of the 8 corners, and something from the middle of each side
  //Top left corner
  findAllSinksFrom((void*) getIOBAt(0, 1), IOPIN, 0, 0, 1); //TOP
  findAllSinksFrom((void*) getIOBAt(1, 0), IOPIN, 0, 1, 0); //LEFT

  //Top right corner
  findAllSinksFrom((void*) getIOBAt(0, get_numCols()), IOPIN, 0, 0, get_numCols()); //TOP
  findAllSinksFrom((void*) getIOBAt(1, get_numCols()+ 1), IOPIN, 0, 1, get_numCols()+ 1); //RIGHT

  //Bottom left corner
  findAllSinksFrom((void*) getIOBAt(get_numRows()+ 1, 1), IOPIN, 0, get_numRows()+ 1, 1); //BOTTOM
  findAllSinksFrom((void*) getIOBAt(get_numRows(), 0), IOPIN, 0, get_numRows(), 0); //LEFT
  
  //Bottom right corner
  findAllSinksFrom((void*) getIOBAt(get_numRows()+ 1, get_numCols()), IOPIN, 0, get_numRows()+ 1, get_numCols()); //BOTTOM
  findAllSinksFrom((void*) getIOBAt(get_numRows(), get_numCols()+ 1), IOPIN, 0, get_numRows(), get_numCols()+ 1); //RIGHT

  //Middle of left side
  findAllSinksFrom((void*) getIOBAt((get_numRows()+ 1) / 2, 0), IOPIN, 0, (get_numRows()+ 1) / 2, 0);
		   
  //Middle of right side
  findAllSinksFrom((void*) getIOBAt((get_numRows()+ 1) / 2, get_numCols()+ 1), IOPIN, 0, (get_numRows()+ 1) / 2, get_numCols()+ 1);

  //Middle of top
  findAllSinksFrom((void*) getIOBAt(0, (get_numCols()+ 1) / 2), IOPIN, 0, 0, (get_numCols()+ 1) / 2);

  //Middle of bottom
  findAllSinksFrom((void*) getIOBAt(get_numRows()+ 1, (get_numCols()+ 1) / 2), IOPIN, 0, get_numRows()+ 1, (get_numCols()+ 1) / 2);
}

//Start a to-all search from some representational logic blocks
void vprFPGA::sweepTimingLogic(){
  //   for(int r = 1; r <= get_numRows(); r++){
  //     for(int c = 1; c <= get_numCols(); c++){
  //       findAllSinksFrom((void*) (getCLBAt(r,c)->getLUT(0)), LB, 0, r, c);
  //     }
  //   }
  //
  //Launch a search across the diagonals of the chip - notice we're assuming it's a square here!
  //Since we are missing IO pins in the corners, we should also perform searches from the
  // 2 strips to either side of the diagonal too.
  if(get_numCols() != get_numRows()){
    ERROR("We can only currently handle timing estimation for square chips!");
  }
  //Top left to bottom right diagonal
  for(int r = 1; r <= get_numRows(); r++){
    findAllSinksFrom((void*) (getCLBAt(r, r)->getLUT(0)), LB, 0, r, r);
  }
  for(int r = 1; r < get_numRows(); r++){
    //Strip just above it
    findAllSinksFrom((void*) (getCLBAt(r, r+1)->getLUT(0)), LB, 0, r, r+1);
    //Strip just below it
    findAllSinksFrom((void*) (getCLBAt(r+1, r)->getLUT(0)), LB, 0, r+1, r);
  }

  //Bottom left to top right diagonal
  for(int r = 1; r <= get_numRows(); r++){
    findAllSinksFrom((void*) (getCLBAt(get_numRows() - r + 1, r)->getLUT(0)), LB, 0, get_numRows() - r + 1, r);
  }
  for(int r = 1; r < get_numRows(); r++){
    //Strip just above it
    findAllSinksFrom((void*) (getCLBAt(get_numRows() - r, r)->getLUT(0)), LB, 0, get_numRows() - r, r);
    //Strip just below it
    findAllSinksFrom((void*) (getCLBAt(get_numRows() - r + 1, r + 1)->getLUT(0)), LB, 0, get_numRows() - r + 1, r + 1);
  }
}

//Start a to-all search from each register bank
void vprFPGA::sweepTimingDelay(){
//   for(int r = 1; r <= get_numRows(); r++){
//     for(int c = 1; r <= get_numCols(); c++){
//       //This only launches a search from the first bank in a delay block.
//       //This is fine until we add interconnect registers in.
//       findAllSinksFrom((void*) (getCLBAt(r,c)->getRegBank(0)), REGISTERBANK, 0, r, c);
//     }
//   }
  //Launch a search across the diagonals of the chip - notice we're assuming it's a square here!
  //Since we are missing IO pins in the corners, we should also perform searches from the
  // 2 strips to either side of the diagonal too.
  if(get_numCols() != get_numRows()){
    ERROR("We can only currently handle timing estimation for square chips!");
  }
  //Top left to bottom right diagonal
  for(int r = 1; r <= get_numRows(); r++){
    findAllSinksFrom((void*) (getCLBAt(r, r)->getRegBank(0)), REGISTERBANK, 0, r, r);
  }
  for(int r = 1; r < get_numRows(); r++){
    //Strip just above it
    findAllSinksFrom((void*) (getCLBAt(r, r+1)->getRegBank(0)), REGISTERBANK, 0, r, r+1);
    //Strip just below it
    findAllSinksFrom((void*) (getCLBAt(r+1, r)->getRegBank(0)), REGISTERBANK, 0, r+1, r);
  }

  //Bottom left to top right diagonal
  for(int r = 1; r <= get_numRows(); r++){
    findAllSinksFrom((void*) (getCLBAt(get_numRows() - r + 1, r)->getRegBank(0)), REGISTERBANK, 0, get_numRows() - r + 1, r);
  }
  for(int r = 1; r < get_numRows(); r++){
    //Strip just above it
    findAllSinksFrom((void*) (getCLBAt(get_numRows() - r, r)->getRegBank(0)), REGISTERBANK, 0, get_numRows() - r, r);
    //Strip just below it
    findAllSinksFrom((void*) (getCLBAt(get_numRows() - r + 1, r + 1)->getRegBank(0)), REGISTERBANK, 0, get_numRows() - r + 1, r + 1);
  }
}

//Perform a to-all search from the source of this block.
void vprFPGA::findAllSinksFrom(void* sourceBlock, e_timingMoveBlockType sourceType, int sourceBank, int sourceRow, int sourceCol){
  priority_queue<pqNode> PQ;
  routingNode *source;
  routingNode *rnode;
  routingNode* nb_rnode;
  routingEdge* nb_edge;
  double pathDelay;
  double newPathDelay;
  list<routingNode*> cleanupList;

#ifdef PARANOID_CHECK_ROUTE
  //Go through the entire routing graph and make sure that all nodes are unexplored before we start
  for(list<routingNode*>::iterator rNodeIter = FPGA->rnodeList.begin(); rNodeIter != FPGA->rnodeList.end(); rNodeIter++){
    if((*rNodeIter)->isExplored == true){
      ERROR("Found explored routing node before routing has begun!");
    }
  }
#endif

  if(sourceType == IOPIN){
    //Pick one of the source routing nodes to be the source for this entire search
    if(((vprIOB*)sourceBlock)->sourceArray->size() == 0){
      ERROR("Input pin has no source!");
    }
    source = ((vprIOB*)sourceBlock)->sourceArray->front();
  }
  else if(sourceType == LB){
    //Pick one of the source routing nodes to be the source for this entire search
    if(!((vprLUT*)sourceBlock)->sourceNode){
      ERROR("Block has no source!");
    }
    source = ((vprLUT*)sourceBlock)->sourceNode;
  }
  else if(sourceType == REGISTERBANK){
    //Pick the first routing node in the bank to be the source for this search
    source = ((registerBank*)sourceBlock)->getFirstRoutingNode();
  }

  //Push the source into the PQ at a cost of 1.0 * intrinsic delay of this node
  //We are trying to find the shortest timing distance between the source and all other nodes so 
  // we will be putting everything in at a cost of just the delay
  //PQ.push(pqNode(source, source->intrinsicDelay, 0.0, 0, source->intrinsicDelay));
  
  //Actually, don't push the source in to the PQ, push the neighbors of the source into the PQ.
  //Just in case the source is also the sink, we want to go through at least 1 edge in the
  // RR graph.
  source->create_fanoutListItr();
  for(nb_edge = source->getnext_fanoutList(); nb_edge != 0; nb_edge = source->getnext_fanoutList()){
    nb_rnode = nb_edge->destNode;
    //Only add unexplored neighbors to the PQ
    if(nb_rnode->isExplored == false){
      newPathDelay = source->intrinsicDelay + nb_rnode->intrinsicDelay + nb_edge->intrinsicDelay;
      PQ.push(pqNode(nb_rnode, newPathDelay, 0.0 ,rnode, newPathDelay));
    }
  }
  source->remove_fanoutListItr();

  //Keep exploring till the PQ is empty
  while(!PQ.empty()){
    rnode=(PQ.top()).rNode;
    
    //We will only pay attention to a routing node the first time we see it
    if(rnode->isExplored == true){
    }
    else{
      //Mark the node explored
      rnode->isExplored=true;
      cleanupList.push_back(rnode);

      pathDelay = (PQ.top()).netDelay;

      //See if this routing node belongs to something interesting (logic block sink, output pin sink, register bank terminal)
      if(rnode->type == SINK){
	//Figure out if this is an I/O pin or a logic block
	if(rnode->col_low == 0 || rnode->col_low == (FPGA->numCols + 1) || 
	   rnode->row_low == 0 || rnode->row_low == (FPGA->numRows+1)){
	  //cout << "Source at " << source->col_low << ", " << source->row_low << " found IO pin at " << rnode->col_low << ", " << rnode->row_low << endl;
	  addTimingTable(pathDelay, sourceType, sourceBank, IOPIN, 0, rnode->row_low - sourceRow, rnode->col_low - sourceCol);
	  incNumHitTable(sourceType, sourceBank, IOPIN, 0, rnode->row_low - sourceRow, rnode->col_low - sourceCol);

	}
	else{
	  addTimingTable(pathDelay, sourceType, sourceBank, LB, 0, rnode->row_low - sourceRow, rnode->col_low - sourceCol);
	  incNumHitTable(sourceType, sourceBank, LB, 0, rnode->row_low - sourceRow, rnode->col_low - sourceCol);
	}
      }
      else if(rnode->type == DELAY){
	//Found a delay node that is not the source
	addTimingTable(pathDelay, sourceType, sourceBank, REGISTERBANK, rnode->getBank(), 
		       rnode->row_low - sourceRow, rnode->col_low - sourceCol);
	incNumHitTable(sourceType, sourceBank, REGISTERBANK, rnode->getBank(), 
		       rnode->row_low - sourceRow, rnode->col_low - sourceCol);

      }
      //If it is not a destination, add it's neighbors to the PQ
      else{
	rnode->create_fanoutListItr();
	for(nb_edge=rnode->getnext_fanoutList(); nb_edge!=0; nb_edge=rnode->getnext_fanoutList()){
	  nb_rnode=nb_edge->destNode;
	  //Only add unexplored neighbors to the PQ
	  if(nb_rnode->isExplored == false){
	    newPathDelay = pathDelay + nb_rnode->intrinsicDelay + nb_edge->intrinsicDelay;
	    PQ.push(pqNode(nb_rnode, newPathDelay, 0.0 ,rnode, newPathDelay));
	  }
	}
	rnode->remove_fanoutListItr();
      }
    }
    PQ.pop();  
  }
  
  //Reset all of the nodes we hit to unexplored
  while(!cleanupList.empty()){
    cleanupList.front()->isExplored=false;
    cleanupList.pop_front();
  }
}



//Set a particular location in the table
//Notice that for the x separation, we can have values
//ranging from negative (num columns + 1) to positive (num columns + 1)
//so we have an offset built in.
void vprFPGA::addTimingTable(double value, e_timingMoveBlockType sourceType, int sourceBank, e_timingMoveBlockType destType, int destBank, int rowSep, int colSep){
//   //Basically this is an uninitialized place, so replace the really big value with the value we want to add in
//   if(timingTable->at(sourceType).at(sourceBank).at(destType).at(destBank).at(colSep + numCols + 1).at(rowSep + numRows + 1) > 1.0E19){
//     timingTable->at(sourceType).at(sourceBank).at(destType).at(destBank).at(colSep + numCols + 1).at(rowSep + numRows + 1) = value;
//   }
//   else
//     timingTable->at(sourceType).at(sourceBank).at(destType).at(destBank).at(colSep + numCols + 1).at(rowSep + numRows + 1) += value;

  //These changes assume that all banks have the same connectivity
  if(timingTable->at(sourceType).at(0).at(destType).at(0).at(colSep + numCols + 1).at(rowSep + numRows + 1) > 1.0E19){
    timingTable->at(sourceType).at(0).at(destType).at(0).at(colSep + numCols + 1).at(rowSep + numRows + 1) = value;
  }
  else
    timingTable->at(sourceType).at(0).at(destType).at(0).at(colSep + numCols + 1).at(rowSep + numRows + 1) += value;

  //These changes assume that A) All types of blocks have the same connectivity (IOB, LB, REG) and B) All
  // registers at different banks have the same connectivity.
  //This is fine until we introducet interconnect registers
//   if(timingTable->at(0).at(0).at(0).at(0).at(colSep + numCols + 1).at(rowSep + numRows + 1) > 1.0E19){
//     timingTable->at(0).at(0).at(0).at(0).at(colSep + numCols + 1).at(rowSep + numRows + 1) = value;
//   }
//   else
//     timingTable->at(0).at(0).at(0).at(0).at(colSep + numCols + 1).at(rowSep + numRows + 1) += value;
}

void vprFPGA::incNumHitTable(e_timingMoveBlockType sourceType, int sourceBank, e_timingMoveBlockType destType, int destBank, int rowSep, int colSep){
  //timingNumHitTable->at(sourceType).at(sourceBank).at(destType).at(destBank).at(colSep + numCols + 1).at(rowSep + numRows + 1) ++;
  //These changes assume that all banks have the same connectivity
  timingNumHitTable->at(sourceType).at(0).at(destType).at(0).at(colSep + numCols + 1).at(rowSep + numRows + 1) ++;
  //These changes assume that A) All types of blocks have the same connectivity (IOB, LB, REG) and B) All
  // registers at different banks have the same connectivity.
  //This is fine until we introducet interconnect registers
  //timingNumHitTable->at(0).at(0).at(0).at(0).at(colSep + numCols + 1).at(rowSep + numRows + 1) ++;
}


void vprFPGA::setTimingTable(double value, e_timingMoveBlockType sourceType, int sourceBank, e_timingMoveBlockType destType, int destBank, int rowSep, int colSep){
  cout << "Timing Table[" << sourceType << "][" << sourceBank << "][" << destType <<"]["  << destBank;
  cout << "]["<< colSep << "][" << rowSep << "] = " << value << endl;
  
  //timingTable->at(sourceType).at(sourceBank).at(destType).at(destBank).at(colSep + numCols + 1).at(rowSep + numRows + 1) = value;
  //These changes assume that all banks have the same connectivity
  timingTable->at(sourceType).at(0).at(destType).at(0).at(colSep + numCols + 1).at(rowSep + numRows + 1) = value;

  //These changes assume that A) All types of blocks have the same connectivity (IOB, LB, REG) and B) All
  // registers at different banks have the same connectivity.
  //This is fine until we introduce interconnect registers
  //timingTable->at(0).at(0).at(0).at(0).at(colSep + numCols + 1).at(rowSep + numRows + 1) = value;
}

//Try to get the CLB at the given location.
//If there isn't one there, create one.
vprCLB* vprFPGA::getOrMakeCLBAt(int row, int col){
  //Make sure we are asking for a CLB location
  if(row < 1 || row > numRows || col < 1 || col > numCols){
    ERROR("Requested invalid CLB location");
  }

  // We might not have that many rows
  if(array.size() <= row){
    array.resize(row + 1);
    array.at(row).resize(col + 1);
    array.at(row).at(col) = (void*) new vprCLB(row, col);
  }
  //Or, we might have a row that doesn't have enough columns
  else if(array.at(row).size() <= col){
    array.at(row).resize(col + 1);
    array.at(row).at(col) = (void *) new vprCLB(row, col);
  }
  //Even if we have enough rows and columns, we might have skipped one in between
  // so we have to fill it in.
  if(!array.at(row).at(col)){
    array.at(row).at(col) = (void *) new vprCLB(row, col);
  }

  return ((vprCLB*) array.at(row).at(col));
}

//Try to get the clb at the given location.
//There might not be one at that location though.
vprCLB* vprFPGA::getCLBAt(int row, int col){
  //Make sure we are asking for a CLB location
  if(row < 1 || row > numRows || col < 1 || col > numCols){
    ERROR("Requested invalid CLB location");
  }
  // We might not have that many rows
  if(array.size() <= row){
    return NULL;
  }
  //Or, we might have a row that doesn't have enough columns
  else if(array.at(row).size() <= col){
    return NULL;
  }
  else{
    return ((vprCLB*) array.at(row).at(col));
  }
}

//Try to get the IOB at the given location.
//If there isn't one there, create one.
vprIOB* vprFPGA::getOrMakeIOBAt(int row, int col){
  short int a = 0;
  short int b = 0;
  short int c = 0;
  short int d = 0;

  //Make sure we are asking for a valid IOB location
  //Basically this is an XOR of the possible combinations
  if(row == 0){
    a = 1;
  }
  else if(row == numRows + 1){
    b = 1;
  }
  if(col == 0){
    c = 1;
  }
  else if(col == numCols + 1){
    d = 1;
  }

  if((a ^ b ^ c ^ d) == 0){
    ERROR("Requested invalid IOB location");
  }

  // We might not have that many rows
  if(array.size() <= row){
    array.resize(row + 1);
    array.at(row).resize(col + 1);
    array.at(row).at(col) = (void*) new vprIOB(row, col, numIOs_perIOB);
  }
  //Or, we might have a row that doesn't have enough columns
  else if(array.at(row).size() <= col){
    array.at(row).resize(col + 1);
    array.at(row).at(col) = (void *) new vprIOB(row, col, numIOs_perIOB);
  }

  return ((vprIOB*) array.at(row).at(col));
}

//Try to get the IOB at the given location.
//There might not be one at that location though.
vprIOB* vprFPGA::getIOBAt(int row, int col){
  short int a = 0;
  short int b = 0;
  short int c = 0;
  short int d = 0;

  //Make sure we are asking for a valid IOB location
  //Basically this is an XOR of the possible combinations
  if(row == 0){
    a = 1;
  }
  else if(row == numRows + 1){
    b = 1;
  }
  if(col == 0){
    c = 1;
  }
  else if(col == numCols + 1){
    d = 1;
  }

  if((a ^ b ^ c ^ d) == 0){
    ERROR("Requested invalid IOB location");
  }

  // We might not have that many rows
  if(array.size() <= row){
    return NULL;
  }
  //Or, we might have a row that doesn't have enough columns
  else if(array.at(row).size() <= col){
    return NULL;
  }
  else{
    return ((vprIOB*) array.at(row).at(col));
  }
}

int vprFPGA::getNumberofBlockType(int blockType){
  switch(blockType){
  case LUT:
    return lutList.size();
  case MULTIPLIER:
    return multList.size();
  case RAM_BLOCK:
    return ramList.size();
  case INPUTPAD:
    ERROR("We shouldn't see an input block here!");
    break;
  case OUTPUTPAD:
    ERROR("We shouldn't see an output block here!");
    break;
  case REGISTER_BANK:
    return registerBankVector.size();
  }
}

registerBank* vprFPGA::getRandomFreeSingleRegBank(){
  list<registerBank *>::iterator iter;
  int i;
  bool singleRegisterAvailable = false;
  int select;

  if(freeRegisterBankList.size() == 0){
    cout << "Ran out of single registers!" <<endl;
    ERROR("Ran out of single registers!");
  }
  else{
    //Figure out if there are any free single register banks available
    for(iter = freeRegisterBankList.begin(); iter != freeRegisterBankList.end(); iter++){
      if((*iter)->getNumPossibleDelay() == 1){
	singleRegisterAvailable = true;
	break;
      }
    }

    if(singleRegisterAvailable){
      while(1){
	select = lrand48() % freeRegisterBankList.size();
	for(iter = freeRegisterBankList.begin(), i = 0; 
	    iter != freeRegisterBankList.end() && i < select; iter++, i ++){
	}
	if((*iter)->getNumPossibleDelay() == 1){
	  return (*iter);
	}
      }
    }
    else{
      // We don't have any single registers available, so let's try a shift register bank
      select = lrand48() % freeRegisterBankList.size();
      for(iter = freeRegisterBankList.begin(), i = 0; 
	  iter != freeRegisterBankList.end() && i < select; iter++, i ++){
      }
      return (*iter);
    }
  }
}

registerBank* vprFPGA::getRandomFreeShiftRegBank(int capacity){
  list<registerBank *>::iterator iter;
  int i;
  bool shiftBankAvailable = false;
  int select;

  if(freeRegisterBankList.size() == 0){
    cout << "Ran out of shift register banks!" << endl;
    ERROR("Ran out of shift register banks!");
  }
  else{
    //Figure out if there are any free shift register banks available
    for(iter = freeRegisterBankList.begin(); iter != freeRegisterBankList.end(); iter++){
      if((*iter)->getNumPossibleDelay() >= capacity){
	shiftBankAvailable = true;
	break;
      }
    }
    if(shiftBankAvailable){
      while(1){
	select = lrand48() % freeRegisterBankList.size();
	for(iter = freeRegisterBankList.begin(), i = 0; 
	    iter != freeRegisterBankList.end() && i < select; iter++, i ++){
	}
	if((*iter)->getNumPossibleDelay() >= capacity){
	  return (*iter);
	}
      }
    }
    else{
      cout << "Ran out of shift register banks!" <<endl;
      ERROR("Ran out of shift register banks!");
    }
  }
}

//Go through all of the free register banks in the system and return the closest one to the
// given coordinates.
registerBank* vprFPGA::getClosestFreeRegBank(int row, int col){
  list<registerBank *>::iterator iter;
  registerBank* closestBank;

  if(freeRegisterBankList.size() == 0){
    return NULL;
  }
  else{
    closestBank = freeRegisterBankList.front();
    for(iter = freeRegisterBankList.begin(); 
	iter != freeRegisterBankList.end(); iter++){
      if(abs(closestBank->hostCLB->row - row) + abs(closestBank->hostCLB->col - col) >
	 abs((*iter)->hostCLB->row - row) + abs((*iter)->hostCLB->col - col)){
	closestBank = *iter;
	if(abs(closestBank->hostCLB->row - row) + abs(closestBank->hostCLB->col - col) == 0){
	  return closestBank;
	}
      }
    }
  }
  return closestBank;
}

vprIOB* vprFPGA::getOtherIOB(int row, int col, int rLim){
  int rowRLim = min(rLim, numRows);
  int colRLim = min(rLim, numCols);
  int side;
  int newRow = row;
  int newCol = col;
  int moveAmt;

  //See if the rLim is still the whole chip
  if(numRows <= rowRLim && numCols <= colRLim){
    while(newRow == row && newCol == col){
      //From VPR:
      /*                              *
       *       +-----1----+           *
       *       |          |           *
       *       |          |           *
       *       0          2           *
       *       |          |           *
       *       |          |           *
       *       +-----3----+           *
       *                              */
      side = lrand48() % 4;
      switch(side){
      case 0:
	newCol = 0;
	newRow = (lrand48() % numRows) + 1;
	break;
      case 1:
	newCol = (lrand48() % numCols) + 1;
	newRow = 0;
	break;
      case 2:
	newCol = numCols + 1;
	newRow = (lrand48() % numRows) + 1;
	break;
      case 3:
	newCol = (lrand48() % numCols) + 1;
	newRow = numRows + 1;
	break;
      }
    }
  }
  else{
    //The idea of windowing, according to VPR is a bit different for
    // IO pins.  If we generate a positive/negative move amount
    // and moving would leave the side of the chip we are
    // currently on, we generate a second move amount value
    // and fix the correct row or column.  I think this is done
    // to provide a smoother transistion between an rLim of
    // numRow and numRow-1.
    moveAmt = 0;

    if(col == 0 || col == numCols + 1){
      //Coming from the left side or right side.
      //Try to find a new row, keep the same column.
      while (moveAmt == 0){
	moveAmt = (lrand48() % ((2 * rowRLim) + 1)) - rowRLim;
      }
      newRow += moveAmt;

      //We might have tried to move up or down too far and fell off of
      // the left/right sides to the top or bottom sides.
      if(newRow < 1 || newRow > numRows){
	//Did we want to move to top to bottom sides?
	if(newRow < 1){
	  newRow = 0;
	}

	else{
	  newRow = numRows + 1;
	}

	//Find a new column.
	if(col == 0){
	  //If we came from left side, the new column should be (1, colRLim)
	  newCol = (lrand48() % colRLim) + 1;
	}
	else{
	  //If we came from the right side, the new column should be (numCols - colRLim + 1, numCols)
	  newCol = numCols - (lrand48()% colRLim);
	}
      }
    }
    else if(row == 0 || row == numRows + 1){
      //Coming from the top or bottom.
      //Try to find a new column, keep the same row
      while (moveAmt == 0){
	moveAmt = (lrand48() % ((2 * colRLim) + 1)) - colRLim;
      }
      newCol += moveAmt;
      
      //We might have tried to move left or right too far and fell off of
      // the top/bottom to the left or right sides.
      if(newCol < 1 || newCol > numCols){
	//Did we want to move to the left or right side?
	if(newCol < 1){
	  newCol = 0;
	}
	else{
	  newCol = numCols + 1;
	}

	//Find a new row
	if(row == 0){
	  //If we came from the top, the new row should be (1, rowRLim)
	  newRow = (lrand48() % rowRLim) + 1;
	}
	else{
	  //If we came from the bottom, the new row should be (numRows - rowRLim + 1, numRows)
	  newRow = numRows - (lrand48()% rowRLim);
	}
      }
    }
  }
  return(getIOBAt(newRow, newCol));
}

vprLUT* vprFPGA::getOtherLUT(int row, int col, int rLim){
  int moveLimitRow = min(numRows, rLim);
  int moveLimitCol = min(numCols, rLim);
  int newRow = row;
  int newCol = col;
  int moveAmtRow = 0;
  int moveAmtCol = 0;
  vprLUT* lut = 0;
  
  while(!lut || (newRow == row && newCol == col)){
    moveAmtRow = (lrand48() % ((2 * moveLimitRow) + 1)) - moveLimitRow;
    moveAmtCol = (lrand48() % ((2 * moveLimitCol) + 1)) - moveLimitCol;

    newRow = row + moveAmtRow;
    newCol = col + moveAmtCol;
    
    //VPR does a wrap around, as commented out, but that doesn't seem right.
    //I'm doing mirroring instead.
    if(newRow < 1){
      //newRow = newRow + numRows;
      newRow = (newRow * -1) + 1;
    }
    else if(newRow > numRows){
      //newRow = newRow - numRows;
      newRow = numRows - (newRow - (numRows + 1));
    }
    if(newCol < 1){
      //newCol = newCol + numCols;
      newCol = (newCol * -1) + 1;
    }
    else if(newCol > numCols){
      //newCol = newCol - numCols;
      newCol = numCols - (newCol - (numCols + 1));
    }
    lut = getCLBAt(newRow, newCol)->getLUT(lrand48() % getCLBAt(newRow, newCol)->getNumLUTs());

    if(lut->get_blockType() != LUT){
      lut = 0;
    }
  }
  return(lut);
}

//Since there is likely large gaps in the arrangements of multipliers, I've got to think of something
// to make windowing work - otherwise we'll just do un-windowed.
//vprLUT* vprFPGA::getOtherMult(int row, int col, int rLim);
//vprLUT* vprFPGA::getOtherRam(int row, int col, int rLim);
registerBank * vprFPGA::getOtherRegBank(int row, int col, int rLim){
  int moveLimitRow = min(numRows, rLim);
  int moveLimitCol = min(numCols, rLim);
  int newRow = row;
  int newCol = col;
  int moveAmtRow;
  int moveAmtCol;
  
  while(newRow == row && newCol == col){
    moveAmtRow = (lrand48() % ((2 * moveLimitRow) + 1)) - moveLimitRow;
    moveAmtCol = (lrand48() % ((2 * moveLimitCol) + 1)) - moveLimitCol;

    newRow = row + moveAmtRow;
    newCol = col + moveAmtCol;
    
    //VPR does a wrap around, as commented out, but that doesn't seem right
    if(newRow < 1){
      newRow = (newRow * -1) + 1;
    }
    else if(newRow > numRows){
      newRow = numRows - (newRow - (numRows + 1));
    }
    if(newCol < 1){
      newCol = (newCol * -1) + 1;
    }
    else if(newCol > numCols){
      newCol = numCols - (newCol - (numCols + 1));
    }
  }
  return(getCLBAt(newRow, newCol)->getRegBank(lrand48() % getCLBAt(newRow, newCol)->getNumRegBanks()));
}

//Make sure that the other register bank that we get is of a certain minimum size and
// has a current occupancy less than X
registerBank * vprFPGA::getOtherRegBank(int row, int col, int rLim, int minSize, int maxCurrOccupancy){
  registerBank *tempRegBank;
  int moveLimitRow = min(numRows, rLim);
  int moveLimitCol = min(numCols, rLim);
  int newRow = row;
  int newCol = col;
  int moveAmtRow;
  int moveAmtCol;
  
  while(newRow == row && newCol == col){
    moveAmtRow = (lrand48() % ((2 * moveLimitRow) + 1)) - moveLimitRow;
    moveAmtCol = (lrand48() % ((2 * moveLimitCol) + 1)) - moveLimitCol;

    newRow = row + moveAmtRow;
    newCol = col + moveAmtCol;
    
    //VPR does a wrap around, as commented out, but that doesn't seem right
    if(newRow < 1){
      newRow = (newRow * -1) + 1;
    }
    else if(newRow > numRows){
      newRow = numRows - (newRow - (numRows + 1));
    }
    if(newCol < 1){
      newCol = (newCol * -1) + 1;
    }
    else if(newCol > numCols){
      newCol = numCols - (newCol - (numCols + 1));
    }
  }

  while(1){
    tempRegBank = getCLBAt(newRow, newCol)->getRegBank(lrand48() % getCLBAt(newRow, newCol)->getNumRegBanks());
    if(tempRegBank->getNumPossibleDelay() >= minSize && tempRegBank->getNumMappedRegs() <= maxCurrOccupancy){
      return (tempRegBank);
    }
  }
}

void vprFPGA::occupyRegBank(registerBank* rb){
  freeRegisterBankList.remove(rb);
  numFreeRegisters -= rb->getNumPossibleDelay();
}

void vprFPGA::vacateRegBank(registerBank* rb){
  freeRegisterBankList.push_back(rb);
  numFreeRegisters += rb->getNumPossibleDelay();
}

//Go though all of the register banks and make sure that the
// empty (and only empty ones) are in the free register list
void vprFPGA::checkFreeRegisters(){
  int count = 0;
  vector<registerBank *>::iterator vIter;
  list<registerBank *>::iterator lIter;
  list<registerBank *>::iterator lIter2;
  
  for(vIter = registerBankVector.begin(); vIter != registerBankVector.end(); vIter++){
    if((*vIter)->isJustBankEmpty()){
      //We should have this register in the free register bank list
      for(lIter = freeRegisterBankList.begin(); lIter != freeRegisterBankList.end(); lIter++){
	if(*lIter == *vIter){
	  break;
	}
      }
      if(lIter == freeRegisterBankList.end()){
	ERROR("Free register mismatch!");
      }
    }
    else{
      //We should not have this register in the free register bank list
      for(lIter = freeRegisterBankList.begin(); lIter != freeRegisterBankList.end(); lIter++){
	if(*lIter == *vIter){
	  break;
	}
      }
      if(lIter != freeRegisterBankList.end()){
	ERROR("Free register mismatch!");
      }
    }
  }

  //Look for duplicate registers in the free register list
  for(lIter = freeRegisterBankList.begin(); lIter != freeRegisterBankList.end(); lIter++){
    //Find the first instance of this register in the list
    for(lIter2 = freeRegisterBankList.begin(); lIter2 != freeRegisterBankList.end(); lIter2++){
      if(*lIter == *lIter2){
	lIter2++;
	break;
      }
    }
    //Look for more
    for( ; lIter2 != freeRegisterBankList.end(); lIter2++){
      if(*lIter == *lIter2){
	ERROR("Duplicate entry!");
      }
    }
  }

  //Make sure that the number of empty registers matches the number of registers in the free list
  for(lIter = freeRegisterBankList.begin(); lIter != freeRegisterBankList.end(); lIter++){
    count += (*lIter)->getNumPossibleDelay();
  }
  if(count != numFreeRegisters){
    ERROR("Free register mismatch!");
  }
}

//We need to find an empty IOB, just grab the first one
void vprFPGA::placeClockPin(int *clockRow, int *clockCol, int *clockPTC){
  vector<vprIOB *>::iterator iter;

  //Look through all the IOBs
  for(iter = iobList.begin(); iter != iobList.end(); iter++){
    //Look at each pin within the IOB
    for(int i = 0; i < (*iter)->get_numIOs(); i++){
      if(!(*iter)->getfrom_currBlockArray(i)){
	*clockRow = (*iter)->get_row();
	*clockCol = (*iter)->get_col();
	*clockPTC = i;
	return;
      }
    }
  }

  for(iter = iobList.begin(); iter != iobList.end(); iter++){
    //Print out the contents of each iob
    for(int i = 0; i < (*iter)->get_numIOs(); i++){
      if((*iter)->getfrom_currBlockArray(i)){
	printf("IOB %d, %d, %d has block %s\n", (*iter)->get_row(),
	       (*iter)->get_col(), i, (*iter)->getfrom_currBlockArray(i)->name.c_str());
      }
      else{
	printf("IOB %d, %d, %d is empty!\n", (*iter)->get_row(),
	       (*iter)->get_col(), i);
      }
    }
  }
  
  ERROR("We need an empty IOB location, but we can't find one!");
}

//Go through all iobs, register banks and LUTs to make sure that,
// if they have a mapping, that the logic block agrees
void vprFPGA::checkArrayMappings(){
  int i;
  vector<logicBlock*>::iterator lbIter;

  for(vector<vprIOB *>::iterator iter = iobList.begin(); iter != iobList.end(); iter++){
    for(lbIter = (*iter)->currBlockArray->begin(), i = 0; 
	lbIter != (*iter)->currBlockArray->end(); lbIter++, i++){
      if(*lbIter){
	if((*lbIter)->row != (*iter)->row || (*lbIter)->col != (*iter)->col || (*lbIter)->ptc_index != i){
	  ERROR("IOB placement mismatch!");
	}
      }
    }
  }

  for(vector<vprLUT *>::iterator iter = lutList.begin(); iter != lutList.end(); iter++){
    if((*iter)->currBlock){
      //We can't check the ptc index, since the physical LUT doesn't keep track of what
      // ptc index it is itself.  We would have to add that info in.
      if((*iter)->currBlock->row != (*iter)->row || (*iter)->currBlock->col != (*iter)->col)
	ERROR("LUT placement mismatch!");
    }
  }

  for(vector<registerBank *>::iterator iter = registerBankVector.begin(); iter != registerBankVector.end(); iter++){
    if(!(*iter)->isJustBankEmpty()){
      //This register bank is occupied by a logic block register bank
      if((*iter)->get_currBlock()){
	if((*iter)->get_currBlock()->row != (*iter)->getHostCLB()->row || 
	   (*iter)->get_currBlock()->col != (*iter)->getHostCLB()->col || 
	   (*iter)->get_currBlock()->ptc_index != (*iter)->getHostBankNum() || 
	   (*iter)->get_currBlock()->rb != *iter){
	  ERROR("Reg bank placement mismatch!");
	}
      }
      //This register bank is occupied by pipelining registers
      else{
	for(list <Terminal *>::iterator prIter = (*iter)->currSinkList.begin(); prIter != (*iter)->currSinkList.end(); prIter++){
	  if((*prIter)->getLogicBlockRow() != (*iter)->getHostCLB()->row || 
	     (*prIter)->getLogicBlockCol() != (*iter)->getHostCLB()->col){
	    ERROR("Reg bank placement mismatch!");
	  }
	}
      }
    }
  }
}

int vprFPGA::getNumberCLBsEmptyLUTandRegs(int requiredLUTs, int requiredRegisters){
  int r, c;
  int count = 0;

  for(r = 1; r <= numRows; r++){
    for(c = 1; c <= numCols; c++){
      if(((vprCLB*) array.at(r).at(c))->getNumEmptyRegBank(1) >= requiredRegisters &&
	 ((vprCLB*) array.at(r).at(c))->getNumEmptyLUTs() >= requiredLUTs){
	count ++;
      }
    }
  }

  return count;
}
