// Copyright (C) 2006 Ken Eguro, Scott Hauck, Akshay Sharma
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

/****************************************************************
This file contains class definitions of the VPR FPGA architecture
and the individual clbs that constitute a VPR-style FPGA. Each
vprCLB object holds information about the source/sink/input-pin/
output-pin routing nodes that it provides. Each vprCLB
owns the track segments in the channels that are north and east
of it. Further, the vprLUT also holds a pointer to the logic
block that is currently placed on it.
****************************************************************/    
#ifndef VPR_ARCH_H
#define VPR_ARCH_H

class logicBlock;
class Netlist;
class c_clb;
class vprCLB;
class Terminal;

//This represents a bank of registers that have exactly 1 input and one output
class registerBank{
 public:
  vector <routingNode *> routingNodeVector;
  list <Terminal *> currSinkList;
  list <Terminal *> currSourceList;
  vprCLB* hostCLB;
  int hostBankNum;
  logicBlock *currBlock;

  //If this register bank is actually a LUT shift register,
  // fill this value in.  Otherwise, it's just NULL.
  vprLUT* linkedLUT;

  registerBank(){
    hostBankNum = -1;
    hostCLB = 0;
    currBlock = 0;
    linkedLUT = 0;
  };
  ~registerBank(){};
  void setHostCLB(vprCLB *clb, int bankNum){hostCLB = clb; hostBankNum = bankNum;};
  vprCLB* getHostCLB(){return hostCLB;};
  int getHostBankNum(){return hostBankNum;};
  
  void addRoutingNode(routingNode *r);
  Terminal* getFirstSink();
  Terminal* getLastSource();
  routingNode* getFirstRoutingNode();
  //Is this register bank empty?
  bool isBankAndLinkedLUTEmpty();
  bool isJustBankEmpty();
  bool isLinkedLUTEmpty();
  int getNumMappedRegs();
  int getNumPossibleDelay();
  int getNumFreeRegs();
  bool bankFull();
  void placeTerminal(Terminal *T);
  void placeTerminalFront(Terminal *T);
  void placeLogicBlock(logicBlock *lb);
  void unPlaceLogicBlock(logicBlock *lb);
  void insertRegisterForwards(Terminal *T, Terminal *parent);
  void insertRegisterBackwards(Terminal *T, Terminal *parent);
  void swapBank(registerBank *otherBank);
  void swapBankAndAssociatedLUT(registerBank *otherBank, double criticalPathDelay);
  void copyContentsToAndEmpty(registerBank *otherBank);
  void updateSinkDelaysAndSlacks();
  void updateSinkDelaysAndSlacksAndAssociatedLUT();
  void updateWireAndTimingCosts(bool timing_driven, double criticalPathDelay);
  double determineDeltaWireCost();
  double determineDeltaTimingCost();
  double determineDeltaTimingCostPrime();

  void migrateFirstNRegs(registerBank *otherBank, int numRegs);
  void migrateLastNRegs(registerBank *otherBank, int numRegs);
  
  void duplicateRegs(Netlist *netlist, registerBank *otherBank, double criticalPathDelay);
  void zipRegs(Netlist *netlist, registerBank *otherBank, double criticalPathDelay);
  void reDuplicateRegs(Netlist *netlist, registerBank *otherBank, double criticalPathDelay);
  
  void removeStartReference(Terminal *T);
  void removeEndReference(Terminal *T);

  void freeBank();

  double returnTotalWireCost();
  double returnTotalTimingCost();
  void updateDeletedCutCost();
  
  void print_info();

  void updateRoutingNodes();

  logicBlock* get_currBlock(){
    return currBlock;
  }

  //Are the LUTs and other registers banks in the clb at the same location empty?
  bool emptyCLB();
  
  void swapEntireStickyMappings(registerBank* otherBank, double criticalPathDelay);
  
  void updateCostAndDeltas(bool timing_driven, double criticalPathDelay, double *deltaWire, 
			   double *deltaTiming);
  void updateCostAndDeltasCostDeltaDelay(bool timing_driven, double criticalPathDelay, double *deltaWire, 
					 double *deltaTiming, double *deltaTimingPrime);
  void updateCostAndDeltasAndAssociatedLUT(bool timing_driven, double criticalPathDelay, 
					   double *deltaWire, double *deltaTiming);
  void updateCostAndDeltasAndAssociatedLUTCostDeltaDelay(bool timing_driven, double criticalPathDelay, 
							 double *deltaWire, double *deltaTiming, double *deltaTimingPrime);
  void updateCostAndDeltasForEntireSticky(registerBank* otherRegisterBank, bool timing_driven, 
					  double criticalPathDelay, double *deltaWire, double *deltaTiming);
  void updateCostAndDeltasForEntireStickyCostDeltaDelay(registerBank* otherRegisterBank, bool timing_driven, 
							double criticalPathDelay, double *deltaWire, 
							double *deltaTiming, double *deltaTimingPrime);
    
  c_clb* returnMappedCLB();

  registerBank* getBestRegBankInParentHostCLBSwap();

  void calculateMaxInputOutputCriticalitiesJustRB(double criticalPathDelay);
  void calculateMaxInputOutputCriticalitiesAndAssociatedLUT(double criticalPathDelay);

  registerBank* getRandomDuplicate();
  bool getRandomMerge(int *moveType, int *numMove, registerBank **parentRegBank, registerBank **childRegBank);
  bool someSinksCloserTo(registerBank *otherRegBank);
  void setLUTAssociation(int lutAssociation);
  bool hasLinkedLUT(){
    if(linkedLUT)
      return true;
    else
      return false;
  }
  
  bool hasMatchingLinkedRegisters(registerBank* otherRegisterBank);

  bool lastSourceHasFanout();
};


class vprBaseBlock{
 public:
  int row,col;  //row and col coordinates of this block
  int magicNum;

  vprBaseBlock(int r,int c);
  ~vprBaseBlock();
  int get_row();  //get the row coord of this block
  int get_col();  //get the col coord of this block
  virtual void print_info();  //print details about this block
  //void print_info();  //print details about this block
};


/****************************************************************
This class represents a LUT in the VPR architecture.
****************************************************************/
class vprLUT : public vprBaseBlock{
 public:
  routingNode* sourceNode;  //source routing node
  routingNode* sinkNode;  //sink routing node
  routingNode* clock_sinkNode;  //sink routing node for clock routing...NOT USED!!
  logicBlock* currBlock;  //the logic block that is currently placed at this LUT
  enum e_genericBlockType blockType;
  int derived_magicNum;
  int index;

  //If this LUT has a shift-register mode, fill this value in.  Otherwise, it's just NULL;
  list <registerBank*> linkedRegisterBanks;

  vprLUT(int r,int c, int ptc_index);
  ~vprLUT();

  vprCLB* getHostCLB();

  void set_currBlock(logicBlock* l);  //set the logic block that is placed at this CLB
  logicBlock* get_currBlock();  //get the logic block that is placed at this CLB
  void set_sourceNode(routingNode* r);  //set the source node of this CLB
  void set_sinkNode(routingNode* r);  //set the sink node of this CLB
  routingNode* get_sourceNode();  //get the source node of this CLB
  routingNode* get_sinkNode();  //get the sink node of this CLB
  void set_blockType(enum e_genericBlockType t);
  enum e_genericBlockType get_blockType();
  void print_info();  //print details about this logic block
  void clearCurrBlock();

  //Is the CLB that this LUT belongs to empty? (LUTs and registers)
  bool emptyCLB();

  //Are the LUTs in this CLB empty?
  bool emptyCLBLUTS();

  //Is this LUT empty?
  bool isLUTAndLinkedRegistersEmpty(){
    if(currBlock)
      return false;
    else{
      //Check any associated register bank
      for(list <registerBank*>::iterator rbIter = linkedRegisterBanks.begin(); rbIter != linkedRegisterBanks.end(); rbIter++){
	if(!(*rbIter)->isJustBankEmpty()){
	  return false;
	}
      }
      return true;
    }
  };
  bool isJustLUTEmpty(){
    if(currBlock)
      return false;
    else
      return true;
  };
  bool isLinkedRegistersEmpty(){
    if(currBlock)
      //If the LUT part is occupied, the register banks must be empty.
      return true;
    else{
      //Check any associated register bank
      for(list <registerBank*>::iterator rbIter = linkedRegisterBanks.begin(); rbIter != linkedRegisterBanks.end(); rbIter++){
	if(!(*rbIter)->isJustBankEmpty()){
	  return false;
	}
      }
      return true;
    }
  };

  bool matchingLinkedRegisters(vprLUT* otherLUT);

  void swapLUTAndAssociatedRegBanks(vprLUT* otherLUT);
  void swapSingleLUTMappingAndAssociatedRegBanks(logicBlock* myLB, vprLUT* otherLUT, logicBlock* otherLB, double criticalPathDelay);
  void swapLUTGroupMappingAndAssociatedRegBanks(vprLUT* otherLUT, double criticalPathDelay);
  void swapStickyMapping(vprLUT* otherLUT, double criticalPathDelay);

  void updateSinkDelaysAndSlacks();
  void updateWireAndTimingCosts(bool timing_driven, double criticalPathDelay);
  double determineDeltaWireCost();
  double determineDeltaTimingCost();
  double determineDeltaTimingCostPrime();

  void updateCostAndDeltas(bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming);
  void updateCostAndDeltasCostDeltaDelay(bool timing_driven, double criticalPathDelay, double *deltaWire, 
					  double *deltaTiming, double *deltaTimingPrime);
  void updateCostAndDeltasForSticky(vprLUT* otherLUT, bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming);
  void updateCostAndDeltasForStickyCostDeltaDelay(vprLUT* otherLUT, bool timing_driven, double criticalPathDelay, double *deltaWire, 
						  double *deltaTiming, double *deltaTimingPrime);
  void updateCostAndDeltasForLUTGroupAndAssociatedRegBanks(vprLUT* otherLUT, bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming);
  void updateCostAndDeltasForLUTGroupAndAssociatedRegBanksCostDeltaDelay(vprLUT* otherLUT, bool timing_driven, double criticalPathDelay, double *deltaWire, 
									 double *deltaTiming, double *deltaTimingPrime);

  c_clb* returnMappedCLBFromLUTsAndFF();
  c_clb* returnMappedCLBFromLUTs();

  void setRegisterAssociation(registerBank* associatedRegisterBank);
  void occupyAllLinkedRegisterBanks();
  void vacateAllLinkedRegisterBanks();
  bool hasOccupiedLinkedRegisterBank();
  void transferAllLinkedRegisters(vprLUT * otherLUT);
  void calculateMaxInputOutputCriticality(double criticalPathDelay);
};

//This represents a group of 1 or more LUTs & register banks
class vprCLB :public vprBaseBlock{
 public:
  vector<registerBank *> registerBankVector;
  vector<vprLUT *> lutVector;
  int derived_magicNum;

  vprCLB(int row, int col);
  ~vprCLB(){
    //Delete the luts and registers
    for(vector<registerBank *>::iterator rbIter = registerBankVector.begin(); rbIter != registerBankVector.end(); rbIter++){
      delete *rbIter;
    }
    for(vector<vprLUT *>::iterator lutIter = lutVector.begin(); lutIter != lutVector.end(); lutIter++){
      delete *lutIter;
    }
  }

  void addAllRegBanksToRegisterBankVector();
  void addAllRegBanksToFreeRegisterBankList();

  void addRoutingNodeToRegisterBank(routingNode* r, int bank);
  void addRoutingNodeToRegisterBank(routingNode* r, int bank, int regNum, int lutAssociation);
  
  int getNumLUTs(){return lutVector.size();};
  int getNumRegBanks(){return registerBankVector.size();};

  //Return the number of empty luts in this CLB
  int getNumEmptyLUTs();  

  //Return the index of the first empty register bank;
  int firstEmptyRegisterIndex();

  //Return the number of empty register banks that are at least of a certain size in this CLB
  int getNumEmptyRegBank(int minSize);  
  //Return a random empty register bank that is at least of a certain size in this CLB
  registerBank *getRandomEmptyRegBank(int minSize);
  //Return the number of register banks in this CLB that are at least of a certain size, and are not driven by something
  // in this CLB.
  int getNumUnrelatedRegBank(int minSize, int maxCurrOccupancy, vprLUT *linkedLUT);
  //Return the number of register banks in this CLB that are at least of a certain size, and are not driven by something
  // in this CLB or include the specified sink.
  int getNumUnrelatedRegBank(int minSize, int maxCurrOccupancy, Terminal *driver, vprLUT *linkedLUT);
  //Return a random register bank that is at least of a certain size and not driven by something  in this CLB
  registerBank *getRandomUnrelatedRegBank(int minSize, int maxCurrOccupancy, vprLUT *linkedLUT);
  //Return a random register bank that is at least of a certain size and not driven by something in this CLB
  //   or include the specified sink.
  registerBank *getRandomUnrelatedRegBank(int minSize, int maxCurrOccupancy, Terminal *driver, vprLUT *linkedLUT);
  //Return the number of register banks that are at least of a certain size in this CLB
  int getNumMinSizeRegBank(int minSize, int maxCurrOccupancy, vprLUT *linkedLUT);
  //Return the number of register banks that are at least of a certain size in this CLB
  // and do not contain a specific driver
  int getNumMinSizeRegBank(int minSize, int maxCurrOccupancy, Terminal *driver, vprLUT *linkedLUT);
  //Return a random register bank that is at least of a certain size in this CLB
  registerBank *getRandomMinSizeRegBank(int minSize, int maxCurrOccupancy, vprLUT *linkedLUT);
  //Return a random register bank that is at least of a certain size in this CLB
  // that does not contain a specific driver
  registerBank *getRandomMinSizeRegBank(int minSize, int maxCurrOccupancy, Terminal *driver, vprLUT *linkedLUT);
 
  registerBank* getRegBank(int i){
    return registerBankVector.at(i);
  }
  
  vprLUT* getLUT(int i){
    return lutVector.at(i);
  }

  vprLUT* getOrMakeLUT(int i);

  //Are all of the register banks in this delay block empty?
  bool areRegistersEmpty();
  bool areLUTsEmpty();

  registerBank* getFirstEmptyUnrelatedRegisterBank();
  registerBank* getFirstEmptySingleOrShiftRegisterBank();
  registerBank* getFirstEmptyShiftRegisterBank(int capacity);

  void swapAllLUTs(vprCLB *otherArrayCLB);
  void swapAllLUTsAndFFs(vprCLB *otherArrayCLB);
  void swapAllLUTsAndAssociatedFFs(vprCLB *otherArrayCLB);
  void updateSinkDelaysAndSlacksForAllLUTsAndAssociatedFFs();
  void updateSinkDelaysAndSlacksForAllNonAssociatedFFs();
  void updateCostAndDeltasForAllLUTsAndAssociatedFFs(bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming);
  void updateCostAndDeltasForAllLUTsAndAssociatedFFsCostDeltaDelay(bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming, double *deltaTimingPrime);
  void updateCostAndDeltasForAllNonAssociatedFFs(bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming);
  void updateCostAndDeltasForAllNonAssociatedFFsCostDeltaDelay(bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming, double *deltaTimingPrime);

  void calculateMaxInputOutputCriticalitiesForAllLUTsAndAssociatedFFs(double criticalPathDelay);
  void calculateMaxInputOutputCriticalitiesForAllNonAssociatedFFs(double criticalPathDelay);
};

/****************************************************************
This class represents an IOB in the VPR architecture. Note that
there are multiple IO elements in a single IOB, and hence there
are multiple source and sink nodes in each IOB.
****************************************************************/
class vprIOB : public vprBaseBlock{
 public:
  vector<routingNode*>* sourceArray;  //ptr to array of source routing nodes
  vector<routingNode*>* sinkArray;  //ptr to array of sink routing nodes
  vector<logicBlock*>* currBlockArray;  //ptr to array of logic blocks placed at this IOB
  int numIOs;  //number of IOs per IOB
  int derived_magicNum;

  vprIOB(int r,int c);
  vprIOB(int r,int c,int n);
  ~vprIOB();
  void addto_sourceArray(routingNode* r,int i);  //add source node to sourceArray at index 'i'
  void addto_sinkArray(routingNode* r,int i);  //add sink node to sinkArray at index 'i'
  void addto_currBlockArray(logicBlock* l,int i);  //place logic block at 'ith' IO location 
  routingNode* getfrom_sourceArray(int i);  //get node in sourceArray at index 'i'
  routingNode* getfrom_sinkArray(int i);  //get node in sinkArray at index 'i'
  logicBlock* getfrom_currBlockArray(int i);  //get logic block placed at 'ith' IO location
  void print_info();  //print details about this IOB
  int get_numIOs(); //get the number of IOs in this IOB
  void lineup_IOs();
  void clearCurrBlockArray(int i);
  int getRandomIndex(){
    return (lrand48() % numIOs);
  };

  void swapMapping(int myIndex, logicBlock* myLB, vprIOB* otherIOB, int otherIndex, logicBlock* otherLB, double criticalPathDelay);
};


class vprFPGA{
 public:
  int numRows,numCols;  //num CLB rows and columns respectively
  int numCLBs,numIOs;  //num CLBs and IOBs in this architecture
  int numIOs_perIOB;  //number of separate IOs per IOB
  vector<vector<void*> > array;  //2-D array of CLBs and IOBs  
  list<routingNode*> rnodeList;  //List that holds ALL routing nodes in the architecture.
                                 //This list will be populated from the rr_graph.echo file
                                 //dumped by VPR. The routing edges that connect nodes are
                                 //'owned' by the routing nodes at which the edges originate.
  list<routingNode*>::iterator* rnode_itr;  //iterator for list of routing nodes
  int magicNum;
  vector <vector <vector <vector <vector <vector <double> > > > > > *timingTable;
  vector <vector <vector <vector <vector <vector <int> > > > > > *timingNumHitTable;

  vector<vprIOB *> iobList;
  vector<vprLUT *> lutList;
  //I'll convert this to a vprCLB later
  vector<vprLUT *> multList;
  vector<vprLUT *> ramList;

  vector<registerBank *> registerBankVector;
  list<registerBank *> freeRegisterBankList;
  int numFreeRegisters;

  vprFPGA(int num_rows,int num_cols,int num_ios);
  ~vprFPGA();

  int get_numRows(){
#ifdef PLACER_DEBUG
    if(magicNum!=23111)
      ERROR("trying to double delete node here...");
#endif
    return numRows;
  }

  int get_numCols(){
#ifdef PLACER_DEBUG
    if(magicNum!=23111)
      ERROR("trying to double delete node here...");
#endif
    return numCols;
  }

  int get_numCLBs();  //get num CLBs in this architecture
  int get_numIOs();  //get num IOBs in this architecture
  int countLUT();
  int countMult();
  int countRegBanks();
  int countRAM();
  void addto_rnodeList(routingNode* r);  //add routing nodes to rnodeList
  routingNode* getnext_rnodeList();  //get next routing node in rnodeList
  void create_rnodeListItr();  //create an iterator for rnodeList
  void remove_rnodeListItr();  //remove iterator for rnodeList

  //count (and optionally print) the total occupancy of congested routing nodes in this FPGA
  long int countCongestedNodeOccupancy(int opt);

  //update the history costs of all routing nodes in this FPGA
  void updateHistoryCosts(int iter);

  void print_info();  //print arch info about this FPGA

  void lineup_IOBs();  //line up IOs in each IOB's vectors 

  //Timing-driven routing stuff
  double calculateAverageIntrinsicDelay();
  void scaleBaseCosts(double averageIntrinsicDelay);

  //Pipeline stuff
  void initPlacementDS(bool testInitialPlacementOnly);

  int countDelayReg();
  void zeroOccupancy();
  long int countNewSharing(int opt);

  //Pipeline placement
  void create_block_type_lists();

  void createTimingTable();
  void initializeTimingTable();
  void initializeTimingNumHitTable();
  void calculateTimingTable();
  void emptyTimingTable();
  void emptyTimingNumHitTable();

  vector <double> cutCostVector;  
  void createCutCostVector();
  void clearCutCostVector();

  void sweepTimingIOPins();
  void sweepTimingLogic();
  void sweepTimingDelay();
  void findAllSinksFrom(void* sourceBlock, e_timingMoveBlockType sourceType, int sourceBank, int sourceRow, int sourceCol);
  void addTimingTable(double value, e_timingMoveBlockType sourceType, int sourceBank, e_timingMoveBlockType destType, int destBank, int rowSep, int colSep);
  void incNumHitTable(e_timingMoveBlockType sourceType, int sourceBank, e_timingMoveBlockType destType, int destBank, int rowSep, int colSep);
  void setTimingTable(double value, e_timingMoveBlockType sourceType, int sourceBank, e_timingMoveBlockType destType, int destBank, int rowSep, int colSep);
  double getTimingTable(e_timingMoveBlockType sourceType, int sourceBank, e_timingMoveBlockType destType, int destBank, int rowSep, int colSep){
    //return timingTable->at(sourceType).at(sourceBank).at(destType).at(destBank).at(colSep + numCols + 1).at(rowSep + numRows + 1);
    
    //These changes assume that all banks have the same connectivity    
    //if(timingTable->at(sourceType).at(0).at(destType).at(0).at(colSep + numCols + 1).at(rowSep + numRows + 1) > 1.0){
    //  cout << "Something went wrong!" << endl;
    //}
    return timingTable->at(sourceType).at(0).at(destType).at(0).at(colSep + numCols + 1).at(rowSep + numRows + 1);
  }

  vprCLB* getOrMakeCLBAt(int row, int col);
  vprCLB* getCLBAt(int row, int col);
  vprIOB* getOrMakeIOBAt(int row, int col);
  vprIOB* getIOBAt(int row, int col);

  int getNumberofBlockType(int blockType);
  
  vprIOB* getRandomIOB(){return iobList.at(lrand48() % iobList.size());};
  vprLUT* getRandomLUT(){return lutList.at(lrand48() % lutList.size());};
  vprLUT* getRandomMult(){return multList.at(lrand48() % multList.size());};
  vprLUT* getRandomRam(){return ramList.at(lrand48() % ramList.size());};

  registerBank* getRandomRegBank(){return registerBankVector.at(lrand48() % registerBankVector.size());};
  registerBank* getRandomFreeSingleRegBank();
  registerBank* getRandomFreeShiftRegBank(int capacity);
  registerBank* getClosestFreeRegBank(int row, int col);

  vprIOB* getOtherIOB(int row, int col, int rLim);
  vprLUT* getOtherLUT(int row, int col, int rLim);
  //vprLUT* getOtherMult(int row, int col, int rLim);
  //vprLUT* getOtherRam(int row, int col, int rLim);
  registerBank* getOtherRegBank(int row, int col, int rLim);
  registerBank* getOtherRegBank(int row, int col, int rLim, int minSize, int maxCurrOccupancy);

  void occupyRegBank(registerBank* rb);
  void vacateRegBank(registerBank* rb);

  void checkFreeRegisters();

  void placeClockPin(int *clockRow, int *clockCol, int *clockPTC);

  void checkArrayMappings();

  int getNumberCLBsEmptyLUTandRegs(int requiredLUTs, int requiredRegisters);
};

#endif

